import json
import pyotp
import requests
import logging

from datetime import datetime, timedelta
from pathlib import Path
from urllib.parse import unquote
from pymemcache.client import base

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import JsonResponse, HttpResponseBadRequest
from django.shortcuts import HttpResponse
from django.contrib import messages
from django.db.models import Q
from django.db import connections
from django.utils.translation import pgettext, gettext as _
from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError

from admin_panel.utils import ip_split
from admin_panel.utils import ip_validate
from main.common import get_filtered_logs
from main.decorators import time_of_function
from main.models.config import ConfigCleaningTable
from main.models.logs import Logs
from main.models.my_waf import LicKeys
from main.models.scan import ScannerReport
from main.models.wafuser import WAFUSER
from main.models.white_list import WhitelistRequest
from main.utils import convert_from_punycode
from main.utils import convert_to_punycode
from main.utils import get_list_messages
from main.utils import get_proxy_conf_dict
from my_waf.utils import send_set_settings
from utils.db import check_db


logger = logging.getLogger(__name__)

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))
mclient = base.Client(cache_address)

new_table_exist = check_db(logger)


def remove_spaces(x):
    return x if x is None else x.replace(' ', '')


def add_separator(x):
    return x if x is None else x.replace(' ', ',')


@login_required
@time_of_function
def add_user(request):
    if request.method == 'POST':
        email = request.POST.get('email')
        waf_alias = request.POST.get('waf_alias')
        password = request.POST.get('password')
        role = request.POST.get('role')
        status = request.POST.get('status')
        domain = request.POST.get('domain')
        pdf_report = request.POST.get('pdf_report', False)
        csv_report = request.POST.get('csv_report', False)
        mt_toggle = request.POST.get('mt_toggle', False)
        ws_toggle = request.POST.get('ws_toggle', False)
        oa_toggle = request.POST.get('oa_toggle', False)
        ec_toggle = request.POST.get('ec_toggle', False)
        acl = request.POST.get('acl', '').replace(' ', ',').replace(',,', ',')

        if acl:
            for ip in ip_split(acl):
                if not ip_validate(ip):
                    return HttpResponseBadRequest('Invalid ip data')

        if role == _('User'):
            role_id = 0
        else:
            role_id = 2

        if status == _('Active'):
            active = True
        else:
            active = False

        if pdf_report == 'on':
            pdf_report = True

        if csv_report == 'on':
            csv_report = True

        if oa_toggle == 'on':
            oa_toggle = True

        if ws_toggle == 'on':
            ws_toggle = True

        if mt_toggle == 'on':
            mt_toggle = True

        if ec_toggle == 'on':
            ec_toggle = True

        if waf_alias == '':
            waf_alias = '*'

        if password and email:
            user = WAFUSER()
            user.set_role(role_id)
            user.email = convert_to_punycode(remove_spaces(email))
            user.waf_alias = waf_alias.replace(' ', ', ').replace(',,', ',')
            user.is_active = bool(active)
            user.combined = False
            user.domain = convert_to_punycode(add_separator(domain).replace('*.', 'sub_').replace(',,', ','))
            user.acl = acl
            user.pdf_report = pdf_report
            user.csv_report = csv_report
            user.mt_toggle = mt_toggle
            user.ws_toggle = ws_toggle
            user.oa_toggle = oa_toggle
            user.ec_toggle = ec_toggle
            user.last_update = datetime.now()
            user.set_password(password)
            try:
                user.save()
            except IntegrityError:
                messages.error(request, pgettext('messages', 'User with that name already exists'))
        else:
            messages.error(request, pgettext('messages', 'Invalid email or password'))

        return HttpResponse(status=200)


@login_required
@time_of_function
def edit_user(request):
    if request.method == 'POST':
        user_id = request.POST.get('user_id')
        email = request.POST.get('email')
        waf_alias = request.POST.get('waf_alias')
        password = request.POST.get('password')
        role = request.POST.get('role')
        status = request.POST.get('status')
        domain = request.POST.get('domain')
        pdf_report = request.POST.get('pdf_report', False)
        csv_report = request.POST.get('csv_report', False)
        mt_toggle = request.POST.get('mt_toggle', False)
        ws_toggle = request.POST.get('ws_toggle', False)
        oa_toggle = request.POST.get('oa_toggle', False)
        ec_toggle = request.POST.get('ec_toggle', False)
        two_factor_toggle = request.POST.get('two_factor_toggle', False)
        acl = request.POST.get('acl', '').replace(' ', ',').replace(',,', ',')

        if acl:
            for ip in ip_split(acl):
                if not ip_validate(ip):
                    return HttpResponseBadRequest('Invalid ip data')

        if role == _('User'):
            role_id = 0
        else:
            role_id = 2

        if status == _('Active'):
            active = True
        else:
            active = False

        if pdf_report == 'on':
            pdf_report = True

        if csv_report == 'on':
            csv_report = True

        if oa_toggle == 'on':
            oa_toggle = True

        if ws_toggle == 'on':
            ws_toggle = True

        if ec_toggle == 'on':
            ec_toggle = True

        if mt_toggle == 'on':
            mt_toggle = True

        if two_factor_toggle == 'on':
            two_factor_toggle = True
        else:
            two_factor_toggle = False

        if email and int(user_id) > 0:
            user = WAFUSER.objects.get(id=user_id)
            user.set_role(role_id)
            user.email = convert_to_punycode(remove_spaces(email))
            user.waf_alias = waf_alias.replace(' ', ', ').replace(',,', ',')
            user.is_active = bool(active)
            user.combined = False
            user.domain = convert_to_punycode(add_separator(domain).replace('*.', 'sub_').replace(',,', ','))
            user.pdf_report = pdf_report
            user.csv_report = csv_report
            user.mt_toggle = mt_toggle
            user.ec_toggle = ec_toggle
            if two_factor_toggle and not user.is_2fa_active:
                user.secret = pyotp.random_base32()
            user.is_2fa_active = two_factor_toggle
            user.ws_toggle = ws_toggle
            user.oa_toggle = oa_toggle
            user.last_update = datetime.now()
            user.acl = acl
            if password:
                user.set_password(password)
            try:
                user.save()
            except IntegrityError:
                messages.error(request, pgettext('messages', 'User with that name already exists'))
        else:
            messages.error(request, pgettext('messages', 'Invalid email or password'))

        return HttpResponse(status=200)


@login_required
@time_of_function
def delete_user(request):
    if request.method == 'POST':
        user_id = request.POST.get('user_id')
        WAFUSER.objects.get(id=user_id).delete()
        return HttpResponse(status=200)


@login_required
@time_of_function
def manage_request(request):
    if request.method == 'POST':
        req_id = request.POST.get('req_id')
        admin_comment = request.POST.get('admin_comment')
        whitelist_requests = WhitelistRequest.objects.filter(log__request_id=req_id)
        for wl_request in whitelist_requests:
            wl_request.done = True
            wl_request.admin_comment = admin_comment
            wl_request.save()
        return HttpResponse(status=200)


@login_required
@time_of_function
def remove_request(request):
    if request.method == 'POST':
        req_id = request.POST.get('req_id')
        whitelist_requests = WhitelistRequest.objects.filter(log__request_id=req_id)
        for wl_request in whitelist_requests:
            wl_request.is_deleted = True
            wl_request.save()
        return HttpResponse(status=200)


@login_required
@time_of_function
def remove_scan(request):
    if request.method == 'POST':
        scan_id = request.POST.get('scan_id')
        scanner_report = ScannerReport.objects.using('default').get(id=int(scan_id))
        scanner_report.delete()
        return HttpResponse(status=200)


@login_required
@time_of_function
def remove_task(request):
    if request.method == 'POST':
        task_id = request.POST.get('task_id', None)
        if task_id is None:
            return HttpResponse(status=404)
        task = ConfigCleaningTable.objects.filter(id=int(task_id))
        if task.exists():
            task.first().delete()
        return HttpResponse(status=200)


@login_required
@time_of_function
def cleaning_run(request):
    if request.method == 'POST':
        if request.POST.get('run_tasks') is not None:
            if request.POST.get('run_tasks') == '2':
                data = json.loads(request.POST.get('task_data', (None, None)))
                tasks = ConfigCleaningTable.objects.filter(id=data['id_task'], desc=data['desc'])
            else:
                tasks = ConfigCleaningTable.objects.filter(enable=True)

            count = 0
            q_all = Q()
            str_query_all = ""
            for task in tasks:
                q = Q()
                str_query = ""
                if task.time_interval != 0:
                    datetime_sub_interval = datetime.today() - timedelta(days=task.time_interval)
                    q &= Q(timestamp__lt=datetime_sub_interval)
                    str_query += "timestamp < '{}'".format(datetime_sub_interval)
                if task.ip != '*':
                    q_ip = Q()
                    for ip in task.ip.replace(' ', '').split(','):
                        q_ip |= Q(ip=str(ip))
                    q &= q_ip
                    if len(str_query) != 0:
                        str_query += " AND "
                    str_query += "ip IN [{}]".format("','".join(task.ip.replace(' ', '').split(',')))
                if task.vhost != '*':
                    q_vhost = Q()
                    str_vhost = ""
                    for i in task.vhost.replace(' ', '').split(','):
                        if len(str_vhost) != 0:
                            str_vhost += " AND "
                        if i[0] in ('*', '.'):
                            q_vhost |= Q(vhost__iendswith=i[1:])
                            str_vhost += "vhost ILIKE '%{}'".format(i[1:])
                        else:
                            q_vhost |= Q(vhost=i)
                            str_vhost += "vhost = '{}'".format(i)
                    q &= q_vhost
                    if len(str_query) != 0:
                        str_query += " AND "
                    str_query += str_vhost
                if task.bt != '*':
                    if ',' in task.bt or task.bt.isdigit():
                        list_bt = task.bt.replace(' ', '').split(',')
                        q &= Q(bt__in=list_bt)
                        if len(str_query) != 0:
                            str_query += " AND "
                        str_query += "bt IN ({})".format(",".join(list_bt))
                if task.url != '*':
                    q &= Q(url=task.url)
                    if len(str_query) != 0:
                        str_query += " AND "
                    str_query += "url = '{}'".format(task.url)
                q_all |= q
                if len(str_query_all) != 0:
                    str_query_all += " OR "
                str_query_all += "({})".format(str_query)
            if tasks.count() != 0:
                attacks_list = Logs.objects.using('waf').only('id').filter(q_all)
                count = attacks_list.count()
                if count != 0:
                    r_sql = 'DELETE FROM attack '
                    try:
                        if len(str_query_all) > 0:
                            str_query_all = 'WHERE ' + str_query_all
                        r = '{}{}'.format(r_sql, str_query_all)
                        with connections['waf'].cursor() as cursor:
                            cursor.execute(r)
                    except Exception as e:
                        logger.error(
                            "An error occurred while connection to db. The connection to the database was rejected. "
                            f"Access rights are too low. Or there is no connection to the database{e}"
                        )

            return HttpResponse(count, status=200)

    return HttpResponse()


@login_required
@time_of_function
def get_rule_wl(request):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain
    row_id = request.GET.get('row_id', '')
    row = Logs.objects.using('waf').filter(id=row_id)
    try:
        row = get_filtered_logs(row, user_alias, domain_from_wafuser, None).values('vhost', 'rule_id', 'mz',
                                                                                   'alias').first()
    except IndexError:
        row = None

    if not row:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    rules_file = Path(path)
    res = {}
    if rules_file.is_file():
        res['wl_id'] = row['rule_id']
        res['domain'] = convert_from_punycode(row['vhost'])
        res['mz'] = str(row['mz']).lower()
    else:
        res = {"status": "empty"}
    res['waf_id'] = row['alias']

    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    for key in keys:
        try:
            response_check_key = requests.post(
                settings.URI_AUTH,
                proxies=get_proxy_conf_dict(),
                headers={'User-agent': settings.USER_AGENT},
                timeout=30,
                data={'key': key.key}
            )

            if response_check_key.status_code == 201:
                check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
                if check_key_result['status'] == 'success':
                    if check_key_result['waf_id'] == res['waf_id']:
                        res['key_id'] = key.id
        except Exception as e:
            logger.error(
                'An error occurred while getting information about the key: {key}. Error information: {err}'
                .format(key=key.get_formated_key(), err=e)
            )

    if type(res.get('key_id')) is not int and not str(res.get('key_id')).isnumeric():
        res['error'] = 'Invalid WAF ID. Perhaps this attack does not exist.'
        messages.error(request, pgettext('messages', 'Invalid WAF ID. Perhaps this attack does not exist.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    return JsonResponse(res)


@login_required
@time_of_function
def add_wl(request):
    get_form = request.POST.get('form')
    get_extension_list = request.POST.getlist('extension[]') or []
    get_form_list = str(get_form).split('&') or []
    get_mz = []
    for el in get_form_list:
        if 'mz=' in el:
            el = str(el).replace('mz=', '')
            for x, y in ("body", "BODY"), ("url", "URL"), ("args", "ARGS"), ("headers", "HEADERS"), \
                        ("mla", "MLA"), ("no", "No"), ('ban', 'Ban'):
                el = el.replace(x, y)
            get_mz.append(el)
    mz = '|'.join(get_mz)
    get_form_dict = {}
    for el in get_form_list:
        el = el.split('=')
        get_form_dict[el[0]] = el[1]
    extension_list = []
    for el in get_extension_list:
        el = str(el).split(';')
        if el[0] != 'none':
            if el[0].lower() == 'body' or el[0].lower() == 'url' \
                    or el[0].lower() == 'args' or el[0].lower() == 'headers':
                zone = '$' + el[0].upper()
            else:
                zone = '$' + el[0]
            if el[2] == 'true':
                zone += '_X:'
            else:
                zone += ':'
            extension_list.append(zone + el[1])
    if 'active' in get_form_dict and get_form_dict['active'] == 'true':
        get_form_dict['active'] = True
    else:
        get_form_dict['active'] = False

    if len(extension_list) > 0:
        extension = '|'.join(extension_list)
        get_form_dict['extension'] = extension
    get_form_dict['mz'] = mz
    if 'other_zone' in get_form_dict:
        if get_form_dict['other_zone'] != '':
            other_zone = '|'.join(unquote(get_form_dict['other_zone']).split(','))
            if get_form_dict['mz'] != '':
                get_form_dict['mz'] = get_form_dict['mz'] + '|' + other_zone
            else:
                get_form_dict['mz'] = other_zone
            get_form_dict.pop('other_zone')
        else:
            get_form_dict.pop('other_zone')

    if get_form_dict.get('domain'):
        get_form_dict['domain'] = convert_to_punycode(unquote(get_form_dict['domain']))

    selected_key = get_form_dict.get('lic_keys_select')
    if not selected_key:
        messages.error(request, pgettext('messages', 'Invalid WAF ID. Perhaps this attack does not exist.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    try:
        key = LicKeys.objects.get(id=selected_key).key
    except ObjectDoesNotExist:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    get_form_dict.pop('lic_keys_select')
    full_settings_dict = {"key": key, 'add': get_form_dict}
    send_status = send_set_settings('WL', full_settings_dict)
    if send_status[0] == 'success':
        return HttpResponse('{"status": "success"}', content_type='application/json')
    else:
        return HttpResponse('{"status": "error"}', content_type='application/json')
