from django.conf import settings
from django.shortcuts import HttpResponseRedirect
from django.utils.translation import gettext as _
from django.utils.translation import pgettext
from django.views.generic import TemplateView
from django.contrib.auth.mixins import LoginRequiredMixin

from main.common import get_reason_by_model
from main.common import get_user_key_waf_id
from main.decorators import time_of_function
from main.models.config import Config
from main.models.config import ConfigCleaningTable
from main.models.scan import ScannerReport
from main.models.wafuser import WAFUSER
from main.models.white_list import WhitelistRequest
from main.utils import convert_from_punycode
from utils.db import check_db

new_table_exist = check_db(None)


class IndexView(LoginRequiredMixin, TemplateView):
    template_name = 'index.html'

    @time_of_function
    def get(self, request, *args, **kwargs):
        if request.user.get_role() > 1:
            users = WAFUSER.objects.order_by('email').all()
            if settings.API_URI is not None:
                keys_waf_ids = get_user_key_waf_id(request.user)
            else:
                keys_waf_ids = None
            whitelist_requests = None
            whitelist_requests_list = None
            scanner_reports = None
            for user in users:
                if user.is_admin:
                    user.role = _("Administrator")
                else:
                    user.role = _("User")

                if user.is_active:
                    user.status = _('Active')
                else:
                    user.status = _('Inactive')

                if user.acl is None:
                    user.acl = ''

                user.waf_alias = str(user.waf_alias).replace(' ', '')
                user.domain = convert_from_punycode(str(user.domain).replace('sub_', '*.'))
                user.email = convert_from_punycode(str(user.email))

                whitelist_requests_list = []
                group_whitelist_requests_list = []
                rule_ids_dict = {}
                mz_group_dict = {}
                request_ids = []
                whitelist_requests = WhitelistRequest.objects.filter(is_deleted=False).order_by('done', '-id',
                                                                                                '-record_date')
                for whitelist_request in whitelist_requests:
                    if whitelist_request.done:
                        whitelist_request.status = _('Completed')
                    else:
                        whitelist_request.status = _('Incompleted')

                    whitelist_request.request_date = whitelist_request.log.timestamp.__format__('%d.%m.%Y %H:%M:%S')
                    whitelist_request.record_date = whitelist_request.record_date.__format__('%d.%m.%Y %H:%M:%S')

                    whitelist_request.reason = get_reason_by_model(whitelist_request.log.bt)
                    whitelist_request.log.other_headers = str(whitelist_request.log.other_headers).strip()
                    whitelist_request.body = bytes(whitelist_request.log.body).decode('utf-8', 'ignore')
                    whitelist_request.log.vhost = convert_from_punycode(whitelist_request.log.vhost)

                    if whitelist_request.log.request_id in request_ids:
                        rule_ids_dict[whitelist_request.log.request_id] = \
                            str(rule_ids_dict[whitelist_request.log.request_id]) + ", " + str(
                                whitelist_request.log.rule_id)
                        mz_group_dict[whitelist_request.log.request_id] = \
                            str(mz_group_dict[whitelist_request.log.request_id]) + ", " + str(whitelist_request.log.mz)
                    else:
                        request_ids.append(whitelist_request.log.request_id)
                        rule_ids_dict[whitelist_request.log.request_id] = str(whitelist_request.log.rule_id)
                        mz_group_dict[whitelist_request.log.request_id] = str(whitelist_request.log.mz)
                        group_whitelist_requests_list.append(whitelist_request)

                for whitelist_request in group_whitelist_requests_list:
                    if whitelist_request.log.request_id in rule_ids_dict:
                        whitelist_request.log.rule_id = str(rule_ids_dict[whitelist_request.log.request_id]).split(', ')
                        whitelist_requests_list.append(whitelist_request)

                    if whitelist_request.log.request_id in mz_group_dict:
                        whitelist_request.log.mz_group = str(mz_group_dict[whitelist_request.log.request_id]).lower()

                scanner_reports = ScannerReport.objects.using('default').all()
                for scanner in scanner_reports:
                    scanner.scan_date = scanner.scan_date.__format__('%d.%m.%Y %H:%M:%S')

            # getting config data
            config = self._get_config_data()

            tasks_cleaning = self._get_tasks_cleaning()

            context = self.get_context_data(
                home=settings.HOME_PAGE,
                st_exist=new_table_exist,
                config=config,
                t=True,
                tasks_cleaning=tasks_cleaning,
                users=users,
                whitelist_requests=whitelist_requests,
                whitelist_requests_list=whitelist_requests_list,
                scanner_reports=scanner_reports,
                keys_waf_ids=keys_waf_ids,
                user=request.user
            )

            return self.render_to_response(context)
        else:
            return HttpResponseRedirect(settings.HOME_PAGE)

    @staticmethod
    def _get_config_data():
        config = Config.objects.first()
        if config:
            config.send_email_text = pgettext('py', 'Enable') if config.send_email else pgettext('py', 'Disable')
        else:
            config = Config()
            config.send_email_text = pgettext('py', 'Disable')
            config.save()
        return config

    @staticmethod
    def _get_tasks_cleaning():
        configs = ConfigCleaningTable.objects.all().order_by('time_interval')
        for config in configs:
            if config.time_interval == 0:
                config.interval_text = '*'
            else:
                config.interval_text = '{} {}'.format(config.time_interval, _("day(s)"))
            config.bt1 = False
            config.bt2 = False
            config.bt3 = False
            config.bt4 = False
            config.bt5 = False
            config.bt6 = False
            config.bt7 = False
            config.bt8 = False
            config.bt9 = False
            config.bt10 = False
            config.bt14 = False
            config.bt16 = False
            if config.bt != '*':
                split_bt = config.bt.split(',')
                if '1' in split_bt:
                    config.bt1 = True
                if '2' in split_bt:
                    config.bt2 = True
                if '3' in split_bt:
                    config.bt3 = True
                if '4' in split_bt:
                    config.bt4 = True
                if '5' in split_bt:
                    config.bt5 = True
                if '6' in split_bt:
                    config.bt6 = True
                if '7' in split_bt:
                    config.bt7 = True
                if '8' in split_bt:
                    config.bt8 = True
                if '9' in split_bt:
                    config.bt9 = True
                if '10' in split_bt:
                    config.bt10 = True
                if '14' in split_bt:
                    config.bt14 = True
                if '16' in split_bt:
                    config.bt16 = True

        return configs
