"""
 Django settings for Nemesida WAF Cabinet project.
"""

import os
import sys
from pathlib import Path

from tzlocal import get_localzone
from cabinet.settings import ALLOWED_HOSTS, SECRET_KEY, HTTP_PROXY_CONF, DB_NAME_CABINET, DB_USER_CABINET, \
    DB_PASS_CABINET, DB_HOST_CABINET, DB_NAME_CONF, DB_USER_CONF, DB_PASS_CONF, DB_HOST_CONF
from django.contrib.messages import constants as messages

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
# BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BASE_DIR = Path(__file__).resolve().parent.parent.parent
sys.path.append('/var/www/app/cabinet')

try:
    from cabinet.settings import API_URI

    if API_URI not in ('', None):
        if 'nw-api' not in API_URI:
            API_URI += '/nw-api/'

        if not API_URI.endswith('/'):
            API_URI += '/'
    else:
        API_URI = None

except (ImportError, NameError):
    API_URI = None

try:
    from cabinet.settings import DB_PORT_CABINET

    if DB_PORT_CABINET == '':
        DB_PORT_CABINET = '5432'
except (ImportError, NameError):
    DB_PORT_CABINET = '5432'

try:
    from cabinet.settings import DB_PORT_CONF

    if DB_PORT_CONF == '':
        DB_PORT_CONF = '5432'
except (ImportError, NameError):
    DB_PORT_CONF = '5432'

try:
    from cabinet.settings import DEBUG
except (ImportError, NameError):
    DEBUG = False

try:
    from cabinet.settings import API_PROXY
except (ImportError, NameError):
    API_PROXY = ''

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'admin_panel',
    'main',
    'my_waf',
    'model_tuning',
    'report',
    # 'nginx',

]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware'
]

if DEBUG:
    MIDDLEWARE.append('cabinet.middlewares.memory_profiler_mw.MemoryProfilerMiddleware')

INTERNAL_IPS = [
    # ...
    "127.0.0.1",
    # ...
]

ROOT_URLCONF = 'cabinet.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.request',
            ],
        },
    },
]

WSGI_APPLICATION = 'cabinet.wsgi.application'

# CACHE SETTINGS
try:
    from cabinet.settings import MEMCACHED_HOST, MEMCACHED_PORT

    location_memc = '127.0.0.1:11211'
    if MEMCACHED_HOST not in (None, '') and MEMCACHED_PORT not in (None, ''):
        location_memc = MEMCACHED_HOST + ":" + MEMCACHED_PORT
except (ImportError, NameError):
    location_memc = '127.0.0.1:11211'

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.PyMemcacheCache',
        'LOCATION': location_memc,
        'TIMEOUT': 600,
    }
}

# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases
DATABASE_ROUTERS = ['main.router.OtherRouter']

try:
    if DB_PORT_CONF == '':
        DB_PORT_CONF = '5432'
except NameError:
    DB_PORT_CONF = '5432'

try:
    if DB_PORT_CABINET == '':
        DB_PORT_CABINET = '5432'
except NameError:
    DB_PORT_CABINET = '5432'

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

DATABASES = {
    'waf': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': DB_NAME_CONF,
        'HOST': DB_HOST_CONF,
        'PORT': DB_PORT_CONF,
        'USER': DB_USER_CONF,
        'PASSWORD': DB_PASS_CONF,
    },
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': DB_NAME_CABINET,
        'HOST': DB_HOST_CABINET,
        'PORT': DB_PORT_CABINET,
        'USER': DB_USER_CABINET,
        'PASSWORD': DB_PASS_CABINET,
    },
}

# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

PASSWORD_RESET_TIMEOUT = 240

# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/
LANGUAGE_CODE = 'en'

LANGUAGES = (
    ('en', 'English'),
    ('ru', 'Russian'),
)

LANGUAGE_COOKIE_NAME = 'django_language_code'

# Locate locale files
LOCALE_PATHS = (
    os.path.join(BASE_DIR, 'locale/'),
    # os.path.join(PROJECT_DIR, 'locale'),
)

TIME_ZONE = str(get_localzone())

USE_I18N = True

USE_L10N = True

USE_TZ = False

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.10/howto/static-files/


if os.getenv('DEV'):
    # for short_task django dev server add variable DEV to environments
    HOME_PAGE_FOR_STATIC = ''
    HOME_PAGE = ''

    INSTALLED_APPS.append("debug_toolbar")
    INSTALLED_APPS.append("rosetta")
    MIDDLEWARE.append("debug_toolbar.middleware.DebugToolbarMiddleware")
else:
    HOME_PAGE_FOR_STATIC = 'waf/personal/'
    HOME_PAGE = '/waf/personal/'

STATIC_ROOT = os.path.join(BASE_DIR, HOME_PAGE_FOR_STATIC + 'static/')
STATIC_URL = HOME_PAGE + 'static/'

STATICFILES_DIRS = (
    os.path.join(BASE_DIR, 'main', 'static'),
    os.path.join(BASE_DIR, 'admin_panel', 'static'),
    os.path.join(BASE_DIR, 'my_waf', 'static'),
)

LOGIN_URL = HOME_PAGE + "login/"
LOGIN_REDIRECT_URL = HOME_PAGE
LOGOUT_REDIRECT_URL = HOME_PAGE

AUTH_USER_MODEL = 'main.WAFUSER'

DEFAULT_DATE_FILTER = 5
DEFAULT_PAGE_SIZE = 100
DEFAULT_MT_PAGE_SIZE = 250

# Logging if execution time is longer

LOG_FOR_TIMEOUT = 120

# URI to remote API

URI_RL = 'https://nemesida-security.com/nw/rl'
URI_RL_TS = URI_RL + '/ts'
URI_AUTH = 'https://nemesida-security.com/nw/auth/v3'
URI_GEOIP = 'https://geoip.nemesida-security.com/geoip/v3?ip='
URI_GEOIP_BATCH = "https://geoip.nemesida-security.com/geoip/batch/v1?geo_token="
URI_GEOIP_TIMEOUT = 10

V_GEOIP = 'v3'
V_APP = 'v3.1.0'

# Security settings
SECURE_CONTENT_TYPE_NOSNIFF = True
SECURE_BROWSER_XSS_FILTER = True

USER_AGENT = 'Nemesida WAF Cabinet'

MESSAGE_TAGS = {
    messages.DEBUG: 'alert-secondary',
    messages.INFO: 'alert-info',
    messages.SUCCESS: 'alert-success',
    messages.WARNING: 'alert-warning',
    messages.ERROR: 'alert-danger',
}

# Paths

PATH_DIR_TASKS = os.path.join(BASE_DIR, 'tasks')

# Another params

DEBUG_MEMORY = False
