import ipaddress
import json
import logging
import multiprocessing
from collections import Counter
from contextlib import contextmanager
from datetime import datetime
from typing import (
    Optional,
    Union
)
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from multiprocessing import Lock as PLock
from threading import Lock as TLock

import requests
from django.conf import settings
from django.core.cache import cache
from django.db.models import Max
from django.db.models import Q
from django.db.models import QuerySet
from django.db.models.aggregates import Count
from django.utils.translation import gettext as _

from main.models import Logs, ScanRequest, LicKeys, MtTable, MtExport, WhitelistRequest, Geo_Tokens
from main.models.machine_learning import Training
from main.utils import convert_from_punycode, get_proxy_conf_dict
from my_waf.utils import get_lic_keys


logger = logging.getLogger(__name__)

fields_equal = [
    "body_equal",
    "referer_equal",
    "method_equal",
    "url_equal",
    "param_equal",
    "ua_equal",
    "cookie_equal",
]

vGeoIP = settings.V_GEOIP

# Proxy settings
proxy_dict_conf = get_proxy_conf_dict()

user_agent = {'User-agent': settings.USER_AGENT}


def chunks(li, n):
    for i in range(0, len(li), n):
        chunk = li[i:i + n]

        yield chunk


def chunks_for_queryset(queryset: QuerySet, n):
    count_q = queryset.count()
    if count_q and n > count_q:
        n = count_q

    for i in range(0, count_q, n):
        chunk = queryset[i:i + n]

        yield chunk


@contextmanager
def dummy_lock():
    yield


def run_multithreading_process(
    f_target,
    data,
    *args,
    len_chunks=100,
    f_chunking=chunks,
    max_workers=None,
    **kwargs
):
    """ Run multi threads func with chunking data """

    with ThreadPoolExecutor(max_workers=max_workers) as pool:
        for i in f_chunking(data, len_chunks):
            pool.submit(f_target, i, *args, **kwargs)

        pool.shutdown(wait=True)


def run_multiprocess_process(
    f_target,
    data,
    *args,
    len_chunks=100,
    f_chunking=chunks,
    max_workers=None,
    **kwargs
):
    """ Run multiprocess func with chunking data """

    with ProcessPoolExecutor(max_workers=max_workers) as pool:
        for i in f_chunking(data, len_chunks):
            pool.submit(f_target, i, *args, **kwargs)

        pool.shutdown(wait=True)


def get_ml_table():
    ml_table = {}
    for ml in Training.objects.using("waf").all():
        ml_table['{}-{}'.format(ml.waf_id, ml.vhost)] = ml
    return ml_table


def get_tags(main_base, lvl=1, bt: [list, int] = 0):
    if main_base.ordered:
        list_q = list(main_base.values_list('id', flat=True))
        main_base = Logs.objects.using('waf').filter(id__in=list_q)
    if lvl == 1:
        main_tags = dict()
        indicator = 0
        main_base = main_base.values('vhost').annotate(
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            mla=Count('bt', filter=Q(bt=3)),
            mlc=Count('bt', filter=Q(bt=8)),
            ie=Count('bt', filter=Q(bt=5)),
            bf=Count('bt', filter=Q(bt=7)),
            fl=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
            av=Count('bt', filter=Q(bt__in=[4, 6])),
            oapi=Count('bt', filter=Q(bt=14)),
            blip=Count('bt', filter=Q(bt=16)),
        ).values('sa', 'mla', 'mlc', 'av', 'bf', 'ie', 'fl', 'ddos', 'oapi', 'blip', 'vhost')

        main_base = list(main_base)
        for i in main_base:
            tags = []
            vhost = convert_from_punycode(i.pop('vhost'))
            for k, v in sorted(i.items(), key=lambda item: -item[1]):
                if v == 0:
                    continue
                if k == 'sa':
                    tag = 'Signature'
                    bt = [1, 2]
                elif k == 'mla':
                    tag = 'MLA'
                    bt = [3]
                elif k == 'mlc':
                    tag = 'MLC'
                    bt = [8]
                elif k == 'av':
                    tag = 'Antivirus'
                    bt = [4, 6]
                elif k == 'bf':
                    tag = 'Brute'
                    bt = [7]
                elif k == 'ie':
                    tag = 'Error'
                    bt = [5]
                elif k == 'fl':
                    tag = 'Flood'
                    bt = [9]
                elif k == 'ddos':
                    tag = 'DDoS'
                    bt = [10]
                elif k == 'oapi':
                    tag = 'API Firewall'
                    bt = [14]
                elif k == 'blip':
                    tag = 'Blocked IP'
                    bt = [16]
                else:
                    tag = 'undefined'
                    bt = [0]

                tags.append({'tag': tag, 'dcount': attack_short_count(v, 0), 'vhost': vhost, 'bt': bt, 'id': indicator})

                indicator += 1
            main_tags[vhost] = tags

        return main_tags
    elif lvl == 2:
        bt = list(bt)
        main = []
        if set(bt) & {1, 2} != set():
            main1 = main_base.filter(bt__in=[1, 2]).values('vhost', 'description').annotate(
                dcount=Count('id')).order_by('-dcount')
            for i in main1.iterator():
                i['bt'] = [1, 2]
                main.append(i)
        if set(bt) & {4, 6} != set():
            main46 = main_base.filter(bt__in=[4, 6]).values('vhost', 'description').annotate(
                dcount=Count('id')).order_by('-dcount')
            for i in main46.iterator():
                i['bt'] = [4, 6]
                main.append(i)
        if set(bt) & {5} != set():
            main5 = main_base.filter(bt=5).values('bt', 'vhost', 'description').annotate(dcount=Count('id')).order_by(
                '-dcount')
            for i in main5.iterator():
                i['bt'] = [5]
                main.append(i)
        if set(bt) & {7, 9, 10} != set():
            main7910 = main_base.filter(bt__in=[7, 9, 10]).values('bt', 'vhost', 'possible').annotate(
                dcount=Count('id')).order_by('-dcount')
            for i in main7910.iterator():
                i['bt'] = [i['bt']]
                if i['possible'] in (None, False):
                    i['possible'] = False

                else:
                    i['possible'] = True
                main.append(i)

        indicator = 0
        false_is_exist = []
        true_is_exist = []
        for i in main:
            i['id'] = indicator
            indicator += 1
            if not set(i['bt']).isdisjoint({1, 2, 4, 6, 5}):
                if 5 in i['bt'] and i['description'] in ('', None):
                    i['subtag'] = _('No description')
                    i['description'] = '__nodesc_'
                else:
                    i['subtag'] = i['description']
            elif not set(i['bt']).isdisjoint({7, 9, 10}):
                if i['possible']:
                    i['subtag'] = 'Possible'
                    true_is_exist.append(i)
                else:
                    i['subtag'] = _('excl. Possible')
                    false_is_exist.append(i)

        for num, i in enumerate(false_is_exist):
            if i in main:
                main.remove(i)
            if num > 0:
                false_is_exist[0]['dcount'] = int(i['dcount']) + int(false_is_exist[0]['dcount'])
                false_is_exist[0]['bt'] = i['bt'] + false_is_exist[0]['bt']
        if false_is_exist:
            main.append(false_is_exist[0])
        for num, i in enumerate(true_is_exist):
            if i in main:
                main.remove(i)
            if num > 0:
                true_is_exist[0]['dcount'] = int(i['dcount']) + int(true_is_exist[0]['dcount'])
                true_is_exist[0]['bt'] = i['bt'] + true_is_exist[0]['bt']
        if true_is_exist:
            main.append(true_is_exist[0])

        main = sorted(main, key=lambda d: -int(d['dcount']))

        for i in main:
            i['dcount'] = attack_short_count(int(i.get('dcount')), 0)
        return main


def get_country_for_group(main_base):
    if main_base.ordered:
        list_q = list(main_base.values_list('id', flat=True))
        main_base = Logs.objects.using('waf').filter(id__in=list_q)
    main_base = main_base.values('vhost', 'cc').annotate(dcount=Count('id')).order_by('-dcount')
    main = dict()
    for j in main_base:
        vhost = convert_from_punycode(j.get('vhost'))
        if main.get(vhost) is None:
            main[vhost] = []
        if j.get('cc') is None:
            continue
        else:
            if j.get('cc').strip(' ') == '':
                j['cc'] = 'None'
        if j.get('cc') == 'None':
            for i in main[vhost]:
                if i.get('cc') == 'None':
                    j['dcount'] = j.get('dcount') + int(i.get('dcount').replace('M', '000000').replace('K', '000'))
                    main[vhost].remove(i)
                    break
        j['dcount'] = attack_short_count(j.get('dcount'), 0)
        main[vhost].append(j)
    return main


def get_vhost(main):
    return main.values('vhost').annotate(max_timestamp=Max('timestamp')).annotate(dcount=Count('id')).order_by(
        '-max_timestamp')


def get_groups(main_base):
    if main_base.ordered:
        list_q = list(main_base.values_list('id', flat=True))
        main_base = Logs.objects.using('waf').filter(id__in=list_q)
    vh = {}
    for num, i in enumerate(get_vhost(main_base).values_list('vhost', flat=True)):
        vh[convert_from_punycode(i)] = num

    main = main_base.values('vhost').annotate(dcount=Count('id'))

    indicator = 0
    for i in main:
        i['id'] = indicator
        i['vhost'] = convert_from_punycode(i['vhost'])
        indicator += 1
        i['dcount'] = attack_short_count(i.get('dcount'), 0)
    main = sorted(main, key=lambda x: vh.get(x['vhost'], 999))
    return main


def waf_domain_filtration(main_base, waf_domain=None):
    if main_base.model.__name__ == 'ScannerReport':
        filter_field = 'domain'
    elif main_base.model.__name__ == 'Logs':
        filter_field = 'vhost'
    else:
        filter_field = 'vhost'
    if waf_domain:
        check_domain = waf_domain.replace(' ', '').split(',')
        if waf_domain == '*':
            pass
        elif len(check_domain) > 1:
            q_objects = Q()
            for elem in check_domain:
                if elem[0] == '*':
                    q_objects |= Q(**{filter_field + '__icontains': elem[1:].lstrip('.')})
                elif elem.count('sub_') == 1:
                    q_objects |= Q(**{filter_field + '__icontains': '.' + elem[4:]})
                elif elem.find('.') == 0:
                    q_objects |= Q(**{filter_field + '__icontains': elem[elem.find('.') + 1:]})
                else:
                    q_objects |= Q(**{filter_field: elem})
            main_base = main_base.filter(q_objects)
        elif waf_domain[0] == '*':
            main_base = main_base.filter(**{filter_field + '__icontains': waf_domain[1:]})
        else:
            if waf_domain == '*':
                main_base = main_base.filter(**{filter_field + '__icontains': waf_domain[1:].lstrip('.')})
            elif waf_domain.count('sub_') == 1:
                main_base = main_base.filter(**{filter_field + '__icontains': '.' + waf_domain[4:]})
            elif waf_domain.find('.') == 0:
                main_base = main_base.filter(**{filter_field + '__icontains': waf_domain[waf_domain.find('.') + 1:]})
            else:
                main_base = main_base.filter(**{filter_field: waf_domain})
    return main_base


def get_filtered_logs(main_queryset: QuerySet, waf_id: str, waf_domain=None, q=None) -> QuerySet:
    """ """
    if main_queryset.model.__name__ == 'ScannerReport':
        filter_field = 'waf_alias'
        second_filter_field = 'domain'
    elif main_queryset.model.__name__ == 'Logs':
        filter_field = 'alias'
        second_filter_field = 'vhost'
    elif main_queryset.model.__name__ == 'MtTable':
        filter_field = 'waf_id'
        second_filter_field = 'vhost'
    else:
        filter_field = 'waf_id'
        second_filter_field = 'vhost'
    if waf_id != '*':
        check_waf_id = waf_id.replace(' ', '').split(',')
        if len(check_waf_id) == 1:
            main_queryset = main_queryset.only(str(filter_field), str(second_filter_field)).filter(
                **{filter_field: waf_id})
            main_queryset = waf_domain_filtration(main_queryset, waf_domain)
        else:
            q_objects = Q()
            for elem in check_waf_id:
                q_objects |= Q(**{filter_field: elem})
            main_queryset = main_queryset.only(str(filter_field), str(second_filter_field)).filter(q_objects)
            main_queryset = waf_domain_filtration(main_queryset, waf_domain)
    elif waf_id == '*':
        if q:
            main_queryset = main_queryset.only(str(filter_field), str(second_filter_field)).filter(**{filter_field: q})
        else:
            main_queryset = waf_domain_filtration(main_queryset, waf_domain)
    return main_queryset


def get_ml_learning_progress(
        ml_table, i, ml_learning_progress_full,
        ml_learning_progress_full_with_point,
        ml_learning_progress_full_with_point_and_star
                            ):
    try:
        if ml_learning_progress_full is not None:
            ml_learning_progress = ml_learning_progress_full.progress
        elif ml_learning_progress_full_with_point is not None:
            ml_learning_progress = ml_learning_progress_full_with_point.progress
        elif ml_learning_progress_full_with_point_and_star is not None:
            ml_learning_progress = ml_learning_progress_full_with_point_and_star.progress
        else:
            vhost = i.vhost
            # while tmp:
            if vhost.count('.') == 1:
                position = vhost.find('.')
                vhost = vhost[position:]
                ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, vhost))
                if ml_learning_progress_full is not None:
                    ml_learning_progress = ml_learning_progress_full.progress
                else:
                    ml_learning_progress = None
            else:
                if vhost.count('.') > 2:
                    split_vhost = vhost.split('.')
                    vhost = '.' + split_vhost[-2] + '.' + split_vhost[-1]
                position = vhost.find('.')
                vhost = vhost[position + 1:]
                ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, 'sub_' + vhost))
                if ml_learning_progress_full is not None:
                    ml_learning_progress = ml_learning_progress_full.progress
                else:
                    ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, '.' + vhost))
                    if ml_learning_progress_full is not None:
                        ml_learning_progress = ml_learning_progress_full.progress
                    else:
                        ml_learning_progress = None

        if ml_learning_progress is None:
            ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, 'all'))
            if ml_learning_progress_full is not None:
                ml_learning_progress = ml_learning_progress_full.progress
            else:
                ml_learning_progress = '0'

    except Exception:
        ml_learning_progress = '0'
    return ml_learning_progress


def get_main_logs_new(main_base, ml_table, mcache, page):
    count_ips = dict()
    page_size = settings.DEFAULT_PAGE_SIZE
    start_page = page_size * (page - 1)
    end_page = page_size * page
    main = main_base.defer('cc').only('vhost', 'method', 'alias', 'bt', 'ip', 'body', 'request_id',
                                      'blocked', 'description', 'possible', 'lm',
                                      'mz', 'timestamp').order_by('-timestamp')[start_page:end_page]

    for ip in main_base.values('ip').annotate(dcount=Count('id')).iterator():
        count_ips[ip.get('ip')] = ip.get('dcount')

    ids = [i.id for i in main]
    blocked_request_id = []

    setter_geotokens_for_logs(main)

    getting_geoip_multithread(main, mcache)

    scan_requests = ScanRequest.objects.filter(log__id__in=ids).values('log', 'status', "description")
    scan_statuses = {'Logs_' + str(i['log']): [i['status'], i['description']] for i in scan_requests}
    req_ids = main_base.filter(bt__in=[3, 8]).values('request_id').annotate(dcount=Count('id')) \
        .values_list('request_id', flat=True)

    if len(main) > 0:
        st_entries = MtTable.objects.using('waf').filter(request_id__in=list(req_ids), vhost=main[0].vhost).values()
        st_entries = {i.get('request_id').strip(): i for i in st_entries}
    else:
        st_entries = dict()

    st_export_ids = MtExport.objects.filter(request_id__in=list(req_ids)).values('request_id', 'bt', 'mz',
                                                                                 'vhost').annotate(dcount=Count('id'))
    whitelist_requests = WhitelistRequest.objects.filter(is_deleted=False, log__id__in=list(ids))
    wl_req_dict = dict()
    for i in whitelist_requests:
        wl_req_dict[i.log.id] = {'admin_comment': i.admin_comment, 'user_comment': i.user_comment, 'user': i.user}

    del whitelist_requests, req_ids, scan_requests

    for i in main:
        if wl_req_dict.get(i.id):
            i.unlock_info = wl_req_dict.get(i.id)
        i.fields_equal = {}
        i.cls_log = i._meta.model.__name__
        i.bt_origin = i.bt

        if i._meta.model.__name__ == 'Logs':
            i.fid = 'Logs_' + str(i.id)
            i.agent = None
        else:
            i.fid = 'StTable_' + str(i.id)
            i.blocked = None

        if st_entries.get(i.request_id.strip()) is not None:
            if st_entries.get(i.request_id).get('vhost') == i.vhost \
                    and st_entries.get(i.request_id).get('bt') in (12, 13):
                j = st_entries.get(i.request_id)
                if not set(j.get('mz', '').lower().split('|')).isdisjoint(set(str(i.mz).lower().split('|'))):
                    i.bt = j.get('bt')
                    i.agent = j.get('agent')
                    i.fid = 'Logs_' + str(i.id)

        if i.bt in (12, 13):
            i.is_export = False
            for j in st_export_ids:
                if (i.request_id, i.bt, i.vhost) == (j.get('request_id'), j.get('bt'), j.get('vhost')):
                    if i.mz.lower().strip() in j.get('mz', '').lower().strip():
                        i.is_export = True
                        break

        try:
            status = scan_statuses.get(i.fid)
            if status is not None:
                status = status[0]

                if scan_statuses.get(i.fid)[1] not in ('', None):
                    description_recheck = scan_statuses.get(i.fid)[1][0].lower() + scan_statuses.get(i.fid)[1][1:]

                else:
                    description_recheck = None

            else:
                description_recheck = None
        except KeyError:
            status = None
            description_recheck = None
        i.status = status
        i.body = parse_body(i.body)

        i.mz_for_st = str(i.mz).lower().split('|')
        i.method = str(i.method).strip()
        i.description_recheck = description_recheck
        i.ips = [{
            'ip': i.ip,
            'count': 1,
        }]
        i.timestamp_max = i.timestamp
        i.timestamp_min = i.timestamp
        for ip in i.ips:
            ip['location'] = get_location_by_ip(mcache, ip['ip'])
            ip['count_ip'] = attack_short_count(count_ips.get(ip['ip'], 1))

        ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, i.vhost))

        ml_learning_progress_full_with_point = ml_table.get('{}-{}'.format(i.alias, '.' + i.vhost))

        ml_learning_progress_full_with_point_and_star = ml_table.get('{}-{}'.format(i.alias, '*' + '.'
                                                                                    + i.vhost[i.vhost.find(".") + 1:]))

        ml_learning_progress = get_ml_learning_progress(
            ml_table, i, ml_learning_progress_full,
            ml_learning_progress_full_with_point,
            ml_learning_progress_full_with_point_and_star
        )

        i.ml_learning_progress = ml_learning_progress
        i.reason = get_reason_by_model(i.bt)
        i.vhost = convert_from_punycode(i.vhost)
        if i.bt in (11, 12, 13):
            i.reason = get_reason_by_model(i.bt_origin)

        blocked_request_id.append(i.request_id)

    counter = Counter(blocked_request_id)
    counter_dict = dict(counter)
    repeat_request_ids = []
    no_lm = []

    for k, v in counter_dict.items():
        if v > 1:
            repeat_request_ids.append(k)

    for obj in main:
        if obj.request_id in repeat_request_ids:
            if obj.blocked or obj.blocked is None:
                obj.group_blocked = False
                no_lm.append(obj.request_id)
                obj.lm = False
            if not obj.blocked:
                obj.group_blocked = True
        else:
            obj.group_blocked = False

    for obj in main:
        if obj.request_id in no_lm:
            obj.group_blocked = False
            obj.lm = False
    return main


# SEE name of functions
def get_main_logs(main_base, ml_table, mcache, second_row=False):
    groups = {}
    ids = []
    blocked_request_id = []

    for log in main_base:
        ids.append(log.get('id'))

        ips_count = {log.get('ip'): 1}

        ips = []
        for ip in ips_count:
            ips.append({
                'ip': ip,
                'count': 1,
            })

        groups[log.get('id')] = {
            'ips': sorted(ips, key=lambda x: -x['count']),
            'timestamp_max': log.get('timestamp'),
            'timestamp_min': log.get('timestamp'),
        }

    if not second_row:
        main = Logs.objects.using('waf').defer('cc').only('vhost', 'alias', 'bt', 'ip', 'request_id', 'blocked',
                                                          'description',
                                                          'possible', 'lm'). \
            filter(id__in=ids).order_by('-timestamp')
    else:
        main = Logs.objects.using('waf').defer('cc').filter(id__in=ids).order_by('-timestamp')

    ips_list = []
    for i in main:
        ips_list.append(i.ip)
    processes = []
    for ip in list(chunks(list(set(ips_list)), 1))[:len(ips_list)]:
        token = None
        proc = multiprocessing.Process(target=get_location_from_remote, args=(token, ip, mcache))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    scan_requests = ScanRequest.objects.filter(log__id__in=ids).values('log', 'status', "description")
    scan_statuses = {'Logs_' + str(i['log']): [i['status'], i['description']] for i in scan_requests}

    for i in main:
        i.fields_equal = {}
        i.bt_origin = i.bt

        if i._meta.model.__name__ == Logs._meta.model.__name__:
            i.fid = 'Logs_' + str(i.id)
        else:
            i.fid = 'StTable_' + str(i.id)
            i.blocked = None
            i.timestamp = datetime.fromtimestamp(int(i.timestamp))

        try:

            status = scan_statuses.get(i.fid)

            if status is not None:
                status = status[0]
                if scan_statuses.get(i.fid)[1] not in ('', None):
                    description_recheck = scan_statuses.get(i.fid)[1][0].lower() + scan_statuses.get(i.fid)[1][1:]

                else:
                    description_recheck = None

            else:
                description_recheck = None
        except KeyError:
            status = None
            description_recheck = None
        i.status = status
        i.description_recheck = description_recheck
        i.ips = groups[i.id]['ips']
        i.timestamp_max = groups[i.id]['timestamp_max']
        i.timestamp_min = groups[i.id]['timestamp_min']
        for ip in i.ips:
            ip['location'] = get_location_by_ip(mcache, ip['ip'])

        try:
            ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, i.vhost))
            ml_learning_progress_full_with_point = ml_table.get('{}-{}'.format(i.alias, '.' + i.vhost))
            if ml_learning_progress_full is not None:
                ml_learning_progress = ml_learning_progress_full.progress
            elif ml_learning_progress_full_with_point is not None:
                ml_learning_progress = ml_learning_progress_full_with_point.progress
            else:
                ######################
                vhost = i.vhost
                # while tmp:
                if vhost.count('.') == 1:
                    position = vhost.find('.')
                    vhost = vhost[position:]
                    ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, vhost))
                    if ml_learning_progress_full is not None:
                        ml_learning_progress = ml_learning_progress_full.progress
                    else:
                        ml_learning_progress = None
                else:
                    if vhost.count('.') > 2:
                        split_vhost = vhost.split('.')
                        vhost = '.' + split_vhost[-2] + '.' + split_vhost[-1]
                    position = vhost.find('.')
                    vhost = vhost[position + 1:]
                    ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, 'sub_' + vhost))
                    if ml_learning_progress_full is not None:
                        ml_learning_progress = ml_learning_progress_full.progress
                    else:
                        ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, '.' + vhost))
                        if ml_learning_progress_full is not None:
                            ml_learning_progress = ml_learning_progress_full.progress
                        else:
                            ml_learning_progress = None
                ######################
            if ml_learning_progress is None:
                ml_learning_progress_full = ml_table.get('{}-{}'.format(i.alias, 'all'))
                if ml_learning_progress_full is not None:
                    ml_learning_progress = ml_learning_progress_full.progress
                else:
                    ml_learning_progress = '0'

        except Exception:
            ml_learning_progress = '0'
        i.ml_learning_progress = ml_learning_progress
        i.reason = get_reason_by_model(i.bt)
        if i.bt in (11, 12, 13):
            i.reason = get_reason_by_model(i.bt_origin)

        if second_row:
            i.mz = i.mz.replace("$HEADERS_VAR:", '')
            i.body = parse_body(i.body)
            headers = i.other_headers.lower().replace(' ', '')

            if len(i.body) < 2000:
                i.body_is_large = False
                if 'application/json' in headers and 'content-type:' in headers:
                    if ':' not in headers[
                                  headers.find('content-type:') + len('content-type:'):headers.find('application/json')]:
                        if isinstance(i.body, str) and i.body != "":
                            try:
                                tmp = json.loads(i.body)
                                i.body = json.dumps(tmp, indent=2, sort_keys=True)

                            except json.JSONDecodeError:
                                i.body = i.body
                            i.body = i.body.strip()
            else:
                i.body_is_large = True

        blocked_request_id.append(i.request_id)

    counter = Counter(blocked_request_id)
    counter_dict = dict(counter)
    repeat_request_ids = []
    no_lm = []

    for k, v in counter_dict.items():
        if v > 1:
            repeat_request_ids.append(k)

    for obj in main:
        if obj.request_id in repeat_request_ids:
            if obj.blocked or obj.blocked is None:
                obj.group_blocked = False
                no_lm.append(obj.request_id)
                obj.lm = False
            if not obj.blocked:
                obj.group_blocked = True
        else:
            obj.group_blocked = False

    for obj in main:
        if obj.request_id in no_lm:
            obj.group_blocked = False
            obj.lm = False
    return main


def get_reason_by_model(bt):
    if bt == 3:
        return "Nemesida AI MLA"
    elif bt == 4:
        return "Antivirus"
    elif bt == 7:
        return "Brute-force Analysis"
    elif bt == 8:
        return "Nemesida AI MLC"
    elif bt == 9:
        return "Flood"
    elif bt == 14:
        return "API Firewall"
    elif bt == 16:
        return "Blocked IP"

    return "Signature Analysis"


def write_memcache(mcache, value: str, json_as_str: str) -> None:
    """Write message in memcache"""
    if mcache:
        mcache.set(value + vGeoIP, json_as_str, 86400)
    else:
        cache.set(value + vGeoIP, json_as_str, 86400)


def search_local_ip(i):
    from .utils import list_private_networks
    for value in list_private_networks:
        net = ipaddress.ip_network(value)
        bool_var = ipaddress.ip_address(f"{i}") in net
        if bool_var:
            return True


def setter_geotokens_for_logs(logs):
    geo_tokens = {}
    for token in Geo_Tokens.objects.using('waf').all():
        geo_tokens[token.waf_id] = token.geo_token

    for i in logs:
        if hasattr(i, "alias"):
            alias = i.alias
            i.geotoken = geo_tokens.get(alias)
        else:
            alias = i.get('alias')
            i["geotoken"] = geo_tokens.get(alias)


def adding_token_to_dict_for_geoip(
    logs,
    res_dict: dict,
    lock: Optional[Union[TLock, PLock]] = None
):
    for i in logs:
        if hasattr(i, "geotoken"):
            geotoken = i.geotoken
            ip = i.ip
        else:
            geotoken = i.get('geotoken')
            ip = i.get('ip')

        with lock if lock else dummy_lock():
            if res_dict.get(geotoken):
                res_dict[geotoken].add(ip)
            else:
                res_dict[geotoken] = {ip}


def getting_geoip_multithread(
    logs,
    mcache,
    max_workers=5,
    f_chunking_add_token=chunks,
    f_chunking_geobatch=chunks
):
    """ Getting info by ip. Request to geoip batch and save to mcache result"""
    dict_token_and_ips = {}
    thr_lock = TLock()

    run_multithreading_process(
        adding_token_to_dict_for_geoip, logs, dict_token_and_ips,
        len_chunks=1000, max_workers=max_workers,
        f_chunking=f_chunking_add_token, lock=thr_lock
    )

    for token, ips in dict_token_and_ips.items():
        run_multithreading_process(
            get_location_gtoipv3_batch, list(ips), token, mcache, thr_lock,
            max_workers=max_workers, f_chunking=f_chunking_geobatch
        )


def get_location_gtoipv3_batch(
    arr: list,
    token,
    mcache,
    lock: Optional[Union[TLock, PLock]] = None
):
    """Get ip_info from Geoip when if many ip"""
    if not token:
        return
    try:
        url = settings.URI_GEOIP_BATCH
        r = requests.post(
            url + f"{token}",
            headers=user_agent,
            proxies=proxy_dict_conf,
            timeout=settings.URI_GEOIP_TIMEOUT,
            data=json.dumps(arr),
        )

        data = json.loads(r.content)
        for k, v in data.items():
            if v == 'Private IP':
                v = json.dumps({'status': 'fail', 'description': 'Private IP'})

            with lock if lock else dummy_lock():
                write_memcache(mcache, value=k, json_as_str=json.dumps(v))

    except Exception as e:
        logger.error(f"An error occurred when get GeoIP batch error: {e}")
        logger.debug(f"Request parameters: ips={arr} token={token}")


def get_location_from_remote(
    token: Optional[str],
    ip='',
    mcache=None,
    lock=None
):
    url = settings.URI_GEOIP
    ip_str = ''
    current_datetime = datetime.now().strftime("%d/%b/%Y:%H:%M:%S")
    log_file = open('/var/log/uwsgi/cabinet/ipinfo2memcache.log', 'a')
    for i in ip:
        if lock is not None:
            lock.acquire()
        c = None
        try:
            c = mcache.get(i + vGeoIP)
            if c is not None:
                ct = json.loads(c)
                if ct.get('status') != 'success' and (
                        ct.get('status') != 'fail' and ct.get('description') != 'Private IP'):
                    c = None
        except KeyError:
            break
        finally:
            if lock is not None:
                lock.release()
        if c is None:
            ip_str += i + '|'
    if ip_str:

        try:
            if token:
                r = requests.get(url + ip_str.rstrip().replace('\n', '')[:-1] + f'&geo_token={token}',
                                 headers=user_agent,
                                 proxies=proxy_dict_conf,
                                 timeout=settings.URI_GEOIP_TIMEOUT)

                if "Enter valid IP" not in r.text:
                    try:
                        for row in r.text.split('<br>'):
                            b = json.loads(row)
                            if lock is not None:
                                lock.acquire()
                            try:
                                if mcache:
                                    mcache.set(ip_str[:-1] + vGeoIP, json.dumps(b), 86400)
                                else:
                                    cache.set(ip_str[:-1] + vGeoIP, json.dumps(b), 86400)
                            finally:
                                if lock is not None:
                                    lock.release()
                    except Exception as e:
                        logger.error("{date} An error occurred when get GeoIP data about IP {ip} ".format(
                            ip=ip_str[:-1],
                            date=current_datetime + f'{e}',
                            file=log_file))

        except Exception as er:
            logger.error("{date} An error occurred when get GeoIP data about IP {ip} ".format(
                ip=ip_str[:-1],
                date=current_datetime + f'{er}',
                file=log_file
            ))


def get_location_by_ip(mcache, ip, check_token=None, token=None):
    default_location = {
        'city': 'None', 'country': 'None',
        "cc": "None", 'isp': 'None',
        'org': 'None', "tor": "None",
        "mobile": "None", "proxy": "None",
        "hosting": "None", "crawler": "None"
    }
    location = default_location
    try:
        # check existing geo_token
        if not token and check_token:
            c = None
        else:
            c = mcache.get(ip + vGeoIP)

        if c:
            c = c.decode('utf-8').replace("'", '"')
            ip_info = json.loads(c)

            location.update(ip_info)

    except Exception as e:
        logger.error(
            'An error occurred while retrieving information '
            'about IP {} from memory. Error description: {}'.format(ip, e)
        )
        logger.debug(
            'Parameters for receiving IP data: ', ip, token, check_token
        )

    return location


def parse_body(content):
    try:
        content = bytes(content).decode('utf-8', 'ignore')
    except Exception:
        content = ""

    return content


def are_rows_equal(rows):
    if len(rows) < 2:
        return True

    first_value = rows[0]
    for value in rows[1:]:
        if first_value != value:
            return False

    return True


def attack_short_count(count, n=1):
    if n == 0:
        if count > 1000000:
            return str(int(count / 1000000)) + "M"
        elif count > 1000:
            return str(int(count / 1000)) + "K"
        else:
            return str(count)
    else:
        if count > 1000000:
            return str(round(count / 1000000, n)) + "M"
        elif count > 1000:
            return str(round(count / 1000, n)) + "K"
        else:
            return str(count)


def set_last_filter_in_session(request):
    request.session['current_search'] = request.GET.get('search', '').strip()
    if 'last_search' not in request.session.keys():
        last_search = ''
        return last_search
    else:
        last_search = request.session['last_search']
        return last_search


def update_last_filter_in_session(request):
    request.session['last_search'] = request.session['current_search']


def draw_attackers_top10_tooltip(i, top10, last24h_popular_target_dict, last24h_count_popular_target_dict,
                                 mostly_blocked_count_dict, check_bt, attackers_top10_row, mcache):
    popular_target = top10.values('vhost'). \
                         annotate(host_count=Count('vhost')).order_by('-host_count').filter(ip=i.get('ip'))[:1]

    target = popular_target[0]
    last24h_popular_target_dict[i.get('ip')] = str(target.get('vhost'))
    last24h_count_popular_target_dict[i.get('ip')] = str(target.get('host_count'))

    bt_desc = top10.values('bt').annotate(
        bt_count=Count('bt')).order_by('-bt_count').filter(ip=i.get('ip')).filter(
        bt__in=[1, 2, 3, 4, 7, 8, 9, 10])[:1]

    desc = bt_desc[0]
    mostly_blocked_count_dict[i.get('ip')] = str(desc.get('bt_count'))
    check_bt[i.get('ip')] = str(desc.get('bt'))

    if str(get_location_by_ip(mcache, i.get('ip'))['mobile']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['mobile']).lower() == 'false':
        mobile = 'No'
    else:
        mobile = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['proxy']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['proxy']).lower() == 'false':
        proxy = 'No'
    else:
        proxy = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['hosting']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['hosting']).lower() == 'false':
        hosting = 'No'
    else:
        hosting = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['crawler']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['crawler']).lower() == 'false':
        crawler = 'No'
    else:
        crawler = 'Yes'

    popular_target_percent = (int(last24h_count_popular_target_dict.get(i.get('ip'))) /
                              int(i.get('all_attacks'))) * 100
    if int(popular_target_percent) != 100:
        popular_target_percent = round(popular_target_percent, 1)
    else:
        popular_target_percent = int(popular_target_percent)

    if mostly_blocked_count_dict.get(i.get('ip')) is None or mostly_blocked_count_dict.get(i.get('ip')) == '-':
        mostly_blocked_count = 0
    else:
        mostly_blocked_count = mostly_blocked_count_dict.get(i.get('ip'))
    mostly_blocked_percent = int(((int(mostly_blocked_count) / int(i.get('all_attacks'))) * 100))

    if mostly_blocked_percent != 100:
        mostly_blocked_percent = round(((int(mostly_blocked_count) / int(i.get('all_attacks'))) * 100), 1)

    bt_desc = ''
    if check_bt.get(i.get('ip')) == '1' or check_bt.get(i.get('ip')) == '2':
        bt_desc = 'Signature Analysis'
    if check_bt.get(i.get('ip')) == '3' or check_bt.get(i.get('ip')) == '8':
        bt_desc = 'Nemesida AI MLA'
    if check_bt.get(i.get('ip')) == '7':
        bt_desc = 'Brute-force'
    if check_bt.get(i.get('ip')) == '4':
        bt_desc = 'Antivirus'
    if check_bt.get(i.get('ip')) == '10':
        bt_desc = 'DDoS'
    if check_bt.get(i.get('ip')) == '9':
        bt_desc = 'Flood'

    rldesc = ''
    rule_id_dict = dict()
    rule_id_count_dict = dict()
    if check_bt.get(i.get('ip')) == '1' or check_bt.get(i.get('ip')) == '2':
        rldesc_base = top10.values('description', 'rule_id'). \
                          annotate(description_count=Count('description')).order_by('-description_count').filter(
            ip=i.get('ip')).filter(
            bt__in=[1, 2])[:1]
        desc = rldesc_base[0]
        rldesc = " (" + str(desc.get('description')) + ", "
        rule_id_dict[i.get('ip')] = desc.get('rule_id')

        rule_id_count_base = top10.values('rule_id').annotate(
            rule_id_count=Count('rule_id')).order_by('-rule_id_count').filter(
            ip=i.get('ip')).filter(rule_id=rule_id_dict.get(i.get('ip')))[:1]

        rule_id_count_dict[i.get('ip')] = rule_id_count_base[0].get('rule_id_count')
        if rule_id_count_dict.get(i.get('ip')) is None:
            rule_id_count = 0
        else:
            rule_id_count = int(rule_id_count_dict.get(i.get('ip')))
        rule_id_percent = (rule_id_count / int(i.get('all_attacks')) * 100)
        if rule_id_percent != '100.0':
            rule_id_percent = round(int(rule_id_percent), 1)
        else:
            rule_id_percent = int(rule_id_percent)
        rldesc += str(rule_id_percent) + "%)"

    org = ''
    if len(str(get_location_by_ip(mcache, i.get('ip'))['org'])) > 1:
        org = 'Org: ' + str(get_location_by_ip(mcache, i.get('ip'))['org']) + '\n'
    isp = ''
    if len(str(get_location_by_ip(mcache, i.get('ip'))['isp'])) > 1:
        isp = 'ISP: ' + str(get_location_by_ip(mcache, i.get('ip'))['isp'])

    attackers_top10_row.put([
        i.get('ip'),
        i.get('all_attacks'),
        i.get('sa'),
        i.get('na'),
        i.get('bf'),
        i.get('av'),
        i.get('ddos'),
        i.get('flood'),
        "Main target: " + str(last24h_popular_target_dict.get(i.get('ip'))) + " (" +
        attack_short_count(int(last24h_count_popular_target_dict.get(i.get('ip'))), 0) + ", " +
        str(popular_target_percent) + "%)" + '\n' +
        "Mostly blocked by: " + str(bt_desc) + ', '
        + str(mostly_blocked_percent) + '%' + rldesc + '\n' +
        "Geo: " + str(get_location_by_ip(mcache, i.get('ip'))['city']) + ", " +
        str(get_location_by_ip(mcache, i.get('ip'))['country']) +
        '(' + str(get_location_by_ip(mcache, i.get('ip')).get('cc')) + ')' + '\n' +
        'Mobile: ' + mobile + '\n' +
        'Proxy: ' + proxy + '\n' +
        'Hosting: ' + hosting + '\n' +
        'Crawler: ' + crawler + '\n' +
        org + isp,
        str(get_location_by_ip(mcache, i.get('ip')).get('cc')).lower()
    ])


def draw_top_by_ip_tooltip(i, charts, top_by_ip_popular_target_dict, top_by_ip_count_popular_target_dict,
                           mostly_blocked_count_dict, check_bt, last24h_ip_count_dict, ip_info_queue, mcache):
    top_by_ip_popular_target = charts.values('vhost'). \
                                   annotate(host_count=Count('vhost')).order_by('-host_count').filter(ip=i.get('ip'))[
                               :1]

    for target in top_by_ip_popular_target:
        top_by_ip_popular_target_dict[i.get('ip')] = str(target.get('vhost'))
        top_by_ip_count_popular_target_dict[i.get('ip')] = str(target.get('host_count'))

    bt_desc = charts.values('bt').annotate(
        bt_count=Count('bt')).order_by('-bt_count').filter(ip=i.get('ip')).filter(
        bt__in=[1, 2, 3, 4, 7, 8, 9, 10])[:1]
    for desc in bt_desc:
        mostly_blocked_count_dict[i.get('ip')] = str(desc.get('bt_count'))
        check_bt[i.get('ip')] = str(desc.get('bt'))

    if str(get_location_by_ip(mcache, i.get('ip'))['mobile']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['mobile']).lower() == 'false':
        mobile = 'No'
    else:
        mobile = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['proxy']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['proxy']).lower() == 'false':
        proxy = 'No'
    else:
        proxy = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['hosting']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['hosting']).lower() == 'false':
        hosting = 'No'
    else:
        hosting = 'Yes'

    if str(get_location_by_ip(mcache, i.get('ip'))['crawler']).lower() == 'none' or \
            str(get_location_by_ip(mcache, i.get('ip'))['crawler']).lower() == 'false':
        crawler = 'No'
    else:
        crawler = 'Yes'

    popular_target_percent = (int(top_by_ip_count_popular_target_dict.get(i.get('ip'))) /
                              int(i.get('all_attacks'))) * 100
    if int(popular_target_percent) != 100:
        popular_target_percent = round(popular_target_percent, 1)
    else:
        popular_target_percent = int(popular_target_percent)

    if last24h_ip_count_dict.get(i.get('ip')) is None:
        last24h_count = 0
    else:
        last24h_count = attack_short_count(int(last24h_ip_count_dict.get(i.get('ip'))), 0)

    if mostly_blocked_count_dict.get(i.get('ip')) is None or mostly_blocked_count_dict.get(i.get('ip')) == '-':
        mostly_blocked_count = 0
    else:
        mostly_blocked_count = mostly_blocked_count_dict.get(i.get('ip'))
    mostly_blocked_percent = int(((int(mostly_blocked_count) / int(i.get('all_attacks'))) * 100))

    if mostly_blocked_percent != 100:
        mostly_blocked_percent = round(((int(mostly_blocked_count) / int(i.get('all_attacks'))) * 100), 1)

    bt_desc = ''
    if check_bt.get(i.get('ip')) == '1' or check_bt.get(i.get('ip')) == '2':
        bt_desc = 'Signature Analysis'
    if check_bt.get(i.get('ip')) == '3' or check_bt.get(i.get('ip')) == '8':
        bt_desc = 'Nemesida AI MLA'
    if check_bt.get(i.get('ip')) == '7':
        bt_desc = 'Brute-force'
    if check_bt.get(i.get('ip')) == '4':
        bt_desc = 'Antivirus'
    if check_bt.get(i.get('ip')) == '10':
        bt_desc = 'DDoS'
    if check_bt.get(i.get('ip')) == '9':
        bt_desc = 'Flood'

    rldesc = ''
    rule_id_dict = dict()
    rule_id_count_dict = dict()
    if check_bt.get(i.get('ip')) == '1' or check_bt.get(i.get('ip')) == '2':
        rldesc_base = charts.values('description', 'rule_id'). \
                          annotate(description_count=Count('description')).order_by('-description_count').filter(
            ip=i.get('ip')).filter(
            bt__in=[1, 2])[:1]
        for desc in rldesc_base:
            rldesc = " (" + str(desc.get('description')) + ", "
            rule_id_dict[i.get('ip')] = desc.get('rule_id')

        rule_id_count_base = charts.values('rule_id').annotate(
            rule_id_count=Count('rule_id')).order_by('-rule_id_count').filter(
            ip=i.get('ip')).filter(rule_id=rule_id_dict.get(i.get('ip')))[:1]
        for t in rule_id_count_base:
            rule_id_count_dict[i.get('ip')] = t.get('rule_id_count')
        if rule_id_count_dict.get(i.get('ip')) is None:
            rule_id_count = 0
        else:
            rule_id_count = int(rule_id_count_dict.get(i.get('ip')))
        rule_id_percent = (rule_id_count / int(i.get('all_attacks')) * 100)
        if rule_id_percent != '100.0':
            rule_id_percent = round(int(rule_id_percent), 1)
        else:
            rule_id_percent = int(rule_id_percent)
        rldesc += str(rule_id_percent) + "%)"

    org = ''
    if len(str(get_location_by_ip(mcache, i.get('ip'))['org'])) > 1:
        org = 'Org: ' + str(get_location_by_ip(mcache, i.get('ip'))['org']) + '\n'
    isp = ''
    if len(str(get_location_by_ip(mcache, i.get('ip'))['isp'])) > 1:
        isp = 'ISP: ' + str(get_location_by_ip(mcache, i.get('ip'))['isp'])

    ip_info_queue.put([
        i.get('ip'),
        i.get('all_attacks'),
        i.get('bt1'),
        i.get('bt2'),
        i.get('bt3'),
        i.get('bt4'),
        i.get('bt7'),
        i.get('bt8'),
        i.get('bt9'),
        i.get('bt10'),
        "Main target: " + str(top_by_ip_popular_target_dict.get(i.get('ip'))) + " (" +
        attack_short_count(int(top_by_ip_count_popular_target_dict.get(i.get('ip'))), 0) + ", " +
        str(popular_target_percent) + "%)" + '\n' +
        "Mostly blocked by: " + str(bt_desc) + ', '
        + str(mostly_blocked_percent) + '%' + rldesc + '\n' +
        "Geo: " + str(get_location_by_ip(mcache, i.get('ip'))['city']) + ", " +
        str(get_location_by_ip(mcache, i.get('ip'))['country']) +
        '(' + str(get_location_by_ip(mcache, i.get('ip'))['cc']) + ')' + '\n' +
        'Mobile: ' + mobile + '\n' +
        'Proxy: ' + proxy + '\n' +
        'Hosting: ' + hosting + '\n' +
        'Crawler: ' + crawler + '\n' +
        org + isp,
        '(' + str(last24h_count) + ' in 24 h.)',
        str(get_location_by_ip(mcache, i.get('ip'))['cc']).lower()
    ])


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def get_user_key_waf_id(user, without_lite=False):
    processes = []
    processes_queue = multiprocessing.Queue()
    keys = LicKeys.objects.filter(user__id=user.id).all()

    if settings.API_URI is None:
        return []

    for key in keys:
        proc = multiprocessing.Process(target=get_lic_keys, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    accept_waf_ids = list()

    for key in keys_form_multiprocessing:
        if key.tarif == 'Enterprise' or key.tarif == 'Business' or (key.tarif == 'Light' and not without_lite):
            accept_waf_ids.append(key.waf_id)

    return accept_waf_ids


def get_user_key_waf_id_enter(user):
    processes = []
    processes_queue = multiprocessing.Queue()
    keys = LicKeys.objects.filter(user__id=user.id).all()

    if settings.API_URI is None:
        return []

    for key in keys:
        proc = multiprocessing.Process(target=get_lic_keys, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    accept_waf_ids = list()

    for key in keys_form_multiprocessing:
        if key.tarif == 'Enterprise':
            accept_waf_ids.append(key.waf_id)
            accept_waf_ids.append(key)

    return accept_waf_ids


def get_list_plans(user):
    processes = []
    processes_queue = multiprocessing.Queue()
    keys = LicKeys.objects.filter(user__id=user.id).all()
    if settings.API_URI is None:
        return []

    for key in keys:
        proc = multiprocessing.Process(target=get_lic_keys, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    list_plans = set()
    for key in keys_form_multiprocessing:
        list_plans.add(key.tarif)
    return list(list_plans)
