#!/usr/bin/python3

import asyncio
import json
import json as simplejson
import logging
import os
import psycopg2
import psycopg2.extras
import re
import requests
import sys

from django.conf import settings
from pymemcache.client.base import Client
from threading import Thread
from time import sleep
from urllib.parse import unquote

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)

from main.utils import close_db_connection
from main.utils import convert_from_punycode
from main.utils import get_proxy_conf_dict
from main.utils import init_db_connection
from utils.mail import is_smtp_enabled
from utils.mail import send_mail_from_cabinet

# logging
log_place = '/var/log/uwsgi/cabinet/attack_notification.log'
logging.basicConfig(
    filename=log_place,
    level=logging.INFO,
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%d/%b/%Y:%H:%M:%S %z'
)

if os.environ.get('DJANGO_SETTINGS_MODULE') is None:
    os.environ['DJANGO_SETTINGS_MODULE'] = "cabinet.settings_extra"

# Postgres connect
DB_NAME = settings.DB_NAME_CONF
DB_USER = settings.DB_USER_CONF
DB_PASS = settings.DB_PASS_CONF
DB_HOST = settings.DB_HOST_CONF
try:
    DB_PORT = settings.DB_PORT_CONF
    if DB_PORT == '':
        DB_PORT = '5432'
except NameError:
    DB_PORT = '5432'

try:
    DB_PORT_CABINET = settings.DB_PORT_CABINET
    if DB_PORT_CABINET == '':
        DB_PORT_CABINET = '5432'
except NameError:
    DB_PORT_CABINET = '5432'

user_agent = {'User-agent': settings.USER_AGENT}

vGeoIP = settings.V_GEOIP
cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))

mcache = Client(cache_address)

user_config_map = {
    # todo link to settings
    1: "bt1",
    2: "bt2",
    3: "bt3",
    4: "bt4",
    5: "bt5",
    7: "bt7",
    8: "bt8",
    9: "bt9",
    10: "bt10",
    14: "bt14",
    16: "bt16"
}

fields = {'id',
          'timestamp',
          'ip',
          'vhost',
          'url',
          'lm',
          'rule_id',
          'mz',
          'param',
          'body',
          'method',
          'ua',
          'referer',
          'description',
          'bt',
          'cookie',
          'request_id',
          'waf_id'}

block_types_map = {
    # todo link to settings
    3: 'Nemesida AI MLA',
    4: 'Antivirus',
    5: 'Internal Error',
    7: 'Brute-force Analysis',
    8: 'Nemesida AI MLC',
    9: 'Flood Analysis',
    10: 'DDoS Analysis',
    14: 'API Firewall',
    16: 'Blocked IP'
}


class Notifier:
    def __init__(self):
        while True:
            try:

                # receive last processed ID
                connect = init_db_connection(
                    settings.DB_NAME_CABINET,
                    settings.DB_USER_CABINET,
                    settings.DB_PASS_CABINET,
                    settings.DB_HOST_CABINET,
                    DB_PORT_CABINET
                )
                cursor = connect.cursor()
                cursor.execute("SELECT last_id FROM attack_notify")
                res = cursor.fetchone()
                close_db_connection(cursor, connect)

                # update last processed ID
                self.last_id = res[0] if res else 0

                # finish processing
                break

            except Exception as e:
                logging.error('An error occurred while receiving data from attack_notify: {}'.format(e))
                sleep(15)

    @staticmethod
    def create_body(row, ip_data):
        # subject
        subject_data = [
            'Attack: {}'.format(row.get('vhost')),
            row.get('ip') if not ip_data else '{} ({})'.format(row.get('ip'), ip_data.get('cc'))
        ]

        if row.get('lm'):
            subject_data.append('M')
        if row.get('possible'):
            subject_data.append('Possible')

        subject = ' | '.join(subject_data)

        if row.get('bt') in (7, 8, 9, 10):
            s_word = "Detected"
        elif row.get('blocked') is True:
            s_word = "Blocked"
        else:
            s_word = "Detected"

        # block type
        block_type = block_types_map.get(row.get('bt'), 'Signature Analysis')

        # ip info
        if ip_data:
            ip_info = "{} ({}, {})".format(row.get('ip'), ip_data.get('city', ''), ip_data.get('cc', ''))
        else:
            ip_info = "{}".format(row.get('ip'))

        if row.get('bt') not in range(3, 16):
            main_body = '{rule_tag:12} {rule}\n'.format(
                rule_tag='Rule:',
                rule=str(row.get('rule_id'))
            )
        else:
            main_body = ''

        main_body += '{host_tag:12} {host}\n{ip_info_tag:12} {ip_info}\n{date_tag:12} {date}'.format(
            host_tag='Host:',
            host=row.get('vhost'),
            ip_info_tag='IP:',
            ip_info=ip_info,
            date_tag='Date:',
            date=str(row.get('timestamp').strftime('%H:%M:%S %d.%m.%Y'))
        )

        if row.get('waf_id') not in (None, 'null', 'NULL', '-', ''):
            main_body += '\n{waf_id_tag:12} {waf_id}'.format(
                waf_id_tag='WAF ID:',
                waf_id=row.get('waf_id', '').strip('\n ')
            )
        if row.get('method') not in (None, 'null', 'NULL', '-', ''):
            main_body += '\n{method_tag:12} {method}'.format(
                method_tag='Method:',
                method=row.get('method', '').strip('\n ')
            )
        if row.get('request_id') not in (None, 'null', 'NULL', '-', ''):
            main_body += '\n{request_id_tag:12} {request_id}'.format(
                request_id_tag='Request ID:',
                request_id=row.get('request_id', '').strip('\n ')
            )

        try:
            if bytes(row.get('body')).decode('utf-8', 'ignore')[:50] != 'NULL':
                b = bytes(row.get('body')).decode('utf-8', 'ignore').strip('\n ')
            else:
                b = '-'
        except Exception:
            b = '-'

        zone = row.get('mz', 'None').replace('$HEADERS_VAR:', '').replace('URL', 'Path')

        body = '{s_word} by {block_type}'.format(
            s_word=s_word,
            block_type=block_type
        )

        zone = '' if zone is None else zone
        if zone not in ('None', 'null', '-', '') and row.get('bt') not in (4, 5):
            body += ' {block_in}'.format(block_in="in " + zone)

        if row.get('bt') == 5:
            body += ' ({description})'.format(description=row.get('description'))

        body += '\n\n'

        zone = zone.lower()

        body += '{main_body}\n\n'.format(main_body=main_body)

        if row.get('url'):
            body += '{url_tag:12} {url}\n'.format(url_tag='Path:', url=row.get('url').strip('\n '))

        if row.get('param') not in (None, 'null', 'NULL', '-', ''):
            if block_type not in ('Brute-force', 'Flood') or zone == 'args':
                body += '{args_tag:12} {args}\n'.format(
                    args_tag='ARGS:', args=row.get('param').strip('\n ')
                )

        if row.get('ua') not in (None, 'null', 'NULL', '-', ''):
            body += '{ua_tag:12} {ua}\n'.format(
                ua_tag='UA:',
                ua=row.get('ua').strip('\n ')
            )

        if row.get('cookie') not in (None, 'null', 'NULL', '-', ''):
            body += '{cookie_tag:12} {cookie}\n'.format(
                cookie_tag='Cookie:',
                cookie=row.get('cookie').replace("\n", "; ").replace("<br>", "; ").strip('\n ')
            )

        if zone == 'referer' and row.get(zone):
            body += "{referer_tag:12} {referer}\n".format(
                referer_tag="Referer:",
                referer=row.get(zone).strip('\n ')
            )

        other_headers = row.get('other_headers', '').splitlines()
        if len(other_headers) > 0:
            if not body.endswith('\n\n'):
                body += '\n'
            headers_format = ': '.join(other_headers[0].split(':', 1)).strip('\n ')
            for i in other_headers[1:]:
                headers_format += '\n'
                headers_format += ' ' * 13 + ': '.join(i.split(':', 1)).strip()

            body += '{headers_tag:12} {headers}\n'.format(
                headers_tag='Headers:',
                headers=headers_format
            )

        if b:
            if not body.endswith('\n\n'):
                body += '\n'
            if block_type not in ('Antivirus', 'Brute-force', 'Flood'):
                body += '{body_tag:12} {body}\n'.format(body_tag='Body:', body=b.strip('\n '))
            elif zone == 'body' and block_type != 'Antivirus':
                body += '{body_tag:12} {body}\n'.format(body_tag='Body:', body=b.strip('\n '))
            elif block_type == 'Antivirus':
                body += '{virus_tag:12} {virus}\n'.format(virus_tag='Virus:', virus=b.strip('\n '))

        # url decode body
        body = unquote(unquote(body))

        return subject, body

    @staticmethod
    def create_limited_body(row, ip_data):
        # subject
        subject_data = [
            'Attack: {}'.format(row.get('vhost')),
            row.get('ip') if not ip_data else '{} ({})'.format(row.get('ip'), ip_data.get('cc'))
        ]

        if row.get('lm'):
            subject_data.append('M')
        if row.get('possible'):
            subject_data.append('Possible')

        subject = ' | '.join(subject_data)

        if row.get('bt') in (7, 8, 9, 10):
            s_word = "Detected"
        elif row.get('blocked') is True:
            s_word = "Blocked"
        else:
            s_word = "Detected"

        # block type
        block_type = block_types_map.get(row.get('bt'), 'Signature Analysis')

        # ip info
        if ip_data:
            ip_info = "{} ({}, {})".format(row.get('ip'), ip_data.get('city', ''), ip_data.get('cc', ''))
        else:
            ip_info = "{}".format(row.get('ip'))

        if row.get('bt') not in range(3, 16):
            main_body = '{rule_tag:12} {rule}\n'.format(
                rule_tag='Rule:',
                rule=str(row.get('rule_id'))
            )
        else:
            main_body = ''

        main_body += '{host_tag:12} {host}\n{ip_info_tag:12} {ip_info}\n{date_tag:12} {date}'.format(
            host_tag='Host:',
            host=row.get('vhost'),
            ip_info_tag='IP:',
            ip_info=ip_info,
            date_tag='Date:',
            date=str(row.get('timestamp').strftime('%H:%M:%S %d.%m.%Y'))
        )

        if row.get('request_id') not in (None, 'null', 'NULL', '-', ''):
            main_body += '\n{request_id_tag:12} {request_id}'.format(
                request_id_tag='Request ID:',
                request_id=row.get('request_id', None)
            )

        zone = row.get('mz', 'None').replace('$HEADERS_VAR:', '').replace('URL', 'Path')

        body = '{s_word} by {block_type} '.format(
            s_word=s_word,
            block_type=block_type
        )

        zone = '' if zone is None else zone
        if zone not in ('None', 'null', '-', '') and row.get('bt') not in (4, 5):
            body += '{block_in}\n\n'.format(block_in="in " + zone)
        else:
            body += '\n\n'

        body += '{main_body}\n\n'.format(main_body=main_body)

        # url decode body
        body = unquote(unquote(body))

        return subject, body

    @staticmethod
    def send_mail(subject, body, attack_id, email):
        if not is_smtp_enabled:
            return
        try:
            send_mail_from_cabinet(
                send_to=email,
                subject=subject,
                text=body
            )
            logging.info('An attack notification with an id {} has been sent'.format(str(attack_id)))
        except Exception as e:
            logging.error('An error occurred sending mail for attack with id {}: {}'.format(attack_id, str(e)))

    @staticmethod
    def get_memcache_ip(ip):
        if type(ip) is bytes:
            ip = ip.decode()
        try:
            ip = ip.rstrip()
            ip_discr_in_mem = mcache.get(ip + vGeoIP)
            if ip_discr_in_mem:
                c = ip_discr_in_mem.decode('utf-8').replace('\'', '"')
                d = simplejson.loads(c)
                if d['status'] == 'fail':
                    d = {}
            else:
                d = {}
        except Exception as e:
            logging.error('An error occurred fetching data from memcached: {} {}'.format(str(e), str(type(e).__name__)))
            d = {}
        return d

    @staticmethod
    def get_ip_data(ip, ii):

        def local(inp_ip):
            reg = re.compile(
                r'(^127\.)|(^10\.)|(^172\.1[6-9]\.)|(^172\.2\d\.)|(^172\.3[0-1]\.)|(^192\.168\.)|(^100\.127\.49\.)|'
                r'(^100\.(6[4-9]|[7-9]\d|1[01]\d|12[0-7]))'
            )
            if reg.match(inp_ip):
                return True
            else:
                return False

        def multicast(inp_ip):
            reg = re.compile(
                r'(^22[4-9]\.)|(^23\d\.)')
            if reg.match(inp_ip):
                return True

            if inp_ip == '::1':
                return True

            return False

        url = settings.URI_GEOIP
        try:

            connect = init_db_connection(DB_NAME, DB_USER, DB_PASS, DB_HOST, DB_PORT)
            cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cursor.execute("SELECT geo_token FROM geo_tokens WHERE waf_id = %s", (ii.get('waf_id'),))
            tokens_geo = cursor.fetchall()
            close_db_connection(cursor, connect)

            if not tokens_geo:
                tokens_geo = [['']]
            ip = ip.rstrip()
            if not local(ip) and not multicast(ip):
                r = requests.get(url + ip + f'&geo_token={tokens_geo[0][0]}',
                                 headers=user_agent,
                                 proxies=get_proxy_conf_dict(),
                                 timeout=30
                                 )
                req_json = r.text.replace('<br>', '')
                if r.ok:
                    d = json.loads(req_json)
                else:
                    d = {}

            else:
                d = {}

        except Exception as err:
            d = {}
            logging.error(
                'An unknown error occurred when receiving ip information: {}'
                .format(err)
            )
        return d

    @staticmethod
    def multithreading(f, rows):
        threads = []
        logging.debug('Threads need to create ' + str(len(rows)))
        for i in range(len(rows)):
            logging.debug('Thread {} was started'.format(i))
            t = Thread(target=f, args=[rows[i][0], rows[i][1], rows[i][2]])
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    @staticmethod
    def write_last_pos(lid):
        try:
            connect = init_db_connection(
                settings.DB_NAME_CABINET,
                settings.DB_USER_CABINET,
                settings.DB_PASS_CABINET,
                settings.DB_HOST_CABINET,
                DB_PORT_CABINET)
            cursor = connect.cursor()
            cursor.execute('SELECT last_id FROM attack_notify;')
            if cursor.fetchone() is not None:
                cursor.execute('UPDATE attack_notify SET last_id = %s WHERE true;', (lid,))
            else:
                cursor.execute("INSERT INTO attack_notify VALUES ('%s');", (lid,))
            close_db_connection(cursor, connect)
            return True

        except Exception as e:
            logging.error('An error occurred writing last id: {}'.format(e))
            return False

    @staticmethod
    def get_users(domain, waf_id, bt):
        if bt not in user_config_map:
            return []
        if waf_id in [None, ""]:
            waf_id = ""
        if type(domain) is str:
            if domain.strip(' ') == '':
                domain = '*'
        elif domain in (None, ''):
            domain = '*'
        str_dom = 'domain ILIKE \'%' + domain.strip('.')[-1] \
                  + '%\' OR domain ILIKE \'%*%\' OR domain IS NULL OR domain = \'\''
        str_vh = 'vhosts ILIKE \'%' + domain.strip('.')[-1] \
                 + '%\' OR vhosts ILIKE \'%*%\' OR vhosts IS NULL OR vhosts = \'\''

        nes_query = 'SELECT id, domain, email FROM main_wafuser WHERE {dwhere}'
        if waf_id != '':
            waf_id = 'OR waf_alias LIKE \'%' + waf_id + '%\''
        nes_where = '(' + str_dom + ') AND (waf_alias  IN (\'\',\'*\') ' + waf_id + ')'

        nq = nes_query.format(dwhere=nes_where)
        sql = """
                SELECT email, email_details, domain, vhosts 
                FROM ({query}) AS users JOIN user_config ON user_id = users.id {dwhere}
            """
        dwhere = 'WHERE notification_enable = true AND ' + user_config_map[bt] + ' AND (' + str_vh + ');'
        try:
            connect = init_db_connection(
                settings.DB_NAME_CABINET,
                settings.DB_USER_CABINET,
                settings.DB_PASS_CABINET,
                settings.DB_HOST_CABINET,
                DB_PORT_CABINET)
            cursor = connect.cursor()
            cursor.execute(sql.format(query=nq, dwhere=dwhere))
            users = cursor.fetchall()
            close_db_connection(cursor, connect)
        except Exception as e:
            users = []
            logging.error('An error occurred fetching user settings from db: {}'.format(e))

        excl = []
        domain_reverse = domain[::-1]
        for user in users:
            if user[2] in (None, ''):
                d = '*'
            else:
                d = user[2].replace(' ', '').replace('sub_', '*.').split(',')
            for i in d:
                if '*' in d:
                    break
                if domain.split('.')[-1] in i:
                    if domain != i:
                        tmp = len(domain) if len(domain) < len(i) else len(i)
                        i_reverse = i[::-1]
                        for j in range(0, tmp):
                            if i_reverse[j] != domain_reverse[j]:
                                if i_reverse[j] != '*':
                                    if user not in excl:
                                        excl.append(user)
                                    break
                            elif j == tmp - 1:
                                if i_reverse[j] != '.':
                                    if len(i) == tmp + 1:
                                        if i_reverse[j + 1] != '.':
                                            if user not in excl:
                                                excl.append(user)
                                            break
                                    else:
                                        if user not in excl:
                                            excl.append(user)
                                        break
                        else:
                            if user in excl:
                                excl.remove(user)
                            break

                    else:
                        break

                elif i == d[-1]:
                    if user not in excl:
                        excl.append(user)

            if user in excl:
                continue

            if user[3] in (None, ''):
                vh = '*'
            else:
                vh = user[3].replace(' ', '').replace('sub_', '*.').split(',')
            for i in vh:
                if '*' in vh:
                    break
                if domain.split('.')[-1] in i:
                    if domain != i:
                        tmp = len(domain) if len(domain) < len(i) else len(i)
                        i_reverse = i[::-1]
                        for j in range(0, tmp):
                            if i_reverse[j] != domain_reverse[j]:
                                if i_reverse[j] != '*':
                                    if user not in excl:
                                        excl.append(user)
                                    break
                            elif j == tmp - 1:
                                if i_reverse[j] != '.':
                                    if len(i) == tmp + 1:
                                        if i_reverse[j + 1] != '.':
                                            if user not in excl:
                                                excl.append(user)
                                            break
                                    else:
                                        if user not in excl:
                                            excl.append(user)
                                        break
                        else:
                            if user in excl:
                                excl.remove(user)
                            break

                    else:
                        break

                elif i == vh[-1]:
                    if user not in excl:
                        excl.append(user)

            if user in excl:
                continue

        for i in excl:
            users.remove(i)

        return users

    def main(self):
        try:
            connect = init_db_connection(DB_NAME, DB_USER, DB_PASS, DB_HOST, DB_PORT)
            cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cursor.execute('SELECT id FROM attack WHERE id > %s ORDER BY id DESC;', (self.last_id,))
            last_attack = cursor.fetchone()
            close_db_connection(cursor, connect)
            last_attack = last_attack[0] if last_attack else self.last_id
        except Exception as e:
            last_attack = self.last_id
            logging.error('An error occurred fetching attacks from db: {}'.format(e))
        while last_attack > self.last_id:
            db_available = True
            try:
                connect = init_db_connection(DB_NAME, DB_USER, DB_PASS, DB_HOST, DB_PORT)
                cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cursor.execute('SELECT id, timestamp, ip, vhost, url, lm, rule_id, mz, param, body, method, ua, '
                               'referer, description, bt, cookie, request_id, waf_id, other_headers, possible, blocked '
                               'FROM attack WHERE id <= %s and id > %s ORDER BY id LIMIT 500;',
                               (last_attack, self.last_id,))
                res = cursor.fetchall()
                close_db_connection(cursor, connect)
            except Exception as e:
                res = []
                logging.error('An error occurred fetching attacks from db: {}'.format(e))
            if res:
                try:
                    connect = init_db_connection(DB_NAME, DB_USER, DB_PASS, DB_HOST, DB_PORT)
                    cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
                    cursor.execute(
                        'SELECT vhost, waf_id, bt FROM attack WHERE id <= %s and id > %s GROUP BY vhost, waf_id, bt;',
                        (res[-1].get("id"), self.last_id))
                    resset = cursor.fetchall()
                    close_db_connection(cursor, connect)
                except Exception as e:
                    resset = []
                    logging.error('An error occurred fetching attacks from db: {}'.format(e))

                if resset:
                    dict_set = dict()
                    for ii in resset:
                        # Convert from punycode
                        ii['vhost'] = convert_from_punycode(ii.get('vhost'))

                        users = self.get_users(
                            domain=ii.get('vhost'),
                            waf_id=ii.get('waf_id'),
                            bt=ii.get('bt')
                        )
                        key = '{}|{}|{}'.format(ii.get('vhost'), ii.get('waf_id'), ii.get('bt'))
                        dict_set[key] = users
                    for ii in res:
                        # Convert from punycode
                        ii['vhost'] = convert_from_punycode(ii.get('vhost'))
                        ii['url'] = convert_from_punycode(ii.get('url'))

                        key = '{}|{}|{}'.format(ii.get('vhost'), ii.get('waf_id'), ii.get('bt'))
                        users = dict_set.get(key, [])
                        if not users:
                            continue

                        d = self.get_memcache_ip(ii.get('ip'))
                        if not d:
                            d = self.get_ip_data(ii.get('ip'), ii)

                        # Running users list
                        for user in users:
                            email = user[0]

                            email_details = user[1]

                            if email_details:
                                subject, body = self.create_body(ii, d)
                            else:
                                subject, body = self.create_limited_body(ii, d)

                            self.send_mail(subject=subject, body=body, attack_id=ii['id'], email=email)
                            db_available = self.write_last_pos(ii['id'])
                            if not db_available:
                                break

                        if not db_available:
                            break

            try:
                logging.debug('Try to write last_id')
                if res:
                    self.last_id = res[-1].get('id')
            except Exception as e:
                logging.error('An error occurred when extracting an element by key: {}'.format(e))
            logging.debug('All mails has been sent')
            self.write_last_pos(self.last_id)
            try:
                connect = init_db_connection(DB_NAME, DB_USER, DB_PASS, DB_HOST, DB_PORT)
                cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
                cursor.execute('SELECT id FROM attack WHERE id > %s ORDER BY id DESC LIMIT 1;', (self.last_id,))
                last_attack = cursor.fetchall()
                close_db_connection(cursor, connect)
                if last_attack:
                    last_attack = last_attack[0].get('id')
                else:
                    last_attack = self.last_id
            except Exception as e:
                last_attack = self.last_id
                logging.error('An error occurred fetching attacks from db: {}'.format(e))

        return True


def get_last_attack(loop, a):
    try:
        a.main()
    except Exception as e:
        logging.error('An error occurred when an unknown error occurred when sending a notification: {}'.format(e))
    loop.call_later(30, get_last_attack, loop, a)


def main():
    loop = asyncio.new_event_loop()
    logging.info('Started')

    try:
        a = Notifier()
        loop.call_soon(get_last_attack, loop, a)
        loop.run_forever()
    finally:
        loop.close()
        logging.info('Finished')


if __name__ == '__main__':
    main()
