import asyncio
import json
import logging
import os
import psycopg2
import requests
import sys

from django.conf import settings
from pymemcache.client.base import Client

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)

# Do not change the import
from main.utils import get_proxy_conf_dict


BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)

if os.environ.get('DJANGO_SETTINGS_MODULE') is None:
    os.environ['DJANGO_SETTINGS_MODULE'] = "cabinet.settings_extra"

# logging
log_place = '/var/log/uwsgi/cabinet/ipinfo2memcache.log'

# Memcached settings
vGeoIP = settings.V_GEOIP
cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))
mcache = Client(cache_address)

# Postgres connect
DB_NAME = settings.DB_NAME_CONF
DB_USER = settings.DB_USER_CONF
DB_PASS = settings.DB_PASS_CONF
DB_HOST = settings.DB_HOST_CONF

try:
    DB_PORT = settings.DB_PORT_CONF
    if DB_PORT == '':
        DB_PORT = '5432'
except NameError:
    DB_PORT = '5432'


user_agent = {'User-agent': settings.USER_AGENT}


def send_geo_request(url: str, ips: list):

    # init
    result = {}

    try:

        # get Geo
        response = requests.post(
            url,
            headers=user_agent,
            proxies=get_proxy_conf_dict(),
            timeout=settings.URI_GEOIP_TIMEOUT,
            data=str(ips)
        ).text

        # load result as dict
        result = json.loads(response)

    except requests.exceptions.Timeout as e:
        logging.error('An error occurred while sending request to GeoIP server: {}'.format(e))

    except requests.exceptions.ProxyError as e:
        logging.error('An error occurred while sending request to GeoIP server: {}'.format(e))

    except Exception as e:
        logging.error('An error occurred while sending request to GeoIP server: {}'.format(e))
        return False

    # return the result
    return result


def check_ip_from_memcached(ips):

    from main.utils import is_local

    # init
    result = []

    # processing IPS
    for ip in ips:

        try:
            get_ip = mcache.get(ip + vGeoIP)
        except Exception:
            get_ip = None

        if get_ip is None:
            if is_local(ip) is False:
                result.append(ip)
        else:
            if json.loads(get_ip.decode('utf-8').replace("'", '"')).get('city') == '':
                result.append(ip)

    return result


def get_geo(ips, tokens):
    try:

        # skip empty list
        if not ips:
            return

        # receiving GeoIP data with valid token
        for token in tokens:

            # skip empty
            if not token:
                continue
            else:
                token = str(token[0])

            # logging.debug('Sending request to GeoIP')
            r = send_geo_request(settings.URI_GEOIP_BATCH + token, ips)

            # if result is not empty/False
            if r:

                # update data in the memcached
                for k, v in r.items():

                    # skip private
                    if v == 'Private IP':
                        continue

                    else:

                        # skip non-success
                        if (not isinstance(v, dict)) or (v.get('status') != 'success'):
                            continue

                        # update data in the memcached
                        try:
                            mcache.set(k + vGeoIP, json.dumps(v), 86400)
                        except Exception as e:
                            logging.error('An error occurred while storing data for {}: {}'.format(k, e))

            # incorrect token
            elif r is False:

                # processing with the next token
                continue

            # stop GeoIP processing
            return

    except Exception as e:
        logging.error('An error occurred while receiving data from GeoIP-server: {}'.format(e))


def ip_memcache(loop, tsa, tsm):

    # init
    ips = []

    # connect to the database
    connect = psycopg2.connect(database=DB_NAME, user=DB_USER, password=DB_PASS, host=DB_HOST, port=DB_PORT, connect_timeout=10)
    cursor = connect.cursor()

    cursor.execute("SELECT geo_token FROM geo_tokens;")
    res = cursor.fetchall()
    tokens = [x for x in res if x]

    if tokens:

        # attack: get the IPs
        query = """ SELECT DISTINCT ip FROM attack WHERE timestamp > to_timestamp(%s) """
        cursor.execute(query, (tsa,))
        res = cursor.fetchall()
        ips.extend([item[0] for item in res if item])
        del res

        # attack: get the ts
        query = """ SELECT CAST(EXTRACT(EPOCH FROM timestamp) AS INTEGER) FROM attack ORDER BY timestamp DESC LIMIT 1 """
        cursor.execute(query,)
        res = cursor.fetchone()
        tsa = res[0] if res else tsa

        ##

        # mt_main: get the IPs
        query = """ SELECT DISTINCT ip FROM mt_main WHERE CAST(timestamp AS INTEGER) > %s """
        cursor.execute(query, (tsm,))
        res = cursor.fetchall()
        ips.extend([item[0] for item in res if item])
        del res

        # attack: get the ts
        query = """ SELECT CAST(timestamp AS INTEGER) FROM mt_main ORDER BY id DESC LIMIT 1 """
        cursor.execute(query,)
        res = cursor.fetchone()
        tsm = res[0] if res else tsm

        # processing the IPs
        try:

            # remove duplicates
            ips = set(ips)

            # exclude IP from the memcached
            ips = [x for x in ips if mcache.get(x + vGeoIP) is None]

            # split the list into chunks
            chunks = [ips[x:x + 100] for x in range(0, len(ips), 100)]

            # processing the chunk list
            for chunk in chunks:
                get_geo(chunk, tokens)

        except Exception as e:
            logging.error('An error occurred while processing the IPs in multithreading mode: {}'.format(e))

    # close database connection
    cursor.close()
    connect.close()

    loop.call_later(90, ip_memcache, loop, tsa, tsm)


# Event Loop
def main():
    try:
        logging.basicConfig(
            filename=log_place,
            level=logging.INFO,
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%d/%b/%Y:%H:%M:%S %z')
        logging.info('Started')
        loop = asyncio.new_event_loop()
        try:
            tsa = 0
            tsm = 0
            loop.call_soon(ip_memcache, loop, tsa, tsm)
            loop.run_forever()
        finally:
            loop.close()
            logging.info('Finished')
    except Exception as e:
        logging.error(f'An error occurred while {str(e)} {type(e).__name__}')


if __name__ == '__main__':
    main()
