from django.core.exceptions import ValidationError
from django.db import models
from .wafuser import WAFUSER


class Config(models.Model):
    SSL = 'ssl'
    TLS = 'tls'
    NONE = None
    ENCRYPTION_CHOICES = [
        (SSL, 'SSL'),
        (TLS, 'TLS'),
        (NONE, 'No')
    ]

    smtp_host = models.CharField(
        max_length=255,
        verbose_name='SMTP host name'
    )
    smtp_port = models.IntegerField(default=25)
    smtp_user = models.CharField(
        max_length=255,
        verbose_name='SMTP user name'
    )
    smtp_password = models.CharField(
        max_length=255,
        verbose_name='SMTP password'
    )
    encryption = models.CharField(
        max_length=8,
        null=True,
        blank=True,
        choices=ENCRYPTION_CHOICES, default=NONE
    )
    send_email = models.BooleanField(
        # on/off all mail features
        default=False,
        verbose_name='Send Email'
    )

    class Meta:
        db_table = 'config'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Save old SMTP password
        self._smtp_password = self.smtp_password

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if not self.pk and Config.objects.exists():
            raise ValidationError('There can only be one settings entry')

        if self.smtp_password != self._smtp_password:
            self.smtp_password = self.smtp_password

        return super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields
        )


class ConfigCleaningTable(models.Model):
    time_interval = models.IntegerField(default=0, verbose_name='Time interval')
    ip = models.CharField(max_length=512, default='*', verbose_name="IP")
    url = models.CharField(max_length=1024, default='*', verbose_name='URL')
    bt = models.CharField(max_length=255, default='*', verbose_name='Attack Model')
    vhost = models.CharField(max_length=255, default='*', verbose_name='Domain')
    desc = models.CharField(max_length=255, default='', verbose_name='Description')
    enable = models.BooleanField(
        default=False,
        verbose_name='Enable'
    )

    class Meta:
        db_table = 'config_cleaning_attack'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UserConfig(models.Model):
    user = models.OneToOneField(
        WAFUSER,
        on_delete=models.CASCADE,
        primary_key=True
    )
    vhosts = models.TextField(
        blank=True,
        null=True,
        verbose_name='vhosts'
    )
    notification_enable = models.BooleanField(
        verbose_name='Notification enable',
        default=False
    )
    email_details = models.BooleanField(
        default=True,
        verbose_name='Email Details'
    )

    bt1 = models.BooleanField(
        default=True,
        verbose_name='Enable Signature Analysis'
    )
    bt2 = models.BooleanField(
        default=True,
        verbose_name='Enable Signature Analysis'
    )
    bt3 = models.BooleanField(
        default=True,
        verbose_name='Enable Nemesida AI MLA'
    )
    bt4 = models.BooleanField(
        default=True,
        verbose_name='Enable Antivirus'
    )
    bt5 = models.BooleanField(
        default=True,
        verbose_name='Enable BT5'
    )
    bt7 = models.BooleanField(
        default=True,
        verbose_name='Enable Brute force'
    )
    bt8 = models.BooleanField(
        default=True,
        verbose_name='Enable Nemesida AI MLC'
    )
    bt9 = models.BooleanField(
        default=True,
        verbose_name='Enable Flood'
    )
    bt10 = models.BooleanField(
        default=True,
        verbose_name='Enable DDoS'
    )
    bt14 = models.BooleanField(
        default=True,
        verbose_name='Enable OpenAPI'
    )
    bt16 = models.BooleanField(
        default=True,
        verbose_name='Enable Blocked IP'
    )

    class Meta:
        db_table = 'user_config'
