import os
import json
import base64
import logging

from datetime import datetime

from django.db import models
from django.conf import settings
from django.core import signing
from pymemcache.client.base import Client
from main.utils import remove_zero_from_ip, convert_from_punycode

logger = logging.getLogger(__name__)

mclient = Client((settings.MEMCACHED_HOST, settings.MEMCACHED_PORT))


class TableQuerySet(models.QuerySet):
    BT_CELL_CLASS_MAPPING = {
        11: 'bt-red',
        12: 'bt-blue',
        13: 'bt-green',
    }

    BT_CELL_DEFAULT_CLASS = 'bt-black'

    def signed_rows(self, field):
        for row in self:
            setattr(row, field, signing.dumps(row.id))
        return self

    def decode_body(self):
        for row in self:
            try:
                row.body = base64.b64decode(row.body).decode("utf-8")
            except Exception:
                try:
                    missing_padding = len(row.body) % 4
                    if missing_padding != 0:
                        decode_body = row.body + '=' * (4 - missing_padding)
                    else:
                        decode_body = row.body
                    row.body = base64.b64decode(decode_body).decode("utf-8")
                except Exception:
                    row.body = row.body
        return self

    def logs_format(self, export=False):
        for row in self:
            if row._meta.model.__name__ == 'Logs':
                row.uri = row.url
                row.args = row.param
                row.mz_origin = row.mz.lower()
                row.vhost = convert_from_punycode(row.vhost)
                row.mz = row.mz.lower().replace('url', 'uri')
                row.timestamp = int(row.timestamp.timestamp())
                if row.other_headers:
                    row.other_headers = '\n'.join([i.strip() for i in row.other_headers.split(';')])

                if export:
                    try:
                        row.body = bytes(row.body).decode('utf-8', 'ignore')
                    except Exception as err:
                        row.body = ""
        return self

    def format(self):
        for row in self:
            if row._meta.model.__name__ == 'Logs':
                row.uri = row.url
                row.args = row.param
                row.mz_origin = row.mz.lower()
                row.mz = row.mz.lower().replace('url', 'uri')
                # row.timestamp = row.timestamp.strftime('%H:%M:%S %d.%m.%Y')

            row.vhost = convert_from_punycode(row.vhost)
            if row.timestamp:
                row.timestamp = datetime.fromtimestamp(
                    int(row.timestamp)).strftime('%H:%M:%S %d.%m.%Y')
            if row.other_headers:
                row.other_headers = os.linesep.join([i.strip() for i in row.other_headers.split('; ')])
                row.other_headers = row.other_headers.strip('; ')
            bt_value = getattr(row, 'bt', None)
            row.bt_class = ''
            if row.body:
                row.body = os.linesep.join([s for s in row.body.splitlines() if s])
            if bt_value is not None:
                row.bt_class = self.BT_CELL_CLASS_MAPPING.get(
                    bt_value, self.BT_CELL_DEFAULT_CLASS
                )

        return self

    def cut_fields(self):
        fields = ['args', 'uri', 'cookie', 'ua', 'body', 'referer']

        for row in self:
            if row._meta.model.__name__ == 'Logs':
                row.uri = row.url
                row.args = row.param
            for field in fields:
                if getattr(row, field, None):
                    f = getattr(row, field)
                    f = ' '.join(f.split())
                    if len(f) > 1000:
                        field_attr = f[0:1000] + ' ' + '[cutted]'
                        setattr(row, field, field_attr)
        return self

    def add_ip_country(self, only_ip=False):
        vGeoIP = settings.V_GEOIP
        for row in self:
            try:
                if only_ip:
                    row.ip = row.ip
                else:
                    m_data = mclient.get(remove_zero_from_ip(row.ip) + vGeoIP)
                    if m_data is None:
                        row.ip = row.ip
                    else:
                        m_ip = json.loads(m_data.decode('utf-8').replace("\'", "\""))
                        if m_ip.get('status') != 'fail':
                            row.ip = m_ip

            except json.decoder.JSONDecodeError:
                logger.error(
                    "An error occurred while parsing JSON: {}"
                    .format(mclient.get(remove_zero_from_ip(row.ip) + vGeoIP).decode('utf-8'))
                )
                mclient.delete(remove_zero_from_ip(row.ip) + vGeoIP)
            except ConnectionRefusedError:
                logger.error("An error occurred while connecting to memcached")
            except Exception as e:
                logger.error(
                    "An unknown error occurred when receiving ip information: {}".format(e)
                )
        return self

    def split_zone(self):
        for row in self:
            if type(row.mz) is str:
                row.mz = [i for i in row.mz.split('|') if i != '']
        return self


class MtTable(models.Model):
    id = models.AutoField(primary_key=True)
    bt = models.BigIntegerField(blank=True, null=True)
    timestamp = models.TextField(blank=True, null=True)
    request_id = models.TextField(blank=True, default='')
    mz = models.TextField(blank=True, null=True)
    mz_origin = models.TextField(blank=True, null=True)
    ip = models.TextField(blank=True, null=True)
    vhost = models.TextField(blank=True, null=True)
    method = models.TextField(blank=True, null=True)
    uri = models.TextField(blank=True, null=True)
    args = models.TextField(blank=True, null=True)
    cookie = models.TextField(blank=True, null=True)
    other_headers = models.TextField(blank=True, null=True)
    ua = models.TextField(blank=True, null=True)
    referer = models.TextField(blank=True, null=True)
    body = models.TextField(blank=True, null=True)
    waf_id = models.TextField(blank=True, null=True)
    agent = models.TextField(blank=True, default='-')
    cc = models.CharField(max_length=2, default='', blank=True, null=True, verbose_name='Country Code')
    objects_for_mt = TableQuerySet.as_manager()
    objects = models.Manager()

    class Meta:
        db_table = 'mt_main'

    def check_ip_country(self, only_ip=False):
        vGeoIP = settings.V_GEOIP
        try:
            if only_ip:
                self.ip = self.ip
            else:
                m_data = mclient.get(remove_zero_from_ip(self.ip) + vGeoIP)

                self.ip_info = {}

                if m_data:
                    m_ip = json.loads(m_data.decode('utf-8').replace("\'", "\""))
                    if m_ip.get('status') != 'fail':
                        self.ip_info = m_ip

        except json.decoder.JSONDecodeError:
            logger.error(
                "An error occurred while parsing JSON: {}"
                .format(mclient.get(remove_zero_from_ip(self.ip) + vGeoIP).decode('utf-8'))
            )
            mclient.delete(remove_zero_from_ip(self.ip) + vGeoIP)
        except ConnectionRefusedError:
            logger.error("An error occurred while connecting to memcached")
        except Exception as e:
            logger.error(
                "An unknown error occurred when receiving ip information: {}".format(e)
            )
        return self


class MtExport(models.Model):
    id = models.AutoField(primary_key=True)
    bt = models.BigIntegerField(blank=True, default=12)
    timestamp = models.TextField(blank=True, default='')
    request_id = models.TextField(blank=True, default='')
    mz = models.TextField(blank=True, default=0)
    vhost = models.TextField(blank=True, default='')
    method = models.TextField(blank=True, default='')
    uri = models.TextField(blank=True, default='')
    args = models.TextField(blank=True, default='')
    cookie = models.TextField(blank=True, default='')
    ua = models.TextField(blank=True, default='')
    referer = models.TextField(blank=True, default='')
    body = models.TextField(blank=True, default='')
    other_headers = models.TextField(blank=True, null=True)
    waf_id = models.TextField(blank=True, default='')
    objects_for_mt = TableQuerySet.as_manager()
    objects = models.Manager()

    class Meta:
        db_table = 'mt_export'

    def check_ip_country(self, only_ip=False):
        vGeoIP = settings.V_GEOIP
        try:
            if only_ip:
                self.ip = self.ip
            else:
                m_data = mclient.get(remove_zero_from_ip(self.ip) + vGeoIP)
                if m_data is None:
                    self.ip = self.ip
                else:
                    m_ip = json.loads(m_data.decode('utf-8').replace("\'", "\""))
                    if m_ip.get('status') != 'fail':
                        self.ip = m_ip

        except json.decoder.JSONDecodeError:
            logger.error(
                "An error occurred while parsing JSON: {}"
                .format(mclient.get(remove_zero_from_ip(self.ip) + vGeoIP).decode('utf-8'))
            )
            mclient.delete(remove_zero_from_ip(self.ip) + vGeoIP)
        except ConnectionRefusedError:
            logger.error("An error occurred while connecting to memcached")
        except Exception as e:
            logger.error(
                "An unknown error occurred when receiving ip information: {}".format(e)
            )
        return self


def remove_duplicated_records(model, fields):
    """
    Removes records from `model` duplicated on `fields`
    while leaving the most recent one (biggest `id`).
    """
    duplicates = model.objects.values(*fields)

    # override any model specific ordering (for `.annotate()`)
    duplicates = duplicates.order_by()

    # group by same values of `fields`; count how many rows are the same
    if model.__name__ == 'ExportFn':
        duplicates = duplicates.annotate(
            max_id=models.Max("id"), count_id=models.Count("id")
        )
        cache_name_to_clear = 'data_md5_export_fn'
    else:
        duplicates = duplicates.annotate(
            max_id=models.Max("id"), count_id=models.Count("request_id")
        )
        cache_name_to_clear = 'data_md5_export_fp'

    # leave out only the ones which are actually duplicated
    duplicates = duplicates.filter(count_id__gt=1)

    for duplicate in duplicates:
        to_delete = model.objects.filter(**{x: duplicate[x] for x in fields})

        # leave out the latest duplicated record
        lic_keys = set(to_delete.values_list('waf_id', flat=True))
        lic_keys = set(' '.join(lic_keys).split(' '))
        to_delete = to_delete.exclude(id=duplicate["max_id"])
        model.objects.filter(id=duplicate["max_id"]).update(waf_id=' '.join(lic_keys))
        to_delete.delete()

    mclient.delete(cache_name_to_clear)
