import base64
import json
import yaml
import logging
import requests
from json import JSONDecodeError
from urllib.parse import urlparse
from django.conf import settings
from django.utils.translation import pgettext
from url_normalize import url_normalize

from main.convert_open_api import convert_schema
from main.utils import get_proxy_api_dict


try:
    api_url = settings.API_URI
except AttributeError:
    api_url = 'http://localhost:8080/nw-api/'

USER_AGENT = settings.USER_AGENT
logger = logging.getLogger(__name__)

proxy_dict_api = get_proxy_api_dict()


def get_openapi_schema(key):
    request_json = send_request_get_schema(key)
    if request_json is not None:
        schemas = request_json.get('schemas')
        if schemas is None:
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'An error occurred while get schema. Schemas is not exist for key: {key}'.format(key=format_key))
            return None
        decoded_schemas = {}
        for schema in schemas:
            if not schema.get('schema_name') or not schema.get('schema'):
                logger.error(
                    'An error occurred while get invalid schema: {schema}'.format(schema=schema))
                continue
            try:
                decoded_schemas[schema['schema_name']] = base64.b64decode(schema['schema']).decode('utf-8', 'ignore')
            except Exception as error:
                logger.error(
                    f'An error occurred while get json from schema: {schema} error {error}')
        return decoded_schemas
    else:
        return None


def get_paths(schema_name, schemas):
    formatted_list_path = []
    try:
        schemas = json.loads(schemas)
    except JSONDecodeError:
        logger.error(
            'An error occurred while get json from schema: {schema}'.format(schema=schemas))
        return formatted_list_path

    if schemas is not None and 'paths' in schemas:
        for schema_key, schema_val in schemas['paths'].items():
            if '{' in schema_key:
                methods = schema_val.keys()
                for method in methods:
                    if not schema_val[method].get('parameters'):
                        continue
                    parameters = schema_val[method]['parameters']
                    for param in parameters:
                        if param['in'] == 'path':
                            param_name = param['name']
                            param_type = param['schema'].get('type')
                            if param_type == 'integer':
                                param_type = 'int'
                            else:
                                param_type = 'str'
                            formatted_list_path.append(
                                schema_name + schema_key.replace('{' + param_name + '}', '{' + param_type + '}')
                            )

                    break
            else:
                formatted_list_path.append(schema_name + schema_key)

    return formatted_list_path


def update_paths(key):
    paths = []
    get_schemas = get_openapi_schema(key)
    if get_schemas is not None:
        for schema_key, schema_val in get_schemas.items():
            paths += get_paths(schema_key, schema_val)
        response = send_request_set_openapi_urls(key, paths)
        if response:
            return True
        else:
            return False


def load_schema(key, api_schema, schema_name, old_name):
    # Key as obj LicKeys
    try:
        api_schema = json.loads(api_schema)
    except JSONDecodeError:
        try:
            api_schema = yaml.safe_load(api_schema)
        except Exception as e:
            logger.error(
                'An error occurred while get schema for key {key}: {error}'
                .format(key=key.get_formated_key(), error=str(e))
            )

            return False, pgettext('messages', 'File is not json/yaml')

    try:
        if str(api_schema.get('swagger', '3.0.0'))[0] != '3':
            api_schema = convert_schema(api_schema)
    except Exception as e:
        logger.error(
            'An error occurred while converting schema OpenAPI 2.0 to 3.0 for key {key}: {error}'
            .format(key=key.get_formated_key(), error=str(e))
        )

        return False, pgettext('messages', 'An error occurred while converting schema OpenAPI 2.0 to 3.0')

    if not api_schema.get('servers'):
        api_schema.update({"servers": [{"url": "/"}]})

    if old_name:
        for i in api_schema.get('servers', []):
            i['url'] = i.get('url', old_name).replace(old_name, schema_name)

    try:
        api_schema = json.dumps(api_schema)
    except JSONDecodeError as e:
        logger.error(
            'An error occurred while get schema for key {key}: {error}'
            .format(key=key.get_formated_key(), error=str(e))
        )

        return False, pgettext('messages', 'File is not json/yaml')
    r = send_request_set_schema(key.key, schema_name, api_schema, old_name)

    if not r:
        return False, pgettext('messages', 'An error occurred while saving the schema')

    update_paths(key.key)

    return True, None


def get_paths_info(schema, schema_name):
    path_info_list = []

    try:
        domain = urlparse(schema['servers'][0]['url']).netloc or schema_name
    except (KeyError, TypeError):
        return None
    if schema.get('paths'):
        for paths_key, paths_val in schema['paths'].items():
            path_info = {paths_key: []}
            methods = paths_val.keys()
            for method in methods:
                data = {}
                for key in schema['paths'][paths_key][method]:
                    if key == 'parameters':
                        try:
                            data[key] = parse_info_parameters(schema['paths'][paths_key][method][key])
                        except Exception as error:
                            logger.error(
                                'An error occurred while get parameters from path schema '
                                f'with schema name {domain} and url {paths_key} error {error}')
                    elif key == 'requestBody':
                        try:
                            data[key] = parse_info_req_body(schema['paths'][paths_key][method][key])
                        except Exception as error:
                            logger.error(
                                'An error occurred while get requestBody from path schema '
                                f'with schema name {domain} and url {paths_key} error {error}')

                    else:
                        data[key] = schema['paths'][paths_key][method][key]
                if not data:
                    data = 'Without fields containing information'

                path_info[paths_key].append({method: data})
            path_info_list.append(path_info)

        return {domain: path_info_list}
    else:
        return None


def parse_info_parameters(pars):
    parameters_list = list()
    for i in pars:
        d = dict()
        name = i.get('name')
        if i.get('schema'):
            parameters_type = i.get('schema').get('type')
        elif i.get('type'):
            parameters_type = i.get('type')
        else:
            parameters_type = 'unknown type'
        d[name] = parameters_type
        if parameters_type == 'array' and i.get('schema', {}).get('items'):
            d['items'] = get_vals_from_object(i.get('schema', {}).get('items', {}))
        elif parameters_type == 'object':
            d = get_vals_from_object(i)
        parameters_list.append(d)
    return parameters_list


def get_vals_from_object(obj):
    d = obj
    if obj.get('type') == 'object':
        prop = obj.get('properties', {})
        if not d.get('properties') and prop:
            d['properties'] = {}
        for k, v in prop.items():
            val = get_vals_from_object(v)
            d['properties'][k] = val
    elif obj.get('type') == 'array':
        arr = obj.get('items', {})
        d['type'] = obj.get('type')
        if arr.get('type') == 'object':
            prop = arr.get('properties', {})
            if not d.get('properties') and prop:
                d['properties'] = {}
            for k, v in prop.items():
                val = get_vals_from_object(v)
                d['properties'][k] = val
        else:
            d['items'] = get_vals_from_object(arr)
    else:
        d['type'] = obj.get('type')
    return d


def parse_info_req_body(req_body):
    d = dict()
    content = req_body.get('content')
    d['content'] = dict()
    if content:
        for type_key, body_type in content.items():
            tmp = content.get(type_key)
            dict_args = content.get(type_key).get('schema')
            if dict_args:
                tmp['schema'] = get_vals_from_object(dict_args)
            d['content'][type_key] = tmp

    return d


def get_openapi_paths_schemas(key):
    request_json = send_request_get_openapi_paths(key)
    path_schema_names = {}
    # Temp solution, need replaced by get from api
    paths_in_schemas = []
    existing_schemas = []
    get_schemas = get_openapi_schema(key)
    if get_schemas is not None:
        for schema_name, schemas in get_schemas.items():
            try:
                schemas = json.loads(schemas)
            except JSONDecodeError:
                logger.error(
                    'An error occurred while get json from schema: {schema}'.format(schema=schemas))
                continue

            existing_schemas.append(schema_name)
            if schemas is not None and 'paths' in schemas.keys():
                for url in schemas['paths'].keys():
                    paths_in_schemas.append(schema_name + url)
    if request_json is not None:
        paths = request_json.get('paths', [])
        for path in paths:
            if not path.get('schema_name') or not path.get('url') or not path.get('schema'):
                logger.error(
                    'An error occurred while get invalid path: {path}'.format(path=path))
                continue
            if path['schema_name'] + path['url'] in paths_in_schemas:
                continue
            if path['schema']:
                try:
                    tmp = json.loads(base64.b64decode(path['schema']).decode('utf-8', 'ignore').replace('\'', '\"'))
                except JSONDecodeError:
                    logger.error(
                        'An error occurred while get json from path schema: {schema}'
                        .format(schema=base64.b64decode(path['schema'])
                                .decode('utf-8', 'ignore').replace('\'', '\"')))
                    tmp = dict()
                except Exception as error:
                    logger.error(
                        f"An error occurred while decode b64 from path schema: {path['schema']} error {error}")
                    tmp = dict()
                data = dict()
                if tmp.get(path['schema_name']):
                    if tmp[path['schema_name']].get(path['url']):
                        for i in tmp[path['schema_name']].get(path['url']):
                            for key in tmp[path['schema_name']][path['url']][i]:
                                if not data.get(i):
                                    data[i] = dict()
                                if key == 'parameters':
                                    try:
                                        data[i][key] = parse_info_parameters(
                                            tmp[path['schema_name']][path['url']][i][key])

                                    except Exception as error:
                                        logger.error(
                                            'An error occurred while get parameters from path schema '
                                            f"with schema name {path['schema_name']} and url {path['url']} error "
                                            f"{error}")
                                elif key == 'requestBody':
                                    try:
                                        data[i][key] = parse_info_req_body(
                                            tmp[path['schema_name']][path['url']][i][key])
                                    except Exception as error:
                                        logger.error(
                                            f'An error occurred while get request Body from path schema'
                                            f"with schema name {path['schema_name']} and url {path['url']} error"
                                            f" {error}")

                                else:
                                    data[i][key] = tmp[path['schema_name']][path['url']][i][key]
                    else:
                        data['ERROR'] = {"parameters": ['Error: URL is not exist in schema for path']}
                else:
                    data['ERROR'] = {"parameters": ['Error: Schema\'s name is not exist in schema for path']}
                path['schema'] = data
                methods = [i for i in path['schema']]
                path['methods'] = methods
            if path['schema_name'] in path_schema_names:
                schema_name_list = path_schema_names[path['schema_name']]['list_path']
                schema_name_list.append(path)
                path_schema_names[path['schema_name']]['list_path'] = schema_name_list
            else:
                path_schema_names[path['schema_name']] = {'list_path': [path],
                                                          'exists_schemas': path['schema_name'] in existing_schemas}
        return path_schema_names
    else:
        return None


def send_request_get_schema(key):
    try:
        response = requests.post(
            url_normalize(api_url + '/get_openapi_schemas'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json={"key": key},
        )
    except Exception as e:
        logger.error(
            'An error occurred while get schema: {error}'.format(error=str(e)))
        return None

    if response.ok:
        try:
            return json.loads(response.text)
        except JSONDecodeError:
            return None
    else:
        logger.error(
            'An error occurred while get schema with response code: {code}'.format(code=str(response.status_code)))

        return None


def send_request_set_schema(key, schema_name, schema, old_name=None):
    data = {
        "key": key,
        "set": {
            "schema_name": schema_name,
            "schema": base64.b64encode(schema.encode('utf-8')).decode('utf-8')
        }
    }
    if old_name and old_name != schema_name:
        data['set']['old_name'] = old_name
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_schema'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while set schema: {error}'.format(error=str(e)))
        return None

    if not response.ok:
        logger.error(
            'An error occurred while set schema with response code: {code}'.format(code=str(response.status_code)))
        return False
    elif '"status": "fail"' in response.text:
        try:
            desc = json.loads(response.text).get('description', 'Not description')
        except JSONDecodeError:
            desc = 'Not description'
        logger.error(
            'An error occurred while set schema with response status fail. Error: {er}'.format(er=desc))
        return False
    else:
        return True


def send_request_del_schema(key, schema_name):
    data = {
        "key": key,
        "del": {
            "schema_name": schema_name
        }
    }
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_schema'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while del schema: {error}'.format(error=str(e)))
        return None

    if response.ok:
        return True
    else:
        logger.error(
            'An error occurred while delete schema with response code: {code}'.format(code=str(response.status_code)))

        return False


def send_request_add_update_path(key, schema_name, url, schema):
    data = {
        "key": key,
        'set': {
            "schema_name": schema_name,
            "url": url,
            "schema": base64.b64encode(schema).decode('utf-8')
        }
    }
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_schema_paths'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while set path schema: {error}'.format(error=str(e)))
        return None

    if response.ok:
        return True
    else:
        logger.error(
            'An error occurred while set path with response code: {code}'.format(code=str(response.status_code)))

        return False


def send_request_del_path(key, schema_name, url):
    data = {
        "key": key,
        'del': {
            "schema_name": schema_name,
            "url": url,
        }
    }
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_schema_paths'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while set path schema: {error}'.format(error=str(e)))
        return None

    if response.ok:
        return True
    else:
        logger.error(
            'An error occurred while del path with response code: {code}'.format(code=str(response.status_code)))

        return False


def send_request_set_openapi_url(key):
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_url'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json={"key": key}
        )
    except Exception as e:
        logger.error(
            'An error occurred while set url: {error}'.format(error=str(e)))
        return None
    if response:
        return response
    else:
        return None


def send_request_get_openapi_paths(key):
    try:
        response = requests.post(
            url_normalize(api_url + '/get_openapi_schema_paths'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json={"key": key},
        )
    except Exception as e:
        logger.error(
            'An error occurred while get paths schema: {error}'.format(error=str(e)))
        return None

    if response.ok:
        try:
            return json.loads(response.text)
        except JSONDecodeError:
            return None
    else:
        logger.error(
            'An error occurred while get paths with response code: {code}'.format(code=str(response.status_code)))

        return None


def send_request_get_exclude_path(key):
    try:
        response = requests.post(
            url_normalize(api_url + '/get_openapi_exclude_paths'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json={'key': key},
        )
    except Exception as e:
        logger.error(
            'An error occurred while get exclude paths: {error}'.format(error=str(e)))
        return None
    if response.ok:
        excluded_openapi_urls = json.loads(response.text).get('exclude_paths')
        return excluded_openapi_urls
    else:
        logger.error('An error occurred while get exclude paths with response code: {code}'.format(
            code=str(response.status_code)))

        return None


def send_request_set_exclude_path(key, urls):
    data = {
        "key": key,
        'set': {
            "exclude_paths": urls,
        }
    }
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_exclude_paths'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while set exclude path: {error}'.format(error=str(e)))
        return None

    if response.ok:
        return True
    else:
        logger.error('An error occurred while set exclude paths with response code: {code}'.format(
            code=str(response.status_code)))

        return False


def send_request_set_openapi_urls(key, urls):
    data = {
        "key": key,
        'url': urls
    }
    try:
        response = requests.post(
            url_normalize(api_url + '/set_openapi_url'),
            headers={'User-agent': USER_AGENT},
            proxies=proxy_dict_api,
            timeout=30,
            json=data,
        )
    except Exception as e:
        logger.error(
            'An error occurred while set url: {error}'.format(error=str(e)))
        return None

    if response.ok:
        return True
    else:
        logger.error(
            'An error occurred while set urls with response code: {code}'.format(code=str(response.status_code)))
        return False
