import json
import logging

from datetime import (
    datetime,
    timedelta,
    date
)
from threading import Lock
from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.db import connections
from django.db.models import (
    Count,
    DateTimeField,
    QuerySet,
    Q
)
from django.db.models.functions import Trunc
from django.utils.translation import gettext as _
from pymemcache.client.base import Client
from main.common import (
    get_location_by_ip,
    get_filtered_logs,
    attack_short_count,
    getting_geoip_multithread,
    adding_token_to_dict_for_geoip,
    setter_geotokens_for_logs,
    chunks_for_queryset
)
from main.models import Geo_Tokens
from main.models.logs import Logs
from main.models.wafuser import WAFUSER
from main.search import search_filter
from main.utils import (
    is_local,
    convert_from_punycode,
    convert_to_punycode
)

try:
    from uwsgidecorators import spool

    spoolIsrun = True
except Exception:
    spoolIsrun = False


    def spool(func):
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

logger = logging.getLogger(__name__)

vAPP = settings.V_APP

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))


def set_semaphore_val(semaphore_key, mcache: Client, delta=0):
    ts = datetime.timestamp(datetime.now())
    mcache.set(semaphore_key, ts + delta)


def get_token_by_ip(ip: str, tokens_dict: dict[list]):
    for k, v in tokens_dict.items():
        if ip in v:
            return k


def setter_geotokens(logs):
    geo_tokens = {}
    for token in Geo_Tokens.objects.using('waf').all():
        geo_tokens[token.waf_id] = token.geo_token

    for i in logs:
        if hasattr(i, "alias"):
            aliases_valid = set(i.alias) & set(geo_tokens.keys())
            if len(aliases_valid):
                alias = aliases_valid.pop()
            else:
                alias = None

            i.alias = alias
            i.geotoken = geo_tokens.get(alias)
        else:
            aliases_valid = set(i.get('alias')) & set(geo_tokens.keys())
            if len(aliases_valid):
                alias = aliases_valid.pop()
            else:
                alias = None

            i['alias'] = alias
            i["geotoken"] = geo_tokens.get(alias)


def processing_top10_rows_by_geoip(
        rows,
        res: list,
        mcache: Client,
        lock: Lock,
        tokens_dict: dict[list],
        semaphore_key: str
):
    for i in rows:

        token = get_token_by_ip(i.get('ip'), tokens_dict)

        geo_info = get_location_by_ip(
            mcache, i.get('ip'), token=token, check_token=True
        )
        if str(geo_info['mobile']).lower() == 'none' or \
                str(geo_info['mobile']).lower() == 'false':
            mobile = 'No'
        else:
            mobile = 'Yes'

        if str(geo_info['proxy']).lower() == 'none' or \
                str(geo_info['proxy']).lower() == 'false':
            proxy = 'No'
        else:
            proxy = 'Yes'

        if str(geo_info['tor']).lower() == 'none' or \
                str(geo_info['tor']).lower() == 'false':
            tor = 'No'
        else:
            tor = 'Yes'

        if str(geo_info['hosting']).lower() == 'none' or \
                str(geo_info['hosting']).lower() == 'false':
            hosting = 'No'
        else:
            hosting = 'Yes'

        if str(geo_info['crawler']).lower() == 'none' or \
                str(geo_info['crawler']).lower() == 'false':
            crawler = 'No'
        else:
            crawler = 'Yes'

        org = ''
        if len(str(geo_info['org'])) > 1:
            org = 'Org: ' + str(geo_info['org']) + '\n'
        isp = ''
        if len(str(geo_info['isp'])) > 1:
            isp = 'ISP: ' + str(geo_info['isp'])

        if geo_info['cc'] in (None, 'none', 'None'):
            if not is_local(i.get('ip')):
                geo_info['geo'] = _('Country not defined')
            else:
                geo_info['geo'] = _('Private IP')
        else:
            geo_info['geo'] = "{0}, {1} ({2})".format(
                str(geo_info['city']),
                str(geo_info['country']),
                str(geo_info['cc'])
            )

        if token:
            r = [
                i.get('ip'),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('ca'),
                i.get('ddos'),
                i.get('flood'),

                "Geo: " + geo_info['geo'] + '\n' +
                'Mobile: ' + mobile + '\n' +
                'Tor: ' + tor + '\n' +
                'Proxy: ' + proxy + '\n' +
                'Hosting: ' + hosting + '\n' +
                'Crawler: ' + crawler + '\n' +
                org + isp,
                str(geo_info['cc']).lower()
            ]
        else:
            r = [
                i.get('ip'),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('av'),
                i.get('ddos'),
                i.get('flood'),

                "Geo: " + 'GeoIP information available only for commercial version' + '\n' +
                'Mobile: ' + mobile + '\n' +
                'Proxy: ' + proxy + '\n' +
                'Hosting: ' + hosting + '\n' +
                'Crawler: ' + crawler + '\n' +
                org + isp,
                str('none').lower()
            ]

        with lock:
            res.append(r)

        set_semaphore_val(semaphore_key, mcache)


def get_last_time_rows(main_query: QuerySet):
    res_query = main_query.values(all_attacks=Count('bt'))\
        .order_by('-all_attacks').annotate(
        sa=Count('bt', filter=Q(bt__in=[1, 2])),
        na=Count('bt', filter=Q(bt__in=[3, 8])),
        bf=Count('bt', filter=Q(bt=7)),
        flood=Count('bt', filter=Q(bt=9)),
        ddos=Count('bt', filter=Q(bt=10)),
        av=Count('bt', filter=Q(bt=4)),
        OpenAPI=Count('bt', filter=Q(bt=14)),
        Blocked=Count('bt', filter=Q(bt=16)),
    )

    res = ''
    if res_query.count() > 0:
        for row in res_query:
            res += attack_short_count(row.get('all_attacks')) + ";" + attack_short_count(row.get('sa')) + ";" + \
                   attack_short_count(row.get('na')) + ";" + attack_short_count(row.get('bf')) + ";" + \
                   attack_short_count(row.get('flood')) + ";" + \
                   attack_short_count(row.get('ddos')) + ";" + attack_short_count(row.get('av')) + ";" + \
                   attack_short_count(row.get('OpenAPI')) + ";" + attack_short_count(row.get('Blocked')) + ";"
    else:
        res += '0' + ";" + '0' + ";" + '0' + ";" + '0' + ";" + '0' + ";" + '0' + \
               ";" + '0' + ";" + '0' + ";" + '0' + ";"

    return res


def processing_ip_data_by_geoip(
    rows,
    res: list,
    mcache: Client,
    lock: Lock,
    tokens_dict: dict[list],
    semaphore_key: str
):
    for i in rows:
        token = get_token_by_ip(i.get('ip'), tokens_dict)

        geo_info = get_location_by_ip(
            mcache, i.get('ip'), token=token, check_token=True
            )
        if str(geo_info['mobile']).lower() == 'none' or \
                str(geo_info['mobile']).lower() == 'false':
            mobile = 'No'
        else:
            mobile = 'Yes'

        if str(geo_info['tor']).lower() == 'none' or \
                str(geo_info['tor']).lower() == 'false':
            tor = 'No'
        else:
            tor = 'Yes'

        if str(geo_info['proxy']).lower() == 'none' or \
                str(geo_info['proxy']).lower() == 'false':
            proxy = 'No'
        else:
            proxy = 'Yes'

        if str(geo_info['hosting']).lower() == 'none' or \
                str(
                    geo_info['hosting']
                    ).lower() == 'false':
            hosting = 'No'
        else:
            hosting = 'Yes'

        if str(geo_info['crawler']).lower() == 'none' or \
                str(
                    geo_info['crawler']
                    ).lower() == 'false':
            crawler = 'No'
        else:
            crawler = 'Yes'

        last24h_count = attack_short_count(
            int(i.get('count_24h', 0)), 0
        )

        org = ''
        if len(str(geo_info['org'])) > 1:
            org = 'Org: ' + str(geo_info['org']) + '\n'
        isp = ''
        if len(str(geo_info['isp'])) > 1:
            isp = 'ISP: ' + str(geo_info['isp'])

        if geo_info['cc'] in (None, 'none', 'None'):
            if not is_local(i.get('ip')):
                geo_info['geo'] = _('Country not defined')
            else:
                geo_info['geo'] = _('Private IP')
        else:
            geo_info['geo'] = "{0}, {1} ({2})".format(
                str(geo_info['city']),
                str(geo_info['country']),
                str(geo_info['cc'])
            )
        if token:
            r = [
                i.get('ip'),
                i.get('all_attacks'),
                i.get('bt1'),
                i.get('bt2'),
                i.get('bt3'),
                i.get('bt4'),
                i.get('bt7'),
                i.get('bt8'),
                i.get('bt9'),
                i.get('bt10'),

                "Geo: " + geo_info['geo'] + '\n' +
                'Mobile: ' + mobile + '\n' +
                'Tor: ' + tor + '\n' +
                'Proxy: ' + proxy + '\n' +
                'Hosting: ' + hosting + '\n' +
                'Crawler: ' + crawler + '\n' +
                org + isp,
                '(' + str(last24h_count) + ' in 24 h.)',
                str(geo_info['cc']).lower(),
            ]
        else:
            r = [
                i.get('ip'),
                i.get('all_attacks'),
                i.get('bt1'),
                i.get('bt2'),
                i.get('bt3'),
                i.get('bt4'),
                i.get('bt7'),
                i.get('bt8'),
                i.get('bt9'),
                i.get('bt10'),

                "Geo: " + 'GeoIP information available only for commercial version' + '\n' +
                'Mobile: ' + mobile + '\n' +
                'Proxy: ' + proxy + '\n' +
                'Hosting: ' + hosting + '\n' +
                'Crawler: ' + crawler + '\n' +
                '' + '',
                '(' + str(last24h_count) + ' in 24 h.)',
                str('none').lower(),
            ]

        with lock:
            res.append(r)

        set_semaphore_val(semaphore_key, mcache)


def get_top10(main_query: QuerySet, mcache: Client, semaphore_key):

    query = main_query.values('ip').annotate(
        all_attacks=Count('bt'),
        alias=ArrayAgg('alias', distinct=True),
        sa=Count('bt', filter=Q(bt__in=[1, 2])),
        na=Count('bt', filter=Q(bt__in=[3, 8])),
        bf=Count('bt', filter=Q(bt=7)),
        av=Count('bt', filter=Q(bt=4)),
        flood=Count('bt', filter=Q(bt=9)),
        ddos=Count('bt', filter=Q(bt=10)),
        OpenAPI=Count('bt', filter=Q(bt=14)),
        Blocked=Count('bt', filter=Q(bt=16)),
    ).order_by('-all_attacks')[:10]

    setter_geotokens(query)

    preprocessing_ips_for_geoip(
        query, mcache
    )

    token_dict = {}
    lock = Lock()

    adding_token_to_dict_for_geoip(query, token_dict)

    set_semaphore_val(semaphore_key, mcache)

    res = []
    processing_top10_rows_by_geoip(
        query, res, mcache, lock, token_dict, semaphore_key
    )

    return json.dumps(res)


def get_top10_host(main_query: QuerySet, mcache: Client, semaphore_key):
    host_top10_data = main_query.values('vhost').annotate(
        all_attacks=Count('bt'),
        sa=Count('bt', filter=Q(bt__in=[1, 2])),
        na=Count('bt', filter=Q(bt__in=[3, 8])),
        bf=Count('bt', filter=Q(bt=7)),
        av=Count('bt', filter=Q(bt=4)),
        OpenAPI=Count('bt', filter=Q(bt=14)),
        Blocked=Count('bt', filter=Q(bt=16)),
    ).order_by('-all_attacks')[:17]

    res = list()
    for i in host_top10_data:
        res.append(
            [
                convert_from_punycode(i.get('vhost')),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('av'),
                i.get('OpenAPI'),
                i.get('Blocked'),
            ]
        )

        set_semaphore_val(semaphore_key, mcache)

    return json.dumps(res)


def get_ip_info_data(
    main_query: QuerySet,
    mcache: Client,
    semaphore_key: str,
    yesterday_date,
    end_date
):
    query = main_query.values('ip').annotate(
        all_attacks=Count('bt'),
        alias=ArrayAgg('alias', distinct=True),
        count_24h=Count(
            'bt',
            filter=Q(
                timestamp__gte=yesterday_date, timestamp__lte=end_date
            )
        ),
        bt1=Count('bt', filter=Q(bt=1)),
        bt2=Count('bt', filter=Q(bt=2)),
        bt3=Count('bt', filter=Q(bt=3)),
        bt4=Count('bt', filter=Q(bt=4)),
        bt7=Count('bt', filter=Q(bt=7)),
        bt8=Count('bt', filter=Q(bt=8)),
        bt9=Count('bt', filter=Q(bt=9)),
        bt10=Count('bt', filter=Q(bt=10)),

    ).order_by('-all_attacks')[:15]

    setter_geotokens(query)

    preprocessing_ips_for_geoip(
        query, mcache
    )

    token_dict = {}
    lock = Lock()

    setter_geotokens_for_logs(query)

    adding_token_to_dict_for_geoip(query, token_dict)

    set_semaphore_val(semaphore_key, mcache)

    res = []
    processing_ip_data_by_geoip(
        query, res, mcache, lock, token_dict, semaphore_key
    )

    return json.dumps(res)


def get_linechart(main_query: QuerySet, mcache: Client, semaphore_key):
    linechart_data = main_query.annotate(
        date_line=Trunc('timestamp', 'day', output_field=DateTimeField())
    ).values('date_line').order_by('date_line').annotate(
        all_attacks=Count('bt'),
        sa=Count('bt', filter=Q(bt__in=[1, 2])),
        na=Count('bt', filter=Q(bt__in=[3, 8])),
        bf=Count('bt', filter=Q(bt=7)),
        ddos=Count('bt', filter=Q(bt=10)),
        flood=Count('bt', filter=Q(bt=9)),
        av=Count('bt', filter=Q(bt=4)),
        OpenAPI=Count('bt', filter=Q(bt=14)),
        Blocked=Count('bt', filter=Q(bt=16))
    )

    res = list()
    for i in linechart_data:
        res.append(
            [
                i.get('date_line').strftime("%d.%m.%Y"),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('flood'),
                i.get('ddos'),
                i.get('av'),
                i.get('OpenAPI'),
                i.get('Blocked'),
                0,
            ]
        )

        set_semaphore_val(semaphore_key, mcache)

    return json.dumps(res)


def preprocessing_ips_for_geoip(main_query: QuerySet, mcache: Client):
    getting_geoip_multithread(
        main_query, mcache, max_workers=1,
        f_chunking_add_token=chunks_for_queryset
    )


def finalize_gen_charts(
    res,
    mcache: Client,
    mcache_key,
    semaphore_key,
    time_save=3600
):
    mcache.set(mcache_key, res, time_save)
    mcache.delete(semaphore_key)


@spool
def gen_charts(args):

    if spoolIsrun:
        connections.close_all()

    req_get = json.loads(args['req_get'])

    # DATES
    get_start_date = req_get.get('start')
    get_end_date = req_get.get('end')

    if not get_start_date:
        return

    start_date = datetime.strptime(get_start_date, '%d.%m.%Y').strftime('%Y-%m-%d')
    end_date = datetime.strptime(get_end_date, '%d.%m.%Y').strftime('%Y-%m-%d')

    today_date = date.today()
    formatted_summary_end_date = datetime.strptime(
        end_date + ' 23:59:59.000000', '%Y-%m-%d %H:%M:%S.%f'
    ).date()
    if formatted_summary_end_date > today_date:
        str_end_date = str(today_date) + ' 23:59:59.000000'
    else:
        str_end_date = end_date + ' 23:59:59.000000'

    dt_end_datetime = datetime.strptime(
        str_end_date, '%Y-%m-%d %H:%M:%S.%f'
    )
    dt_end_date = dt_end_datetime.date()
    dt_yesterday_date = dt_end_date - timedelta(days=1)
    dt_last_week = dt_end_date - timedelta(days=7)
    dt_last_month = dt_end_date - timedelta(days=30)

    # Another variables
    user_id = int(args['user_id'])
    state = req_get.get('state', '').strip()
    s = req_get.get('search', '').strip()

    mcache = Client(cache_address)
    semaphore_mcache_key = str(user_id) + 'summary' + vAPP
    mcache_key = str(
        str(user_id) + '|'
        + str(end_date) + '|'
        + str(start_date) + '|'
        + convert_to_punycode(s) + '|'
        + state + '|'
        + vAPP
    ).replace(' ', '_')

    set_semaphore_val(semaphore_mcache_key, mcache)

    q = req_get.get('domain')
    if q == 'None':
        q = None

    # waf id
    wafuser = WAFUSER.objects.get(id=user_id)
    user_alias = wafuser.waf_alias

    # waf domain
    domain_from_wafuser = wafuser.domain

    search_history = wafuser.search_history
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    if s is not None and s != '':
        if 'charts:' + s in search_history_dict.keys():
            reb = dict()
            reb['charts:' + s] = search_history_dict['charts:' + s] + 1
            search_history_dict.pop('charts:' + s)
            reb.update(search_history_dict)
            search_history_dict = reb

        else:
            reb = dict()
            reb['charts:' + s] = 1
            reb.update(search_history_dict)
            search_history_dict = reb
    else:
        for el in search_history_dict.keys():
            if 'charts:' in el:
                s = el.replace('charts:', '')
                break
    if s == '#clean':
        s = ''

    while len(search_history_dict) > 1000:
        search_history_dict.popitem()

    wafuser.search_history = json.dumps(search_history_dict)
    wafuser.save()

    charts_history = []

    set_semaphore_val(semaphore_mcache_key, mcache)

    sortbyfreq = True
    if sortbyfreq:
        tmp = dict()
        last_use = ''
        for el in list(search_history_dict.keys()):
            if 'charts:' in el and 'charts:' + s != el:
                last_use = el
                break

        if last_use in search_history_dict:
            tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
            search_history_dict.pop(last_use)
        tmp.update(
            {k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
        search_history_dict = tmp

    for el in search_history_dict.keys():
        if 'charts:' in el and '#clean' not in el and s.lower() in el.replace('charts:', '').lower():
            charts_history.append(el.replace('charts:', ''))

    charts_history_json = json.dumps(charts_history)

    main = Logs.objects.using('waf').only(
        'vhost', 'alias', 'timestamp', 'ip'
    ).filter(
        timestamp__gte=dt_last_month, timestamp__lte=dt_end_datetime
    ).exclude(possible=True).exclude(bt=5)

    set_semaphore_val(semaphore_mcache_key, mcache)

    main = get_filtered_logs(main, user_alias, domain_from_wafuser, q)

    charts = main.values('vhost', 'alias', 'timestamp').filter(
        timestamp__gte=start_date, timestamp__lte=dt_end_datetime
    ).exclude(possible=True)

    # If the state of the checkbox on the chart has changed Top by IP
    if state:

        s = s.strip()
        if s:
            try:
                charts = search_filter(s, charts)
            except Exception:
                charts = Logs.objects.using('waf').none()

        # if not all checkboxs active
        split_state = []
        if state != 'clear':
            split_state = state.split('-')

        exclude_bt = []
        for i in split_state:
            if i == 'Signature Analysis':
                exclude_bt.append(1)
                exclude_bt.append(2)
            if i == 'Nemesida MLA':
                exclude_bt.append(3)
            if i == 'Bruteforce':
                exclude_bt.append(7)
            if i == 'Antivirus':
                exclude_bt.append(4)
            if i == 'Nemesida MLC':
                exclude_bt.append(8)
            if i == 'flood':
                exclude_bt.append(9)
            if i == 'Ddos':
                exclude_bt.append(10)
            if i == 'OpenAPI':
                exclude_bt.append(14)
            if i == 'Blocked':
                exclude_bt.append(16)

        charts = charts.values('ip').filter(
            bt__in=[1, 2, 3, 4, 7, 8, 9, 10, 14, 16]
        ).exclude(Q(bt__in=exclude_bt))

        ip_info_json = get_ip_info_data(
            charts, mcache, semaphore_mcache_key, dt_yesterday_date,
            dt_end_datetime
        )

        finalize_gen_charts(
            ip_info_json, mcache, mcache_key, semaphore_mcache_key,
            time_save=600
        )
        return ip_info_json

    main_base = main.filter(
        timestamp__gte=dt_last_month, timestamp__lte=dt_end_datetime
    )

    today_summary_base = main_base.filter(timestamp__gte=dt_end_date)
    last24_summary_base = main_base.filter(
        timestamp__gte=dt_yesterday_date, timestamp__lte=dt_end_datetime
    )
    last_week_summary_base = main_base.filter(
        timestamp__gte=dt_last_week, timestamp__lte=dt_end_datetime
    )
    last_month_summary_base = main_base.filter(
        timestamp__gte=dt_last_month, timestamp__lte=dt_end_datetime
    )

    set_semaphore_val(semaphore_mcache_key, mcache)

    s = req_get.get('search', '').strip()
    if s:
        try:
            charts = search_filter(s, charts)
            today_summary_base = search_filter(s, today_summary_base)
            last24_summary_base = search_filter(s, last24_summary_base)
            last_week_summary_base = search_filter(s, last_week_summary_base)
            last_month_summary_base = search_filter(s, last_month_summary_base)
        except Exception:
            charts = Logs.objects.using('waf').none()
            today_summary_base = Logs.objects.using('waf').none()
            last24_summary_base = Logs.objects.using('waf').none()
            last_week_summary_base = Logs.objects.using('waf').none()
            last_month_summary_base = Logs.objects.using('waf').none()

    set_semaphore_val(semaphore_mcache_key, mcache)

    charts = charts.values('ip', 'timestamp', 'vhost', 'alias')
    top10 = last24_summary_base.values('vhost', 'bt', 'alias').annotate(
        count=Count('id')
    ).filter(bt__in=[1, 2, 3, 4, 7, 8, 9, 10, 14, 16])

    # stub for unused data
    today_summary_list = list()
    last24_summary_list = list()
    last_week_summary_list = list()
    last_month_summary_list = list()
    brute_force_summary_list = list()
    sms_flood_summary_list = list()
    ddos_summary_list = list()
    internal_errors_summary_list = list()
    description_summary_list = list()

    today_summary_json = json.dumps(today_summary_list)
    last24_summary_json = json.dumps(last24_summary_list)
    last_week_summary_json = json.dumps(last_week_summary_list)
    last_month_summary_json = json.dumps(last_month_summary_list)
    brute_force_summary_json = json.dumps(brute_force_summary_list)
    sms_flood_summary_json = json.dumps(sms_flood_summary_list)
    ddos_summary_json = json.dumps(ddos_summary_list)
    internal_errors_summary_json = json.dumps(internal_errors_summary_list)
    description_summary_json = json.dumps(description_summary_list)

    # result data

    host_top10_json = get_top10_host(top10, mcache, semaphore_mcache_key)
    attackers_top10_json = get_top10(top10, mcache, semaphore_mcache_key)
    linechart = get_linechart(charts, mcache, semaphore_mcache_key)
    ip_info_json = get_ip_info_data(
        charts, mcache, semaphore_mcache_key, dt_yesterday_date,
        dt_end_datetime
    )
    last_search_json = json.dumps(list(search_history_dict.keys())[0])

    res = str(
        linechart + ';'
        + host_top10_json + ";"
        + attackers_top10_json + ";"
        + today_summary_json + ";"
        + last24_summary_json + ";"
        + last_week_summary_json + ";"
        + last_month_summary_json + ";"
        + brute_force_summary_json + ";"
        + sms_flood_summary_json + ";"
        + internal_errors_summary_json + ";"
    )

    res += get_last_time_rows(today_summary_base)
    res += get_last_time_rows(last24_summary_base)
    res += get_last_time_rows(last_week_summary_base)
    res += get_last_time_rows(last_month_summary_base)

    res += str(
        ip_info_json + ";"
        + description_summary_json + ";"
        + ddos_summary_json + ";"
        + last_search_json + ";"
        + charts_history_json + ";"
    )

    finalize_gen_charts(
        res, mcache, mcache_key, semaphore_mcache_key,
        time_save=3600
    )

    return res
