import idna
import ipaddress
import logging
import math
import psycopg2
import re
import requests

from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models import F
from django.db.models import Min, Max
from time import sleep


logger = logging.getLogger(__name__)


list_private_networks = [
    "0.0.0.0/8",
    "10.0.0.0/8",
    "100.64.0.0/10",
    "127.0.0.0/8",
    "169.254.0.0/16",
    "172.16.0.0/12",
    "192.0.0.0/24",
    "192.0.2.0/24",
    "192.88.99.0/24",
    "192.168.0.0/16",
    "198.18.0.0/15",
    "198.51.100.0/24",
    "203.0.113.0/24",
    "224.0.0.0/4",
    "240.0.0.0/4",
    "192.0.0.0/24",
    "::/128",
    "::1/128",
    "::ffff:0:0/96",
    "::ffff:0:0:0/96",
    "64:ff9b::/96",
    "64:ff9b:1::/48",
    "100::/64",
    "2001:0000::/32",
    "2001:20::/28",
    "2001:db8::/32",
    "2002::/16",
    "fc00::/7",
    "fe80::/10",
    "ff00::/8"
]


def is_local(ip):
    try:
        ip = ipaddress.ip_address(ip)
        for network in list_private_networks:
            if ip in ipaddress.ip_network(network):
                return True
        return False
    except Exception:
        return False


def get_params(group):
    if group:
        params = {"body_equal": ArrayAgg('body'),
                  "referer_equal": ArrayAgg('referer'),
                  "method_equal": ArrayAgg('method'),
                  "url_equal": ArrayAgg('url'),
                  "param_equal": ArrayAgg('param'),
                  "ua_equal": ArrayAgg('ua'),
                  "cookie_equal": ArrayAgg('cookie'),
                  "timestamp__max": Max('timestamp'),
                  "timestamp__min": Min('timestamp'),
                  "timestamp": ArrayAgg('timestamp'),
                  "ip": ArrayAgg('ip'),
                  "id": Max('id')}
    else:
        params = {"body_equal": F('body'),
                  "referer_equal": F('referer'),
                  "method_equal": F('method'),
                  "url_equal": F('url'),
                  "param_equal": F('param'),
                  "ua_equal": F('ua'),
                  "cookie_equal": F('cookie'),
                  "timestamp__max": F('timestamp'),
                  "timestamp__min": F('timestamp'),
                  "timestamp": F('timestamp'),
                  "ip": F('ip'),
                  "id": F('id')}

    return params


def convert_size(size_bytes):
    if size_bytes == 0 or size_bytes is None:
        return "-"
    if size_bytes < 100:
        result = size_bytes * 8
        return '{result} b/s'.format(result=result)
    elif size_bytes < 10000:
        result = round(size_bytes * 0.0078125, 2)
        return '{result} Kb/s'.format(result=result)
    else:
        result = round(size_bytes * 7.6294 * math.pow(10, -6), 2)
        return '{result} Mb/s'.format(result=result)


def filter_upstreams(names, exp=None):
    ups = []
    if (exp is None or
            exp == '*' or
            exp == ''):
        ups = names
    else:
        for d in exp.split(', '):
            for u in names:
                if d[0] == "*":
                    if d[2:] == u:
                        ups.append(u)
                    if d[1:] in u:
                        ups.append(u)
                elif d == u:
                    ups.append(u)
        ups = set(ups)
    return ups


def get_update_time():
    headers = {'User-Agent': settings.USER_AGENT}
    try:
        r = requests.get(settings.URI_RL_TS, headers=headers, proxies=get_proxy_conf_dict(), timeout=30)
        return int(r.text)
    except Exception as e:
        logger.error(
            'An error occurred while getting information about the timestamp of the rules. Error information: {err}'
            .format(err=e)
        )
        return None


def remove_zero_from_ip(ip):
    if type(ipaddress.ip_address(ip)) is ipaddress.IPv4Address:
        no_zero_ip = re.sub(r"\b0+(\d)", r'\1', str(ip))
    else:
        no_zero_ip = re.sub(r"\b0+(\d)", r'\1', str(ip))
    return no_zero_ip


def get_list_messages(storage):
    list_messages = list()
    for message in storage:
        mes = dict()
        mes['message'] = message.message
        mes['tags'] = message.tags
        mes['level'] = message.level
        list_messages.append(mes)
    return list_messages


def convert_to_punycode(text: str) -> str:
    if isinstance(text, str):
        list_cyrillic = find_cyrillic_words(text)

        for word in list_cyrillic:
            try:
                word_punycode = idna.encode(word).decode()
                text = text.replace(word, word_punycode)
            except idna.IDNAError as e:
                logger.error(
                    'An error occurred when converting the {word} string to Punycode. Error information: {err}'
                    .format(err=e, word=word)
                )

    return text


def convert_from_punycode(text: str) -> str:
    if isinstance(text, str):
        punycode_matches = find_punycode_words(text)

        for word_punycode in punycode_matches:
            try:
                word = idna.decode(word_punycode)
                text = text.replace(word_punycode, word)
            except idna.IDNAError as e:
                logger.error(
                    'An error occurred when converting the Punycode {word} string to Cyrillic information: {err}'
                    .format(err=e, word=word_punycode)
                )

    return text


def recursion_convert_to_punycode_obj(obj, with_keys=False):
    if isinstance(obj, str):
        res = convert_to_punycode(obj)
    elif isinstance(obj, list):
        res = []
        for i in obj:
            if isinstance(i, str):
                i = convert_to_punycode(i)
            else:
                i = recursion_convert_to_punycode_obj(i, with_keys=with_keys)

            res.append(i)
    elif isinstance(obj, dict):
        res = {}
        for k, v in obj.items():
            if isinstance(v, str):
                v = convert_to_punycode(v)
            else:
                v = recursion_convert_to_punycode_obj(v, with_keys=with_keys)

            if with_keys:
                k = recursion_convert_from_punycode_obj(k, with_keys=with_keys)

            res[k] = v
    else:
        res = obj

    return res


def recursion_convert_from_punycode_obj(obj, with_keys=False):
    if isinstance(obj, str):
        res = convert_from_punycode(obj)
    elif isinstance(obj, list):
        res = []
        for i in obj:
            if isinstance(i, str):
                i = convert_from_punycode(i)
            else:
                i = recursion_convert_from_punycode_obj(i, with_keys=with_keys)

            res.append(i)
    elif isinstance(obj, dict):
        res = {}
        for k, v in obj.items():
            if isinstance(v, str):
                v = convert_from_punycode(v)
            else:
                v = recursion_convert_from_punycode_obj(v, with_keys=with_keys)

            if with_keys:
                k = recursion_convert_from_punycode_obj(k, with_keys=with_keys)

            res[k] = v
    else:
        res = obj

    return res


def find_cyrillic_words(text):
    cyrillic_template = r"[а-яА-Я]+[а-яА-Я\d]*(?:\.[а-яА-Я\d]+)+\b"

    return re.findall(cyrillic_template, text)


def find_punycode_words(text):
    punycode_pattern = r'xn--[a-zA-Z0-9-]+\.[a-zA-Z0-9-]+\b'

    return re.findall(punycode_pattern, text)


def get_proxy_conf_dict():
    """
    Function to get a main proxy dictionary
    :return: proxy dictionary
    """
    if settings.HTTP_PROXY_CONF == '':
        return {
            "http": None,
            "https": None
        }

    try:
        http_proxy = settings.HTTP_PROXY_CONF
        https_proxy = settings.HTTP_PROXY_CONF
        proxy_dict = {
            "http": http_proxy,
            "https": https_proxy
        }
    except AttributeError as e:
        logger.info('Failed to set proxy with error: {}'.format(e))
        proxy_dict = {
            "http": None,
            "https": None
        }
    return proxy_dict


def get_proxy_api_dict():
    """
    Function to get a proxy API dictionary
    :return: proxy dictionary
    """
    if settings.API_PROXY == '':
        return {
            "http": None,
            "https": None
        }

    try:
        http_proxy = settings.API_PROXY
        https_proxy = settings.API_PROXY
        proxy_dict = {
            "http": http_proxy,
            "https": https_proxy
        }
    except AttributeError as e:
        logger.info('Failed to set proxy with error: {}'.format(e))
        proxy_dict = {
            "http": None,
            "https": None
        }
    return proxy_dict


def init_db_connection(db_name, db_user, db_pass, db_host, db_port):
    while True:
        try:

            connect = psycopg2.connect(
                database=db_name,
                user=db_user,
                password=db_pass,
                host=db_host,
                port=db_port,
                connect_timeout=10
            )
            connect.set_session(autocommit=True)
            return connect

        except Exception as e:
            logging.error('An error occurred while initializing the connection to the database {}: {}'.format(db_name, e))
            sleep(5)


def close_db_connection(cursor, connect):
    try:
        cursor.close()
        connect.close()
    except Exception as e:
        logging.error('An error occurred while closing the connection to the database: {}'.format(e))
