import os
import json
import logging
import queue
import threading
import time
import operator

import requests
import multiprocessing

from datetime import datetime, timedelta, date
from pathlib import Path
from pymemcache.client.base import Client
from url_normalize import url_normalize
from urllib.parse import unquote

from django import db
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db.models import Count, DateTimeField, CharField, Max
from django.db.models.functions import Cast, TruncDate, Trunc, Concat
from django.db.models import Q, F
from django.http import JsonResponse
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.utils.translation import pgettext
from django.db import connections
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist

from main.common import attack_short_count
from main.common import chunks
from main.common import draw_attackers_top10_tooltip
from main.common import draw_top_by_ip_tooltip
from main.common import get_country_for_group
from main.common import get_filtered_logs
from main.common import get_groups
from main.common import get_list_plans
from main.common import get_location_by_ip
from main.common import get_location_from_remote
from main.common import get_main_logs
from main.common import get_main_logs_new
from main.common import get_ml_table
from main.common import get_tags
from main.common import get_user_key_waf_id
from main.common import get_vhost
from main.common import parse_body
from main.decorators import start_end_time
from main.decorators import start_end_time_scanner
from main.decorators import start_end_time_var
from main.decorators import time_of_function
from main.emulator import NWAF_emu
from main.fields import FieldNames
from main.models.config import UserConfig
from main.models.logs import Logs
from main.models.my_waf import Geo_Tokens
from main.models.my_waf import LicKeys
from main.models.scan import ScannerReport
from main.models.scan import ScanRequest
from main.models.wafuser import WAFUSER
from main.models.white_list import WhitelistRequest
from main.search import scan_search_filter
from main.search import search_filter
from main.tasks.summary import gen_charts
from main.utils import convert_from_punycode
from main.utils import convert_to_punycode
from main.utils import get_list_messages
from main.utils import get_params
from main.utils import get_proxy_api_dict
from main.utils import get_proxy_conf_dict
from my_waf.utils import check_keys
from my_waf.utils import get_lic_keys
from my_waf.utils import get_rules_management
from my_waf.utils import get_waf_id
from my_waf.utils import send_set_settings
from my_waf.utils import set_dyn_bl
from utils.db import check_db

from utils.http import check_geo_ip_tokens

logger = logging.getLogger(__name__)

new_table_exist = check_db(logger)

try:
    api_url = settings.API_URI
except AttributeError:
    api_url = 'http://localhost:8080/nw-api/'

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))

PAGE_SIZE = settings.DEFAULT_PAGE_SIZE

vGeoIP = settings.V_GEOIP
vAPP = settings.V_APP

isCCEnable = None
try:
    r = "SELECT * FROM information_schema.columns WHERE column_name = 'cc' AND table_name = 'attack'"
    with connections['waf'].cursor() as cursor:
        cursor.execute(r)
        res_cc = cursor.fetchone()
        if res_cc is None:
            isCCEnable = False
        else:
            isCCEnable = True

except Exception as e:
    logger.error('Warning. Field "cc" is not exist in table "attack": {}'.format(e))
    isCCEnable = False


def scanner_settings(request):
    return render(request, 'main/scanner_settings.html')


def logging(request):
    return render(request, 'main/logging.html')


@time_of_function
def get_recheck_status(request):
    logs_id = json.loads(request.POST.get('logs_id'))
    logs_statuses = dict()
    list_id_logs = dict()
    for k, v in logs_id.items():
        list_id_logs[k.replace('Logs_', '')] = v

    if list_id_logs:
        logs = Logs.objects.select_related("scanrequest").filter(id__in=list_id_logs.keys())
        for log in logs:
            if log.scanrequest.status != list_id_logs[str(log.id)]:
                if log.scanrequest.status == 3:
                    logs_statuses['Logs_' + str(log.id) + '_desc'] = log.scanrequest.description
                logs_statuses['Logs_' + str(log.id)] = log.scanrequest.status
    return JsonResponse(logs_statuses)


@login_required
@time_of_function
def get_rule(request):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain
    rule_id = request.GET.get('rule_id', '')
    row_id = request.GET.get('row_id', '')
    row = Logs.objects.using('waf').filter(id=row_id)
    try:
        row = get_filtered_logs(row, user_alias, domain_from_wafuser, None).values('id', 'vhost', 'rule_id',
                                                                                   'mz').first()
    except IndexError:
        row = None

    if not row:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    rules_file = Path(path)
    if rules_file.is_file():
        emu = NWAF_emu(path)
        res = emu.get_rule(rule_id)

        if row:
            res["plan"] = 0
            res['wl_data'] = {'row_id': row['id']}
        else:
            res["plan"] = -1
            res["wl"] = ""
    else:
        res = {"status": "empty"}
    return JsonResponse(res)


@login_required
@time_of_function
def get_long_desc(request):
    row_id = request.GET.get('row_id', '')
    if not row_id:
        messages.error(request, pgettext('messages', 'Invalid parameters request.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    row = Logs.objects.using('waf').filter(id=row_id).only('long_description').first()
    if row:
        desc = row.long_description
    else:
        desc = None

    return JsonResponse({'long_description': desc})


@login_required
@time_of_function
def get_large_body(request):
    row_id = request.GET.get('row_id', '')
    if not row_id:
        messages.error(request, pgettext('messages', 'Invalid parameters request.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    row = Logs.objects.using('waf').filter(id=row_id).only('body').first()
    if row:
        body = parse_body(row.body)
        try:
            tmp = json.loads(body)
            body = json.dumps(tmp, sort_keys=True, indent=2)
        except json.JSONDecodeError:
            body = body
    else:
        body = None

    return JsonResponse({'body': body})


@login_required
@time_of_function
def get_rule_wl(request):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain
    row_id = request.GET.get('row_id', '')
    row = Logs.objects.using('waf').filter(id=row_id)
    try:
        row = get_filtered_logs(row, user_alias, domain_from_wafuser, None).values('vhost', 'rule_id', 'mz',
                                                                                   'alias').first()
    except IndexError:
        row = None

    if not row:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    rules_file = Path(path)
    res = {}
    if rules_file.is_file():
        res['wl_id'] = row['rule_id']
        res['domain'] = convert_from_punycode(row['vhost'])
        res['mz'] = str(row['mz']).lower()
    else:
        res = {"status": "empty"}
    res['waf_id'] = row['alias']

    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    for key in keys:
        try:
            response_check_key = requests.post(
                settings.URI_AUTH,
                proxies=get_proxy_conf_dict(),
                headers={'User-agent': settings.USER_AGENT},
                timeout=30,
                data={'key': key.key}
            )

            if response_check_key.status_code == 201:
                check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
                if check_key_result['status'] == 'success':
                    if check_key_result['waf_id'] == res['waf_id']:
                        res['key_id'] = key.id
        except Exception as error:
            logger.error(
                f'An error occurred while getting information about the key: {key}. Error information: {error}'
            )

    if type(res.get('key_id')) is not int and not str(res.get('key_id')).isnumeric():
        res['error'] = 'Invalid WAF ID. Perhaps this attack does not exist'
        messages.error(request, pgettext('messages', 'Invalid WAF ID. Perhaps this attack does not exist.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    return JsonResponse(res)


@login_required
@time_of_function
def add_wl(request):
    get_form = request.POST.get('form')
    get_extension_list = request.POST.getlist('extension[]') or []
    get_form_list = str(get_form).split('&') or []
    get_mz = []
    for el in get_form_list:
        if 'mz=' in el:
            el = str(el).replace('mz=', '')
            for x, y in ("body", "BODY"), ("url", "URL"), ("args", "ARGS"), ("headers", "HEADERS"), \
                    ("mla", "MLA"), ("no", "No"), ('ban', 'Ban'):
                el = el.replace(x, y)
            get_mz.append(el)
    mz = '|'.join(get_mz)
    get_form_dict = {}
    for el in get_form_list:
        el = el.split('=')
        get_form_dict[el[0]] = el[1]
    extension_list = []
    for el in get_extension_list:
        el = str(el).split(';')
        if el[0] != 'none':
            if el[0].lower() == 'body' or el[0].lower() == 'url' \
                    or el[0].lower() == 'args' or el[0].lower() == 'headers':
                zone = '$' + el[0].upper()
            else:
                zone = '$' + el[0]
            if el[2] == 'true':
                zone += '_X:'
            else:
                zone += ':'
            extension_list.append(zone + el[1])
    if 'active' in get_form_dict and get_form_dict['active'] == 'true':
        get_form_dict['active'] = True
    else:
        get_form_dict['active'] = False

    if len(extension_list) > 0:
        extension = '|'.join(extension_list)
        get_form_dict['extension'] = extension
    get_form_dict['mz'] = mz
    if 'other_zone' in get_form_dict:
        if get_form_dict['other_zone'] != '':
            other_zone = '|'.join(unquote(get_form_dict['other_zone']).split(','))
            if get_form_dict['mz'] != '':
                get_form_dict['mz'] = get_form_dict['mz'] + '|' + other_zone
            else:
                get_form_dict['mz'] = other_zone
            get_form_dict.pop('other_zone')
        else:
            get_form_dict.pop('other_zone')

    if get_form_dict.get('domain'):
        get_form_dict['domain'] = convert_to_punycode(unquote(get_form_dict['domain']))

    selected_key = get_form_dict.get('lic_keys_select')
    try:
        key = LicKeys.objects.get(id=selected_key).key
    except ObjectDoesNotExist:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    get_form_dict.pop('lic_keys_select')
    get_form_dict.pop('csrfmiddlewaretoken')
    full_settings_dict = {"key": key, 'add': get_form_dict}
    send_status = send_set_settings('WL', full_settings_dict)
    if send_status[0] == 'success':
        return HttpResponse('{"status": "success"}', content_type='application/json')
    else:
        messages.error(request, pgettext('messages', 'An occured error while adding WL rule'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@time_of_function
def ip_info(request):
    # memcached
    mcache = Client(cache_address)

    get_ip = request.GET.get('ip')
    waf_id = request.GET.get('geo_token_id')
    token = None
    for token_obj in Geo_Tokens.objects.using('waf').all():
        if token_obj.waf_id == waf_id:
            token = token_obj.geo_token

    ips_list = list()
    ips_list.append(get_ip)

    for ip in list(chunks(list(set(ips_list)), 1))[:len(ips_list)]:
        get_location_from_remote(token, ip, mcache)
    ips = get_location_by_ip(mcache, get_ip, check_token=True, token=token)
    context = {
        'ips': ips,
    }
    return render(request, 'main/ip_info.html', context)


@login_required
@time_of_function
def set_ip_blocked(request):
    keys = LicKeys.objects.filter(user__id=request.user.id)
    processes = []
    processes_queue = multiprocessing.Queue()

    if settings.API_URI is None:
        return []

    for key in keys:
        proc = multiprocessing.Process(target=get_lic_keys, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    ip = request.POST.get('ip')
    domain = request.POST.get('domain')
    waf_id = request.POST.get('waf_id')
    if domain:
        str_input = 'domain={} {}'.format(domain, ip)
    else:
        str_input = '{}'.format(ip)
    sending_to_api = False

    for key in keys_form_multiprocessing:
        if waf_id == key.waf_id:
            sending_to_api = True
            res = set_dyn_bl(key, [str_input])
            if res[0] is False:
                messages.error(request, res[1])
                storage = messages.get_messages(request)
                list_messages = get_list_messages(storage)
                return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    if not sending_to_api:
        messages.error(request, pgettext('messages', 'No valid keys in bl setting'))

        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    else:
        return HttpResponse(status=201)


@login_required
@start_end_time
@time_of_function
def home(request, start_date, end_date):
    # memcached
    mcache = Client(cache_address)
    # domain cookie
    q = request.GET.get('domain', None)
    if q:
        request.session['domain'] = q
    # waf id
    wafuser = WAFUSER.objects.get(id=request.user.id)
    user_alias = wafuser.waf_alias
    # waf domain
    domain_from_wafuser = wafuser.domain

    search_history = wafuser.search_history

    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    sortbyfreq = True
    if sortbyfreq:
        tmp = dict()
        tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
        search_history_dict = tmp

    attack_history = []
    charts_history = []
    scan_history = []

    for el in search_history_dict.keys():
        if 'attacks:' in el:
            attack_history.append(el.replace('attacks:', ''))
        if 'charts:' in el:
            charts_history.append(el.replace('charts:', ''))
        if 'scan:' in el:
            scan_history.append(el.replace('scan:', ''))

    summary_end_date = end_date
    today_date = date.today()
    formatted_summary_end_date = datetime.strptime(summary_end_date, '%Y-%m-%d %H:%M:%S.%f').date()
    if formatted_summary_end_date > today_date:
        summary_end_date = str(today_date) + ' 23:59:59.000000'
    get_end_date = datetime.strptime(summary_end_date, '%Y-%m-%d %H:%M:%S.%f').date()
    last_month = get_end_date - timedelta(days=30)

    if user_alias:
        # main logs
        if request.GET.get('start', None):
            main_base = Logs.objects.using('waf'). \
                only('ip', 'timestamp', 'vhost', 'alias', 'bt', 'id', 'description', 'rule_id'). \
                filter(timestamp__gte=last_month, timestamp__lte=summary_end_date)
        else:
            main_base = Logs.objects.using('waf'). \
                only('ip', 'timestamp', 'vhost', 'alias', 'bt', 'id', 'description', 'rule_id'). \
                filter(timestamp__gte=last_month, timestamp__lte=summary_end_date)

        main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)

        internal_errors_base = main_base.filter(bt=5)
        brute_force_base = main_base.filter(bt=7)
        sms_flood_base = main_base.filter(bt=9)
        ddos_base = main_base.filter(bt=10)
        main_base = main_base.exclude(bt=5)

        waf_mode = wafuser.waf_mode

        main = Logs.objects.using('waf').only('ip', 'timestamp', 'vhost', 'alias') \
            .filter(timestamp__gte=start_date, timestamp__lte=end_date).exclude(bt=5)
        main = get_filtered_logs(main, user_alias, domain_from_wafuser, q)
        main = main \
            .only('ip', 'timestamp', 'vhost', 'alias')

        yesterday_date = get_end_date - timedelta(days=1)
        last_week = get_end_date - timedelta(days=7)

        description_summary_base = main.filter(bt__in=[1, 2])
        top10 = main_base.filter(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)

        today_summary = main_base.only('vhost', 'bt').annotate(count=Count('id')). \
            filter(bt__in=[1, 2, 3, 4, 7, 8, 9, 10]).filter(timestamp__gte=get_end_date)

        today_summary_data = today_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            flood=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
            ca=Count('bt', filter=Q(bt=4)),
        )

        today_summary_piechart_data = today_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            bt1=Count('bt', filter=Q(bt=1)),
            bt2=Count('bt', filter=Q(bt=2)),
            bt3=Count('bt', filter=Q(bt=3)),
            bt4=Count('bt', filter=Q(bt=4)),
            bt7=Count('bt', filter=Q(bt=7)),
            bt8=Count('bt', filter=Q(bt=8)),
            bt9=Count('bt', filter=Q(bt=9)),
            bt10=Count('bt', filter=Q(bt=10)),
        )

        today_summary_list = list()
        for i in today_summary_piechart_data:
            today_summary_list.append([
                "BT1",
                i.get('bt1'),
                'Signature Analysis',
            ])
            today_summary_list.append([
                "BT2",
                i.get('bt2'),
                'Signature Analysis',
            ])
            today_summary_list.append([
                "BT3",
                i.get('bt3'),
                'Nemesida AI MLA',
            ])
            today_summary_list.append([
                "BT4",
                i.get('bt4'),
                'Antivirus',
            ])
            today_summary_list.append([
                "BT7",
                i.get('bt7'),
                'Brute force',
            ])
            today_summary_list.append([
                "BT8",
                i.get('bt8'),
                'Nemesida AI MLC',
            ])
            today_summary_list.append([
                "BT9",
                i.get('bt9'),
                'Flood',
            ])
            today_summary_list.append([
                "BT10",
                i.get('bt10'),
                'DDoS',
            ])

        last24_summary = main_base.only('vhost', 'bt').annotate(count=Count('id')).filter(
            bt__in=[1, 2, 3, 4, 7, 8, 9, 10]).filter(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)

        last24_summary_data = last24_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            flood=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
            ca=Count('bt', filter=Q(bt=4)),
        )

        last24_summary_piechart_data = last24_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            bt1=Count('bt', filter=Q(bt=1)),
            bt2=Count('bt', filter=Q(bt=2)),
            bt3=Count('bt', filter=Q(bt=3)),
            bt4=Count('bt', filter=Q(bt=4)),
            bt7=Count('bt', filter=Q(bt=7)),
            bt8=Count('bt', filter=Q(bt=8)),
            bt9=Count('bt', filter=Q(bt=9)),
            bt10=Count('bt', filter=Q(bt=10)),
        )

        last24_summary_list = list()
        for i in last24_summary_piechart_data:
            last24_summary_list.append([
                "BT2",
                i.get('bt2'),
                'Signature Analysis',
            ])
            last24_summary_list.append([
                "BT3",
                i.get('bt3'),
                'Nemesida AI MLA',
            ])
            last24_summary_list.append([
                "BT4",
                i.get('bt4'),
                'Antivirus',
            ])
            last24_summary_list.append([
                "BT7",
                i.get('bt7'),
                'Brute force',
            ])
            last24_summary_list.append([
                "BT8",
                i.get('bt8'),
                'Nemesida AI MLC',
            ])
            last24_summary_list.append([
                "BT1",
                i.get('bt1'),
                'Signature Analysis',
            ])
            last24_summary_list.append([
                "BT9",
                i.get('bt9'),
                'Flood',
            ])
            last24_summary_list.append([
                "BT10",
                i.get('bt10'),
                'DDoS',
            ])

        last_week_summary = main_base.only('vhost', 'bt').annotate(count=Count('id')).filter(
            bt__in=[1, 2, 3, 4, 7, 8, 9, 10]).filter(timestamp__gte=last_week, timestamp__lte=summary_end_date)

        last_week_summary_data = last_week_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            flood=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
            ca=Count('bt', filter=Q(bt=4)),
        )

        last_week_summary_piechart_data = last_week_summary.values(all_attacks=Count('bt')).order_by('-all_attacks'). \
            annotate(
            bt1=Count('bt', filter=Q(bt=1)),
            bt2=Count('bt', filter=Q(bt=2)),
            bt3=Count('bt', filter=Q(bt=3)),
            bt4=Count('bt', filter=Q(bt=4)),
            bt7=Count('bt', filter=Q(bt=7)),
            bt8=Count('bt', filter=Q(bt=8)),
            bt9=Count('bt', filter=Q(bt=9)),
            bt10=Count('bt', filter=Q(bt=10)),
        )

        last_week_summary_list = list()
        for i in last_week_summary_piechart_data:
            last_week_summary_list.append([
                "BT2",
                i.get('bt2'),
                'Signature Analysis',
            ])
            last_week_summary_list.append([
                "BT3",
                i.get('bt3'),
                'Nemesida AI MLA',
            ])
            last_week_summary_list.append([
                "BT4",
                i.get('bt4'),
                'Antivirus',
            ])
            last_week_summary_list.append([
                "BT7",
                i.get('bt7'),
                'Brute force',
            ])
            last_week_summary_list.append([
                "BT8",
                i.get('bt8'),
                'Nemesida AI MLC',
            ])
            last_week_summary_list.append([
                "BT1",
                i.get('bt1'),
                'Signature Analysis',
            ])
            last_week_summary_list.append([
                "BT9",
                i.get('bt9'),
                'Flood',
            ])
            last_week_summary_list.append([
                "BT10",
                i.get('bt10'),
                'DDoS',
            ])

        last_month_summary = main_base.only('vhost', 'bt').annotate(count=Count('id')). \
            filter(bt__in=[1, 2, 3, 4, 7, 8, 9, 10]).filter(timestamp__gte=last_month, timestamp__lte=summary_end_date)

        last_month_summary_data = last_month_summary.values(all_attacks=Count('bt')).order_by('-all_attacks').annotate(
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            flood=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
            ca=Count('bt', filter=Q(bt=4)),
        )

        last_month_summary_piechart_data = last_month_summary.values(all_attacks=Count('bt')).order_by(
            '-all_attacks').annotate(
            bt1=Count('bt', filter=Q(bt=1)),
            bt2=Count('bt', filter=Q(bt=2)),
            bt3=Count('bt', filter=Q(bt=3)),
            bt4=Count('bt', filter=Q(bt=4)),
            bt7=Count('bt', filter=Q(bt=7)),
            bt8=Count('bt', filter=Q(bt=8)),
            bt9=Count('bt', filter=Q(bt=9)),
            bt10=Count('bt', filter=Q(bt=10)),
        )

        last_month_summary_list = list()
        for i in last_month_summary_piechart_data:
            last_month_summary_list.append([
                "BT2",
                i.get('bt2'),
                'Signature Analysis',
            ])
            last_month_summary_list.append([
                "BT3",
                i.get('bt3'),
                'Nemesida AI MLA',
            ])
            last_month_summary_list.append([
                "BT4",
                i.get('bt4'),
                'Antivirus',
            ])
            last_month_summary_list.append([
                "BT7",
                i.get('bt7'),
                'Brute force',
            ])
            last_month_summary_list.append([
                "BT8",
                i.get('bt8'),
                'Nemesida AI MLC',
            ])
            last_month_summary_list.append([
                "BT1",
                i.get('bt1'),
                'Signature Analysis',
            ])
            last_month_summary_list.append([
                "BT9",
                i.get('bt9'),
                'Flood',
            ])
            last_month_summary_list.append([
                "BT10",
                i.get('bt10'),
                'DDoS',
            ])

        brute_force_summary = brute_force_base.values('vhost', 'bt')
        brute_force_summary_piechart_data = brute_force_summary.values('bt').order_by('-bt').annotate(
            today=Count('bt', filter=Q(timestamp__gte=get_end_date)),
            last_24=Count('bt', filter=Q(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)),
            last_week=Count('bt', filter=Q(timestamp__gte=last_week, timestamp__lte=summary_end_date)),
            last_month=Count('bt', filter=Q(timestamp__gte=last_month, timestamp__lte=summary_end_date))
        )

        brute_force_summary_list = list()
        for i in brute_force_summary_piechart_data:
            brute_force_summary_list.append([
                'TODAY',
                i.get('today')
            ])
            brute_force_summary_list.append([
                '24 H.',
                i.get('last_24')
            ])
            brute_force_summary_list.append([
                'WEEK',
                i.get('last_week')
            ])
            brute_force_summary_list.append([
                'MONTH',
                i.get('last_month')
            ])

        sms_flood_summary = sms_flood_base.values('vhost', 'bt')
        sms_flood_summary_piechart_data = sms_flood_summary.values('bt').order_by('-bt').annotate(
            today=Count('bt', filter=Q(timestamp__gte=get_end_date)),
            last_24=Count('bt', filter=Q(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)),
            last_week=Count('bt', filter=Q(timestamp__gte=last_week, timestamp__lte=summary_end_date)),
            last_month=Count('bt', filter=Q(timestamp__gte=last_month, timestamp__lte=summary_end_date))
        )

        sms_flood_summary_list = list()
        for i in sms_flood_summary_piechart_data:
            sms_flood_summary_list.append([
                'TODAY',
                i.get('today')
            ])
            sms_flood_summary_list.append([
                '24 H.',
                i.get('last_24')
            ])
            sms_flood_summary_list.append([
                'WEEK',
                i.get('last_week')
            ])
            sms_flood_summary_list.append([
                'MONTH',
                i.get('last_month')
            ])

        ddos_summary = ddos_base.values('vhost', 'bt')
        ddos_summary_piechart_data = ddos_summary.values('bt').order_by('-bt').annotate(
            today=Count('bt', filter=Q(timestamp__gte=get_end_date)),
            last_24=Count('bt', filter=Q(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)),
            last_week=Count('bt', filter=Q(timestamp__gte=last_week, timestamp__lte=summary_end_date)),
            last_month=Count('bt', filter=Q(timestamp__gte=last_month, timestamp__lte=summary_end_date))
        )

        ddos_summary_list = list()
        for i in ddos_summary_piechart_data:
            ddos_summary_list.append([
                'TODAY',
                i.get('today')
            ])
            ddos_summary_list.append([
                '24 H.',
                i.get('last_24')
            ])
            ddos_summary_list.append([
                'WEEK',
                i.get('last_week')
            ])
            ddos_summary_list.append([
                'MONTH',
                i.get('last_month')
            ])

        internal_errors_summary = internal_errors_base.values('vhost', 'bt')
        internal_errors_summary_piechart_data = internal_errors_summary.values('bt').order_by('-bt').annotate(
            today=Count('bt', filter=Q(timestamp__gte=get_end_date)),
            last_24=Count('bt', filter=Q(timestamp__gte=yesterday_date, timestamp__lte=summary_end_date)),
            last_week=Count('bt', filter=Q(timestamp__gte=last_week, timestamp__lte=summary_end_date)),
            last_month=Count('bt', filter=Q(timestamp__gte=last_month, timestamp__lte=summary_end_date))
        )

        internal_errors_summary_list = list()
        for i in internal_errors_summary_piechart_data:
            internal_errors_summary_list.append([
                'TODAY',
                i.get('today')
            ])
            internal_errors_summary_list.append([
                '24 H.',
                i.get('last_24')
            ])
            internal_errors_summary_list.append([
                'WEEK',
                i.get('last_week')
            ])
            internal_errors_summary_list.append([
                'MONTH',
                i.get('last_month')
            ])

        description_summary = description_summary_base.values('description').annotate(
            counts=Count('description')).order_by('-counts')[:10]

        description_summary_list = list()
        for i in description_summary:
            description_summary_list.append(
                [
                    i.get('description'),
                    i.get('counts')
                ])

        today_summary_json = json.dumps(today_summary_list)
        last24_summary_json = json.dumps(last24_summary_list)
        last_week_summary_json = json.dumps(last_week_summary_list)
        last_month_summary_json = json.dumps(last_month_summary_list)
        brute_force_summary_json = json.dumps(brute_force_summary_list)
        sms_flood_summary_json = json.dumps(sms_flood_summary_list)
        ddos_summary_json = json.dumps(ddos_summary_list)
        internal_errors_summary_json = json.dumps(internal_errors_summary_list)
        description_summary_json = json.dumps(description_summary_list)

        # linechart
        linechart_data = main \
            .annotate(date_line=Trunc('timestamp', 'day', output_field=DateTimeField())) \
            .values('date_line') \
            .order_by('date_line') \
            .annotate(all_attacks=Count('bt'),
                      sa=Count('bt', filter=Q(bt__in=[1, 2])),
                      na=Count('bt', filter=Q(bt__in=[3, 8])),
                      bf=Count('bt', filter=Q(bt=7)),
                      flood=Count('bt', filter=Q(bt=9)),
                      ddos=Count('bt', filter=Q(bt=10)),
                      ca=Count('bt', filter=Q(bt=4))
                      )

        rows2 = list()
        for i in linechart_data:
            rows2.append([
                i.get('date_line').strftime("%d.%m.%Y"),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('flood'),
                i.get('ddos'),
                i.get('ca'),
                0,
            ])

        linechart = json.dumps(rows2)

        end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')

        top10 = top10.only('vhost', 'bt', 'alias').annotate(count=Count('id')).filter(bt__in=[1, 2, 3, 4, 7, 8, 9, 10])

        host_top10_data = top10.values('vhost').annotate(
            all_attacks=Count('bt'),
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            ca=Count('bt', filter=Q(bt=4)),
        ).order_by('-all_attacks')[:17]

        attackers_top10_data = top10.values('ip').annotate(
            all_attacks=Count('bt'),
            sa=Count('bt', filter=Q(bt__in=[1, 2])),
            na=Count('bt', filter=Q(bt__in=[3, 8])),
            bf=Count('bt', filter=Q(bt=7)),
            ca=Count('bt', filter=Q(bt=4)),
            flood=Count('bt', filter=Q(bt=9)),
            ddos=Count('bt', filter=Q(bt=10)),
        ).order_by('-all_attacks')[:10]

        ip_info_data = main.values('ip').annotate(
            all_attacks=Count('bt'),
            bt1=Count('bt', filter=Q(bt=1)),
            bt2=Count('bt', filter=Q(bt=2)),
            bt3=Count('bt', filter=Q(bt=3)),
            bt4=Count('bt', filter=Q(bt=4)),
            bt7=Count('bt', filter=Q(bt=7)),
            bt8=Count('bt', filter=Q(bt=8)),
            bt9=Count('bt', filter=Q(bt=9)),
            bt10=Count('bt', filter=Q(bt=10)),
        ).order_by('-all_attacks')[:15]

        host_top10_row = list()
        for i in host_top10_data:
            host_top10_row.append([
                i.get('vhost'),
                i.get('all_attacks'),
                i.get('sa'),
                i.get('na'),
                i.get('bf'),
                i.get('ca'),
            ])

        ips_list = []
        for i in attackers_top10_data:
            ips_list.append(i.get('ip'))

        processes = []
        for ip in list(chunks(list(set(ips_list)), 1))[:len(ips_list)]:
            proc = multiprocessing.Process(target=get_location_from_remote, args=(ip, mcache))
            proc.start()
            processes.append(proc)

        for pr in processes:
            pr.join()

        ips_list2 = []
        for i in ip_info_data:
            ips_list2.append(i.get('ip'))

        processes2 = []
        for ip in list(chunks(list(set(ips_list2)), 1))[:len(ips_list2)]:
            proc = multiprocessing.Process(target=get_location_from_remote, args=(ip, mcache))
            proc.start()
            processes2.append(proc)

        for pr in processes2:
            pr.join()

        last24h_popular_target_dict = dict()
        last24h_count_popular_target_dict = dict()

        last24h_ip_count_dict = dict()

        last24 = top10.values('ip').annotate(
            all_attacks=Count('bt'),
        ).order_by('-all_attacks')
        for i in last24:
            last24h_ip_count_dict[i.get('ip')] = str(i.get('all_attacks'))

        mostly_blocked_count_dict = dict()
        check_bt = dict()

        attackers_top10_processes = []
        attackers_top10_queue = multiprocessing.Queue()
        # Before multiprocessing, we close connections to the database so that there is no error
        # DatabaseError SSL error: decryption failed or bad record
        db.connections.close_all()
        for i in attackers_top10_data:
            proc = multiprocessing.Process(target=draw_attackers_top10_tooltip, args=(i, top10,
                                                                                      last24h_popular_target_dict,
                                                                                      last24h_count_popular_target_dict,
                                                                                      mostly_blocked_count_dict,
                                                                                      check_bt,
                                                                                      attackers_top10_queue, mcache))
            proc.start()
            attackers_top10_processes.append(proc)

        for proc in attackers_top10_processes:
            proc.join()

        attackers_top10_row = list()
        while not attackers_top10_queue.empty():
            attackers_top10_row.append(attackers_top10_queue.get())

        ip_info_row = list()

        top_by_ip_popular_target_dict = dict()
        top_by_ip_count_popular_target_dict = dict()
        mostly_blocked_count_dict = dict()

        check_bt = dict()

        ip_info_processes = []
        ip_info_queue = multiprocessing.Queue()
        for i in ip_info_data:
            proc = multiprocessing.Process(target=draw_top_by_ip_tooltip, args=(i, main,
                                                                                top_by_ip_popular_target_dict,
                                                                                top_by_ip_count_popular_target_dict,
                                                                                mostly_blocked_count_dict, check_bt,
                                                                                last24h_ip_count_dict, ip_info_queue,
                                                                                mcache))
            proc.start()
            ip_info_processes.append(proc)

        for proc in ip_info_processes:
            proc.join()

        while not ip_info_queue.empty():
            ip_info_row.append(ip_info_queue.get())

        host_top10_json = json.dumps(host_top10_row)
        attackers_top10_json = json.dumps(attackers_top10_row)
        ip_info_json = json.dumps(ip_info_row)

        context = {
            'home': settings.HOME_PAGE,
            'mode': waf_mode,
            'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date': end_date_q,
            'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'linechart': linechart,
            'host_top10': host_top10_json,
            'attackers_top10': attackers_top10_json,
            'ip_info': ip_info_json,
            'today_summary': today_summary_data,
            'last24_summary': last24_summary_data,
            'last_week': last_week_summary_data,
            'last_month': last_month_summary_data,
            'today_summary_json': today_summary_json,
            'last24_summary_json': last24_summary_json,
            'last_week_summary_json': last_week_summary_json,
            'last_month_summary_json': last_month_summary_json,
            'brute_force_summary_json': brute_force_summary_json,
            'sms_flood_summary_json': sms_flood_summary_json,
            'ddos_summary_json': ddos_summary_json,
            'internal_errors_summary_json': internal_errors_summary_json,
            'description_summary_json': description_summary_json,
            'attack_history': attack_history[:25],
            'charts_history': charts_history[:25],
            'scan_history': scan_history[:25],
            'user': request.user,
            'domain': q,
        }

        return render(request, 'main/summary.html', context)

    else:
        context = {
            'user': request.user
        }

        return render(request, 'main/summary.html', context)


@login_required
@start_end_time
@time_of_function
def get_group_tags(request, start_date, end_date):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    q = request.GET.get('domain', None)

    if q == 'None':
        q = None
    if request.method == 'POST':
        if request.POST.get('data') is None:
            messages.error(request, pgettext('messages', 'Invalid data'))

            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
        data = json.loads(request.POST.get('data'))
        lvl = data.get('lvl')
        possible = list(data.get('possible', []))
        description = list(data.get('description', []))

        vhost = data.get('vhost')

        if vhost:
            if type(vhost) is list:
                vhost = [convert_to_punycode(v) for v in vhost]
            else:
                vhost = convert_to_punycode(vhost)
        bt = list(data.get('bt', []))
        cc = list(data.get('cc', []))
        s = data.get('search', '').strip()
        user_alias = wafuser.waf_alias
        domain_from_wafuser = wafuser.domain
        if user_alias and lvl not in [None, '']:
            lvl = int(lvl)
            qu = Q()
            for i in range(0, len(bt)):
                q_tmp = Q()
                if description:
                    if description[i] not in (None, '', 'None'):
                        if description[i] == '__nodesc_':
                            description[i] = ''
                            q_tmp &= (Q(description=description[i]) | Q(description__isnull=True))
                        else:
                            q_tmp &= Q(description=description[i])
                if bt[i] not in (None, '', 'None'):
                    bt_tmp = [int(x) for x in bt[i].strip('[]').split(',')]
                    bt_tmp = Q(bt__in=bt_tmp)
                    q_tmp &= bt_tmp
                qu |= q_tmp
            if cc:
                if cc != [] and isCCEnable:
                    qt = Q()
                    if 'None' in cc:
                        qt &= Q(cc__in=['  ', 'None'] + cc)
                    else:
                        qt &= Q(cc__in=cc)
                    if '#null' in cc:
                        qt |= Q(cc__isnull=True)

                    qu &= qt
            if possible:
                if not set(possible).issuperset({'True', 'False'}):
                    if set(possible) & {'false', 'False', False, None, 'None', ''}:
                        qu &= ~Q(possible=True)
                    elif set(possible) & {'true', 'True', True}:
                        qu &= Q(possible=True)

            main_base = Logs.objects.using('waf').only('vhost', 'alias', 'timestamp') \
                .filter(timestamp__gte=start_date, timestamp__lte=end_date).filter(qu)
            if vhost not in (None, '', 'None'):
                main_base = main_base.filter(vhost__in=vhost).only('vhost', 'alias', 'timestamp')

            main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)
            if s:
                try:
                    main_base = search_filter(s, main_base)
                except Exception:
                    main_base = Logs.objects.using('waf').none()
            bt_tmp = bt
            bt = []
            for i in bt_tmp:
                for x in i.strip('[]').split(','):
                    if x != '':
                        bt.append(int(x))

            main_base = main_base.only('vhost', 'bt', 'description', 'possible')
            main = get_tags(main_base, bt=bt, lvl=lvl)

            if lvl == 2:
                template = 'main/upload_tags.html'
                context = {
                    'home': settings.HOME_PAGE,
                    'table': main,
                    'user': request.user,
                    'id_table': '',
                    'lvl': lvl,
                }

                return render(request, template, context)
            elif lvl == 1:
                return JsonResponse(main)
    messages.error(request, pgettext('messages', 'Access error'))

    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)
    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


def give_license(request) -> object:
    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)
    list_key_for_requests = [i for i in lic_keys if i.tarif == 'Enterprise']
    threads = []
    threads_queue = queue.Queue()
    for key in lic_keys:
        thread = threading.Thread(target=get_rules_management, args=(key, threads_queue))
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

    keys_form_threading = []
    while not threads_queue.empty():
        keys_form_threading.append(threads_queue.get())

    keys_form_threading = sorted(keys_form_threading, key=operator.attrgetter('description'))
    dict_waf_id_key = {}
    for key in keys_form_threading:
        dict_waf_id_key[get_waf_id(key)] = key.key
        if not key.description:
            key.description = get_waf_id(key)
    return keys_form_threading, list_key_for_requests, dict_waf_id_key


def give_plans(request) -> object:
    # Get keys waf ids for user
    keys_waf_ids = get_user_key_waf_id(request.user)
    plans = get_list_plans(request.user)
    return keys_waf_ids, plans


def level_list_for_EC(list_event) -> dict:
    """return list with level for mainbox in EC"""
    level_list = {}
    for i in list_event:
        if i['level'] not in level_list.keys():
            level_list[i['level']] = 1
        else:
            level_list[i['level']] = level_list[i['level']] + 1
    for i in level_list:
        level_list[i] = str(level_list[i])
    return level_list


def agent_for_EC(list_event) -> set:
    """return list with agent for mainbox in EC"""
    agent_list = set([i['agent'] for i in list_event])
    return agent_list


def waf_id_list_for_EC(list_event) -> set:
    """return list with waf_id for mainbox in EC"""
    waf_id_list = set([i['waf_id'] for i in list_event])
    return waf_id_list


def IP_for_EC(list_event) -> set:
    """return list with ip for mainbox in EC"""
    ip_list = set([i['ip'] for i in list_event])
    return ip_list


def category_for_EC(list_event) -> set:
    """return list with category for mainbox in EC"""
    category_list = set([i['category'] for i in list_event])
    return category_list


def delete_event_for_EC(delete: object, key: list) -> None:
    for i in key:
        data = {'key': i.key, "id": list(delete)}

        try:
            requests.post(
                url_normalize(api_url + '/del_event'),
                headers={'User-agent': settings.USER_AGENT, 'Content-type': 'application/json'},
                proxies=get_proxy_api_dict(),
                timeout=30,
                json=data,
            )
        except Exception as error:
            logger.error(f'An error occurred error when deleting {error}')


def request_with_limit_and_offset_for_EC(key: object, start_unixtime: float,
                                         end_unixtime: float, quant_events_page: str, paginator: str, listing=None
                                         ) -> object:
    if paginator == 'end':
        paginator = 0
    data = {'key': key.key, "startTime": round(start_unixtime),
            "endTime": round(end_unixtime), 'limit': quant_events_page, 'offset': paginator}
    if not listing:
        ...
    else:
        for i in listing:
            for key, value in i.items():
                data[key] = value
    try:
        return requests.post(
            url_normalize(api_url + '/get_event'),
            headers={'User-agent': settings.USER_AGENT, 'Content-type': 'application/json'},
            proxies=get_proxy_api_dict(),
            timeout=30,
            json=data,
        )
    except Exception as error:
        logger.error('An error occurred error when get events: {}'.format(error))


def request_without_limit_and_offset_for_EC(key: object, start_unixtime: float,
                                            end_unixtime: float, listing=None) -> object:

    data = {'key': key.key, "startTime": round(start_unixtime),
            "endTime": round(end_unixtime)}
    if not listing:
        ...
    else:
        for i in listing:
            for key, value in i.items():
                data[key] = value
    try:
        return requests.post(
            url_normalize(api_url + '/get_event'),
            headers={'User-agent': settings.USER_AGENT, 'Content-type': 'application/json'},
            proxies=get_proxy_api_dict(),
            timeout=30,
            json=data,
        )
    except Exception as error:
        logger.error('An error occurred error when get events: {}'.format(error))


def logick_fill_answer_list(key: list, start_unixtime: float, end_unixtime: float,
                            quant_events_page: str, paginator: str, answer_list: list, listing=None) -> int:
    count = 0
    if len(key) > 1:
        list_waf_id = []
        for i in key:
            response = request_with_limit_and_offset_for_EC(i, start_unixtime,
                                                            end_unixtime, quant_events_page, paginator, listing)
            answer_json = json.loads(response.content.decode())
            # uniqueness check waf_id
            if answer_json['waf_id'] not in list_waf_id:
                list_waf_id.append(answer_json['waf_id'])
                for value in answer_json['events']:
                    value['waf_id'] = answer_json['waf_id']
                    value['timestamp'] = datetime.fromtimestamp(value['timestamp'])
                    value['timestamp'] = value['timestamp'].strftime('%d-%m-%Y %H:%M:%S')
                    answer_list.append(value)

            count = count + answer_json['count']

    else:
        response = request_with_limit_and_offset_for_EC(key[0], start_unixtime,
                                                        end_unixtime, quant_events_page, paginator, listing)
        answer_json = json.loads(response.content.decode())
        for i in answer_json['events']:
            i['waf_id'] = answer_json['waf_id']
            answer_list.append(i)
            i['timestamp'] = datetime.fromtimestamp(i['timestamp'])
            i['timestamp'] = i['timestamp'].strftime('%d-%m-%Y %H:%M:%S')
        count = answer_json['count']
    return count


def logick_EC_management_if_attr(request: object,
                                 quant_answer: int, start_unixtime: float,
                                 end_unixtime: float, key_list: list, key_for_menu: list, paginator=None,
                                 quant_events_page=None, ) -> tuple:
    """logick_EC_management_if_attr return tuple with lists"""

    answer = {'level': request.GET.get('level'), 'agent': request.GET.get('agent'),
              'ip': request.GET.get('ip'), 'category': request.GET.get('cat'), 'waf_id': request.GET.get('waf_id')}
    listing_str = answer
    listing = [{key: value.split(',')} for key, value in answer.items() if value]
    for i in list(answer):
        a = listing_str[i].split(',')
        listing_str[f'len_{i}'] = len(a) - 1

    answer_list = []
    count = logick_fill_answer_list(
        key_list,
        start_unixtime,
        end_unixtime,
        quant_events_page,
        paginator,
        answer_list,
        listing
    )
    # calculate page paginator
    if paginator != 'end':
        if int(paginator) == 0:
            calculate = count - (int(quant_events_page) * 1)
        else:
            calculate = count - (int(quant_events_page) * (int(paginator) + 1))
        if calculate <= 0:
            paginator = "end"

    list_event = answer_list[:quant_answer]
    answer_if_attr = []
    if len(key_for_menu) > 1:
        list_waf_id = []
        for i in key_for_menu:
            response = request_without_limit_and_offset_for_EC(i, start_unixtime, end_unixtime)
            answer_not_attr = json.loads(response.content.decode())
            # uniqueness check waf_id
            if answer_not_attr['waf_id'] not in list_waf_id:
                list_waf_id.append(answer_not_attr['waf_id'])
                for value in answer_not_attr['events']:
                    value['waf_id'] = answer_not_attr['waf_id']
                    answer_if_attr.append(value)
    else:
        resp = request_without_limit_and_offset_for_EC(key_list[0], start_unixtime, end_unixtime)
        answer_not_attr = json.loads(resp.content.decode())
        for i in answer_not_attr['events']:
            i['waf_id'] = answer_not_attr['waf_id']
            answer_if_attr.append(i)
    level_list = level_list_for_EC(answer_if_attr)
    agent_list = agent_for_EC(answer_if_attr)
    waf_id_list = waf_id_list_for_EC(answer_if_attr)
    ip_list = IP_for_EC(answer_if_attr)
    category_list = category_for_EC(answer_if_attr)
    return list_event, level_list, agent_list, waf_id_list, ip_list, category_list, listing_str, paginator, answer_list


@login_required
@start_end_time_var
def EC_management(request, start_date, end_date):
    try:
        start_d = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f')
        start_unixtime = time.mktime(start_d.timetuple())
        end_d = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
        end_unixtime = time.mktime(end_d.timetuple())
        keys_form_threading, key, dict_waf_id_key = give_license(request)
        quant_event = request.GET.get('quant_event')
        page_pagination = request.GET.get('page_pagination')
        delete = request.GET.get('delete')
        last_page = request.GET.get('last_page')
        page_last = 0
        if last_page:
            page_last = int(last_page)
        if delete:
            delete_event_for_EC(json.loads(request.POST.get('list')), key)
        quant_events_page = '100'
        if quant_event:
            quant_events_page = str(quant_event.replace(' ', ''))
        paginator = '0'
        if page_pagination:
            paginator = str(page_pagination)

        answer_list = []

        count = logick_fill_answer_list(
            key,
            start_unixtime,
            end_unixtime,
            quant_events_page,
            paginator,
            answer_list
        )
        # calculate events
        if quant_event:
            quant_answer = int(quant_events_page)
        else:
            quant_answer = 100
        list_quant_for_page = [i for i in [100, 200, 300, 500] if i != quant_answer]

        # calculate quant to the end list
        if int(paginator) == 0:
            calculate = count - (int(quant_events_page) * 1)
        else:
            calculate = count - (int(quant_events_page) * (int(paginator) + 1))
        if calculate <= 0:
            paginator = "end"

        attr = request.GET.get('attr')
        search = request.GET.get('search')

        if not attr:
            answer_if_not_attr = []
            list_waf_id = []
            if len(key) > 1:
                for i in key:
                    response = request_without_limit_and_offset_for_EC(i, start_unixtime, end_unixtime)
                    answer_not_attr = json.loads(response.content.decode())
                    # uniqueness check waf_id
                    if answer_not_attr['waf_id'] not in list_waf_id:
                        list_waf_id.append(answer_not_attr['waf_id'])
                        for value in answer_not_attr['events']:
                            value['waf_id'] = answer_not_attr['waf_id']
                            answer_if_not_attr.append(value)
            else:
                resp = request_without_limit_and_offset_for_EC(key[0], start_unixtime, end_unixtime)
                answer_not_attr = json.loads(resp.content.decode())
                for i in answer_not_attr['events']:
                    i['waf_id'] = answer_not_attr['waf_id']
                    answer_if_not_attr.append(i)
            listing_str = []
            list_event = answer_list[:quant_answer]
            level_list = level_list_for_EC(answer_if_not_attr)
            agent_list = agent_for_EC(answer_if_not_attr)
            waf_id_list = waf_id_list_for_EC(answer_if_not_attr)
            ip_list = IP_for_EC(answer_if_not_attr)
            category_list = category_for_EC(answer_if_not_attr)
        else:
            waf_id = request.GET.get('waf_id')
            if waf_id:
                keys = []
                for i in key:
                    for waf in waf_id.split(','):
                        if waf != '':
                            if i.key == dict_waf_id_key[waf]:
                                keys.append(i)
            else:
                keys = key
            list_event, level_list, agent_list, \
                waf_id_list, ip_list, category_list, \
                listing_str, paginator, answer_list = logick_EC_management_if_attr(request,
                                                                                   quant_answer, start_unixtime,
                                                                                   end_unixtime, keys, key, paginator,
                                                                                   quant_events_page,
                                                                                   )

        if search:
            list_event = [i for i in answer_list
                          for key, value in i.items()
                          if str(search) in str(value)]

        context = {
            'home': settings.HOME_PAGE,
            'home_origin': settings.HOME_PAGE,
            'lic_keys': keys_form_threading,
            'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'list_event': list_event,
            'level_list': level_list,
            'agent_list': agent_list,
            'ip_list': ip_list,
            'category_list': category_list,
            'listing_str': listing_str,
            'quant_answer': quant_answer,
            'list_quant_for_page': list_quant_for_page,
            'paginator': paginator,
            'page_last': page_last,
            'waf_id_list': waf_id_list
        }
        if not hasattr(request, 'variable'):
            return render(request, 'main/EC_management.html', context)
        else:
            return context
    except Exception as error:
        logger.error(f'An error occurred error when run func EC_management {error}')


@login_required
@start_end_time
@time_of_function
def tab_EC_management(request, start_date, end_date):
    request.variable = True
    try:
        context = EC_management(request, start_date, end_date)
        return render(request, 'main/tab_EC_management.html', context)
    except Exception as error:
        logger.error(f'An error occurred error when run func tab_EC_management {error}')


@login_required
@start_end_time
@time_of_function
def get_country_tags(request, start_date, end_date):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None
    if request.method == 'POST':
        if request.POST.get('data') is None:
            messages.error(request, pgettext('messages', 'Invalid data'))

            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
        data = json.loads(request.POST.get('data'))
        possible = list(data.get('possible', []))
        description = list(data.get('description', []))
        vhost = data.get('vhost')
        if vhost:
            if type(vhost) is list:
                vhost = [convert_to_punycode(v) for v in vhost]
            else:
                vhost = convert_to_punycode(vhost)
        bt = list(data.get('bt', []))
        cc = list(data.get('cc', []))
        s = data.get('search', '').strip()
        user_alias = wafuser.waf_alias
        domain_from_wafuser = wafuser.domain
        if user_alias:
            qu = Q()
            for i in range(0, len(bt)):
                q_tmp = Q()
                if description:
                    if description[i] not in (None, '', 'None'):
                        if description[i] == '__nodesc_':
                            description[i] = ''
                            q_tmp &= (Q(description=description[i]) | Q(description__isnull=True))
                        else:
                            q_tmp &= Q(description=description[i])
                if bt[i] not in (None, '', 'None'):
                    bt_tmp = [int(x) for x in bt[i].strip('[]').split(',')]
                    bt_tmp = Q(bt__in=bt_tmp)
                    q_tmp &= bt_tmp
                qu |= q_tmp
            if cc:
                if cc != [] and isCCEnable:
                    qt = Q()
                    if 'None' in cc:
                        qt &= Q(cc__in=['  ', 'None'] + cc)
                    else:
                        qt &= Q(cc__in=cc)
                    if '#null' in cc:
                        qt |= Q(cc__isnull=True)

                    qu &= qt
            if possible:
                if not set(possible).issuperset({'True', 'False'}):
                    if set(possible) & {'false', 'False', False, None, 'None', ''}:
                        qu &= ~Q(possible=True)
                    elif set(possible) & {'true', 'True', True}:
                        qu &= Q(possible=True)

            main_base = Logs.objects.using('waf').only('vhost', 'alias', 'timestamp') \
                .filter(timestamp__gte=start_date, timestamp__lte=end_date).filter(qu)

            if vhost not in (None, '', 'None'):
                main_base = main_base.filter(vhost__in=vhost)
            main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)

            if s:
                try:
                    main_base = search_filter(s, main_base)
                except Exception:
                    main_base = Logs.objects.using('waf').none()

            main_base = main_base.only('vhost', 'cc')
            main = get_country_for_group(main_base)
            return JsonResponse(main)
    messages.error(request, pgettext('messages', 'Access error'))
    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)

    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


def check_name_attacks_and_reasons(row: list) -> str:
    """ return name of attack from page attack table """
    if row.bt == 7 or row.bt == 9 or row.bt == 10:
        if row.possible == 1:
            ...
    if row.reason == 'Nemesida AI MLA':
        name = 'Nemesida AI MLA'
    elif row.reason == 'Nemesida AI MLC':
        name = 'Nemesida AI MLC'
    elif 'Brute' in row.reason:
        if row.possible == 1:
            name = 'brute-force'
        else:
            name = 'Brute-force'
    elif 'OpenAPI' in row.reason:
        name = 'OpenAPI'
    elif 'Blocked IP' in row.reason:
        name = 'Blocked IP'
    elif 'Flood' in row.reason:
        if row.possible == 1:
            name = 'flood'
        else:
            name = 'Flood'
    else:
        if 'Virus' in row.description:
            name = 'Virus'
        else:
            if row.bt == 5:
                name = 'Internal error'
            elif row.bt == 10:
                name = 'DDoS'
            else:
                name = row.description
    return name


def recalculating_number_attacks(main: list) -> tuple:
    """ USE in get_logs_group and combines logic if use mode groups ip addres """
    template = 'main/three_lvl_group.html'

    dict_ip = {ip.ip: [] for ip in main}
    for value in main:
        if value.ip in dict_ip:
            dict_ip[value.ip].append(value)

    tags_group = {value: {} for value in dict_ip.keys()}

    for value in dict_ip:
        tags_group[value] = {'ips': dict_ip[value][0].ips,
                             'ip': value,
                             'vhost': dict_ip[value][0].vhost,
                             'timestamp_max': dict_ip[value][0].timestamp_max,
                             'timestamp_min': dict_ip[value][0].timestamp_min,
                             'bt': dict_ip[value][0].bt,
                             'alias': dict_ip[value][0].alias,
                             'geotoken': dict_ip[value][0].geotoken,
                             'ml_learning_progress': dict_ip[value][0].ml_learning_progress,
                             'tags': {},
                             'head_group_ip': True
                             }
    # recalculating number of tags
    for i in dict_ip:
        for b in dict_ip[i]:
            name = check_name_attacks_and_reasons(b)
            if name not in tags_group[i]['tags']:
                tags_group[i]['tags'].update({name: 1})
            else:
                tags_group[i]['tags'][name] = tags_group[i]['tags'][name] + 1

    for i in tags_group:
        if i in dict_ip.keys():
            if tags_group[i]['tags'][next(iter(tags_group[i]['tags']))] == 1:
                dict_ip[i][0].display = True
            else:
                dict_ip[i].insert(0, tags_group[i])

    return tags_group, dict_ip, template


@login_required
@start_end_time
@time_of_function
def get_logs_group(request, start_date, end_date):

    wafuser = WAFUSER.objects.get(id=request.user.id)
    q = request.POST.get('domain', None)

    # Get keys waf ids for user
    keys_waf_ids = get_user_key_waf_id(request.user)
    plans = get_list_plans(request.user)
    if 'Enterprise' in plans:
        user_plans = 0
    elif 'Business' in plans or 'Light' in plans:
        user_plans = 1
    else:
        user_plans = -1

    if q == 'None':
        q = None
    if request.method == 'POST':
        if request.POST.get('data') is None:
            messages.error(request, pgettext('messages', 'Invalid data'))

            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

        data = json.loads(request.POST.get('data'))

        data['vhost'] = convert_to_punycode(data['vhost'])

        upload = data.get('upload', None)
        id_log = data.get('first_log')
        lvl = data.get('lvl')
        possible = list(data.get('possible', []))
        description = list(data.get('description', []))
        vhost = data.get('vhost')
        if vhost:
            if type(vhost) is list:
                vhost = [convert_to_punycode(v) for v in vhost]

        bt = list(data.get('bt', []))
        cc = list(data.get('cc', []))
        s = data.get('search', '').strip()
        user_alias = wafuser.waf_alias
        domain_from_wafuser = wafuser.domain
        if user_alias and lvl not in [None, '']:
            lvl = int(lvl)
            qu = Q()
            for i in range(0, len(bt)):
                q_tmp = Q()
                if description:
                    if description[i] not in (None, '', 'None'):
                        if description[i] == '__nodesc_':
                            description[i] = ''
                            q_tmp &= (Q(description=description[i]) | Q(description__isnull=True))
                        else:
                            q_tmp &= Q(description=description[i])
                if bt[i] not in (None, '', 'None'):
                    bt_tmp = [int(x) for x in bt[i].strip('[]').split(',')]
                    bt_tmp = Q(bt__in=bt_tmp)
                    q_tmp &= bt_tmp
                qu |= q_tmp
            if cc:
                if cc != [] and isCCEnable:
                    qt = Q()
                    if 'None' in cc:
                        qt &= Q(cc__in=['  ', 'None'] + cc)
                    else:
                        qt &= Q(cc__in=cc)
                    if '#null' in cc:
                        qt |= Q(cc__isnull=True)

                    qu &= qt
            if possible:
                if not set(possible).issuperset({'True', 'False'}):
                    if set(possible) & {'false', 'False', False, None, 'None', ''}:
                        qu &= ~Q(possible=True)
                    elif set(possible) & {'true', 'True', True}:
                        qu &= Q(possible=True)

            main_base = Logs.objects.using('waf').only('vhost', 'alias', 'timestamp', 'alias', 'ip') \
                .filter(vhost=vhost, timestamp__gte=start_date, timestamp__lte=end_date).filter(qu)

            main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)

            if id_log not in (None, '', 'None'):
                main_base = main_base.filter(id__gte=id_log)
            if s:
                try:
                    main_base = search_filter(s, main_base)
                except Exception:
                    main_base = Logs.objects.using('waf').none()

            mcache = Client(cache_address)
            ml_table = get_ml_table()

            count_all = main_base.count()
            page = int(data.get('page', 1))
            template = 'main/three_lvl.html'
            if upload not in (None, '', 'None'):
                template = 'main/scroll_attacks.html'

            main = get_main_logs_new(main_base, ml_table, mcache, page)

            count_after = len(main)
            if count_all <= (page - 1) * PAGE_SIZE + count_after:
                page = 0

            for i in main:
                if not i.geotoken:
                    i.ips[0]['location'] = {'status': 'fail', 'description': 'Private IP'}

            dict_ip = {}
            tags_group = {}

            if 'group' in data:
                if data['group'] == '':
                    ...
                else:
                    tags_group, dict_ip, template = recalculating_number_attacks(main)

            context = {
                'home': settings.HOME_PAGE,
                'table': main,
                'user': request.user,
                'id_table': '',
                'st_exist': new_table_exist,
                'page': page,
                'names': FieldNames().__dict__,
                'lvl': lvl,
                'user_plans': user_plans,
                'keys_waf_ids': keys_waf_ids,
                'dict_ip': dict_ip,
                # 'tags_group': tags_group # not used
            }
            return render(request, template, context)
    messages.error(request, pgettext('messages', 'Access error'))

    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)

    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time
@time_of_function
def second_row(request, start_date, end_date):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None
    get_id = request.GET.get('id')
    mcache = Client(cache_address)
    s = request.GET.get('search', '').strip()
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain
    if user_alias:
        main_base = Logs.objects.using('waf').filter(pk=get_id).defer('cc')
        main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)
        ws = main_base
        if s:
            try:
                ws = search_filter(s, main_base)
            except Exception:
                ws = Logs.objects.using('waf').none()

        # group attacks
        ips_group = request.GET.get('g', None)
        if ips_group == 'false':
            request.session['group'] = False
        elif ips_group == 'true':
            request.session['group'] = True
        else:
            request.session['group'] = False

        group = True
        if ips_group == 'true' or request.session.get('group', False) is True:
            params = get_params(True)
        elif ips_group == 'false' or request.session.get('group', False) is False:
            params = get_params(False)
            group = False
        else:
            params = get_params(True)

        main = ws.values('id', 'ip', 'timestamp').order_by('-timestamp')
        ml_table = get_ml_table()

        if isinstance(ml_table, bytes):
            ml_table = ml_table.decode('utf-8', 'ignore').replace("'", "\"")
            ml_table = json.loads(ml_table)

        main = get_main_logs(main, ml_table, mcache, True)

        context = {
            'home': settings.HOME_PAGE,
            'main': main,
            'user': request.user,
            'group': group,
            'params': params,
        }

        return render(request, 'main/second_row.html', context)
    messages.error(request, pgettext('messages', 'Access error'))

    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)

    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time
@time_of_function
def charts(request, start_date, end_date):
    get_start_date = request.GET.get('start')
    get_end_date = request.GET.get('end')
    start_date = datetime.strptime(get_start_date, '%d.%m.%Y').strftime('%Y-%m-%d')
    end_date = datetime.strptime(get_end_date, '%d.%m.%Y').strftime('%Y-%m-%d')

    state = request.GET.get('state', '').strip()
    s = request.GET.get('search', '').strip()
    clear_cache = request.GET.get('clear_cache', 'false')
    mcache = Client(cache_address)
    ts = datetime.timestamp(datetime.now())
    semaphore_mcache_key = str(request.user.id) + 'summary' + vAPP
    mcache_key = str(
        str(request.user.id) + '|'
        + str(end_date) + '|'
        + str(start_date) + '|'
        + convert_to_punycode(s) + '|'
        + state + '|'
        + vAPP
    ).replace(' ', '_')

    last_ts = mcache.get(semaphore_mcache_key)

    if clear_cache == 'true':
        mcache.delete(mcache_key)

    if clear_cache == 'false':
        res = mcache.get(mcache_key)

        if res is not None:
            if mcache.get('gen_chart_for_' + str(request.user.id)):
                mcache.delete('gen_chart_for_' + str(request.user.id))
            return JsonResponse(res.decode('utf-8'), safe=False)
        else:
            if last_ts:

                if (ts - float(last_ts.decode('utf-8'))) < 45:

                    res = mcache.get(mcache_key)
                    start_gen = ts
                    while res is None:
                        if last_ts is None:
                            continue
                        if (ts - float(last_ts.decode('utf-8'))) > 45 or ts - start_gen > 45:
                            break
                        ts = datetime.timestamp(datetime.now())
                        last_ts = mcache.get(semaphore_mcache_key)
                        res = mcache.get(mcache_key)
                    else:
                        if mcache.get('gen_chart_for_' + str(request.user.id)):
                            mcache.delete('gen_chart_for_' + str(request.user.id))
                        if res and (not isinstance(res, str)):
                            res = res.decode('utf-8')
                        return JsonResponse(res, safe=False)
                if (ts - float(last_ts.decode('utf-8'))) > 45:
                    mcache.delete(semaphore_mcache_key)

                    if mcache.get('gen_chart_for_' + str(request.user.id)):
                        if os.path.exists(
                                mcache.get('gen_chart_for_' + str(request.user.id) + '|' + vAPP,
                                           bytes('', 'utf-8')).decode(
                                    'utf-8')):
                            os.remove(
                                mcache.get('gen_chart_for_' + str(request.user.id) + '|' + vAPP,
                                           bytes('', 'utf-8')).decode(
                                    'utf-8'))
                        mcache.delete('gen_chart_for_' + str(request.user.id))
                else:
                    return HttpResponse("LOADING, TRY AGAIN LATER")

    request_get = json.dumps(request.GET).encode('utf-8')

    if mcache.get('gen_chart_for_' + str(request.user.id)) is None:
        res = gen_charts(
            user_id=str(request.user.id), req_get=request_get
        )
    else:
        res = mcache.get('gen_chart_for_' + str(request.user.id))

    if res and (not isinstance(res, str)):
        res = res.decode('utf-8')

    if os.path.isfile(res):
        mcache.set(semaphore_mcache_key, ts)
        if mcache.get('gen_chart_for_' + str(request.user.id)) is None:
            mcache.set('gen_chart_for_' + str(request.user.id) + '|' + vAPP, res)

        last_ts = mcache.get(semaphore_mcache_key)
        ts = datetime.timestamp(datetime.now())

        if last_ts:
            if (ts - float(last_ts.decode('utf-8'))) < 45:
                res = mcache.get(mcache_key)
                start_gen = ts
                while res is None:
                    if last_ts is None:
                        continue
                    if (ts - float(last_ts.decode('utf-8'))) > 45 or ts - start_gen > 45:
                        break
                    ts = datetime.timestamp(datetime.now())
                    last_ts = mcache.get(semaphore_mcache_key)
                    res = mcache.get(mcache_key)
                else:
                    if mcache.get('gen_chart_for_' + str(request.user.id)):
                        mcache.delete('gen_chart_for_' + str(request.user.id))
                    return JsonResponse(res.decode('utf-8'), safe=False)
        res = mcache.get(mcache_key)
        if res is not None:
            if mcache.get('gen_chart_for_' + str(request.user.id)):
                mcache.delete('gen_chart_for_' + str(request.user.id))
            return JsonResponse(res.decode('utf-8'), safe=False)

        return HttpResponse("LOADING, TRY AGAIN LATER")
    else:
        logger.error(
            'Warning. The uwsgidecorators module cannot be imported. Therefore, generation is performed synchronously'
        )
        if res == 0:
            return HttpResponse("LOADING, TRY AGAIN LATER")
        if res and (not isinstance(res, str)):
            res = res.decode('utf-8')
        return JsonResponse(res, safe=False)


@login_required
@start_end_time
@time_of_function
def check_logs(request, start_date, end_date):
    wafuser = WAFUSER.objects.get(id=request.user.id)
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None
    last_id = request.GET.get('id')
    mcache = Client(cache_address)
    s = request.GET.get('search', '').strip()
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain

    if user_alias and not last_id == '-Infinity':
        '''check this '''
        main_base = Logs.objects.using('waf').all().filter(timestamp__gte=start_date, timestamp__lte=end_date).exclude(
            bt=7).exclude(bt=5)
        main_base = get_filtered_logs(main_base, user_alias, domain_from_wafuser, q)
        ws = main_base.filter(pk__gt=last_id)
        if s:
            try:
                ws = search_filter(s, ws)
            except Exception:
                ws = Logs.objects.using('waf').none()

        # group attacks
        ips_group = request.GET.get('g', None)
        if ips_group == 'false':
            request.session['group'] = False
        elif ips_group == 'true':
            request.session['group'] = True
        else:
            request.session['group'] = False

        group = True
        if ips_group == 'true' or request.session.get('group', False) is True:
            params = get_params(True)
        elif ips_group == 'false' or request.session.get('group', False) is False:
            params = get_params(False)
            group = False
        else:
            params = get_params(True)

        main = ws \
            .values('vhost', 'bt', 'mz', 'rule_id') \
            .annotate(**params).order_by('-timestamp__max')

        ml_table = get_ml_table()

        if isinstance(ml_table, bytes):
            ml_table = ml_table.decode('utf-8', 'ignore').replace("'", "\"")
            ml_table = json.loads(ml_table)

        main = get_main_logs(main, ml_table, mcache, True)

        context = {
            'home': settings.HOME_PAGE,
            'main': main,
            'user': request.user,
            'group': group
        }

        return render(request, 'main/newlogs.html', context)
    messages.error(request, pgettext('messages', 'Access error'))

    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)

    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time
@time_of_function
def logs_html(request, start_date, end_date):
    q = request.GET.get('domain', None)
    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    if q == 'None':
        q = None
    s = request.GET.get('search', '').strip()
    loading = request.GET.get('loading')
    user = WAFUSER.objects.get(id=request.user.id)
    search_history = user.search_history

    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    if s is not None and s != '':
        if 'attacks:' + s in search_history_dict.keys():
            reb = dict()
            reb['attacks:' + s] = search_history_dict['attacks:' + s] + 1
            search_history_dict.pop('attacks:' + s)
            reb.update(search_history_dict)
            search_history_dict = reb

        else:
            reb = dict()
            reb['attacks:' + s] = 1
            reb.update(search_history_dict)
            search_history_dict = reb
    else:
        for el in search_history_dict.keys():
            if 'attacks:' in el:
                s = el.replace('attacks:', '')
                break
    if s == '#clean':
        s = ''

    while len(search_history_dict) > 1000:
        search_history_dict.popitem()

    user.search_history = json.dumps(search_history_dict)
    user.save()

    attack_history = []
    history_id_list = 'attack_searh_history_list'

    sortbyfreq = True
    if sortbyfreq:
        tmp = dict()
        last_use = ''
        for el in list(search_history_dict.keys()):
            if 'attacks:' in el and '#clean' not in el and 'attacks:' + s != el:
                last_use = el
                break

        if last_use in search_history_dict:
            tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
            search_history_dict.pop(last_use)
        tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
        search_history_dict = tmp

    for el in search_history_dict.keys():
        if 'attacks:' in el and '#clean' not in el:
            attack_history.append(el.replace('attacks:', '')) and s.lower() in el.replace('attacks:', '').lower()

    # get user alias
    user_alias = WAFUSER.objects.get(id=request.user.id).waf_alias
    domain_from_wafuser = WAFUSER.objects.get(id=request.user.id).domain
    main = Logs.objects.using('waf').only('vhost', 'alias', 'timestamp') \
        .filter(timestamp__gte=start_date, timestamp__lte=end_date)
    main = get_filtered_logs(main, user_alias, domain_from_wafuser, q)
    if s:
        try:
            main = search_filter(s, main)
        except Exception as error:
            logger.error(
                f"An error occurred when loading attacks with the filter: \"{s}\". Error description: {error}"
            )
            messages.warning(request, pgettext('messages', 'Invalid search string'))
            main = Logs.objects.using('waf').none()

    # group attacks
    group = False

    groups_attack = True

    if groups_attack:
        vhosts_list = get_vhost(main).values_list('vhost', flat=True)
        count_all = vhosts_list.count()
        page = int(request.GET.get('page', 1))
        start_page = PAGE_SIZE * (page - 1)
        end_page = PAGE_SIZE * page
        vhosts_list = vhosts_list[start_page:end_page]
        main = main.only('vhost', 'bt').filter(vhost__in=list(vhosts_list))
        main = get_groups(main)
    else:
        mcache = Client(cache_address)
        ml_table = get_ml_table()

        count_all = len(main)
        main = main.values('id', 'ip', 'timestamp').order_by('-timestamp')
        paginator = Paginator(main, PAGE_SIZE)
        page = int(request.GET.get('page', 1))
        try:
            main = paginator.page(page)
        except PageNotAnInteger:
            # If page is not an integer, deliver first page.
            main = paginator.page(1)
        except EmptyPage:
            messages.error(request, pgettext('messages', 'Error generating the page'))

            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
        main = check_geo_ip_tokens(main, filed="waf_id")

        main = get_main_logs(main, ml_table, mcache)

    if not hasattr(request.user, 'userconfig'):
        UserConfig.objects.create(user=request.user)

    if count_all == len(main):
        page = 0

    # Passing data to template
    context = {
        'home': settings.HOME_PAGE,
        'table': main,
        'user': request.user,
        'attack_history': attack_history[:25],
        'history_id_list': history_id_list,
        'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'end_date': end_date_q,
        'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'group': group,
        'names': FieldNames().__dict__,
        'domain': q,
        'st_exist': new_table_exist,
        'groups_attack': groups_attack,
        'last_filter': s,
        'page': page,
        'loading': loading,
    }

    return render(request, 'main/attacks.html', context)


@login_required
@time_of_function
def wl_request(request):
    log = request.POST.get('log_id')
    comment = request.POST.get('comment')
    record_date = datetime.now()
    try:
        log_obj = Logs.objects.using('waf').defer('cc').get(id=log)
    except Logs.DoesNotExist:
        messages.error(request,
                       pgettext('messages', 'Error. Record does not exist. '))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    log_def, created = Logs.objects.using('default').get_or_create(
        id=log_obj.id,
        defaults={'timestamp': log_obj.timestamp,
                  'vhost': log_obj.vhost,
                  'ip': log_obj.ip,
                  'rule_id': log_obj.rule_id,
                  'method': log_obj.method,
                  'url': log_obj.url,
                  'lm': log_obj.lm,
                  'mz': log_obj.mz,
                  'param': log_obj.param,
                  'ua': log_obj.ua,
                  'other_headers': log_obj.other_headers,
                  'referer': log_obj.referer,
                  'description': log_obj.description,
                  'bt': log_obj.bt,
                  'cookie': log_obj.cookie,
                  'request_id': log_obj.request_id,
                  'alias': log_obj.alias,
                  'body': log_obj.body}
    )

    WhitelistRequest.objects.create(
        record_date=record_date,
        user_comment=comment,
        log=log_def,
        user=request.user
    )
    return JsonResponse({'user': request.user.email, 'comment': comment})


@login_required
@time_of_function
def scan_request(request):
    log = request.POST.get('log_id')
    # comment = request.POST.get('comment')

    try:
        log_obj = Logs.objects.using('waf').get(id=log)
    except Logs.DoesNotExist:
        messages.error(request,
                       pgettext('messages', 'Error. Record does not exist. '))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    if log_obj.alias == 'NEMESIDAWAFFREE' or log_obj.alias == 'COMMUNITYEDITION':
        messages.error(request,
                       pgettext('messages', 'Not available in Community Edition.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    log_def, created = Logs.objects.using('default').get_or_create(
        id=log_obj.id,
        defaults={'timestamp': log_obj.timestamp,
                  'vhost': log_obj.vhost,
                  'ip': log_obj.ip,
                  'rule_id': log_obj.rule_id,
                  'method': log_obj.method,
                  'url': log_obj.url,
                  'lm': log_obj.lm,
                  'mz': log_obj.mz,
                  'param': log_obj.param,
                  'ua': log_obj.ua,
                  'referer': log_obj.referer,
                  'description': log_obj.description,
                  'bt': log_obj.bt,
                  'cookie': log_obj.cookie,
                  'request_id': log_obj.request_id,
                  'alias': log_obj.alias,
                  'cc': log_obj.cc,
                  'body': log_obj.body}
    )
    try:
        sr, cr = ScanRequest.objects.get_or_create(log=log_def)
        sr.status = False
        sr.save()
    except Exception as error:
        logger.error(f'An error occurred while writing to "ScanRequest": {error}')
        messages.error(request, pgettext('messages', 'An error occurred while writing to \"ScanRequest\"'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    return HttpResponseRedirect(settings.HOME_PAGE)


# A request to scan the scanner's report for a false
# When in the admin panel it is confirmed that it is false to expose ScanReport.false = True
@login_required
@time_of_function
def scanner_request(request):
    report_id = request.GET.get('report_id')
    try:
        scan_obj = ScannerReport.objects.using('waf').get(id=report_id)
    # except (Websites.DoesNotExist, ScannerReport.DoesNotExist):
    except ScannerReport.DoesNotExist:
        messages.error(request,
                       pgettext('messages', 'Error. Report does not exist.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    ScannerReport.objects.using('default').create(
        id=report_id,
        scan_date=scan_obj.scan_date,
        domain=scan_obj.domain,
        method=scan_obj.method,
        param=scan_obj.param,
        payload=scan_obj.payload,
        type=scan_obj.type,
        url=scan_obj.url,
        data=scan_obj.data,
        waf_alias=scan_obj.waf_alias
    )
    return HttpResponse(status=200)


@login_required
@start_end_time_scanner
@time_of_function
def scanner_html(request, start_date, end_date):
    dom = request.GET.get('domain', None)
    if dom == 'None':
        dom = None
    s = request.GET.get('search')
    if s is not None:
        s = s.strip()

    if request.GET.get('start') is not None:
        new_start_date = datetime.strptime(request.GET.get('start'), '%d.%m.%Y').strftime('%Y-%m-%d %H:%M:%S.%f')
    else:
        new_start_date = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d')
        new_start_date += ' 00:00:00.000000'

    if request.GET.get('start') is not None:
        new_end_date = datetime.strptime(request.GET.get('end'), '%d.%m.%Y').strftime('%Y-%m-%d')
        new_end_date += ' 23:59:59.000000'
    else:
        new_end_date = end_date

    wafuser = WAFUSER.objects.get(id=request.user.id)
    search_history = wafuser.search_history
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain

    new_vesion = True

    if user_alias:
        if s is not None and s != '':
            if 'scan:' + s in search_history_dict.keys():
                reb = dict()
                reb['scan:' + s] = search_history_dict['scan:' + s] + 1
                search_history_dict.pop('scan:' + s)
                reb.update(search_history_dict)
                search_history_dict = reb

            else:
                reb = dict()
                reb['scan:' + s] = 1
                reb.update(search_history_dict)
                search_history_dict = reb
        else:
            for el in search_history_dict.keys():
                if 'scan:' in el:
                    s = el.replace('scan:', '')
                    break
        if s == '#clean' or s is None:
            s = ''

        while len(search_history_dict) > 1000:
            search_history_dict.popitem()

        wafuser.search_history = json.dumps(search_history_dict)
        wafuser.save()

        sortbyfreq = True
        if sortbyfreq:
            tmp = dict()
            last_use = ''
            for el in list(search_history_dict.keys()):
                if 'scan:' in el and '#clean' not in el and 'scan:' + s != el:
                    last_use = el
                    break

            if last_use in search_history_dict:
                tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
                search_history_dict.pop(last_use)
            tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
            search_history_dict = tmp

        loc_scanner_report = [i.id for i in ScannerReport.objects.all()]
        scan_report = ScannerReport.objects.using('waf').filter(scan_date__gte=new_start_date,
                                                                scan_date__lte=new_end_date)
        scan_report = get_filtered_logs(scan_report, user_alias, domain_from_wafuser, dom)
        scan_report = scan_report.exclude(id__in=loc_scanner_report)
        if s:
            try:
                scan_report = scan_search_filter(s, scan_report)
            except Exception as error:
                logger.error(
                    f"An error occurred when loading attacks with the filter: \"{s}\". Error description: {error}"
                )
                messages.warning(request, pgettext('messages', 'Invalid search string'))

                scan_report = Logs.objects.using('waf').none()
        if new_vesion:
            today = datetime.strptime(new_end_date, '%Y-%m-%d %H:%M:%S.%f').replace(hour=0, minute=0, second=0,
                                                                                    microsecond=0)
            scan_types = scan_report.exclude(type='SDE').values('domain', 'type').annotate(dcount=Count('id'))
            scan_types_today = scan_report.exclude(type='SDE').filter(scan_date__gte=today).values('domain',
                                                                                                   'type').annotate(
                dcount=Count('id'))
            domain_list = list(
                scan_report.exclude(type='SDE').values('domain').annotate(dcount=Count('id')).order_by('-dcount'))
            scan_types_today_dict = dict()
            for i in scan_types_today:
                domain_tmp = convert_from_punycode(i.get('domain'))
                if scan_types_today_dict.get(domain_tmp) is None:
                    scan_types_today_dict[domain_tmp] = []
                scan_types_today_dict[domain_tmp].append({'type': i.get('type'), 'count': i.get('dcount')})
            scan_types_dict = dict()
            for i in scan_types:
                domain_tmp = convert_from_punycode(i.get('domain'))
                if scan_types_dict.get(domain_tmp) is None:
                    scan_types_dict[domain_tmp] = []
                add_today = None
                for j in scan_types_today_dict.get(domain_tmp, []):
                    if j.get('type') == i.get('type'):
                        add_today = attack_short_count(j.get('count'))
                scan_types_dict[domain_tmp].append({
                    'type': i.get('type'),
                    'count': attack_short_count(i.get('dcount')),
                    'add_today': add_today
                })
            main = []
            for i in domain_list:
                domain_tmp = convert_from_punycode(i.get('domain'))
                main.append({
                    'domain': domain_tmp,
                    'types': scan_types_dict.get(domain_tmp),
                    'dcount': attack_short_count(i.get('dcount'))
                })
        else:

            main = scan_report.exclude(type='SDE').order_by('-scan_date')

        sde_base = scan_report.filter(type='SDE').annotate(
            date=Cast(
                TruncDate('scan_date', DateTimeField()), CharField()
            )
        ).order_by('-date')
        sde_reports = sde_base.values('url', 'domain', 'date').order_by('-date').distinct('url', 'domain', 'date')

        list_date = sde_base.values_list('date', flat=True).distinct('date')

        sde_files_info = dict()
        list_date = tuple(list_date)
        sde = dict()
        sde_domains = set()
        for i in sde_reports:
            domain_tmp = convert_from_punycode(i.get('domain'))
            i['full_url'] = domain_tmp + i.get('url')
            i['dir'] = '/'.join(i.get('full_url').split('/')[0:-1]) + '/'
            if not sde_files_info.get(i.get('full_url')) and i.get('date') in list_date \
                    and list_date.index(i.get('date')) <= 3:

                if domain_tmp not in sde_domains:
                    sde[domain_tmp] = dict()
                sde_domains.add(domain_tmp)

                sde_files_info[i.get('full_url')] = {
                    'name_file': i.get('full_url').replace(i['dir'], ''),
                    'status': 'new' if list_date.index(i.get('date')) == 0 else 'old',
                    'date': i.get('date')
                }

                if i['dir'] not in sde[domain_tmp]:
                    sde[domain_tmp][i['dir']] = list()

                sde[domain_tmp][i['dir']].append(sde_files_info[i.get('full_url')])
            else:
                sde_base.filter(date=i.get('date'), domain=i.get('domain'), url=i.get('url')).delete()

        attack_history = []
        scan_history = []
        for el in search_history_dict.keys():
            if 'scan:' in el and '#clean' not in el and s.lower() in el.replace('scan:', '').lower():
                scan_history.append(el.replace('scan:', ''))
                attack_history.append(el.replace('scan:', ''))
        context = {
            'home': settings.HOME_PAGE,
            'scan_report': main,
            'attack_history': attack_history[:25],
            'scan_history': scan_history[:25],
            'user': request.user,
            'new_version': new_vesion,
            'last_filter': s,
            'domain': dom,
            'st_exist': new_table_exist,
            'sde': sde,
            'start_date': datetime.strptime(new_start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date_short': datetime.strptime(new_end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        }

        if request.GET.get('dyn') is None:
            return render(request, 'main/scanner.html', context)
        else:
            return render(request, 'main/scanner_search.html', context)
    else:
        messages.warning(request, pgettext('messages', 'Access error'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time_scanner
@time_of_function
def scanner_compare_html(request, start_date, end_date):
    dom = request.GET.get('domain', None)
    if dom == 'None':
        dom = None
    s = request.GET.get('search')
    if s is not None:
        s = s.strip()

    if request.GET.get('start') is not None:
        new_start_date = datetime.strptime(request.GET.get('start'), '%d.%m.%Y').strftime('%Y-%m-%d %H:%M:%S.%f')
    else:
        new_start_date = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d')
        new_start_date += ' 00:00:00.000000'

    if request.GET.get('start') is not None:
        new_end_date = datetime.strptime(request.GET.get('end'), '%d.%m.%Y').strftime('%Y-%m-%d')
        new_end_date += ' 23:59:59.000000'
    else:
        new_end_date = end_date

    wafuser = WAFUSER.objects.get(id=request.user.id)
    search_history = wafuser.search_history
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain

    new_vesion = True

    if user_alias:
        if s is not None and s != '':
            if 'scan:' + s in search_history_dict.keys():
                reb = dict()
                reb['scan:' + s] = search_history_dict['scan:' + s] + 1
                search_history_dict.pop('scan:' + s)
                reb.update(search_history_dict)
                search_history_dict = reb

            else:
                reb = dict()
                reb['scan:' + s] = 1
                reb.update(search_history_dict)
                search_history_dict = reb
        else:
            for el in search_history_dict.keys():
                if 'scan:' in el:
                    s = el.replace('scan:', '')
                    break
        if s == '#clean' or s is None:
            s = ''

        while len(search_history_dict) > 1000:
            search_history_dict.popitem()

        wafuser.search_history = json.dumps(search_history_dict)
        wafuser.save()

        sortbyfreq = True
        if sortbyfreq:
            tmp = dict()
            last_use = ''
            for el in list(search_history_dict.keys()):
                if 'scan:' in el and '#clean' not in el and 'scan:' + s != el:
                    last_use = el
                    break

            if last_use in search_history_dict:
                tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
                search_history_dict.pop(last_use)
            tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
            search_history_dict = tmp

        loc_scanner_report = [i.id for i in ScannerReport.objects.all()]
        scan_report = ScannerReport.objects.using('waf').filter(scan_date__gte=new_start_date,
                                                                scan_date__lte=new_end_date)
        scan_report = get_filtered_logs(scan_report, user_alias, domain_from_wafuser, dom)
        scan_report = scan_report.exclude(id__in=loc_scanner_report)
        if s:
            try:
                scan_report = scan_search_filter(s, scan_report)
            except Exception as error:
                logger.error(
                    f"An error occurred when loading attacks with the filter: \"{s}\". Error description: {error}"
                )
                messages.warning(request, pgettext('messages', 'Invalid search string'))

                scan_report = Logs.objects.using('waf').none()

        if new_vesion:
            scan_types = scan_report.exclude(type='SDE').values('domain', 'type').annotate(dcount=Count('id'))

            list_date = scan_report.annotate(date=TruncDate('scan_date', DateTimeField())).order_by('-date') \
                .values_list('date', flat=True).distinct('date')
            list_date = list(list_date)
            scan_report_for_compare = scan_report.exclude(type='SDE') \
                .annotate(date=TruncDate('scan_date', DateTimeField()))

            fields_name = [i.name for i in ScannerReport._meta.fields]
            if 'id' in fields_name:
                fields_name.remove('id')
            if 'scan_date' in fields_name:
                fields_name.remove('scan_date')
            scan_dist = scan_report_for_compare.values(*fields_name + ['date', 'scan_date']) \
                .annotate(dupe_id=Concat(*[F(f) for f in fields_name],
                                         output_field=CharField())) \
                .annotate(dupe_count=Count('dupe_id'))

            scan_dist_dict = dict()
            for i in scan_dist:
                key_dict = i.get('dupe_id')
                if key_dict not in scan_dist_dict:
                    scan_dist_dict[key_dict] = list()
                scan_dist_dict[key_dict].append(i)

            domain_list = list(
                scan_report.exclude(type='SDE').values('domain').annotate(dcount=Count('id')).order_by('-dcount'))
            dec_dict = dict()
            inc_dict = dict()
            count_for_domain_dict = dict()

            for list_dup in scan_dist_dict.values():
                date_min = list_date[0]
                date_max = list_date[-1]
                for i in list_dup:
                    if i.get('date') < date_min:
                        date_min = i.get('date')
                    if i.get('date') > date_max:
                        date_max = i.get('date')

                if count_for_domain_dict.get(list_dup[0].get('domain')) is None:
                    count_for_domain_dict[list_dup[0].get('domain')] = dict()
                    count_for_domain_dict[list_dup[0].get('domain')]['_all'] = 0
                count_for_domain_dict[list_dup[0].get('domain')]['_all'] += 1

                if count_for_domain_dict[list_dup[0].get('domain')].get(list_dup[0].get('type')) is None:
                    count_for_domain_dict[list_dup[0].get('domain')][list_dup[0].get('type')] = 0
                count_for_domain_dict[list_dup[0].get('domain')][list_dup[0].get('type')] += 1

                if dec_dict.get(list_dup[0].get('domain')) is None:
                    dec_dict[list_dup[0].get('domain')] = dict()
                if inc_dict.get(list_dup[0].get('domain')) is None:
                    inc_dict[list_dup[0].get('domain')] = dict()

                if date_max < list_date[0] or len(list_date) <= 1:
                    if inc_dict[list_dup[0].get('domain')].get(list_dup[0].get('type')) is None:
                        inc_dict[list_dup[0].get('domain')][list_dup[0].get('type')] = 0
                    inc_dict[list_dup[0].get('domain')][list_dup[0].get('type')] += 1
                elif date_min >= list_date[-1]:
                    if dec_dict[list_dup[0].get('domain')].get(list_dup[0].get('type')) is None:
                        dec_dict[list_dup[0].get('domain')][list_dup[0].get('type')] = 0
                    dec_dict[list_dup[0].get('domain')][list_dup[0].get('type')] += 1
            scan_types_dict = dict()
            for i in scan_types:
                if scan_types_dict.get(i.get('domain')) is None:
                    scan_types_dict[i.get('domain')] = []
                inc_vulns = None
                dec_vulns = None
                if i.get('domain') in dec_dict and i.get('type') in dec_dict.get(i.get('domain')):
                    dec_vulns = attack_short_count(dec_dict.get(i.get('domain'), dict()).get(i.get('type')))
                if i.get('domain') in inc_dict and i.get('type') in inc_dict.get(i.get('domain')):
                    inc_vulns = attack_short_count(inc_dict.get(i.get('domain'), dict()).get(i.get('type')))
                scan_types_dict[i.get('domain')].append({
                    'type': i.get('type'),
                    'count': attack_short_count(count_for_domain_dict.get(i.get('domain')).get(i.get('type'))),
                    'inc_vulns': inc_vulns,
                    'dec_vulns': dec_vulns
                })
            main = []
            for i in domain_list:
                main.append({
                    'domain': i.get('domain'),
                    'types': scan_types_dict.get(i.get('domain')),
                    'dcount': attack_short_count(count_for_domain_dict.get(i.get('domain')).get('_all'))
                })
        else:

            main = scan_report.exclude(type='SDE').order_by('-scan_date')

        attack_history = []
        scan_history = []
        for el in search_history_dict.keys():
            if 'scan:' in el and '#clean' not in el and s.lower() in el.replace('scan:', '').lower():
                scan_history.append(el.replace('scan:', ''))
                attack_history.append(el.replace('scan:', ''))
        context = {
            'home': settings.HOME_PAGE,
            'scan_report': main,
            'attack_history': attack_history[:25],
            'scan_history': scan_history[:25],
            'user': request.user,
            'last_filter': s,
            'domain': dom,
            'st_exist': new_table_exist,
            'start_date_compare': datetime.strptime(new_start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d'),
            'end_date_compare': datetime.strptime(new_end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d'),
            'start_date': datetime.strptime(new_start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date_short': datetime.strptime(new_end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        }

        return render(request, 'main/scanner_compare.html', context)


@login_required
@start_end_time_scanner
@time_of_function
def get_scan_logs(request, start_date, end_date):
    dom = request.GET.get('dom', None)
    if dom == 'None':
        dom = None
    if request.method == 'POST':
        if request.POST.get('data') is None:
            messages.error(request, pgettext('messages', 'Invalid data'))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
        data = json.loads(request.POST.get('data'))
        upload = data.get('upload', None)
        page = data.get('page', 1)
        id_log = data.get('first_log')
        types = list(data.get('type', []))
        if '' in types:
            types.remove('')
        domain = data.get('domain')
        if domain:
            if type(domain) is list:
                domain = [convert_to_punycode(d) for d in domain]
            else:
                domain = convert_to_punycode(domain)
        is_compare = data.get('is_compare', False)
        s = data.get('search', '').strip()
        wafuser = WAFUSER.objects.get(id=request.user.id)
        if data.get('start') is not None:
            new_start_date = datetime.strptime(data.get('start'), '%d.%m.%Y').strftime('%Y-%m-%d %H:%M:%S.%f')
        else:
            new_start_date = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d')
            new_start_date += ' 00:00:00.000000'

        if data.get('end') is not None:
            new_end_date = datetime.strptime(data.get('end'), '%d.%m.%Y').strftime('%Y-%m-%d')
            new_end_date += ' 23:59:59.000000'
        else:
            new_end_date = end_date

        user_alias = wafuser.waf_alias
        domain_from_wafuser = wafuser.domain

        new_vesion = True

        if s == '#clean' or s is None:
            s = ''

        loc_scanner_report = [i.id for i in ScannerReport.objects.all()]
        scan_report = ScannerReport.objects.using('waf').exclude(type='SDE').filter(domain=domain,
                                                                                    scan_date__gte=new_start_date,
                                                                                    scan_date__lte=new_end_date)
        if types:
            scan_report = scan_report.filter(type__in=types)
        if id_log not in (None, '', 'None'):
            scan_report = scan_report.filter(id__gte=id_log)
        scan_report = get_filtered_logs(scan_report, user_alias, domain_from_wafuser, dom)
        scan_report = scan_report.exclude(id__in=loc_scanner_report)

        if s:
            try:
                scan_report = scan_search_filter(s, scan_report)
            except Exception as error:
                logger.error(
                    f"An error occurred when loading attacks with the filter: \"{s}\". Error description: {error}"
                )
                messages.warning(request, pgettext('messages', 'Invalid search string'))

                scan_report = Logs.objects.using('waf').none()

        if is_compare:
            list_date = ScannerReport.objects.using('waf').exclude(type='SDE') \
                .filter(scan_date__gte=new_start_date,
                        scan_date__lte=new_end_date) \
                .annotate(date=TruncDate('scan_date', DateTimeField())).order_by('-date') \
                .values_list('date', flat=True).distinct('date')
            list_date = list(list_date)
            scan_report_for_compare = scan_report.exclude(type='SDE') \
                .annotate(date=TruncDate('scan_date', DateTimeField()))

            fields_name = [i.name for i in ScannerReport._meta.fields]
            if 'id' in fields_name:
                fields_name.remove('id')
            if 'scan_date' in fields_name:
                fields_name.remove('scan_date')
            scan_dist = scan_report_for_compare.annotate(max_date=Max('scan_date'), max_id=Max('id')) \
                .values(*fields_name + ['date', 'max_date', 'max_id']) \
                .annotate(dupe_id=Concat(*[F(f) for f in fields_name],
                                         output_field=CharField())) \
                .annotate(dupe_count=Count('dupe_id'))

            scan_dist_dict = dict()
            for i in scan_dist:
                key_dict = i.get('dupe_id')
                if key_dict not in scan_dist_dict:
                    scan_dist_dict[key_dict] = list()
                scan_dist_dict[key_dict].append(i)
            main_inc = []
            main_dec = []
            for list_dup in scan_dist_dict.values():
                vuln = list_dup[0]
                if len(list_date) > 1:
                    date_min = list_date[0]
                    date_max = list_date[-1]
                    for i in list_dup:
                        if i.get('date') < date_min:
                            date_min = i.get('date')
                        if i.get('date') > date_max:
                            date_max = i.get('date')

                    vuln['dec_vuln'] = False
                    vuln['inc_vuln'] = False

                    if date_max < list_date[0] or len(list_date) <= 1:
                        vuln['inc_vuln'] = True
                    elif date_min >= list_date[-1]:
                        vuln['dec_vuln'] = True

                if vuln.get('dec_vuln'):
                    main_dec.append(vuln)
                else:
                    main_inc.append(vuln)

            main_dec = sorted(main_dec, key=lambda x: x.get('max_date'), reverse=True)
            main_inc = sorted(main_inc, key=lambda x: x.get('max_date'), reverse=True)

            main = main_inc + main_dec

            count_all = len(main)
        else:
            main = scan_report.order_by('-scan_date')

            count_all = main.count()
        start_page = PAGE_SIZE * (page - 1)
        end_page = PAGE_SIZE * page
        main = main[start_page:end_page]
        count_after = len(main)
        if count_all <= (page - 1) * PAGE_SIZE + count_after:
            page = 0
        template = 'main/scan_logs.html'
        if upload not in (None, '', 'None'):
            template = 'main/scanner_table.html'
        context = {
            'home': settings.HOME_PAGE,
            'scan_report': main,
            'user': request.user,
            'new_version': new_vesion,
            'domain': dom,
            'page': page,
            'is_compare': is_compare,
            'start_date': datetime.strptime(new_start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date_short': datetime.strptime(new_end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        }

        return render(request, template, context)
    else:
        messages.warning(request, pgettext('messages', 'Access error'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time_scanner
@time_of_function
def get_scan_chart(request, start_date, end_date):
    s = request.GET.get('search')
    if s is not None:
        s = s.strip()

    if request.GET.get('start') is not None:
        new_start_date = datetime.strptime(request.GET.get('start'), '%d.%m.%Y').strftime('%Y-%m-%d %H:%M:%S.%f')
    else:
        new_start_date = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%Y-%m-%d')
        new_start_date += ' 00:00:00.000000'

    if request.GET.get('start') is not None:
        new_end_date = datetime.strptime(request.GET.get('end'), '%d.%m.%Y').strftime('%Y-%m-%d')
        new_end_date += ' 23:59:59.000000'
    else:
        new_end_date = end_date

    wafuser = WAFUSER.objects.get(id=request.user.id)

    user_alias = wafuser.waf_alias

    if user_alias:
        if s == '#clean' or s is None:
            s = ''

        loc_scanner_report = [i.id for i in ScannerReport.objects.all()]
        scan_report = ScannerReport.objects.using('waf').filter(scan_date__gte=new_start_date,
                                                                scan_date__lte=new_end_date)
        if s:
            try:
                scan_report = scan_search_filter(s, scan_report)
            except Exception:
                scan_report = Logs.objects.using('waf').none()
        scan_report = scan_report.exclude(id__in=loc_scanner_report)

        scan_linechart_data = scan_report.annotate(
            date_line=Trunc('scan_date', 'day', output_field=DateTimeField())) \
            .values('date_line', 'domain').order_by('date_line').annotate(dcount=Count('id'))
        domain_list = list(scan_report.values('domain').annotate(dcount=Count('id')).order_by('-dcount').values_list(
            'domain', flat=True))

        scan_linechart = dict()
        scan_date = list()
        if len(domain_list) > 15:
            domain_list = domain_list[:15] + ['other']
        for i in range(len(domain_list)):
            domain_list[i] = convert_from_punycode(domain_list[i])

        for i in scan_linechart_data:
            domain_tmp = convert_from_punycode(i.get('domain'))

            if scan_linechart.get(i.get('date_line').strftime("%d.%m.%Y")) is None:
                scan_linechart[i.get('date_line').strftime("%d.%m.%Y")] = dict()
            if i.get('domain') not in domain_list:
                i['domain'] = 'other'
            if i.get('domain') == 'other' and scan_linechart[i.get('date_line').strftime("%d.%m.%Y")].get(
                    i.get('domain')) is not None:
                scan_linechart[i.get('date_line').strftime("%d.%m.%Y")][domain_tmp] += i.get('dcount')
            else:
                scan_linechart[i.get('date_line').strftime("%d.%m.%Y")][domain_tmp] = i.get('dcount')
            if i.get('date_line').strftime("%d.%m.%Y") not in scan_date:
                scan_date.append(i.get('date_line').strftime("%d.%m.%Y"))

        data_line = []

        for i in scan_date:
            tmp_list = [i]
            for j in domain_list:
                if scan_linechart.get(i).get(j) is None:
                    tmp_list.append(0)
                else:
                    tmp_list.append(scan_linechart.get(i).get(j))
            data_line.append(tmp_list)

        return JsonResponse({'domain_list': domain_list, 'data_chart': data_line})
    messages.error(request, pgettext('messages', 'Access error'))
    storage = messages.get_messages(request)
    list_messages = get_list_messages(storage)

    return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@login_required
@start_end_time_scanner
@time_of_function
def get_last_filter(request, start_date, end_date):
    user = WAFUSER.objects.get(id=request.user.id)
    search_history = user.search_history
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1
    last_search = ''
    num = request.GET.get('num', "1").strip()
    num = int(num)
    tab = 'attacks'
    if request.GET.get('charts', None) is not None:
        tab = 'charts'
    if num == 0 and '#clean' in list(search_history_dict.keys())[0]:
        last_search = ''
    else:
        for el in list(search_history_dict.keys())[num:]:
            if tab + ':' in el:
                last_search = el.replace(tab + ':', '')
                break

    last_search_json = json.dumps(last_search)
    return JsonResponse(last_search_json, safe=False)


@login_required
@start_end_time_scanner
@time_of_function
def get_last_scan(request, start_date, end_date):
    user = WAFUSER.objects.get(id=request.user.id)
    search_history = user.search_history
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1
    last_search = ''
    num = request.GET.get('num', "1").strip()
    num = int(num)
    if num == 0 and '#clean' in list(search_history_dict.keys())[0]:
        last_search = ''
    else:
        for el in list(search_history_dict.keys())[num:]:
            if 'scan:' in el:
                last_search = el.replace('scan:', '')
                break

    last_scan_json = json.dumps(last_search)
    return JsonResponse(last_scan_json, safe=False)


@login_required
@start_end_time_scanner
@time_of_function
def get_list_filters(request, start_date, end_date):
    user = WAFUSER.objects.get(id=request.user.id)
    search_history = user.search_history
    s = request.GET.get('search', '').strip()
    if s == '#clean':
        s = ''
    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    sortbyfreq = True
    if sortbyfreq:
        tmp = dict()
        last_use = ''
        i = 0
        for el in list(search_history_dict.keys()):
            if request.GET.get('tab', 'attacks') + ':' in el and (i != 0 or request.GET.get('reset', None)):
                last_use = el
                break
            i += 1
        if last_use in search_history_dict:
            tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
            search_history_dict.pop(last_use)
        tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
        search_history_dict = tmp
    list_filter = list()
    for el in search_history_dict.keys():
        if request.GET.get('tab', 'attacks') + ':' in el and '#clean' not in el and s.lower() in el.replace(
                request.GET.get('tab', 'attacks') + ':', '').lower():
            list_filter.append(el.replace(request.GET.get('tab', 'attacks') + ':', ''))

    last_search_json = json.dumps(list_filter[:25])
    return JsonResponse(last_search_json, safe=False)
