import io
import json
import base64
import operator

import jsonref
import logging

from datetime import datetime
from json import JSONDecodeError
from wsgiref.util import FileWrapper
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http.response import JsonResponse, HttpResponse
from django.views.decorators.http import require_POST, require_GET
from django.contrib import messages
from django.shortcuts import render, redirect
from django.utils.translation import pgettext

from main.decorators import start_end_time
from main.decorators import time_of_function
from main.models.my_waf import LicKeys
from main.open_api_utils import get_openapi_paths_schemas
from main.open_api_utils import get_openapi_schema
from main.open_api_utils import get_paths_info
from main.open_api_utils import load_schema
from main.open_api_utils import send_request_del_path
from main.open_api_utils import send_request_del_schema
from main.open_api_utils import send_request_get_exclude_path
from main.open_api_utils import send_request_get_openapi_paths
from main.open_api_utils import send_request_set_exclude_path
from main.open_api_utils import send_request_set_schema
from main.open_api_utils import update_paths
from main.utils import get_list_messages
from my_waf.utils import check_keys
from my_waf.utils import get_waf_id
from utils.db import check_db

logger = logging.getLogger(__name__)

new_table_exist = check_db(logger)


@login_required
@start_end_time
@time_of_function
def open_api_settings(request, start_date, end_date):
    if not new_table_exist or not settings.API_URI:
        return redirect('index')
    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None

    keys = LicKeys.objects.filter(user__id=request.user.id)
    keys_list = []
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    for key in lic_keys:
        waf_id = get_waf_id(key.key)
        if waf_id is not None and key.tarif != 'Light':
            key.waf_id = waf_id
            keys_list.append(key)
    keys_list = sorted(keys_list, key=operator.attrgetter('description'))

    for key in keys_list:
        get_schemas = get_openapi_schema(key.key)
        if get_schemas is not None:
            for schema_key, schema_val in get_schemas.items():
                try:
                    get_schemas[schema_key] = get_paths_info(jsonref.loads(schema_val), schema_key)
                except JSONDecodeError as e:
                    logger.error(
                        'An error occurred when parsing path information in json format in the schema: {schema_name} '
                        'for the key: {key} with an error: {error}'
                        .format(schema_name=schema_key, key=key.get_formated_key(), error=str(e))
                    )
                    messages.error(
                        request,
                        pgettext('messages',
                                 'An error occurred when parsing path information in json format in the schema')
                    )

                    get_schemas[schema_key] = None
        key.get_schemas = get_schemas
    context = {
        'keys_list': keys_list,
        'user': request.user,
        'home': settings.HOME_PAGE,
        'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'end_date': end_date_q,
        'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'domain': q,
    }

    return render(request, 'main/openapi.html', context)


@require_POST
@login_required
def load_openapi_schema(request):
    key = request.POST.get('key')
    schema_name = request.POST.get('schema_name')
    old_name = request.POST.get('old_name')
    is_edit = request.POST.get('edit')
    key = LicKeys.objects.get(id=int(key))
    try:
        api_schema = request.FILES['api_schema'].read().decode('utf-8')
    except Exception:
        if not is_edit:
            logger.error(
                'An error occurred while uploading a file with a schema for the key: {key}'
                .format(key=key.get_formated_key())
            )
            messages.error(request, pgettext('messages', 'File not valid or file not exist'))

            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

        else:
            schemas = get_openapi_schema(key.key)
            api_schema = schemas.get(old_name)
            if api_schema is None:
                logger.error(
                    'An error occurred while processing an invalid schema for the key: {key}'
                    .format(key=key.get_formated_key())
                )
                messages.error(request, pgettext('messages', 'Schema with this name is not exist'))

                storage = messages.get_messages(request)
                list_messages = get_list_messages(storage)

                return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    rload = load_schema(key, api_schema, schema_name, old_name)
    if not rload[0]:
        messages.error(request, rload[1])

        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    if not old_name:
        messages.success(request,
                         pgettext('messages', 'The {} schema has been successfully added').format(schema_name))
    else:
        messages.success(request,
                         pgettext('messages', 'The {} schema has been successfully updated').format(schema_name))

    return JsonResponse(data={'status': 'success'})


@login_required
@start_end_time
@time_of_function
def get_openapi_paths(request, start_date, end_date):
    if not new_table_exist or not settings.API_URI:
        return redirect('index')
    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None

    keys = LicKeys.objects.filter(user__id=request.user.id)
    keys_list = []
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    for key in lic_keys:
        waf_id = get_waf_id(key.key)
        if waf_id is not None and key.tarif != 'Light':
            key.waf_id = waf_id
            keys_list.append(key)
    keys_list = sorted(keys_list, key=operator.attrgetter('description'))
    for key in keys_list:
        try:
            openapi_paths_schemas = get_openapi_paths_schemas(key.key)
            key.get_schemas = openapi_paths_schemas
        except Exception as e:
            logger.error(
                'An error occurred while get schema paths for key {key}: {error}'
                .format(key=key.get_formated_key(), error=str(e))
            )
            messages.error(request, pgettext('messages', 'An error occurred while get schema paths'))

    context = {
        'keys_list': keys_list,
        'user': request.user,
        'home': settings.HOME_PAGE,
        'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'end_date': end_date_q,
        'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'domain': q,
    }

    return render(request, 'main/openapi_paths.html', context)


@login_required
@start_end_time
@time_of_function
def get_openapi_exclude_paths(request, start_date, end_date):
    if not new_table_exist or not settings.API_URI:
        return redirect('index')
    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None

    keys = LicKeys.objects.filter(user__id=request.user.id)
    keys_list = []
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    for key in lic_keys:
        waf_id = get_waf_id(key.key)
        if waf_id is not None and key.tarif != 'Light':
            key.waf_id = waf_id
            keys_list.append(key)
    keys_list = sorted(keys_list, key=operator.attrgetter('description'))

    for key in keys_list:
        openapi_exclude_paths = send_request_get_exclude_path(key.key)
        if openapi_exclude_paths is not None:
            dict_exclude_paths = dict()
            for i in openapi_exclude_paths:
                try:
                    
                    if i.get('schema_name') not in dict_exclude_paths.keys():
                        dict_exclude_paths[i.get('schema_name', '-')] = list()
                    dict_exclude_paths[i.get('schema_name', '-')].append(i.get('url', '-'))
                except (TypeError, KeyError, AttributeError):
                    continue
            if dict_exclude_paths is not None:
                for i in dict_exclude_paths.keys():
                    dict_exclude_paths[i] = sorted(dict_exclude_paths[i])
            key.get_excl_paths = dict_exclude_paths
    context = {
        'keys_list': keys_list,
        'user': request.user,
        'home': settings.HOME_PAGE,
        'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'end_date': end_date_q,
        'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'domain': q,
    }

    return render(request, 'main/openapi_exclude_paths.html', context)


@require_POST
@login_required
@time_of_function
def merge_schema(request):
    key = request.POST.get('key')
    key = LicKeys.objects.get(id=int(key))
    schema_name = request.POST.get('schema_name')
    try:
        schema = get_openapi_schema(key.key)[schema_name]
    except KeyError:
        schema = '{{"openapi": "3.0.3","info": {{"description": "Title", "title": "Title", "version": "1.0.0"}}, ' \
                 '"servers": [{{"url":"{url}"}}], "paths": {{}}}}'.format(url=schema_name)

    get_schema_paths = dict()
    get_path = send_request_get_openapi_paths(key.key)
    for i in get_path['paths']:
        if not get_schema_paths.get(i['schema_name']):
            get_schema_paths[i['schema_name']] = dict()
        try:
            get_schema_paths[i['schema_name']][i['url']] = base64.b64decode(i['schema']).decode('utf-8', 'ignore')
        except (NameError, KeyError):
            logger.error(
                'An error occurred while get invalid schema paths for key {key}: "schema" is not exist in path'
                .format(key=key.get_formated_key())
            )
            messages.warning(request,
                             'An error occurred while get invalid schema paths for key {key}: '
                             '"schema" is not exist in path'
                             .format(key=key.get_formated_key())
                             )

        except Exception as e:
            logger.error(
                'An error occurred while get schema paths for key {key}: {error}'
                .format(key=key.get_formated_key(), error=str(e))
            )
            messages.warning(request,
                             'An error occurred while get schema paths for key {key}: {error}'
                             .format(key=key.get_formated_key(), error=str(e))
                             )
    try:
        schema = json.loads(schema.replace('\'', '\"'))
    except JSONDecodeError:
        logger.error(
            'An error occurred while parsing schemas paths in json format with key {key}.'
            .format(key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', 'An error occurred while parsing schemas paths in json format'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    paths = request.POST.getlist('paths[]', [])
    invalid_path = []
    for path in paths:
        if schema_name not in get_schema_paths.keys():
            logger.warning('Warning. Invalid schema name in the path schema: {}'.format(path))
            invalid_path.append(path)
            continue
        path_schema = json.loads(get_schema_paths[schema_name][path].replace('\'', '\"'))
        if path not in path_schema[schema_name].keys():
            logger.warning('Warning. Incorrect path schema, missing URL: {}'.format(path))
            invalid_path.append(path)
            continue
        for method in path_schema[schema_name][path].keys():
            default_response = {
                "200": {
                    "description": "Default response",
                    "content": {}
                }
            }
            path_schema[schema_name][path][method]["responses"] = default_response
        schema['paths'].update(path_schema[schema_name])

    if schema.get('paths'):
        response = send_request_set_schema(key.key, schema_name, json.dumps(schema))
    else:
        response = None
    for path in invalid_path:
        if path in paths:
            paths.remove(path)
    if response:
        for i in paths:
            response = send_request_del_path(key.key, schema_name, i)
            if not response:
                break
        if response:
            update_paths(key.key)
            schema = get_paths_info(schema, schema_name)
            messages.success(
                request,
                pgettext(
                    'messages', 'The paths are successfully merged into the {} schema'
                ).format(schema_name)
            )

            return JsonResponse(data={'status': 'ok', 'schema': schema, 'schema_name': schema_name})

        else:
            logger.error(
                'An error occurred while deleting paths for the key: {key}'
                .format(key=key.get_formated_key())
            )
            messages.error(request, pgettext('messages', "An error occurred while deleting paths"))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    else:
        logger.error(
            'An error occurred while adding the schema for the key: {key}'
            .format(key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', "An error occurred while adding the schema"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@login_required
@time_of_function
def delete_openapi_schema(request):
    schema_name = request.POST.get('schema_name')
    key = request.POST.get('key')
    key = LicKeys.objects.get(id=int(key))
    response = send_request_del_schema(key.key, schema_name)
    update_paths(key.key)
    if response:
        return JsonResponse(data={'status': 'ok'})
    else:
        logger.error(
            'An error occurred while deleting schema for the key: {key}'
            .format(key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', "An error occurred while deleting schema"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@login_required
@time_of_function
def delete_openapi_path(request):
    schema_name = request.POST.get('schema_name')
    url = request.POST.get('url')
    key = request.POST.get('key')
    key = LicKeys.objects.get(id=int(key))
    response = send_request_del_path(key.key, schema_name, url)
    if response:
        return JsonResponse(data={'status': 'ok'})
    else:
        logger.error(
            'An error occurred while deleting paths for the key: {key}'
            .format(key=key.get_formated_key())
        )

        messages.error(request, pgettext('messages', "An error occurred while deleting paths"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@login_required
@time_of_function
def delete_url_from_openapi_schema(request):
    data = request.POST.get('data')
    if data is None:
        messages.error(request, pgettext('messages', 'Invalid data'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    data = json.loads(data)
    schema_name = data.get('schema_name')
    url = data.get('url')
    key = data.get('key')
    key = LicKeys.objects.get(id=int(key))
    get_schemas = get_openapi_schema(key.key)
    schema = get_schemas.get(schema_name)
    if schema:
        schema = json.loads(schema)
        for i in url:
            if schema['paths'].get(i) is not None:
                del schema['paths'][i]
        if len(schema['paths'].keys()) > 0:
            response = send_request_set_schema(key.key, schema_name, json.dumps(schema))
            update_paths(key.key)
        else:
            response = send_request_del_schema(key.key, schema_name)
            update_paths(key.key)

        if response:
            return JsonResponse(data={'status': 'ok'})
        else:
            logger.error(
                'An error occurred while deleting schema for the key: {key}'
                .format(key=key.get_formated_key())
            )
            messages.error(request, pgettext('messages', "An error occurred while deleting schema"))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    else:
        return JsonResponse(data={'status': 'ok'})


@require_POST
@login_required
@time_of_function
def exclude_openapi_path(request):
    path = request.POST.get('url')
    key = request.POST.get('key')
    schema_name = request.POST.get('schema_name')
    key = LicKeys.objects.get(id=int(key))
    urls = send_request_get_exclude_path(key.key)
    if urls:
        urls = list(urls)
    else:
        urls = list()

    urls.append({'schema_name': schema_name, 'url': path})

    urls = [dict(s) for s in set(frozenset(d.items()) for d in urls)]

    response_exclude = send_request_set_exclude_path(key.key, list(urls))
    if response_exclude:
        response = send_request_del_path(key.key, schema_name, path)
        if response:
            return JsonResponse(data={'status': 'ok'})
    else:
        logger.error(
            'An error occurred while set exclude paths for the key: {key}'
            .format(key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', "An error occurred while set exclude paths"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@login_required
@time_of_function
def remove_exclude_openapi_path(request):
    paths = request.POST.get('paths')
    paths = json.loads(paths)
    key = request.POST.get('key')
    schema_name = request.POST.get('schema_name')
    key = LicKeys.objects.get(id=int(key))
    excl_paths = send_request_get_exclude_path(key.key)
    dict_exclude_paths = dict()
    for i in excl_paths:
        if i['schema_name'] not in dict_exclude_paths.keys():
            dict_exclude_paths[i.get('schema_name', '-')] = list()
        dict_exclude_paths[i.get('schema_name', '-')].append(i.get('url', '-'))
    if schema_name in dict_exclude_paths.keys():
        for i in paths:
            if i.get('url', '-') in dict_exclude_paths[schema_name]:
                dict_exclude_paths[schema_name].remove(i.get('url', '-'))
    excl_paths = list()
    for k, v in dict_exclude_paths.items():
        for i in v:
            excl_paths.append({'schema_name': k, 'url': i})

    response = send_request_set_exclude_path(key.key, list(excl_paths))
    if response:
        return JsonResponse(data={'status': 'ok'})
    else:
        logger.error(
            'An error occurred while deleting exclude paths for the key: {key}'
            .format(key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', "An error occurred while deleting exclude paths"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_GET
@login_required
@time_of_function
def download_open_api_schema(request):
    key = request.GET.get('key')
    schema_name = request.GET.get('schema_name')
    key = LicKeys.objects.get(id=int(key)).key
    get_schema = get_openapi_schema(key)[schema_name]
    get_schema = json.loads(get_schema)
    json_file = io.StringIO()
    json.dump(get_schema, json_file, indent=2)
    json_file.seek(0)
    wrapper = FileWrapper(json_file)
    response = HttpResponse(wrapper, content_type='application/json')
    response['Content-Disposition'] = 'attachment; filename={schema_name}.json'.format(schema_name=schema_name)
    return response


@require_POST
@login_required
@time_of_function
def remove_schema_path(request):
    key = request.POST.get('key')
    schema_name = request.POST.get('schema_name')
    path_name = request.POST.get('path_name')
    get_schema = get_openapi_schema(key)[schema_name]
    get_schema = json.loads(get_schema)
    try:
        del get_schema[schema_name]['paths'][path_name]
    except KeyError:
        logger.error(
            'An error occurred while remove paths paths from schema for the key: {key}'
            .format(key='xxx' + key[-6:])
        )

        messages.error(request, pgettext('messages', "An error occurred while remove paths from schema"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    send_request_set_schema(key, schema_name, get_schema)
    update_paths(key)
    return JsonResponse(data={'status': 'ok'})
