import json
import logging

from datetime import datetime
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.views.generic import TemplateView
from django.contrib import messages
from django.utils.translation import pgettext

from pymemcache.client.base import Client

from main.common import get_main_logs, get_filtered_logs, get_ml_table, get_groups, get_vhost
from main.decorators import start_end_time, start_end_time_decorator, time_of_function
from main.models.wafuser import WAFUSER
from main.models.logs import Logs
from django.db import connections

from main.search import search_filter
from main.utils import get_params, get_list_messages

logger = logging.getLogger(__name__)

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))

PAGE_SIZE = settings.DEFAULT_PAGE_SIZE

vGeoIP = settings.V_GEOIP

isCCEnable = None
try:
    r = "SELECT * FROM information_schema.columns WHERE column_name = 'cc' AND table_name = 'attack'"
    with connections['waf'].cursor() as cursor:
        cursor.execute(r)
        res = cursor.fetchone()
        if res is None:
            isCCEnable = False
        else:
            isCCEnable = True
except Exception:
    isCCEnable = False


class ScrollAttacksView(LoginRequiredMixin, TemplateView):
    template_name = 'main/scroll_attacks.html'

    @start_end_time_decorator
    @time_of_function
    def get(self, request, start_date, end_date, *args, **kwargs):
        q = request.GET.get('domain', None)
        end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
        if q == 'None':
            q = None
        s = request.GET.get('search', '').strip()

        user = WAFUSER.objects.get(id=request.user.id)
        search_history = user.search_history

        try:
            if search_history is None:
                search_history = ''
            search_history_dict = json.loads(search_history)
        except json.JSONDecodeError:
            search_history_dict = dict()
            for i in search_history.split(','):
                search_history_dict[i.strip()] = 1

        if s is not None and s != '':
            if 'attacks:' + s in search_history_dict.keys():
                reb = dict()
                reb['attacks:' + s] = search_history_dict['attacks:' + s] + 1
                search_history_dict.pop('attacks:' + s)
                reb.update(search_history_dict)
                search_history_dict = reb

            else:
                reb = dict()
                reb['attacks:' + s] = 1
                reb.update(search_history_dict)
                search_history_dict = reb
        else:
            for el in search_history_dict.keys():
                if 'attacks:' in el:
                    s = el.replace('attacks:', '')
                    break
        if s == '#clean':
            s = ''

        while len(search_history_dict) > 1000:
            search_history_dict.popitem()

        user.search_history = json.dumps(search_history_dict)
        user.save()

        attack_history = []

        sortbyfreq = True
        if sortbyfreq:
            tmp = dict()
            last_use = ''
            for el in list(search_history_dict.keys()):
                if 'attacks:' in el and '#clean' not in el and 'attacks:' + s != el:
                    last_use = el
                    break

            if last_use in search_history_dict:
                tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
                search_history_dict.pop(last_use)
            tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
            search_history_dict = tmp

        for el in search_history_dict.keys():
            if 'attacks:' in el and '#clean' not in el and s.lower() in el.replace('attacks:', '').lower():
                attack_history.append(el.replace('attacks:', ''))

        # get user alias
        user_alias = WAFUSER.objects.get(id=request.user.id).waf_alias
        domain_from_wafuser = WAFUSER.objects.get(id=request.user.id).domain

        if user_alias:
            main = Logs.objects.using('waf').filter(timestamp__gte=start_date, timestamp__lte=end_date)
            main = get_filtered_logs(main, user_alias, domain_from_wafuser, q)

            if s:
                try:
                    main = search_filter(s, main)
                except Exception:
                    main = Logs.objects.using('waf').none()

            # group attacks
            group = False
            params = get_params(False)

            main = main.values('vhost', 'bt', 'mz', 'rule_id').annotate(
                **params
            ).order_by('-timestamp__max')

            paginator = Paginator(main, PAGE_SIZE)
            page = int(request.GET.get('page', 1))
            try:
                main = paginator.page(page)
            except PageNotAnInteger:
                # If page is not an integer, deliver first page.
                main = paginator.page(1)
            except EmptyPage:
                return HttpResponse('')

            mcache = Client(cache_address)
            ml_table = get_ml_table()
            main = get_main_logs(main, ml_table, mcache)

            # Passing data to template
            context = self.get_context_data(
                home=settings.HOME_PAGE,
                table=main,
                user=request.user,
                start_date=datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
                end_date=end_date_q,
                end_date_short=datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
                group=group,
                domain=q,
            )

            if 'last_search' not in request.session.keys():
                request.session['current_search'] = request.GET.get('search', '').strip()
                request.session['last_search'] = ''
            else:
                request.session['last_search'] = request.session['current_search']
                request.session['current_search'] = request.GET.get('search', '').strip()

            return self.render_to_response(context)
        else:
            return HttpResponse("Access closed", 200)


class ScrollAttacksSuggestionsView(LoginRequiredMixin, TemplateView):
    template_name = 'main/scroll_attacks.html'

    @time_of_function
    def get(self, request, *args, **kwargs):
        s = request.GET.get('search', '').strip()

        user = WAFUSER.objects.get(id=request.user.id)
        search_history = user.search_history

        try:
            if search_history is None:
                search_history = ''
            search_history_dict = json.loads(search_history)
        except json.JSONDecodeError:
            search_history_dict = dict()
            for i in search_history.split(','):
                search_history_dict[i.strip()] = 1

        request_count = 0
        if s is not None and s != '':
            if 'attacks:' + s in search_history_dict.keys():
                reb = dict()
                reb['attacks:' + s] = search_history_dict['attacks:' + s] + 1
                search_history_dict.pop('attacks:' + s)
                reb.update(search_history_dict)
                search_history_dict = reb

            else:
                reb = dict()
                reb['attacks:' + s] = 1
                reb.update(search_history_dict)
                search_history_dict = reb
            for el in search_history_dict.keys():
                if 'attacks:' in el:
                    request_count += 1
                    if request_count == 25:
                        search_history_dict.pop(el)
        else:
            for el in search_history_dict.keys():
                if 'attacks:' in el:
                    s = el.replace('attacks:', '')
                    break
        if s == '#clean':
            s = ''

        user.search_history = json.dumps(search_history_dict)
        user.save()

        attack_history = []

        sortbyfreq = True
        if sortbyfreq:
            tmp = dict()
            last_use = ''
            for el in list(search_history_dict.keys()):
                if 'attacks:' in el and '#clean' not in el and 'attacks:' + s != el:
                    last_use = el
                    break

            if last_use in search_history_dict:
                tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
                search_history_dict.pop(last_use)
            tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
            search_history_dict = tmp

        for el in search_history_dict.keys():
            if 'attacks:' in el and '#clean' not in el and s.lower() in el.replace('attacks:', '').lower():
                attack_history.append(el.replace('attacks:', ''))

        # get user alias
        user_alias = WAFUSER.objects.get(id=request.user.id).waf_alias
        if user_alias:
            # Passing data to template
            context = self.get_context_data(
                home=settings.HOME_PAGE,
                user=request.user,
                attack_history=attack_history[:25],
                history_id_list='attack_searh_history_list',
            )

            return self.render_to_response(context)
        else:
            return HttpResponse("Access closed", 200)


@login_required
@start_end_time
@time_of_function
def scroll_attacks(request, start_date, end_date):
    q = request.GET.get('domain', None)
    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    if q == 'None':
        q = None
    s = request.GET.get('search', '').strip()

    user = WAFUSER.objects.get(id=request.user.id)
    search_history = user.search_history

    try:
        if search_history is None:
            search_history = ''
        search_history_dict = json.loads(search_history)
    except json.JSONDecodeError:
        search_history_dict = dict()
        for i in search_history.split(','):
            search_history_dict[i.strip()] = 1

    if s is not None and s != '':
        if 'attacks:' + s in search_history_dict.keys():
            reb = dict()
            reb['attacks:' + s] = search_history_dict['attacks:' + s] + 1
            search_history_dict.pop('attacks:' + s)
            reb.update(search_history_dict)
            search_history_dict = reb

        else:
            reb = dict()
            reb['attacks:' + s] = 1
            reb.update(search_history_dict)
            search_history_dict = reb
    else:
        for el in search_history_dict.keys():
            if 'attacks:' in el:
                s = el.replace('attacks:', '')
                break
    if s == '#clean':
        s = ''

    while len(search_history_dict) > 1000:
        search_history_dict.popitem()

    user.search_history = json.dumps(search_history_dict)
    user.save()

    attack_history = []
    history_id_list = 'attack_searh_history_list'

    sortbyfreq = True
    if sortbyfreq:
        tmp = dict()
        last_use = ''
        for el in list(search_history_dict.keys()):
            if 'attacks:' in el and '#clean' not in el and 'attacks:' + s != el:
                last_use = el
                break

        if last_use in search_history_dict:
            tmp[last_use] = search_history_dict[last_use] if last_use != '' else '#clear'
            search_history_dict.pop(last_use)
        tmp.update({k: v for k, v in sorted(search_history_dict.items(), key=lambda item: item[1], reverse=True)})
        search_history_dict = tmp

    for el in search_history_dict.keys():
        if 'attacks:' in el and '#clean' not in el and s.lower() in el.replace('attacks:', '').lower():
            attack_history.append(el.replace('attacks:', ''))

    # get user alias
    user_alias = WAFUSER.objects.get(id=request.user.id).waf_alias
    domain_from_wafuser = WAFUSER.objects.get(id=request.user.id).domain

    if user_alias:
        main = Logs.objects.using('waf').filter(timestamp__gte=start_date, timestamp__lte=end_date)
        main = get_filtered_logs(main, user_alias, domain_from_wafuser, q)

        if s:
            try:
                main = search_filter(s, main)
            except Exception as e:
                logger.error(
                    "An error occurred when loading attacks with the filter: \"{filter}\". Error description: {error}"
                    .format(filter=s, error=e)
                )
                messages.warning(request, pgettext('messages', 'Invalid search string'))
                storage = messages.get_messages(request)
                list_messages = get_list_messages(storage)

                Logs.objects.using('waf').none()
                return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

        # group attacks
        group = False

        mcache = Client(cache_address)
        ml_table = get_ml_table()

        groups_attack = True

        if groups_attack:
            page = int(request.GET.get('page', 1))
            start_page = PAGE_SIZE * (page - 1)
            end_page = PAGE_SIZE * page
            vhosts_list = get_vhost(main).values_list('vhost', flat=True)
            count_all = vhosts_list.count()
            vhosts_list = vhosts_list[start_page:end_page]
            main = main.filter(vhost__in=list(vhosts_list))
            main = get_groups(main)
            template = 'main/scroll_groups.html'
        else:
            main = main.values('vhost', 'bt', 'mz', 'rule_id').order_by('-timestamp')

            template = 'main/scroll_attacks.html'
            count_all = len(main)
            paginator = Paginator(main, PAGE_SIZE)
            page = int(request.GET.get('page', 1))
            try:
                main = paginator.page(page)
            except PageNotAnInteger:
                # If page is not an integer, deliver first page.
                main = paginator.page(1)
            except EmptyPage:
                return HttpResponse('')
            main = get_main_logs(main, ml_table, mcache)

        if count_all == len(main):
            page = 0
        # Passing data to template
        context = {
            'home': settings.HOME_PAGE,
            'table': main,
            'user': request.user,
            'attack_history': attack_history[:25],
            'history_id_list': history_id_list,
            'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'end_date': end_date_q,
            'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
            'group': group,
            'domain': q,
            'groups_attack': groups_attack,
            'page': page,
        }

        return render(request, template, context)
    else:
        return HttpResponse("Access closed", 200)
