import multiprocessing
import operator
import qrcode
import base64
import pyotp

from io import BytesIO
from datetime import datetime

from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.password_validation import validate_password
from django.core.mail import BadHeaderError
from django.template.loader import render_to_string
from django.db.models.query_utils import Q
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from django.contrib.auth.tokens import default_token_generator
from django.utils.encoding import force_bytes
from django.forms import Form, BooleanField, CharField
from django.http import HttpResponse, JsonResponse
from django.contrib.auth.decorators import login_required
from django.contrib import auth
from django.conf import settings
from django.shortcuts import render, redirect
from django.views.generic.edit import BaseCreateView
from django.utils.translation import gettext as _

from main.decorators import start_end_time
from main.decorators import time_of_function
from main.models import Config
from main.models import LicKeys
from main.models import UserConfig
from main.models import WAFUSER
from my_waf.utils import get_lic_keys
from utils.db import check_db
from utils.mail import send_mail_from_cabinet

new_table_exist = check_db(None)


class UserConfigForm(Form):
    notification_enable = BooleanField(required=False)
    email_details = BooleanField(required=False)
    bt1 = BooleanField(required=False)
    bt2 = BooleanField(required=False)
    bt3 = BooleanField(required=False)
    bt4 = BooleanField(required=False)
    bt5 = BooleanField(required=False)
    bt7 = BooleanField(required=False)
    bt8 = BooleanField(required=False)
    bt9 = BooleanField(required=False)
    bt10 = BooleanField(required=False)
    bt14 = BooleanField(required=False)
    bt16 = BooleanField(required=False)


class DomainsConfigForm(Form):
    vhosts = CharField()


class NotificationConfigView(LoginRequiredMixin, BaseCreateView):
    model = UserConfig
    form_class = UserConfigForm
    template_name = 'main/user_config.html'
    success_url = 'edit_user_config/'

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            user_config = request.user.userconfig
            for field, value in form.cleaned_data.items():
                if field in request.POST.keys():
                    setattr(user_config, field, value)
                    if field == 'bt1':
                        setattr(user_config, 'bt2', value)
            user_config.save()
            return HttpResponse(status=200)
        else:
            return self.form_invalid(form)


class DomainsConfigView(LoginRequiredMixin, BaseCreateView):
    model = UserConfig
    form_class = DomainsConfigForm
    template_name = 'main/user_config.html'
    success_url = 'edit_domain_config/'

    def post(self, request, *args, **kwargs):
        form = self.get_form()

        if form.is_valid():
            user_config = request.user.userconfig
            for field, value in form.cleaned_data.items():
                if field in request.POST.keys():
                    if type(value) is str:
                        if value.strip(' ') == '':
                            value = '*'
                    setattr(user_config, field, value)
            user_config.save()
            return HttpResponse(status=200)
        else:
            return self.form_invalid(form)


@login_required
@time_of_function
def edit_user_password(request):
    context = dict()

    if request.method == 'POST':
        user_id = request.POST.get('user_id')
        new_password = request.POST.get('password')
        cur_password = request.POST.get('currentPassword')

        if int(user_id) > 0:
            user = WAFUSER.objects.get(id=user_id)
            if user.check_password(cur_password) is False:
                context["error"] = _("The current password was entered incorrectly")
                return JsonResponse(context)
            try:
                validate_password(new_password)
                if new_password:
                    user.set_password(new_password)
                    auth.login(request, user)
                user.save()
            except Exception:
                context["error"] = _("Password invalid. Password must contain at least 8 characters, at least 1 "
                                     "number, letters and special characters.")
                return JsonResponse(context)

        return HttpResponse(status=200)


def switch_2fa_status(user: WAFUSER, switch_2fa_to: bool, secret_key=None):
    """ on/off 2fa for user"""
    user.is_2fa_active = switch_2fa_to
    if secret_key:
        user.secret = secret_key
    user.save()


@login_required
def check_and_turn_on_2fa(request) -> HttpResponse:
    """ Additional check before enabling 2FA """
    # todo refactor move to user
    user = WAFUSER.objects.get(email=request.user.email)
    secret_2FA_key = request.POST.get('secret_2FA_key')
    code_input = request.POST.get('code_input')
    check_status = pyotp.TOTP(secret_2FA_key).verify(code_input)
    if check_status:
        switch_2fa_status(user, switch_2fa_to=True, secret_key=secret_2FA_key)
        status_code = 200
    else:
        status_code = 406  # means that the wrong code was entered
    return HttpResponse(status=status_code)


@login_required
@time_of_function
def render_2fa(request):
    # todo refactoring: move to user
    if request.method == 'POST':
        user = WAFUSER.objects.get(email=request.user.email)
        value = request.POST.get('box_value')

        if value == 'false':
            switch_2fa_status(user, switch_2fa_to=False)
            return JsonResponse({'deleted': 'done'}, status=200)
        else:
            secret_2FA_key = pyotp.random_base32()
            totp = pyotp.TOTP(secret_2FA_key)
            otp_auth = totp.provisioning_uri(name=user.email, issuer_name="nemesida")
            qr = qrcode.make(otp_auth)
            buf = BytesIO()
            qr.save(buf, format="JPEG")
            qr_byte = buf.getvalue()
            qr_b64 = base64.b64encode(qr_byte).decode('utf-8')
            return JsonResponse({"qrcode": "data:image/png;base64,{}".format(qr_b64),
                                 'secret_2FA_key': secret_2FA_key},
                                status=200)

    if request.method == 'GET':
        user = WAFUSER.objects.get(email=request.user.email)
        if user.is_2fa_active and user.secret:
            totp = pyotp.TOTP(user.secret)
            otp_auth = totp.provisioning_uri(name=user.email, issuer_name="nemesida")
            qr = qrcode.make(otp_auth)
            buf = BytesIO()
            qr.save(buf, format="JPEG")
            qr_byte = buf.getvalue()
            qr_b64 = base64.b64encode(qr_byte).decode('utf-8')
            return JsonResponse({"qrcode": "data:image/png;base64,{}".format(qr_b64)}, status=200)
        else:
            return JsonResponse({'status': 'done'}, status=200)


@login_required
@start_end_time
@time_of_function
def user_settings(request, start_date, end_date):
    # todo refactoring: move to user
    email_en = Config.objects.only("send_email").first()
    if not email_en:
        email_en = False
    else:
        email_en = email_en.send_email

    end_date_q = datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f')
    q = request.GET.get('domain', None)
    if q == 'None':
        q = None

    keys = LicKeys.objects.filter(user__id=request.user.id).all()

    processes = []
    processes_queue = multiprocessing.Queue()
    for key in keys:
        proc = multiprocessing.Process(target=get_lic_keys, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    keys_form_multiprocessing = sorted(keys_form_multiprocessing, key=operator.attrgetter('description'))
    context = {
        'user': request.user,
        'lic_keys': keys_form_multiprocessing,
        'user_config': request.user.userconfig,
        'home': settings.HOME_PAGE,
        'start_date': datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'end_date': end_date_q,
        'end_date_short': datetime.strptime(end_date, '%Y-%m-%d %H:%M:%S.%f').strftime('%d.%m.%Y'),
        'st_exist': new_table_exist,
        'domain': q,
        'email_enable': email_en,
    }

    return render(request, 'main/user_config.html', context)


@time_of_function
def password_reset_request(request):
    context = dict()
    if request.method == "POST":
        password_reset_form = PasswordResetForm(request.POST)

        if password_reset_form.is_valid():
            data = password_reset_form.cleaned_data['email']
            associated_users = WAFUSER.objects.filter(Q(email=data))

            if associated_users.exists():
                for user in associated_users:
                    subject = "Nemesida WAF Cabinet Password Reset"
                    email_template_name = "registration/password_reset_email.html"
                    c = {
                        'domain': request.get_host(),
                        'home': settings.HOME_PAGE[:-1],
                        "uid": urlsafe_base64_encode(force_bytes(user.pk)),
                        "user": user,
                        'token': default_token_generator.make_token(user),
                        'protocol': 'http',
                    }
                    email = render_to_string(email_template_name, c)
                    try:
                        send_mail_from_cabinet(user.email, subject, email)
                    except BadHeaderError:
                        return HttpResponse('Invalid header found.')
                    return redirect(settings.HOME_PAGE + "password_reset/done/")
        context["error"] = _("Email invalid")

    return render(request=request, template_name="registration/password_reset_form.html", context=context)


@time_of_function
def password_reset_confirm(request, uidb64, token):
    user_id = urlsafe_base64_decode(uidb64).decode()
    context = dict()
    user = WAFUSER.objects.get(id=user_id)
    if request.method == "POST":

        new_password1 = request.POST.get('new_password1')
        new_password2 = request.POST.get('new_password2')

        if new_password1 != new_password2:
            context["error"] = _("Password mismatch")
        else:
            try:
                validate_password(new_password1)
                if new_password1:
                    user.set_password(new_password1)
                    auth.login(request, user)
                user.save()
                return redirect(settings.HOME_PAGE + "reset/done/")
            except Exception:
                context["error"] = _("Password invalid. Password must contain at least 8 characters, at least 1 "
                                     "number, letters and special characters.")

    if not default_token_generator.check_token(user, token):
        context['validlink'] = False
    else:
        context['validlink'] = True
    return render(request=request, template_name="registration/password_reset_confirm.html", context=context)
