import re
from datetime import datetime
from django.db.models import Q

from main.utils import convert_to_punycode


def search_date(query, table_cls):
    ids = []
    rows = table_cls.objects.values('id', 'timestamp')
    for row in rows:
        date = datetime.utcfromtimestamp(
            int(row['timestamp'])).strftime('%H:%M:%S %d.%m.%y')
        if query in str(date):
            ids.append(row['id'])

    return ids


class QueryParser:
    SEARCH_FIELDS_MAPPING = {
        'h': 'vhost',
        'url': 'uri',
        'path': 'uri',
        'headers': 'other_headers',
        'body': '-',
    }

    def __init__(self, query, table_cls):
        self.query = query.strip()
        self.table_cls = table_cls
        self.query_dict = {}
        self.counter = 0
        self.fields = [
            f.name
            for f in table_cls._meta.get_fields()
            if f.name not in {'id'}
        ]

    @staticmethod
    def parse_expression(query):
        """ split query by 'and' or 'or' expression """
        or_expr = re.match(r'^.+\sor', query)
        and_expr = re.match(r'^.+\sand', query)
        if or_expr is not None:
            query = query.split(' or ', 1)
            query.append('OR')
        elif and_expr is not None:
            query = query.split(' and ', 1)
            query.append('AND')
        else:
            query = [query, None, 'AND']
        return query

    def parse_search(self, query, conn='AND'):
        """
        parse user input to dict of fields
        example: {q1:['bt', 'value', 'OR']}
        """
        fields = query.split(':', 1)
        if len(fields) == 1:
            for f in self.fields:
                self.counter += 1
                self.query_dict['q' + str(self.counter)] = [f, fields[0], 'OR']
            return self.query_dict
        else:
            param = fields[0]
            param_with_negative = False
            if param.startswith('!'):
                param = self.SEARCH_FIELDS_MAPPING.get(param[1:], param[1:])
                param_with_negative = True
                check_param = param
            else:
                param = self.SEARCH_FIELDS_MAPPING.get(param, param)
                check_param = param
            fields = self.parse_expression(fields[1])
            self.counter += 1

            if check_param in self.fields:
                if param_with_negative:
                    self.query_dict['q' + str(self.counter)] = ['!' + param, fields[0], conn]
                else:
                    self.query_dict['q' + str(self.counter)] = [param, fields[0], conn]
            else:
                for f in self.fields:
                    self.counter += 1
                    query_string = query.split(':', 1)[0] + ":" + fields[0]
                    self.query_dict['q' + str(self.counter)] = [f, query_string, 'OR']
            if fields[1] is not None:
                self.parse_search(fields[1], fields[2])
            return self.query_dict

    def search_gen(self):
        """ generate Q search object """
        res = self.parse_search(self.query)
        search_object = Q()
        for key, value in res.items():
            if value[1].startswith('bt:'):
                bt_num = value[1].split(':')[-1]
                if bt_num != '11':
                    search_object.add(Q(id__isnull=True), 'AND')
            elif value[1] != '':
                if value[0] == 'timestamp':
                    query = search_date(value[1], self.table_cls)
                    search_object.add(Q(**{'id__in': query}), value[2])
                elif value[0] == 'vhost':
                    search_q = convert_to_punycode(value[1])
                    negation_search_filter = value[0].startswith('!')
                    if negation_search_filter:
                        if re.match(r'^\.', search_q):
                            search_object.add(~Q(**{value[0][1:] + '__icontains': search_q}), value[2])
                        else:
                            search_object.add(~Q(**{value[0][1:] + '__istartswith': search_q}), value[2])
                    else:
                        if re.match(r'^\.', search_q):
                            search_object.add(Q(**{value[0] + '__icontains': search_q}), value[2])
                        else:
                            search_object.add(Q(**{value[0] + '__istartswith': search_q}), value[2])
                elif value[0] == 'waf_id':
                    search_q = value[1]
                    negation_search_filter = value[0].startswith('!')
                    if negation_search_filter:
                        search_object.add(~Q(**{value[0][1:] + '__iexact': search_q}), value[2])
                    else:
                        search_object.add(Q(**{value[0] + '__iexact': search_q}), value[2])
                else:
                    search_q = convert_to_punycode(value[1])
                    negation_search_filter = value[0].startswith('!')
                    if negation_search_filter:
                        search_object.add(~Q(**{value[0][1:] + '__icontains': search_q.replace('url', 'uri').replace('path', 'uri')}),
                                          value[2])
                    else:
                        search_object.add(Q(**{value[0] + '__icontains': search_q.replace('url', 'uri').replace('path', 'uri')}), value[2])
        return search_object
