import os
import json
import copy
import threading
import time
import base64
import logging

from pathlib import Path
from urllib.parse import unquote, unquote_plus
from pymemcache.client.base import Client
from pymemcache.client import base

from django.contrib import messages
from django.utils.translation import pgettext
from django.shortcuts import render, redirect
from django.views.decorators.http import require_http_methods
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger, Page
from django.http import JsonResponse, HttpResponse
from django.conf import settings
from django.core import serializers
from django.http import HttpResponseRedirect
from django.contrib.auth.decorators import login_required
from django.db.models import F

from .search import QueryParser
from .utils import format_logs_mt
from .utils import split_zone_logs
from main.common import chunks
from main.common import get_filtered_logs
from main.common import get_list_plans
from main.common import get_location_by_ip
from main.common import get_location_from_remote
from main.common import getting_geoip_multithread
from main.decorators import time_of_function
from main.emulator import NWAF_emu
from main.fields import ExportFields
from main.fields import FieldNames
from main.models import Geo_Tokens
from main.models import Logs
from main.models import MtExport
from main.models import MtTable
from main.models import remove_duplicated_records
from main.models import WAFUSER
from main.utils import get_list_messages
from main.utils import get_update_time
from main.utils import remove_zero_from_ip
from utils.db import check_db
from utils.http import check_geo_ip_tokens

logger = logging.getLogger(__name__)

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))
mclient = base.Client(cache_address)

dict_replace = {'url': 'uri', 'path': 'uri', 'param': 'args', 'alias': 'waf_id'}
vGeoIP = settings.V_GEOIP

new_table_exist = check_db(logger)


@login_required
@time_of_function
def search_in_table(request):
    query = request.GET.get('q', '')
    table = request.GET.get('t', 'p')
    if table == 'p':
        rows = MtExport.objects_for_mt.using('waf').order_by('-timestamp')
    else:
        rows = MtTable.objects_for_mt.using('waf').order_by('-timestamp')
    if table == 'p':
        table_cls = MtExport
    else:
        table_cls = MtTable

    # search
    if query:
        qp = QueryParser(query.lower(), table_cls)
        search_object = qp.search_gen()
        rows = rows.filter(search_object)

    if request.GET.get('only_num', False):
        return JsonResponse({'count': rows.count()}, safe=False)

    # paginator
    page = request.GET.get('page', 1)
    # this EDIT
    paginator = Paginator(rows, settings.DEFAULT_MT_PAGE_SIZE)

    try:
        numbers = paginator.page(page)
    except PageNotAnInteger:
        numbers = paginator.page(1)
    except EmptyPage:
        numbers = paginator.page(paginator.num_pages)
    field_names = FieldNames().__dict__

    numbers = split_zone_logs(format_logs_mt(numbers))

    return render(request, 'search.html', {'rows': numbers, 't': table, 'names': field_names, 'q': query})


@login_required
@time_of_function
def index_mt(request):
    if not new_table_exist:
        return redirect('index')
    query = request.GET.get('q', '')
    table = request.GET.get('t', '')
    search = request.GET.get('s', None)
    wafuser = WAFUSER.objects.get(id=request.user.id)
    user_alias = wafuser.waf_alias
    domain_from_wafuser = wafuser.domain
    plans = get_list_plans(request.user)
    if 'Enterprise' in plans:
        user_plans = 0
    elif 'Business' in plans or 'Light' in plans:
        user_plans = 1
    else:
        user_plans = -1

    # get table querysets
    if table == 'p':
        rows = MtExport.objects_for_mt.using('waf').order_by('-timestamp')
        rows = get_filtered_logs(rows, user_alias, domain_from_wafuser)
        table_cls = MtExport
    else:
        rows = MtTable.objects_for_mt.using('waf').order_by('-timestamp')
        rows = get_filtered_logs(rows, user_alias, domain_from_wafuser)
        table_cls = MtTable

    # search
    if query:
        qp = QueryParser(query.lower(), table_cls)
        search_object = qp.search_gen()
        if table == 'n':
            rows = rows.filter(search_object)
        elif table == 'p':
            rows = rows.filter(search_object)
        else:
            rows = rows.filter(search_object)

    # paginator
    page = request.GET.get('page', 1)
    paginator = Paginator(rows, settings.DEFAULT_MT_PAGE_SIZE)
    try:
        rows = paginator.page(page)
    except PageNotAnInteger:
        rows = paginator.page(1)
    except EmptyPage:
        rows = Paginator([], settings.DEFAULT_MT_PAGE_SIZE).page(1)

    field_names = FieldNames().__dict__
    rows = split_zone_logs(format_logs_mt(rows))
    rows = check_geo_ip_tokens(main=rows, filed="waf_id")
    if table_cls is MtTable:
        getting_geoip_multithread(rows, mclient)

        for i in rows:
            i.check_ip_country()

    if table == 'p':
        if search:
            template = 'smain/_search.html'
        else:
            template = 'smain/false_positive.html'
        return render(request, template, {'rows': rows, 'names': field_names, 'user_plans': user_plans,
                                          'home': settings.HOME_PAGE, 'q': query, 't': table})
    else:
        if search:
            template = 'smain/_search.html'
        else:
            template = 'smain/index_mt.html'
        return render(request, template, {'rows': rows, 'names': field_names, 'user_plans': user_plans,
                                          'home': settings.HOME_PAGE, 'q': query, 't': table})


@login_required
@time_of_function
def detail_row(request, row_id):
    table = request.GET.get('t')
    cls_log = request.GET.get('cls_log', 'MtTable')
    edit = request.GET.get('edit', 'false')
    if table == 'p':
        rows = MtExport.objects_for_mt.using('waf').filter(id=row_id).format()
    else:
        if cls_log == 'Logs':
            row_orinal = Logs.objects.using('waf').filter(id=row_id).first()
            rows = MtTable.objects_for_mt.using('waf').filter(request_id=row_orinal.request_id, vhost=row_orinal.vhost,
                                                              mz__icontains=row_orinal.mz).logs_format()
        else:
            rows = MtTable.objects_for_mt.using('waf').filter(id=row_id).format()

    if rows.count() == 0:
        return JsonResponse({"row_count": 0}, safe=False)
    row_id = rows.first().id
    qs_json = serializers.serialize('json', rows)
    row_data = json.loads(qs_json)
    row_data_fields = dict()
    for i in row_data[0]['fields']:
        row_data_fields[dict_replace.get(i, i)] = row_data[0]['fields'].get(i)
    row_data[0]['fields'] = row_data_fields
    row_data[0]['fields']['cls_log'] = cls_log
    row_data[0]['fields']['row_id'] = row_id
    for key, value in row_data_fields.items():
        if key != 'mz_origin':
            try:
                row_data_fields[key] = unquote(unquote(row_data_fields[key]))
            except Exception:
                row_data_fields[key] = row_data_fields[key]

    if table != 'p':
        res = row_data[0]['fields']
        mz_origin = rows[0].mz_origin
        if mz_origin is not None and edit == 'false':
            mz_origin = mz_origin.replace("\'", "\"")
            try:
                json_mz_origin = json.loads(mz_origin)
            except json.JSONDecodeError:
                logger.error("An error occurred while parse MZ in details")
                json_mz_origin = dict()
            for key, val in json_mz_origin.items():
                json_mz_origin[key] = base64.b64decode(val).decode('utf-8', 'ignore')
                try:
                    json_mz_origin[key] = unquote_plus(json_mz_origin[key])
                except Exception:
                    json_mz_origin[key] = json_mz_origin[key]
                if key.lower() == 'other_headers':
                    json_mz_origin[key] = os.linesep.join([i.strip() for i in json_mz_origin[key].split('; ')])
                    json_mz_origin[key] = json_mz_origin[key].strip('; ')
            for key, val in json_mz_origin.items():
                key = str(key).lower().replace('url', 'uri')
                res[key] = val
    return JsonResponse(row_data[0]['fields'], safe=False)


@login_required
@require_http_methods(["POST"])
@time_of_function
def delete_rows(request):
    table = request.POST.get('table')
    rows = request.POST.get('data')
    if rows == 'all':
        if table == 'p':
            MtExport.objects_for_mt.using('waf').all().delete()
            mclient.delete('data_md5_export_fp')
        else:
            MtTable.objects.using('waf').all().delete()
    else:
        del_rows = json.loads(rows)
        if table == 'p':
            MtExport.objects_for_mt.using('waf').filter(id__in=del_rows).delete()
            mclient.delete('data_md5_export_fp')
        else:
            MtTable.objects_for_mt.using('waf').filter(id__in=del_rows).delete()
    return JsonResponse({'status': 'ok'})


@login_required
@require_http_methods(["POST"])
@time_of_function
def export_rows(request):
    data = request.POST.get('data')
    checked_fields = json.loads(data)
    row_id = checked_fields['row_id']
    cls_log = checked_fields.get('cls_log', 'MtTable')
    if cls_log == 'Logs':
        if Logs.objects.using('waf').filter(id=row_id).count() == 0:
            return HttpResponse(status=200)
        row_origin = Logs.objects.using('waf').filter(id=row_id)[0]
        export_row = MtTable.objects_for_mt.using('waf').filter(request_id=row_origin.request_id,
                                                                vhost=row_origin.vhost,
                                                                mz__icontains=row_origin.mz) \
            .logs_format(export=True).split_zone()
        if export_row.count() == 0:
            return HttpResponse(status=200)
        else:
            export_row = export_row[0]

    else:
        row_origin = None
        if MtTable.objects_for_mt.using('waf').filter(id=row_id).logs_format().split_zone().count() == 0:
            return HttpResponse(status=200)
        export_row = MtTable.objects_for_mt.using('waf').filter(id=row_id).logs_format(export=True).split_zone()[0]

    export_dict = {}
    # requirement request type
    if checked_fields['bt'] in ('3', '8', '12', '13'):

        fp_fields = [f.name for f in MtTable._meta.get_fields()]
        if 'scanrequest' in fp_fields:
            fp_fields.remove('scanrequest')

        if 'id' in fp_fields:
            fp_fields.remove('id')
        if 'ip' in fp_fields:
            fp_fields.remove('ip')
        if 'req_p' in fp_fields:
            fp_fields.remove('req_p')
        if 'schema' in fp_fields:
            fp_fields.remove('schema')
        checked_fields['fields'].extend(fp_fields)
    if 'agent' in checked_fields['fields']:
        checked_fields['fields'].remove('agent')

    # remove fields not in zone for bt12
    if checked_fields['bt'] in ('12', '13'):
        export_block = []
        for i in checked_fields['fields']:
            if i in ExportFields.fields:
                if i not in export_row.mz:
                    export_block.append(i)
        for e in export_block:
            checked_fields['fields'].remove(e)
    if cls_log == 'Logs':
        row_origin = Logs.objects.using('waf').filter(id=row_id)[0]
        export_row = MtTable.objects_for_mt.using('waf').filter(request_id=row_origin.request_id,
                                                                vhost=row_origin.vhost,
                                                                mz__icontains=row_origin.mz).logs_format(
            export=True)[0]
    else:
        export_row = MtTable.objects_for_mt.using('waf').filter(id=row_id).logs_format(export=True)[0]

    fp_export_fields = [f.name for f in MtExport._meta.get_fields()]

    for i in checked_fields['fields']:
        if dict_replace.get(i, i) in fp_export_fields:
            export_dict[dict_replace.get(i, i)] = getattr(export_row, i)

    if checked_fields['bt'] in ('12', '13'):
        split_waf_id = export_dict['waf_id'].split(' ')
        if len(split_waf_id) > 1:
            export_dict['waf_id'] = split_waf_id[0]
        for i in export_row.mz.split('|'):
            if cls_log == 'Logs' and row_origin is not None:
                if i.lower() not in row_origin.mz.lower():
                    continue
            export_dict_zone = copy.deepcopy(export_dict)
            export_dict_zone['mz'] = i
            split_waf_id = export_dict_zone['waf_id'].split(' ')
            if len(split_waf_id) > 1:
                export_dict_zone['waf_id'] = split_waf_id[0]
            for k in export_dict:
                if k in ExportFields.fields and k not in export_dict_zone['mz']:
                    export_dict_zone.pop(k, None)
                split_waf_id = export_dict_zone['waf_id'].split(' ')
                if len(split_waf_id) > 1:
                    export_dict_zone['waf_id'] = split_waf_id[0]

            if 'mz_origin' in export_dict_zone.keys():
                del export_dict_zone['mz_origin']

            MtExport.objects_for_mt.using('waf').create(**export_dict_zone)
            try:
                mclient.delete('data_md5_export_fp')
            except Exception:
                time.sleep(0.1)
                mclient.delete('data_md5_export_fp')
        # remove duplicates
        fields = ['vhost', 'method', 'uri', 'bt']
        fields += ['ua', 'body', 'args', 'cookie', 'referer', 'other_headers', 'waf_id']
        remove_duplicated_records(MtExport, fields)
    if cls_log == 'Logs':
        if row_origin.mz.lower() == export_row.mz.lower():
            MtTable.objects.using('waf').filter(id=export_row.id).delete()
        else:
            mz_list = MtExport.objects.using('waf').filter(vhost=row_origin.vhost,
                                                           request_id=row_origin.request_id).values_list('mz',
                                                                                                         flat=True)
            mz_list = [i.lower() for i in mz_list]
            if len(mz_list) == export_row.mz.lower().split('|'):
                is_full_export = False
                for i in list(mz_list):
                    if i in export_row.mz.lower().split('|'):
                        is_full_export = True
                if is_full_export:
                    MtTable.objects.using('waf').filter(id=export_row.id).delete()
    else:
        MtTable.objects.using('waf').filter(id=export_row.id).delete()

    return JsonResponse({'status': 'ok'})


@login_required
@require_http_methods(["POST"])
@time_of_function
def check_rows(request):
    field_id = request.POST.get('data')
    cls_logs = request.POST.get('cls_log', 'MtTable')
    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    # creates an object with a list of rules from the rules.bin file
    rules_file = Path(path)
    if rules_file.is_file():
        emu = NWAF_emu(path)
        if cls_logs == 'Logs':
            row = Logs.objects_for_mt.using('waf').filter(id=field_id).annotate(uri=F('url'), args=F('param')).values(
                'uri', 'cookie', 'ua', 'referer', 'body', 'args', 'other_headers')
        else:
            row = MtTable.objects_for_mt.using('waf').filter(id=field_id).values(
                'uri', 'cookie', 'ua', 'referer', 'body', 'args', 'other_headers')
        if row.count() == 0:
            return JsonResponse({"row_count": 0}, safe=False)
        row = row[0]
        res = emu.custom_request(row)
        return JsonResponse(res)
    else:
        return JsonResponse({})


def split_headers(headers):
    headers = headers.splitlines()
    fields = ['cookie', 'content-type', 'referer', 'user-agent']
    try:
        headers_dict = {i.split(":")[0].lower(): i.split(":")[1] for i in headers if i.split(":")[0].lower() in fields}
        other_headers = [i for i in headers if i.split(":")[0].lower() not in fields]
        other_headers = ' '.join(other_headers)
    except IndexError:
        headers_dict = {}
        other_headers = ''
    return headers_dict, other_headers


@login_required
@require_http_methods(["POST"])
@time_of_function
def custom_check(request):
    checked_field = request.POST.get('data')
    fields = json.loads(checked_field)
    headers, other_headers = split_headers(fields["other_headers"])
    fields["cookie"] = headers.get("cookie", '')
    fields["ua"] = headers.get("user-agent", '')
    fields["referer"] = headers.get("referer", '')
    fields["other_headers"] = other_headers
    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    # creates an object with a list of rules from the rules.bin file
    rules_file = Path(path)
    if rules_file.is_file():
        emu = NWAF_emu(path)
        res = emu.custom_request(fields)
        return JsonResponse(res)
    else:
        return JsonResponse({})


@login_required
@require_http_methods(["GET"])
@time_of_function
def check_rules():
    path = str(settings.BASE_DIR) + '/rules/rules.bin'
    rules_file = Path(path)
    if rules_file.is_file():
        if os.stat(path).st_size == 0:
            res = {'status': 'warning'}
        else:
            stat = os.stat(path)
            update_time = get_update_time()
            if update_time is None:
                res = {'status': 'error'}
            elif update_time - stat.st_mtime > 21600:
                res = {'status': 'old'}
            else:
                res = {'status': 'ok'}
    else:
        res = {'status': 'danger'}
    return JsonResponse(res)


@login_required
@require_http_methods(["POST"])
@time_of_function
def edit_fp(request):
    row = {}
    row_id = request.POST.get('row_id', '')

    row['uri'] = request.POST.get('uri', '')
    row['args'] = request.POST.get('args', '')
    row['body'] = request.POST.get('body', '')
    row['ua'] = request.POST.get('ua', '')
    row['cookie'] = request.POST.get('cookie', '')
    row['referer'] = request.POST.get('referer', '')
    row['other_headers'] = request.POST.get('other_headers', '')
    MtExport.objects_for_mt.using('waf').filter(id=row_id).update(**row)
    mclient.delete('data_md5_export_fp')
    return HttpResponseRedirect(settings.HOME_PAGE + 'model_tuning/?t=p')


@login_required
@require_http_methods(["POST"])
@time_of_function
def edit_export(request):
    row = {}
    row_id = request.POST.get('row_id', '')

    row['uri'] = request.POST.get('uri', '')
    row['args'] = request.POST.get('args', '')
    row['body'] = request.POST.get('body', '')
    row['ua'] = request.POST.get('ua', '')
    row['cookie'] = request.POST.get('cookie', '')
    row['referer'] = request.POST.get('referer', '')
    row['other_headers'] = request.POST.get('other_headers', '')

    MtTable.objects_for_mt.using('waf').filter(id=row_id).update(**row)
    if MtTable.objects_for_mt.using('waf').filter(id=row_id).split_zone().count() == 0:
        return redirect('index')
    export_row = MtTable.objects_for_mt.using('waf').filter(id=row_id).logs_format(export=True).split_zone()[0]
    fp_fields = [f.name for f in MtTable._meta.get_fields()]
    fp_export_fields = [f.name for f in MtExport._meta.get_fields()]
    fp_fields.remove('id')
    fp_fields.remove('ip')
    if 'mz_origin' in fp_fields:
        fp_fields.remove('mz_origin')

    params = ['uri', 'args', 'body', 'ua', 'cookie', 'referer', 'other_headers']
    export_dict = {}
    status_fields = [x.lower() for x in export_row.mz]
    export_row = MtTable.objects_for_mt.using('waf').filter(id=row_id).logs_format(export=True)[0]
    for i in fp_fields:
        if dict_replace.get(i, i) in fp_export_fields:
            if dict_replace.get(i, i) in params:
                if dict_replace.get(i, i) in status_fields:
                    export_dict[dict_replace.get(i, i)] = getattr(export_row, i)
            else:
                export_dict[dict_replace.get(i, i)] = getattr(export_row, i)
    split_waf_id = export_dict['waf_id'].split(' ')
    if len(split_waf_id) > 1:
        export_dict['waf_id'] = split_waf_id[0]
    for i in export_row.mz.split('|'):
        if export_dict[i.lower()]:
            export_dict_zone = copy.deepcopy(export_dict)
            export_dict_zone['mz'] = i
            split_waf_id = export_dict_zone['waf_id'].split(' ')
            if len(split_waf_id) > 1:
                export_dict_zone['waf_id'] = split_waf_id[0]
            for k in export_dict:
                if k in ExportFields.fields and k not in export_dict_zone['mz']:
                    export_dict_zone.pop(k, None)
            MtExport.objects_for_mt.using('waf').create(**export_dict_zone)
            mclient.delete('data_md5_export_fp')
    fields = ['vhost', 'method', 'uri', 'bt']
    fields += ['ua', 'body', 'args', 'cookie', 'referer', 'other_headers', 'waf_id']
    remove_duplicated_records(MtExport, fields)
    MtTable.objects.using('waf').filter(id=row_id).delete()
    if request.POST.get('is_logs'):
        return HttpResponse(status=201)
    else:
        return redirect(request.META.get('HTTP_REFERER', '/'))


@login_required
@time_of_function
def ip_info(request):
    # memcached
    mcache = Client(cache_address)

    get_ip = request.GET.get('ip')
    waf_id = request.GET.get('geo_token_id')
    token = None

    for token_obj in Geo_Tokens.objects.using('waf').all():
        if token_obj.waf_id == waf_id:
            token = token_obj.geo_token

    ips_list = list()
    ips_list.append(remove_zero_from_ip(get_ip))

    for ip in list(chunks(list(set(ips_list)), 1))[:len(ips_list)]:
        get_location_from_remote(token, ip, mcache=mcache)
    ips = get_location_by_ip(mcache, get_ip, check_token=True, token=token)

    if ips:
        return render(request, 'smain/ip_info.html', {'ip': ips})
    else:
        return JsonResponse({'status': 'no data'})


@login_required
@time_of_function
def remove_all_by_ip(request):
    ip = request.POST.get('ip')
    try:
        MtTable.objects.using('waf').filter(ip=ip).delete()
    except Exception as e:
        logger.error('An error occurred when deleting records from IP {ip}: {e}'.format(ip=ip, e=e))
        messages.error(request, pgettext('messages', 'An error occurred while deleting.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    return JsonResponse({'status': 'success'})
