import re
import ast
import json
import base64
import logging
import requests
import multiprocessing

from url_normalize import url_normalize
from functools import wraps
from netaddr import IPNetwork, IPRange, IPAddress, AddrFormatError, AddrConversionError
from url_normalize import url_normalize

from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponse

from main.models import LicKeys
from my_waf.default_values import brute, ddos, nwaf_limit
from main.models.wafuser import WAFUSER
from main.utils import convert_to_punycode, convert_from_punycode, get_proxy_api_dict, get_proxy_conf_dict, \
    recursion_convert_to_punycode_obj, recursion_convert_from_punycode_obj

logger = logging.getLogger(__name__)

try:
    api_url = settings.API_URI
except AttributeError:
    api_url = 'http://localhost:8080/nw-api/'

USER_AGENT = settings.USER_AGENT
logger = logging.getLogger(__name__)
auth_url = settings.URI_AUTH


def del_ip_bs(request, ip, key) -> object:
    try:
        data = {'key': key, 'ip-address': ip}
        response = requests.post(
            url_normalize(api_url + '/del_dyn_sync_ban'),
            headers={'Content-type': 'application/json', 'User-agent': USER_AGENT},
            proxies=get_proxy_api_dict(),
            timeout=30,
            json=data,
        )
        if response.ok:
            response = json.loads(response.text.replace('\'', '\"'))
            return response

    except Exception as e:
        logger.error(f'An error occurred while confirming '
                     f'information about the key: {key}. Error information: {e}')


def response_for_bs_management(key: str, ) -> dict:
    """Response for bs_management"""
    try:
        response = requests.post(
            url_normalize(api_url + '/get_dyn_sync_ban'),
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_api_dict(),
            timeout=30,
            data={'key': key},
        )
        if response.ok:
            response = json.loads(response.text.replace('\'', '\"'))
            return response

    except Exception as e:
        logger.error(f'An error occurred while confirming '
                     f'information about the key: {key}. Error information: {e}')


def my_waf_access(function):
    @wraps(function)
    def wrap(request, *args, **kwargs):
        if request.user.is_anonymous:
            return HttpResponseRedirect('/')

        wafuser = WAFUSER.objects.get(id=request.user.id)
        if wafuser.ws_toggle and settings.API_URI:
            return function(request, *args, **kwargs)
        else:
            return HttpResponseRedirect('/')

    return wrap


def check_plans_for_tab_mlc(keys):
    for key in keys:
        if key.tarif != 'Light':
            return True
    return False


def get_waf_id(key):
    if isinstance(key, str):
        if key and len(key) > 6:
            format_key = 'xxx' + key[-6:]
        else:
            format_key = key
    else:
        format_key = key.get_formated_key()
        key = key.key
    try:
        response_check_key = requests.post(
            auth_url,
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': key}
        )
        if response_check_key.ok:
            check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success':
                return check_key_result['waf_id']
    except Exception as e:
        logger.error(
            'An error occurred while getting waf_id by key: {key}. Error information: {err}'
            .format(key=format_key, err=e)
        )
        return None


def get_settings(module, key):
    """
    Function to get settings
    :param module: Module name
    :param key: Sha256 license key
    :return: Received settings
    """
    url = ''
    if key and len(key) > 6:
        format_key = 'xxx' + key[-6:]
    else:
        format_key = key
    if module == 'MLC':
        url = '/v2/get_mlc_settings'
    if module == 'DYN':
        url = '/get_dyn_settings'
    if module == 'ERL':
        url = '/v2/get_dyn_erl'
    if module == 'WL':
        url = '/get_dyn_wl?extended=yes'
    if module == 'BL':
        url = '/get_dyn_bl?extended=yes'

    response = requests.post(
        url_normalize(api_url + url),
        headers={'User-agent': USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': key},
    )
    if response.ok:
        resp_text = response.text
        resp_text = convert_from_punycode(resp_text)

        if module == 'WL':
            response_list = resp_text.split('\n')
            get_rules = ast.literal_eval(response_list[0])
            if get_rules['status'] == 'success':
                return response_list[1:]
        elif module == 'ERL':
            get_rules = recursion_convert_from_punycode_obj(response.json())
        elif module == 'BL':
            response_list = resp_text.split('\n')
            get_rules = ast.literal_eval(response_list[0])
            if len(response_list) > 1:
                get_rules['bl_list'] = response_list[1:]
        elif module == 'MLC':
            get_rules = recursion_convert_from_punycode_obj(response.json())
            if get_rules.get('ddos__wl_ip', False):
                get_rules['ddos__wl_ip'] = [item for item in get_rules.get('ddos__wl_ip')]
            if get_rules.get('brute__bf_detect', False):
                get_rules['brute__bf_detect'] = [item for item in get_rules.get('brute__bf_detect')]
        else:
            get_rules = ast.literal_eval(resp_text)
        if get_rules['status'] == 'fail':
            error_description = get_rules['description']
            logging.error(
                'Get status fail when get {setting} settings for key: {key} with description: {description}'
                .format(setting=module, key=format_key, description=str(error_description))
            )
        if get_rules['status'] == 'none':
            logging.info(
                'Status "none" has been received while processing {setting} settings for key: {key}'
                .format(setting=module, key=format_key)
            )
            return None
        if get_rules['status'] == 'success':
            return get_rules
    else:
        logging.info(
            'Response code {code} has been received while processing {setting} settings for key: {key}'
            .format(code=str(response.status_code), setting=module, key=format_key)
        )
        return None


def send_set_settings(module, data) -> tuple:
    """
    Function to set settings
    :param module: Module name
    :param data: Sha256 license key
    :return: Response
    """
    url = ''
    if module == 'MLC':
        url = '/v2/set_mlc_settings'
        if 'ddos__wl_ip' in data.get('set', False):
            if data['set']['ddos__wl_ip'] == '':
                data['set']['ddos__wl_ip'] = []
            else:
                data['set']['ddos__wl_ip'] = data.get('set').get('ddos__wl_ip').split(' ')

        if 'brute__wl_ip' in data.get('set', False):
            if data['set']['brute__wl_ip'] == '':
                data['set']['brute__wl_ip'] = []
            else:
                data['set']['brute__wl_ip'] = data.get('set').get('brute__wl_ip').split(' ')
        if 'brute__bf_detect' in data.get('set', False):
            data['set']['brute__bf_detect'] = json.loads(data['set']['brute__bf_detect'].replace('`', '\''))
            
    if module == 'DYN':
        url = '/set_dyn_settings'
    if module == 'ERL':
        url = '/v2/set_dyn_erl'
    if module == 'WL':
        url = '/set_dyn_wl'

    if 'upd' in data:
        if 'extension' not in data['upd']:
            data['upd']['extension'] = ''

    data = recursion_convert_to_punycode_obj(data)

    set_response = requests.post(
        url_normalize(api_url + url),
        headers={'User-agent': USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        json=data,
    )
    if set_response.ok:
        api_response = ast.literal_eval(set_response.text)
        if api_response['status'] == 'fail':
            error_description = api_response['description']
            body = json.loads(set_response.request.body.decode('utf-8'))
            body['key'] = '{}{}'.format('X' * 26, body.get('key')[-6:])
            logging.error(
                'Get status fail when set {setting} settings '
                'with description: {description} and data: {data}'
                .format(
                    setting=module, description=str(error_description),
                    data=json.dumps(body)
                )
            )
            if not error_description:
                error_description = 'Get status fail when set {setting} settings'
            return 'fail', error_description
        else:
            if module == 'WL':
                if 'rid' in api_response:
                    return 'success', api_response['rid']
                else:
                    return 'success', None
            elif module == 'ERL':
                if 'id' in api_response:
                    return 'success', api_response['id']
                else:
                    return 'success', None
            else:
                return 'success', None
    else:
        body = json.loads(set_response.request.body.decode('utf-8'))
        body['key'] = '{}{}'.format('X' * 26, body.get('key')[-6:])

        logging.info(
            'Response code {code} has been received while processing {setting} settings with data: {data}'
            .format(code=str(set_response.status_code), setting=module, data=json.dumps(body))
        )
        return 'fail', "Something's went wrong"


def check_keys(keys, check_waf_id):
    """
    Checks keys and their waf id
    :param keys: List keys
    :param check_waf_id: Condition for checking waf id
    :return: List of checked keys and waf ids
    """
    checked_key = []
    checked_key_ids = []
    waf_ids = []

    processes = []
    checked_key_queue = multiprocessing.Queue()
    checked_key_ids_queue = multiprocessing.Queue()

    for key in keys:
        proc = multiprocessing.Process(target=send_request_check_key,
                                       args=(key, check_waf_id, waf_ids, checked_key_queue, checked_key_ids_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    while not checked_key_queue.empty():
        checked_key.append(checked_key_queue.get())

    while not checked_key_ids_queue.empty():
        checked_key_ids.append(checked_key_ids_queue.get())

    return checked_key, checked_key_ids


def send_request_check_key(key, check_waf_id, waf_ids, checked_key_queue, checked_key_ids_queue):
    """
    Sends a request to check the key
    :param key: License key
    :param check_waf_id:
    :param waf_ids: List waf id
    :param checked_key_queue: Queue with keys
    :param checked_key_ids_queue: Queue with waf ids
    """
    try:
        response_check_key = requests.post(
            auth_url,
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': key.key}
        )

        if response_check_key.ok:
            check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success' and str(check_key_result['r_asa_disable']) == 'False':
                tarif_id = int(check_key_result['tarif_id'])
                if tarif_id == 0:
                    key.tarif = 'Light'
                elif tarif_id == 1:
                    key.tarif = 'Business'
                elif tarif_id == 2:
                    key.tarif = 'Enterprise'
                if check_waf_id:
                    if check_key_result['waf_id'] not in waf_ids:
                        waf_ids.append(check_key_result['waf_id'])
                        checked_key_queue.put(key)
                        checked_key_ids_queue.put(key.id)
                else:
                    checked_key_queue.put(key)
                    checked_key_ids_queue.put(key.id)
        else:
            logger.info(
                'Response code {code} as been received while validating the key: {key}'
                .format(code=str(response_check_key.status_code), key=key.get_formated_key())
            )
    except Exception as e:
        logger.error('An error occurred while confirming the key: {key}. Error information: {err}'
                     .format(key=key.get_formated_key(), err=e))


def get_dyn_settings(key, queue):
    """
    Processes received DYN settings
    :param key: License key
    :param queue: Multiprocessor queue
    """
    getting_dyn_settings = get_settings('DYN', key.key)
    if getting_dyn_settings is not None:
        for dyn_key, value in getting_dyn_settings.items():

            if dyn_key == 'nwaf_limit':
                nwaf_limits = str(value).split(', ')
                nwaf_limits_list = []
                for limit in nwaf_limits:
                    limit_dict = {}
                    for el in limit.split(' '):
                        if 'domain' in el:
                            limit_dict['domain'] = el.replace('domain=', '')
                        if 'rate' in el:
                            limit_dict['rate'] = el.replace('rate=', '')
                        if 'block_time' in el:
                            limit_dict['block_time'] = el.replace('block_time=', '')
                    nwaf_limits_list.append(limit_dict)
                key.nwaf_limit = nwaf_limits_list
            else:
                setattr(key, dyn_key, value)

    queue.put(key)


def get_mlc_settings(key, queue):
    """
    Processes received MLC settings
    :param key: License key
    :param queue: Multiprocessor queue
    """
    response_list_vhost = requests.post(
        url_normalize(api_url + '/get_vhosts_list'),
        headers={'User-agent': USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': key.key},
    )
    if response_list_vhost.ok:
        response_list_vhost_dict = ast.literal_eval(response_list_vhost.text)
        try:
            res_list_vhost = str(response_list_vhost_dict['vhosts_list']).strip('][').replace('\"', '')
            tmp_list_vhost = ''
            for num, el in enumerate(res_list_vhost.split(',')):
                tmp_list_vhost += convert_from_punycode(el)
                if num != len(res_list_vhost.split(',')) - 1:
                    tmp_list_vhost += ','
            key.vhost_list = tmp_list_vhost
        except KeyError:
            key.vhost_list = ''
            logger.error('Error with get vhost lists: ' + key.get_formated_key())
    else:
        key.vhost_list = ''
        logger.info(
            'Response code {code} has been received from /get_vhosts_list for key: {key}'
            .format(code=str(response_list_vhost.status_code), key=key.get_formated_key())
        )

    getting_mlc_settings = get_settings('MLC', key.key)
    
    if getting_mlc_settings is not None:
        for setting_key, value in getting_mlc_settings.items():
            if setting_key == 'ddos__wl_ip':
                key.ddos__wl_ip = str(value).strip('][').replace('\'', '')
            elif setting_key == 'brute__wl_ip':
                key.brute__wl_ip = str(value).strip('][').replace('\'', '')
            elif setting_key == 'ddos__wl_url':
                key.ddos__wl_url = str(value).strip('][').replace('\'', '')
            elif setting_key == 'brute__brute_detect':
                key.brute__brute_detect = str(value).strip('][').replace('\'', '')
            elif setting_key == 'brute__flood_detect':
                key.brute__flood_detect = str(value).strip('][').replace('\'', '')
            elif setting_key == 'brute__wl_host':
                key.brute__wl_host = str(value).strip('][').replace('\'', '')
            elif setting_key == 'openapi__vhosts_list':
                key.openapi__vhosts_list = str(value).strip('][').replace('\'', '')
            elif setting_key == 'brute__bf_detect':
                if not value:
                    continue

                key.brute__bf_detect = json.dumps(value)
                value_list = []
                
                for item in value:
                    url = item.get('url', False)
                    if not url:
                        break
                    if item.get('brute', False):
                        bf_type = 'Brute'
                    else:
                        bf_type = 'Flood'
                    if (item.get('zone', False)):
                        zone = ' | '.join(item.get('zone')).upper()
                    else:
                        zone = ''
                    condition = []
                    if not item.get('condition', False):
                        break
                    if (item.get('condition').get('args', False)):
                        if len(item.get("condition").get("args")) > 0:
                            condition.append(f'ARGS<sup>{len(item.get("condition").get("args"))}</sup>')
                    if (item.get('condition').get('body', False)):
                        if len(item.get("condition").get("body")) > 0:
                            condition.append(f'BODY<sup>{len(item.get("condition").get("body"))}</sup>')

                    zone_str = ', <b>zone:</b> BODY | ARGS'
                    if zone:
                        zone_str = f', <b>zone:</b> {zone}'

                    condition_str = ''
                    if len(condition) > 0:
                        condition_str = f', <b>condition:</b> {" | ".join(condition)}'
                    
                    value_list.append(f'{url} (<b>type:</b> {bf_type}{zone_str}{condition_str})')
                key.brute__bf_detect_text = '<br />'.join(value_list)
            else:
                setattr(key, setting_key, value)

    response_list_models = requests.post(
        url_normalize(api_url + '/get_models_list_uri'),
        headers={'User-agent': USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': key.key},
    )
    if response_list_models.ok:
        tmp_list_models = response_list_models.text[1:-1].replace('\'', '').replace('sub_', '*.').split(', ')

        key.models = []

        for i in tmp_list_models:
            key.models.append(convert_from_punycode(i))

    else:
        key.models = []
        logger.info(
            'Response code {code} has been received from /get_models_list_uri for key: {key}'
            .format(code=str(response_list_models.status_code), key=key.get_formated_key())
        )

    model_list = []
    try:
        vhost_list = [i.strip() for i in key.vhost_list.split(', ')]
        for md in key.models:
            vhost = None

            if md in vhost_list:
                vhost = md
            elif md.lower() == 'all' and '*' in vhost_list:
                vhost = '*'
            else:
                for i in vhost_list:
                    if i.startswith('.') and (i in md or i[1:] == md):
                        vhost = i
                        break

            if vhost is None:
                continue

            vhost = convert_to_punycode(vhost)
            response_training_status = requests.post(
                url_normalize(api_url + '/get_training_uri'),
                headers={'User-agent': USER_AGENT},
                proxies=get_proxy_api_dict(),
                timeout=30,
                data={'key': key.key,
                      'vhost': vhost},
            )

            if response_training_status.ok:
                response_training_status_dict = ast.literal_eval(response_training_status.text)
                if response_training_status_dict['status'] == 'success':
                    model = {
                        'name': md,
                        'duration': response_training_status_dict['duration'],
                        'complete': str(response_training_status_dict['complete']),
                    }
                    if str(response_training_status_dict['progress']).isdigit():
                        model['progress'] = str(response_training_status_dict['progress'])
                    else:
                        model['progress'] = 'none'
                else:
                    logger.info(
                        'Status "{status}" has been received while receiving model {model} with key: {key}'
                        .format(status=str(response_training_status_dict['status']), model=md, key=key.get_formated_key())
                    )
                    model = {
                        'name': md,
                        'duration': 'none',
                        'complete': 'none',
                        'progress': 'none'
                    }
                model_list.append(model)
            else:
                logger.info(
                    'Response code {code} has been received while receiving model {model} with key: {key}'
                    .format(code=str(response_training_status.status_code), model=md, key=key.get_formated_key())
                )
    except Exception as e:
        logger.error('Error with get models with: {exception}'.format(exception=str(e)))
    key.models = model_list

    queue.put(key)


def get_lic_keys(key, queue):
    """
    Processes received license key
    :param key: License key
    :param queue: Multiprocessor queue
    """
    try:
        response_check_key = requests.post(
            auth_url,
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': key.key}
        )
        if response_check_key.ok:
            check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success':
                key.remote_asa_disable = str(check_key_result['r_asa_disable'])
                key.models_num = int(check_key_result['models_num'])
                key.days_left = str(check_key_result['days_left']).replace('days', '')
                key.waf_id = str(check_key_result['waf_id'])
                tarif_id = int(check_key_result['tarif_id'])
                if tarif_id == 0:
                    key.tarif = 'Light'
                elif tarif_id == 1:
                    key.tarif = 'Business'
                elif tarif_id == 2:
                    key.tarif = 'Enterprise'
                queue.put(key)
        else:
            logger.info(
                'Response code {code} has been received while validating the key: {key}'
                .format(code=str(response_check_key.status_code), key=key.get_formated_key())
            )
    except Exception as e:
        logger.error('An error occurred while processing the key: {key}. Error information: {err}'
                     .format(key=key.get_formated_key(), err=e))


def get_rules_management(key, queue):
    """
    Processes received ERL and WL rules and BL
    :param key: License key
    :param queue: Multiprocessor queue
    """
    getting_erl_settings: dict = get_settings('ERL', key.key)
    if getting_erl_settings is not None:
        get_erl = getting_erl_settings.get('erl', [])
        for erl in get_erl:
            for erl_key, erl_value in erl.items():
                try:
                    if erl_key == 'other_headers' or erl_key == 'other_headers_x':
                        other_headers = []
                        for el in erl_value:
                            name = el[:el.find(':')]
                            values = el[el.find(':') + 1:]
                            if '!!' in name:
                                name = name.replace('!!', '')
                                name = base64.b64decode(str(name).encode('utf-8')).decode('utf-8')
                                name = '!!' + name
                            else:
                                name = base64.b64decode(str(name).encode('utf-8')).decode('utf-8')
                            result = []
                            last_end_op = 0
                            last_op = " "
                            for start_idx, end_idx in parse_values_erl_headers(values):
                                if start_idx != 0:
                                    text = values[last_end_op: start_idx]
                                    r_val = base64.b64decode(
                                        str(text).encode('utf-8')
                                    ).decode('utf-8')

                                    result.append([last_op, r_val])

                                last_op = values[start_idx: end_idx]
                                last_end_op = end_idx
                            else:
                                text = values[last_end_op:]
                                r_val = base64.b64decode(
                                    str(text).encode('utf-8')
                                ).decode('utf-8')
                                result.append([last_op, r_val])
                            header = {'key': name, 'value': result}
                            other_headers.append(header)
                        erl[erl_key] = other_headers
                    else:
                        erl_value = str(erl_value)
                        erl[erl_key] = list(
                            filter(None, erl_value.replace('|!', ',|!').replace('|&', ',|&').replace('|', ',|')
                                   .replace('&', ',&').replace('!', '!').split(',')))
                        result = []
                        for el in erl[erl_key]:
                            if el[0] == '|' and el[1] == '!':
                                result.append(['|!', base64.b64decode(str(el[2:]).encode('utf-8')).decode('utf-8')])
                            elif el[0] == '&' and el[1] == '!':
                                result.append(['&!', base64.b64decode(str(el[2:]).encode('utf-8')).decode('utf-8')])
                            elif el[0] == '|' and el[1] != '!':
                                result.append(['|', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                            elif el[0] == '&' and el[1] != '!':
                                result.append(['&', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                            elif el[0] == '!':
                                result.append(['!', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                            else:
                                result.append([' ', base64.b64decode(str(el).encode('utf-8')).decode('utf-8')])
                        if len(result) > 0:
                            erl[erl_key] = result
                        else:
                            erl[erl_key] = ''
                except Exception:
                    erl[erl_key] = erl_value

                erl[erl_key] = recursion_convert_from_punycode_obj(erl[erl_key])

        get_erl = sorted(get_erl, key=lambda d: d['id'], reverse=True)
        key.get_erl = get_erl

    getting_wl_rules = get_settings('WL', key.key)

    get_rules = []
    if getting_wl_rules is not None:
        for rule in getting_wl_rules:
            get_rules.append(rule)

    getting_bl_rules = get_settings('BL', key.key)
    if getting_bl_rules is not None:
        bl_list = getting_bl_rules.get('bl_list', list())
        bl_dict = dict()
        for i in bl_list:
            if 'domain=' in i:
                domain = i.split(' ')[0].replace('domain=', '')
                domain = convert_from_punycode(domain)
                ip = i.split(' ')[1]
            else:
                domain = '__ALL__'
                ip = i
            ip = ip.strip()
            domain = domain.strip()
            if bl_dict.get(domain) is None:
                bl_dict[domain] = list()
            bl_dict[domain].append(ip)

        key.bl_all = bl_dict.get('__ALL__')
        key.bl_all_count = len(bl_dict.get('__ALL__', []))
        if '__ALL__' in bl_dict:
            del bl_dict['__ALL__']
        key.bl_dicts = bl_dict
        key.bl_dicts_count = dict()
        for k, v in bl_dict.items():
            key.bl_dicts_count[k] = len(v)

    list_wl_dicts = []
    if get_rules is not None:
        for el in get_rules:
            rule_dict = {}
            try:
                rule_dict['rid'] = str(re.findall(r'RID: \d+', el)[0]).replace('RID: ', '')
            except Exception:
                rule_dict['rid'] = ''
                logger.error('Error with parsing RID in rule:\n' + el)

            try:
                rule_dict['wl_id'] = str(re.findall(r'WL ID:\d+|\*', el)[0]).replace('WL ID:', '')
            except Exception:
                rule_dict['wl_id'] = ''
                logger.error('Error with parsing WL ID in rule:\n' + el)

            try:
                rule_dict['active'] = str(re.findall(r'Active: True|False', el)[0]).replace('Active: ', '')
            except Exception:
                rule_dict['active'] = ''
                logger.error('Error with parsing Active in rule:\n' + el)

            domain = re.findall(r'(domain=)([-a-zA-Zа-яА-Я\d*@:%_+.~#?&/=]{2,256}\b)', el)

            if len(domain) > 0:
                rule_dict['domain'] = str(re.findall(r'(domain=)([-a-zA-Zа-яА-Я\d*@:%_+.~#?&/=]{2,256}\b)', el)[0][1])
                rule_dict['domain'] = convert_from_punycode(rule_dict['domain'])

            mz_and_extension = re.findall(r'Z:(.+)\"', el)
            mz_and_extension = str(mz_and_extension[0])
            find_mz = mz_and_extension.find('$')
            if find_mz == -1:
                rule_dict['mz'] = str(mz_and_extension).rstrip('|').upper()
                rule_dict['mz'] = rule_dict['mz'].replace('COOKIE', 'Cookie').replace('URL', 'PATH')
                rule_dict['extension'] = ''
            else:
                rule_dict['mz'] = mz_and_extension[:-(len(mz_and_extension) - find_mz + 1)]
                rule_dict['mz'] = rule_dict['mz'].upper()
                rule_dict['mz'] = rule_dict['mz'].replace('COOKIE', 'Cookie').replace('URL', 'PATH')
                extension = mz_and_extension[find_mz:]
                extension_str = re.sub(r'\|\$', '|', extension)
                rule_dict['extension'] = extension_str[1:].replace('URL', 'PATH')
                extension = extension.split('|$')
                extension_dict = {}
                for ext in extension:
                    extension_dict[ext[:ext.rfind(':')].replace('$', '')] = ext[ext.find(':') + 1:]
                rule_dict['extension_dict'] = extension_dict

            list_wl_dicts.append(rule_dict)
            list_wl_dicts = sorted(list_wl_dicts, key=lambda d: d['rid'], reverse=True)
            key.wl_dicts = list_wl_dicts
    else:
        key.wl_dicts = list_wl_dicts

    queue.put(key)


def set_default_values(hashed_key):
    """
    Sets default values for MLC and Dyn when new key is added
    :param hashed_key: Hashed license key
    """
    mlc_settings_dict = {
        'ddos__enable': ddos.enable,
        'ddos__interval': ddos.interval,
        'ddos__latest_only': ddos.latest_only,
        'ddos__send_possible': ddos.send_possible,
        'brute__enable': brute.enable,
        'brute__interval': brute.interval,
        'brute__max_val': brute.max_val,
        'brute__latest_only': brute.latest_only,
        'brute__send_possible': brute.send_possible,
    }
    mlc_full_settings_dict = {"key": hashed_key, "set": mlc_settings_dict}
    send_set_settings('MLC', mlc_full_settings_dict)

    dyn_settings_dict = {
        'nwaf_limit': nwaf_limit.nwaf_limit
    }
    dyn_full_settings_dict = {"key": hashed_key, "set": dyn_settings_dict}
    send_set_settings('DYN', dyn_full_settings_dict)


def get_erl_rule(key, erl_id):
    getting_erl_settings: dict = get_settings('ERL', key.key)
    if getting_erl_settings is not None:
        get_erl = getting_erl_settings.get('erl', [])
        for erl in get_erl:
            if erl['id'] == int(erl_id):
                return erl


def set_dyn_bl(key, list_bl):
    get_bl = get_settings('BL', key.key)
    if get_bl:
        general_list = get_bl.get('bl_list', list())
        list_dupl = set()
        for i in general_list.copy():

            if i in list_bl:
                list_dupl.add(i)
            else:
                api_in_i = 'api=' in i
                domain_in_i = i.split(' ')[0] if 'domain=' in i else None

                if 'domain=' in i:
                    i_tmp = i.split(' ')[1]
                elif 'api=' in i:
                    i_tmp = i.split(' ')[0]
                else:
                    i_tmp = i

                for j in list_bl:
                    api_in_j = 'api=' in j
                    domain_in_j = j.split(' ')[0] if 'domain=' in j else None

                    if 'domain=' in j:
                        j_tmp = j.split(' ')[1]
                    elif 'api=' in j:
                        j_tmp = j.split(' ')[0]
                    else:
                        j_tmp = j

                    if i_tmp == j_tmp:
                        if (not domain_in_j and domain_in_i and api_in_i == api_in_j) \
                                or (domain_in_j == domain_in_i and api_in_i != api_in_j):
                            if i in general_list:
                                general_list.remove(i)
                        elif domain_in_j and domain_in_i and domain_in_j == domain_in_i and api_in_i == api_in_j:
                            list_dupl.add(j)
                        elif domain_in_j and not domain_in_i:
                            list_dupl.add(j)

        list_bl += general_list
        list_dupl = list(list_dupl)
        for i in list_dupl:
            list_bl.remove(i)
    else:
        list_dupl = []

    data = {
        "key": key.key,
        "set": {
            "active": True,
            "bl": list_bl
        }
    }

    response = requests.post(
        url_normalize(api_url + '/set_dyn_bl'),
        headers={'User-agent': USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        json=data,
    )

    if response.ok:
        d = json.loads(response.text.replace('\'', '\"'))
        if d['status'] == 'success':
            return list_bl, list_dupl
        else:
            logger.info('An error occurred while set settings BL with {key} and description error: {desc}'.
                        format(code=str(response.status_code), key=key.get_formated_key(), desc=d['description']))
            return False, d['description']
    else:
        logger.info('Response code {code} has been received from /set_dyn_bl for key: {key}'.
                    format(code=str(response.status_code), key=key.get_formated_key()))
        return False, 'Unknown error'


def get_and_search_bl(key: LicKeys, search: list):
    bl_settings: dict = get_settings('BL', key.key)
    if bl_settings:
        get_bl = bl_settings.get('bl_list', list())
    else:
        get_bl = []

    if search:
        searched_list = list()
        for i in search:
            if not i:
                continue
            for j in get_bl:
                if 'domain=' in j:
                    ip_j = j.split(' ')[1]
                elif 'api=' in j:
                    ip_j = j.split(' ')[0]
                else:
                    ip_j = j

                if '/' in i:
                    ip_net = IPNetwork(i)
                    if '/' in ip_j:
                        ip_net2 = IPNetwork(ip_j)
                        if ip_net2 in ip_net or ip_net in ip_net2:
                            searched_list.append(j)
                    elif '-' in ip_j:
                        ip_range = IPRange(ip_j.split('-')[0], ip_j.split('-')[1]).cidrs()

                        for c in ip_range:
                            if c in ip_net or ip_net in c:
                                searched_list.append(j)
                                break
                    else:
                        ip = IPAddress(ip_j)
                        if ip in ip_net:
                            searched_list.append(j)
                elif '-' in i:
                    try:
                        ip_range = IPRange(i.split('-')[0], i.split('-')[1]).cidrs()
                    except (AddrFormatError, AddrConversionError):
                        if i in ip_j or ip_j in i:
                            searched_list.append(j)
                        continue
                    if '/' in ip_j:
                        ip_net2 = IPNetwork(ip_j)
                        for c in ip_range:
                            if ip_net2 in c or c in ip_net2:
                                searched_list.append(j)
                                break
                    elif '-' in ip_j:
                        ip_range2 = IPRange(ip_j.split('-')[0], ip_j.split('-')[1]).cidrs()

                        for c in ip_range:
                            for m in ip_range2:
                                if c in m or m in c:
                                    searched_list.append(j)
                                    break
                    else:
                        ip = IPAddress(ip_j)
                        for c in ip_range:
                            if ip in c:
                                searched_list.append(j)
                                break
                else:
                    try:
                        ip = IPAddress(i)
                    except (AddrFormatError, AddrConversionError):
                        if i in ip_j or ip_j in i:
                            searched_list.append(j)
                        continue
                    if '/' in ip_j:
                        ip_net = IPNetwork(ip_j)
                        if ip in ip_net:
                            searched_list.append(j)
                    elif '-' in ip_j:
                        ip_range = IPRange(ip_j.split('-')[0], ip_j.split('-')[1]).cidrs()
                        for c in ip_range:
                            if ip in c:
                                searched_list.append(j)
                                break
                    else:
                        ip2 = IPAddress(ip_j)
                        if ip == ip2:
                            searched_list.append(j)

        get_bl = searched_list

    return get_bl


def bl_postprocessing_ips(
    bl_list: list,
    domain: str,
    search_type: int = 0
):
    res_list = list()

    count = 0
    for i in bl_list:
        if search_type == 1 and 'api=yes' not in i:
            continue
        elif search_type == 2 and 'api=yes' in i:
            continue
        if domain == '__ALL__' and 'domain' not in i:
            res_list.append(i)
            count += 1
        elif 'domain=' + domain in i:
            res_list.append(i.replace('domain=' + domain, '').strip())
            count += 1
        elif 'domain' not in i:
            res_list.append(i + '__ALL__')

    return res_list, count


def del_dyn_bl(key, search, mode=0):
    get_bl: dict = get_settings('BL', key.key)
    if get_bl:
        bl_list = get_bl.get('bl_list', list())
        if mode == 0:
            for i in search:
                bl_list.remove(i)
        elif mode == 1:
            if search != "__ALL__":
                rem_list = [i for i in get_bl.get('bl_list', list()) if "domain="+search in i]
            else:
                rem_list = [i for i in get_bl.get('bl_list', list()) if "domain=" not in i]

            for i in rem_list:
                bl_list.remove(i)
        else:
            return False
    else:
        return True

    data = {
        "key": key.key,
        "set": {
            "active": True,
            "bl": bl_list
        }
    }

    response = requests.post(
        url_normalize(api_url + '/set_dyn_bl'),
        headers={'User-agent': USER_AGENT, 'Content-Type': 'application/json'},
        proxies=get_proxy_api_dict(),
        timeout=30,
        json=data,
    )
    if response.ok:
        d = json.loads(response.text.replace('\'', '\"'))
        if d['status'] == 'success':
            return True
        else:
            return False
    else:
        logger.info('Response code {code} has been received from /set_dyn_bl for key: {key}'.
                    format(code=str(response.status_code), key=key.get_formated_key()))
        return False


def sort_bl(bl_list):
    res = list()
    for type_ip in ('-', '/', 'only'):
        tmp = list()
        for i in sorted(bl_list):
            if type_ip == 'only':
                t = [p if len(p) == 3 else '0' * (3 - len(p)) + p for p in i.replace('__ALL__', '').split('.')]
                num = ''.join(t)
                tmp.append({'num': num,
                            'ip': i})
            elif type_ip in i:
                tmp.append(i)
        for i in tmp:
            if type_ip == 'only':
                bl_list.remove(i['ip'])
            else:
                bl_list.remove(i)
        if type_ip == 'only':
            for ip in sorted(tmp, key=lambda x: x['num']):
                res.append(ip['ip'])
        else:
            res += tmp

    return res


def parse_values_erl_headers(values: str):
    i = 0
    while i < len(values):
        c = values[i]
        two_op = False
        if c in ("!", "|", "&"):
            curr_op = True
        else:
            curr_op = False

        if curr_op and i + 1 < len(values):
            next_c = values[i + 1]
            if next_c in ("!", "|", "&"):
                two_op = True

        if not curr_op:
            i += 1
        elif two_op:
            yield i, i + 2
            i += 2
        else:
            yield i, i + 1
            i += 1


def b64_encode(val: str) -> str:
    return base64.b64encode(val.encode('utf-8')).decode('utf-8')


def b64_decode(val: str) -> str:
    return base64.b64decode(val.encode('utf-8')).decode('utf-8')
