import ast
import json
import queue
import base64
import logging
import datetime
import operator
import requests
import threading
import ipaddress
import multiprocessing
import re
from urllib.parse import unquote
from netaddr import IPNetwork, IPRange
from django.contrib.auth.decorators import login_required
from django.http import JsonResponse, HttpResponse
from django.utils.translation import pgettext, gettext as _
from django.shortcuts import render, redirect
from django.utils import translation
from django.conf import settings
from django.urls import resolve
from django.contrib import messages
from django.views.decorators.http import require_POST
from django.templatetags.static import static
from django.contrib import auth
from django.core.exceptions import ObjectDoesNotExist
from url_normalize import url_normalize

from main.decorators import time_of_function
from main.models.my_waf import LicKeys, ErlEndOfLife, ModelInterval
from main.models.wafuser import WAFUSER

from main.utils import convert_from_punycode
from main.utils import convert_to_punycode
from main.utils import get_list_messages
from main.utils import get_proxy_api_dict
from main.utils import get_proxy_conf_dict
from main.utils import recursion_convert_from_punycode_obj

from my_waf.utils import check_keys
from my_waf.utils import b64_decode
from my_waf.utils import b64_encode
from my_waf.utils import bl_postprocessing_ips
from my_waf.utils import del_dyn_bl
from my_waf.utils import del_ip_bs
from my_waf.utils import get_and_search_bl
from my_waf.utils import get_dyn_settings
from my_waf.utils import get_erl_rule
from my_waf.utils import get_mlc_settings
from my_waf.utils import get_rules_management
from my_waf.utils import get_waf_id
from my_waf.utils import my_waf_access
from my_waf.utils import parse_values_erl_headers
from my_waf.utils import response_for_bs_management
from my_waf.utils import send_set_settings
from my_waf.utils import set_dyn_bl
from my_waf.utils import sort_bl

my_waf_url = 'waf_settings'
logger = logging.getLogger(__name__)
auth_url = settings.URI_AUTH

try:
    api_url = settings.API_URI
except AttributeError:
    api_url = 'http://localhost:8080/nw-api/'


@login_required
@time_of_function
def logout_view(request):
    request.LANGUAGE_CODE = translation.get_language()
    translation.activate(request.LANGUAGE_CODE)
    response = redirect(settings.HOME_PAGE)
    response.set_cookie(settings.LANGUAGE_COOKIE_NAME, request.LANGUAGE_CODE)
    auth.logout(request)
    return response


@my_waf_access
@login_required
@time_of_function
def dyn_tab(request):
    """
    Dyn page
    :param request: User request
    """

    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    processes = []
    processes_queue = multiprocessing.Queue()
    for key in lic_keys:
        proc = multiprocessing.Process(target=get_dyn_settings, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    keys_form_multiprocessing = sorted(keys_form_multiprocessing, key=operator.attrgetter('description'))
    for key in keys_form_multiprocessing:
        if not key.description:
            key.description = get_waf_id(key)

    context = {
        'lic_keys': keys_form_multiprocessing,
        'home_origin': settings.HOME_PAGE,
        'home': settings.HOME_PAGE + my_waf_url
    }

    return render(request, 'my_waf/tab_dyn_settings.html', context)


@my_waf_access
@login_required
@time_of_function
def mlc_tab(request):
    """
    MLC page
    :param request: User request
    """

    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    processes = []
    processes_queue = multiprocessing.Queue()
    for key in lic_keys:
        proc = multiprocessing.Process(target=get_mlc_settings, args=(key, processes_queue))
        proc.start()
        processes.append(proc)

    for pr in processes:
        pr.join()

    keys_form_multiprocessing = []
    while not processes_queue.empty():
        keys_form_multiprocessing.append(processes_queue.get())

    keys_form_multiprocessing = sorted(keys_form_multiprocessing, key=operator.attrgetter('description'))
    for key in keys_form_multiprocessing:
        if not key.description:
            key.description = get_waf_id(key)

    context = {
        'lic_keys': keys_form_multiprocessing,
        'home_origin': settings.HOME_PAGE,
        'home': settings.HOME_PAGE + my_waf_url
    }
    return render(request, 'my_waf/tab_mlc_settings.html', context)


@my_waf_access
@login_required
@time_of_function
def rule_management_tab(request):
    """
    Page with WL and ERL rules
    :param request: User request
    """
    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)

    threads = []
    threads_queue = queue.Queue()
    for key in lic_keys:
        thread = threading.Thread(target=get_rules_management, args=(key, threads_queue))
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

    keys_form_threading = []
    while not threads_queue.empty():
        keys_form_threading.append(threads_queue.get())

    keys_form_threading = sorted(keys_form_threading, key=operator.attrgetter('description'))
    for key in keys_form_threading:
        if not key.description:
            key.description = get_waf_id(key)

    context = {
        'home': settings.HOME_PAGE + my_waf_url,
        'home_origin': settings.HOME_PAGE,
        'lic_keys': keys_form_threading,
    }

    return render(request, 'my_waf/tab_rule_management.html', context)


def bs_management_if_domain(response: dict, int_row: int) -> tuple:
    """Logick from bs_management.view if in request key domain"""
    domain = response['ban_data'][0]['domain']
    list_ip = [{'ip': i['ip-address'], 'ttl': i['TTL']} for i in response['ban_data'] if i['domain'] == domain]
    len_list = len(list_ip)
    list = [list_ip[i:i + int_row] for i in range(0, len(list_ip), int_row)]
    quan_ip = 0
    if len(list) > 3:
        num = 3
    else:
        num = len(list)
    for i in range(0, num):
        quan_ip = quan_ip + len(list[i])
    if quan_ip < len_list:
        button = 'button'
    else:
        button = None
    return domain, list_ip, len_list, list, button


def bs_management_if_not_domain(response: dict, int_row: int, page: int, domain: str) -> tuple:
    """Logick from bs_management.view if not in request key domain"""
    list_ip = [{'ip': i['ip-address'], 'ttl': i['TTL']}
               for i in response['ban_data'] if i['domain'] == domain]
    len_list = len(list_ip)
    list = [list_ip[i:i + int_row + int_row * int(page)]
            for i in range(0, len(list_ip), int_row + int_row * int(page))]
    quan_ip = 0
    if len(list) > 3:
        num = 3
    else:
        num = len(list)
    for i in range(0, num):
        quan_ip = quan_ip + len(list[i])
    if quan_ip < len_list:
        button = 'button'
    else:
        button = None
    return domain, list_ip, len_list, list, button


def give_license(request) -> object:
    keys = LicKeys.objects.filter(user__id=request.user.id).all()
    lic_keys, checked_ids = check_keys(keys, check_waf_id=True)
    threads = []
    threads_queue = queue.Queue()
    for key in lic_keys:
        thread = threading.Thread(target=get_rules_management, args=(key, threads_queue))
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

    keys_form_threading = []
    while not threads_queue.empty():
        keys_form_threading.append(threads_queue.get())

    keys_form_threading = sorted(keys_form_threading, key=operator.attrgetter('description'))
    for key in keys_form_threading:
        if not key.description:
            key.description = get_waf_id(key)
    if keys_form_threading:
        return keys_form_threading, keys_form_threading[-1]
    else:
        return [], None


@my_waf_access
@login_required
@time_of_function
def bs_management(request, var=None):
    try:
        # Need exceptions this
        keys_form_threading, key = give_license(request)

        int_row = 5
        search = request.GET.get('search')
        domain = request.GET.get('domain')
        del_ip = request.GET.get('del_ip')
        page = request.GET.get('page')
        dell_all = request.GET.get('dell_all')
        key_user = request.GET.get('key_user')

        domain = convert_to_punycode(domain)

        if not key_user:
            key_user = 1

        if dell_all:
            del_ip = '*'
            del_ip_bs(request, del_ip, key=keys_form_threading[int(key_user) - 1].key)
        if not page:
            page = 0
        if del_ip:
            del_ip_bs(request, del_ip, key=keys_form_threading[int(key_user) - 1].key)

        response = response_for_bs_management(key=keys_form_threading[int(key_user) - 1].key)
        response['ban_data'] = [i for i in response['ban_data'] if i['is_banned'] == '1']

        list_domain = set([i['domain'] for i in response['ban_data']])
        len_ip_dict = {}
        for i in response['ban_data']:
            if i['domain'] not in len_ip_dict.keys():
                len_ip_dict[i['domain']] = 1
            else:
                len_ip_dict[i['domain']] = len_ip_dict[i['domain']] + 1

        user_desc = keys_form_threading[int(key_user) - 1].description
        if not domain:

            if response['ban_data']:

                domain, list_ip, len_list, list, button = bs_management_if_domain(response, int_row)
            else:
                domain, list_ip, len_list, list, button = '', [], 0, [], ''

        else:

            if response['ban_data']:
                domain, list_ip, len_list, list, button = bs_management_if_not_domain(response, int_row,
                                                                                      page, domain)
            else:
                domain, list_ip, len_list, list, button = '', [], 0, [], ''

        if search:
            list = [[i] for i in list_ip if i['ip'] == search]

        domain = recursion_convert_from_punycode_obj(domain)
        list_domain = recursion_convert_from_punycode_obj(list_domain, with_keys=True)
        len_ip_dict = recursion_convert_from_punycode_obj(len_ip_dict, with_keys=True)

        context = {
            'home': settings.HOME_PAGE + my_waf_url,
            'home_origin': settings.HOME_PAGE,
            'lic_keys': keys_form_threading,
            'domain': domain,
            'len_list': len_list,
            'list_domain': list_domain,
            'list_ip': list_ip,
            'page': page,
            'list': list,
            'button': button,
            'len_ip_dict': len_ip_dict,
            'user_descr': user_desc
        }
        if not var:
            return render(request, 'my_waf/bs_management.html', context)
        else:
            return domain, len_list, list_domain, list_ip, page, list, button, len_ip_dict, user_desc
    except Exception as e:
        logger.error(f'An error occurred while confirming information about the key: '
                     f'{key}. Error information: {e}')
        return {}


@my_waf_access
@login_required
@time_of_function
def tab_bs_management(request):
    try:
        keys_form_threading, key = give_license(request)
        domain, len_list, list_domain, list_ip, page, list, button, len_ip_dict, user_desc = bs_management(
            request, var=True)

        context = {
            'home': settings.HOME_PAGE + my_waf_url,
            'home_origin': settings.HOME_PAGE,
            'lic_keys': keys_form_threading,
            'domain': domain,
            'len_list': len_list,
            'list_domain': list_domain,
            'list_ip': list_ip,
            'page': page,
            'list': list,
            'button': button,
            'len_ip_dict': len_ip_dict,
            'user_descr': user_desc

        }

        return render(request, 'my_waf/tab_bs_management.html', context)
    except Exception as e:
        logger.error(f'An error occurred while confirming information about the key: '
                     f'{key}. Error information: {e}')

        context = {
            'home': settings.HOME_PAGE + my_waf_url,
            'home_origin': settings.HOME_PAGE,
        }
        return render(request, 'my_waf/tab_bs_management.html', context)


@require_POST
@login_required
@time_of_function
def add_key(request):
    key = str(request.POST.get('key'))
    description = str(request.POST.get('description'))
    user = WAFUSER.objects.get(id=request.user.id)
    waf_ids = []
    row_count = int(LicKeys.objects.filter(user=user).count()) + 1
    tarif = ''
    if key and len(key) > 6:
        format_key = 'xxx' + key[-6:]
    else:
        format_key = key

    try:
        check_key = requests.post(
            auth_url,
            headers={'User-agent': settings.USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': key},
        )
        if check_key.status_code == 401:
            logger.info('Response code 401 has been received while validating the key: {}'.format(format_key))
            response_dict = '{"status": "not valid"}'
            return HttpResponse(response_dict, content_type='application/json')
        elif check_key.ok:
            check_key_result = json.loads(check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success':
                keys = LicKeys.objects.filter(user__id=request.user.id).all()
                for k in keys:
                    response_check_key = requests.post(
                        auth_url,
                        headers={'User-agent': settings.USER_AGENT},
                        proxies=get_proxy_conf_dict(),
                        timeout=30,
                        data={'key': k.key}
                    )
                    if response_check_key.ok:
                        check_key = json.loads(response_check_key.text.replace('\'', '\"'))
                        if check_key['status'] == 'success' and str(check_key['r_asa_disable']) == 'False':
                            waf_ids.append(check_key['waf_id'])
                if check_key_result['waf_id'] in waf_ids:
                    return HttpResponse('{"status": "waf id exist"}', content_type='application/json')
                else:
                    k = LicKeys.objects.create(user=user, key=key, description=description, activated=True)
                    k.save()
                    tarif_id = int(check_key_result['tarif_id'])
                    if tarif_id == 0:
                        tarif = 'Light'
                    elif tarif_id == 1:
                        tarif = 'Business'
                    elif tarif_id == 2:
                        tarif = 'Enterprise'
                    key = {'remote_asa_disable': str(check_key_result['r_asa_disable']),
                           'models_num': str(check_key_result['models_num']),
                           'days_left': str(check_key_result['days_left']).replace('days', ''),
                           'waf_id': str(check_key_result['waf_id']),
                           'tarif': tarif,
                           'row_count': row_count,
                           'description': description,
                           'id': k.id,
                           }
                    context = {
                        'key': key,
                    }
                    return render(request, 'my_waf/new_row_key.html', context)
            else:
                logger.error('Get status fail when get get information for key: ' + format_key)

                messages.error(request, pgettext('messages', "An error occurred while verifying the key"))
                storage = messages.get_messages(request)
                list_messages = get_list_messages(storage)

                return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    except Exception as e:
        logger.error('An error occurred while confirming information about the key: {key}. Error information: {err}'
                     .format(key=format_key, err=e))


@require_POST
@login_required
@time_of_function
def edit_key(request):
    """
    Key editing
    :param request: User request
    :return: Request status
    """
    selected_key = request.POST.get('key_id')
    description = request.POST.get('description')
    key = LicKeys.objects.get(id=int(selected_key)).key
    waf_id = ''
    related_ids = []
    if key and len(key) > 6:
        format_key = 'xxx' + key[-6:]
    else:
        format_key = key

    try:
        response_check_key = requests.post(
            auth_url,
            headers={'User-agent': settings.USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': key}
        )
        if response_check_key.ok:
            check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success' and str(check_key_result['r_asa_disable']) == 'False':
                waf_id = check_key_result['waf_id']

        user_id = request.user.id

        lic_keys = LicKeys.objects.filter(user__id=user_id).all()
        for key in lic_keys:
            response_check_key = requests.post(
                auth_url,
                headers={'User-agent': settings.USER_AGENT},
                proxies=get_proxy_conf_dict(),
                timeout=30,
                data={'key': key.key}
            )
            if response_check_key.ok:
                check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
                if check_key_result['status'] == 'success' and str(check_key_result['r_asa_disable']) == 'False':
                    if check_key_result['waf_id'] == waf_id:
                        related_ids.append(key.id)

        LicKeys.objects.filter(id__in=related_ids).update(description=description)

        related_ids = [str(i) for i in related_ids]
        return HttpResponse('{{"status": "success", "keys": [{}] }}'.
                            format(','.join(related_ids)), content_type='application/json')
    except Exception as e:
        logger.error('An error occurred while getting information about the key: {key}. Error information: {err}'
                     .format(key=format_key, err=e))
        return HttpResponse('{{"status": "fail"}}', content_type='application/json')


@require_POST
@login_required
@time_of_function
def remove_key(request):
    """
    Deleting a key
    :param request: User request
    :return: Request status
    """
    selected_key = request.POST.get('key_id')
    LicKeys.objects.filter(id=selected_key).delete()
    return HttpResponse(status=200)


@require_POST
@my_waf_access
@login_required
@time_of_function
def set_settings(request):
    """
    Setting settings for MLC
    :param request: User request
    :return: Request status
    """
    selected_key = request.POST.get('key')
    setting_name = request.POST.get('setting_name')
    setting_text = request.POST.get('setting_text')
    key = LicKeys.objects.get(id=selected_key).key

    settings_dict = {setting_name: setting_text, 'active': True}

    if setting_name in ('ddos__wl_url', 'brute__brute_detect',
                        'brute__flood_detect', 'brute__wl_host',
                        'openapi__vhosts_list'):
        if settings_dict[setting_name] == '':
            settings_dict[setting_name] = []
        else:
            settings_dict[setting_name] = str(settings_dict[setting_name]).split(', ')

    if setting_name in ('ddos__interval', 'brute__interval', 'brute__max_val') \
            and setting_text.isdigit():
        settings_dict[setting_name] = int(setting_text)

    if setting_name in ('ddos__wl_ip', 'brute__wl_ip'):
        settings_dict[setting_name] = str(settings_dict[setting_name]).replace(',', '')

    if settings_dict[setting_name] == 'true':
        settings_dict[setting_name] = True

    if settings_dict[setting_name] == 'false':
        settings_dict[setting_name] = False

    full_settings_dict = {"key": key, "set": settings_dict}
    send_status = send_set_settings('MLC', full_settings_dict)
    if send_status[0] == 'success':
        return HttpResponse('{"status": "success"}', content_type='application/json')
    else:
        messages.error(request, pgettext('messages', send_status[1]))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(
            data={'messages': list_messages, 'status': 'error'},
            status=255
        )


@my_waf_access
@login_required
@time_of_function
def func_search_domain(request):
    """ Domain search function by address from the page BL html - mywaf/asa_bl.html
    """
    key_id = request.POST.get('key')
    search = request.POST.get('search', '')

    key = LicKeys.objects.get(id=key_id)

    search_list = list()
    if search:
        if ';' in search:
            search = search.split(';')
        else:
            search = [search]
        for i in search:
            search_list.append(i)

    bl_list = get_and_search_bl(key, search_list)

    res = dict()
    for i in bl_list:
        i_pieces = i.strip().split()
        if len(i_pieces) == 3:
            i_domain = i_pieces[0].replace('domain=', '')
            i_ip = i_pieces[1]
            i_api = True if i_pieces[2] and 'api=' in i_pieces[2] else False
        elif len(i_pieces) == 2:
            i_domain = i_pieces[0].replace('domain=', '') if 'domain=' in i_pieces[0] else None
            if not i_domain:
                i_ip = i_pieces[0]
                i_api = True if i_pieces[1] and 'api=' in i_pieces[1] else False
            else:
                i_ip = i_pieces[1]
                i_api = False
        elif len(i_pieces) == 1:
            i_ip = i_pieces[0]
            i_domain = None
            i_api = False
        else:
            continue

        if not i_domain:
            i_domain = '__ALL__'

        ip_data = {
            'ip': i_ip,
            'api': i_api
        }
        if not res.get(i_domain):
            res[i_domain] = {
                'ips': [],
                'count': 0
            }

        res[i_domain]['count'] = res[i_domain]['count'] +  1
        res[i_domain]['ips'].append(ip_data)

    if not res.get('__ALL__') and not search:
        res['__ALL__'] = {
            'ips': [],
            'count': 0
        }

    return JsonResponse(data=res)


@require_POST
@my_waf_access
@login_required
@time_of_function
def dyn_settings(request):
    """
    Setting settings for Dyn
    :param request: User request
    :return: Request status
    """
    selected_key = request.POST.get('key')
    setting_name = request.POST.get('setting_name')
    if setting_name == 'nwaf_limit':
        setting_text = request.POST.getlist('nwaf_limit_array[]')
        setting_text = ', '.join(setting_text)
    else:
        setting_text = request.POST.get('setting_text')
    hashed_key = LicKeys.objects.get(id=selected_key).key
    settings_dict = {setting_name: setting_text, 'active': True}

    full_settings_dict = {"key": hashed_key, "set": settings_dict}
    send_status = send_set_settings('DYN', full_settings_dict)
    if send_status[0] == 'success':
        return HttpResponse('{"status": "success"}', content_type='application/json')
    else:
        messages.error(request, pgettext('messages', send_status[1]))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def erl_settings(request):
    """
    Adding, editing and deleting ERL rules
    :param request: User request
    :return: Request status
    """
    url_name = resolve(request.path).url_name
    if url_name == 'Edit Erl':
        action = 'upd'
    elif url_name == 'Add Erl':
        action = 'add'
    else:
        action = 'del'
    erl_id = request.POST.get('erl_id')
    selected_key = request.POST.get('lic_keys_select')
    hashed_key = LicKeys.objects.get(id=selected_key).key

    if action == 'del':
        try:
            ErlEndOfLife.objects.get(rid=erl_id, key=LicKeys.objects.get(id=selected_key)).delete()
        except ErlEndOfLife.DoesNotExist:
            pass
        full_settings_dict = {"key": hashed_key, action: {'id': int(erl_id)}}
        send_status = send_set_settings('ERL', full_settings_dict)
        if send_status[0] == 'success':
            return HttpResponse(status=200)
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    if action == 'upd':
        settings_dict = {key: value for key, value in request.POST.items()}
    else:
        settings_dict = {key: value for key, value in request.POST.items() if value is not None and value != ''}

    settings_dict.pop('csrfmiddlewaretoken')
    settings_dict.pop('lic_keys_select')
    end_of_life_interval = settings_dict.pop('end_of_life')
    if end_of_life_interval != 'none' and action == 'upd':
        now_datetime = datetime.datetime.now()
        end_timestamp = now_datetime + datetime.timedelta(hours=int(end_of_life_interval))
        ErlEndOfLife.objects.update_or_create(
            rid=erl_id, key=LicKeys.objects.get(id=selected_key),
            defaults={'interval': end_of_life_interval, 'end_timestamp': datetime.datetime.timestamp(end_timestamp)},
        )
    if end_of_life_interval == 'none' and action == 'upd':
        try:
            ErlEndOfLife.objects.get(rid=erl_id, key=LicKeys.objects.get(id=selected_key)).delete()
        except ErlEndOfLife.DoesNotExist:
            pass
    original_other_headers = ''
    for key, value in settings_dict.items():
        if key not in ('ip', 'active', 'other_headers', 'other_headers_x', 'api', 'lm', 'noban', 'no_cookie', 'no_ua', 'no_referer', 'block_proxy', 'block_tor'):
            if key == 'domain':
                value = convert_to_punycode(value)

            if key in ('domain', 'country'):
                func_encode = lambda x: x # noqa
            else:
                func_encode = b64_encode

            if '|' in value or '&' in value:
                split_pipe = re.split(r'(?<!\\)\|', value)
                for val_index, val in enumerate(split_pipe):
                    if '&' in val:
                        split_ampersand = re.split(r'(?<!\\)&', val)
                        for el_index, el in enumerate(split_ampersand):
                            if '!' in el:
                                el = el.replace('\\|', '|')
                                el = el.replace('\\&', '&')
                                if '\\!' in el:
                                    split_ampersand[el_index] = func_encode(
                                        str(el).replace('\\!', '!')
                                    )
                                else:
                                    split_ampersand[el_index] = func_encode(
                                        str(el).replace('!', '')
                                    )
                                    split_ampersand[el_index] = '!' + split_ampersand[el_index]
                            else:
                                el = el.replace('\\&', '&')
                                el = el.replace('\\|', '|')
                                split_ampersand[el_index] = func_encode(str(el))
                        split_pipe[val_index] = '&'.join(split_ampersand)
                        continue
                    if '!' in val and '&' not in val:
                        val = val.replace('\\|', '|')
                        val = val.replace('\\&', '&')
                        if '\\!' in val:
                            split_pipe[val_index] = func_encode(
                                str(val).replace('\\!', '!')
                            )
                        else:
                            split_pipe[val_index] = func_encode(
                                str(val).replace('!', '')
                            )
                            split_pipe[val_index] = '!' + split_pipe[val_index]
                    else:
                        val = val.replace('\\|', '|')
                        val = val.replace('\\&', '&')
                        val = val.replace('\\!', '!')
                        split_pipe[val_index] = func_encode(str(val))
                value = '|'.join(split_pipe)
            elif '!' in value:
                if '\\!' in value:
                    value = func_encode(str(value).replace('\\!', '!'))
                else:
                    value = func_encode(str(value).replace('!', ''))
                    value = '!' + value
            else:
                value = func_encode(str(value))
            settings_dict[key] = value

    if 'active' in settings_dict and settings_dict['active'] == 'true':
        settings_dict['active'] = True
    else:
        settings_dict['active'] = False
    if 'api' in settings_dict and settings_dict['api'] == 'true':
        settings_dict['api'] = True
    else:
        settings_dict['api'] = False
    if 'lm' in settings_dict and settings_dict['lm'] == 'true':
        settings_dict['lm'] = True
    else:
        settings_dict['lm'] = False
    if 'noban' in settings_dict and settings_dict['noban'] == 'true':
        settings_dict['noban'] = True
    else:
        settings_dict['noban'] = False
    if 'no_cookie' in settings_dict and settings_dict['no_cookie'] == 'true':
        settings_dict['no_cookie'] = True
    else:
        settings_dict['no_cookie'] = False
    if 'no_ua' in settings_dict and settings_dict['no_ua'] == 'true':
        settings_dict['no_ua'] = True
    else:
        settings_dict['no_ua'] = False
    if 'no_referer' in settings_dict and settings_dict['no_referer'] == 'true':
        settings_dict['no_referer'] = True
    else:
        settings_dict['no_referer'] = False
    if 'block_proxy' in settings_dict and settings_dict['block_proxy'] == 'true':
        settings_dict['block_proxy'] = True
    else:
        settings_dict['block_proxy'] = False
    if 'block_tor' in settings_dict and settings_dict['block_tor'] == 'true':
        settings_dict['block_tor'] = True
    else:
        settings_dict['block_tor'] = False

    if settings_dict['block_proxy'] is True and settings_dict['block_proxy'] is False:
        settings_dict['block_tor'] = True

    for oh_key in ('other_headers', 'other_headers_x'):
        if oh_key in settings_dict and settings_dict[oh_key] != '':
            get_other_headers = str(settings_dict[oh_key]).split(',')
            original_other_headers = get_other_headers
            other_headers = []

            for el in get_other_headers:
                el = el.strip('"')
                name = el[:el.find(':')]
                values = el[el.find(':') + 1:]
                values = values.replace('\\:', ':')
                values = values.replace('\\"', '"')

                if '!!' in name:
                    name = name.replace('!!', '')
                    name = b64_encode(str(name))
                    name = '!!' + name
                else:
                    name = b64_encode(str(name))

                res_value = ""
                last_end_op = 0
                for start_idx, end_idx in parse_values_erl_headers(values):
                    if start_idx != 0:
                        text = values[last_end_op: start_idx]
                        res_value += b64_encode(str(text))

                    res_value += values[start_idx: end_idx]

                    last_end_op = end_idx
                else:
                    text = values[last_end_op:]
                    res_value += b64_encode(str(text))

                other_headers.append(name + ':' + res_value)
            settings_dict[oh_key] = other_headers

        if oh_key in settings_dict and settings_dict[oh_key] == '':
            settings_dict[oh_key] = []

    if action == 'upd':
        settings_dict.pop('erl_id')
        settings_dict['id'] = erl_id

    full_settings_dict = {"key": hashed_key, action: settings_dict}
    send_status = send_set_settings('ERL', full_settings_dict)

    for key, value in settings_dict.items():
        if key not in ('ip', 'active', 'other_headers', 'other_headers_x', 'api', 'lm', 'noban', 'no_cookie', 'no_ua', 'no_referer', 'block_proxy', 'block_tor', 'id'):

            if key in ('domain', 'country'):
                func_decode = lambda x: x # noqa
            else:
                func_decode = b64_decode

            if '|' in value or '&' in value:
                split_pipe = re.split(r'(?<!\\)\|', value)
                for val_index, val in enumerate(split_pipe):
                    if '&' in val:
                        split_ampersand = re.split(r'(?<!\\)&', val)
                        for el_index, el in enumerate(split_ampersand):
                            if '!' in el:
                                el = el.replace('\\|', '|')
                                el = el.replace('\\&', '&')
                                if '\\!' in el:
                                    split_ampersand[el_index] = func_decode(
                                        str(el).replace('\\!', '!')
                                    )
                                else:
                                    split_ampersand[el_index] = func_decode(
                                        str(el).replace('!', '')
                                    )
                                    split_ampersand[el_index] = '!' + split_ampersand[el_index]
                            else:
                                split_ampersand[el_index] = func_decode(str(el))
                                split_ampersand[el_index] = split_ampersand[el_index].replace('&', '\\&')
                                split_ampersand[el_index] = split_ampersand[el_index].replace('|', '\\|')
                                split_ampersand[el_index] = split_ampersand[el_index].replace('!', '\\!')
                        split_pipe[val_index] = '&'.join(split_ampersand)
                        continue
                    if '!' in val and '&' not in val:
                        val = val.replace('\\|', '|')
                        val = val.replace('\\&', '&')
                        if '\\!' in val:
                            split_pipe[val_index] = func_decode(
                                str(val).replace('\\!', '!')
                            )
                        else:
                            split_pipe[val_index] = func_decode(
                                str(val).replace('!', '')
                            )
                            split_pipe[val_index] = '!' + split_pipe[val_index]
                    else:
                        split_pipe[val_index] = func_decode(str(val))
                        split_pipe[val_index] = split_pipe[val_index].replace('|', '\\|')
                        split_pipe[val_index] = split_pipe[val_index].replace('&', '\\&')
                        split_pipe[val_index] = split_pipe[val_index].replace('!', '\\!')
                value = '|'.join(split_pipe)
            elif '!' in value:
                if '\\!' in value:
                    value = func_decode(str(value).replace('\\!', '!'))
                else:
                    value = func_decode(str(value).replace('!', ''))
                    value = '!' + value
                    value = value.replace('|', '\\|')
                    value = value.replace('&', '\\&')
            else:
                value = func_decode(str(value))
                value = value.replace('|', '\\|')
                value = value.replace('&', '\\&')
                value = value.replace('!', '\\!')
            settings_dict[key] = value
        if key == 'other_headers' or key == 'other_headers_x':
            settings_dict[key] = ',\n'.join(original_other_headers)

    key = {'id': selected_key}

    # ??? what it?
    if len(settings_dict) >= 1:
        settings_dict[list(settings_dict.keys())[0]] = convert_from_punycode(settings_dict[list(settings_dict.keys())[0]])
    if action == 'add':
        if 'ip' not in settings_dict:
            settings_dict['ip'] = ''
        if 'country' not in settings_dict:
            settings_dict['country'] = ''
        if send_status[0] == 'success':
            if end_of_life_interval != 'none':
                now_datetime = datetime.datetime.now()
                end_timestamp = now_datetime + datetime.timedelta(hours=int(end_of_life_interval))
                ErlEndOfLife.objects.update_or_create(
                    rid=send_status[1], key=LicKeys.objects.get(id=selected_key),
                    defaults={'interval': end_of_life_interval,
                              'end_timestamp': datetime.datetime.timestamp(end_timestamp)},
                )
            return HttpResponse('{"status": "success"}', content_type='application/json')
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    else:
        if send_status[0] == 'success':
            row = json.dumps(settings_dict)
            key = json.dumps(key)
            return HttpResponse('{{"status": "success", "row": {}, "key": {} }}'.
                                format(row, key), content_type='application/json')
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def bl_settings(request):
    """
    Adding, editing and deleting bl rules
    :param request: User requests
    :return: Requests status
    """
    url_name = resolve(request.path).url_name
    if url_name == 'Add BL':
        action = 'add'
    elif url_name == 'Delete BL':
        action = 'del'
    elif url_name == 'Get BL':
        action = 'get'
    else:
        return HttpResponse(status=401)

    key_id = request.POST.get('key_id')
    key = LicKeys.objects.get(id=key_id)

    domain = request.POST.get('domain')
    search = request.POST.get('search')
    search_type = int(request.POST.get('select_type', 0))

    if action == 'add':
        ips = request.POST.get('ips', '').strip(',; ')
        send_api = request.POST.get('api')

        list_ip = list()
        if ';' in ips:
            ips = [i.strip(', ') for i in ips.split(';') if i.strip(', ')]
        else:
            ips = [ips.strip(',; ')]

        invalid_ips = list()
        for ip in ips:
            try:
                if '/' in ip:
                    IPNetwork(ip)
                elif '-' in ip:
                    IPRange(ip.split('-')[0], ip.split('-')[1])
                else:
                    ipaddress.ip_network(ip)
            except Exception:
                invalid_ips.append(ip)
        if invalid_ips:
            return JsonResponse(
                status=401,
                data={'status': 'error', 'description': invalid_ips}
            )
        if domain and domain != "__ALL__" and not re.match(r'^(\*\.|\.)?([a-z\dа-я\d][a-z\dа-я\d-]*[a-z\dа-я\d]\.)+[a-zа-я]+$', domain):
            formatted_domain = []
            for num, i in enumerate(domain.split('.')):
                if not re.match(r'^(\*)|([a-z\dа-я\d][a-z\dа-я\d-]*[a-z\dа-я\d])+$ ', i):
                    formatted_domain.append('<span style="color: red;">' + i + '</span>')
                else:
                    if len(domain.split('.')) - 1 == num and not re.match(r'([a-z]){2,}$', i):
                        formatted_domain.append('<span style="color: red;">' + i + '</span>')
                    else:
                        formatted_domain.append(i)
            return JsonResponse(
                status=401,
                data={'status': 'error', 'description': '.'.join(formatted_domain)}
            )

        for i in ips:
            if not i:
                continue
            if domain == '__ALL__':
                if send_api == 'yes':
                    list_ip.append(i + ' api=yes')
                else:
                    list_ip.append(i)
            else:
                if send_api == 'yes':
                    list_ip.append('domain=' + domain + ' ' + i + ' api=yes')
                else:
                    list_ip.append('domain=' + domain + ' ' + i)

        result_del = set_dyn_bl(key, list_ip)

        if result_del[0] is False:
            return JsonResponse(
                status=401,
                data={'status': 'error', 'description': result_del[1]}
            )
        else:
            search_list = []
            if search:
                if ';' in search:
                    search = search.split(';')
                else:
                    search = [search]
                for i in search:
                    search_list.append(i)

            bl_list = get_and_search_bl(key, search_list)
            res_list, count_ips = bl_postprocessing_ips(
                bl_list, domain, search_type=search_type
            )
            ips_list = sort_bl(res_list)
            data = {
                'count': count_ips,
                'domain': domain,
                'ips': ips_list
            }
            if result_del[1]:
                data['duplicates'] = result_del[1]
            del result_del
            return JsonResponse(status=201, data=data)
    elif action == 'del':
        ip = request.POST.get('ip')
        del_domain = request.POST.get('del_domain')

        search_list = list()
        if del_domain:
            result_del = del_dyn_bl(key, domain, 1)
        elif search:

            if ';' in search:
                search = search.split(';')
            else:
                search = [search]
            for i in search:
                search_list.append(i)

            bl_list = get_and_search_bl(key, search_list)
            res_list, count_ips = bl_postprocessing_ips(
                bl_list, domain, search_type=search_type
            )
            list_to_del = list()
            for i in res_list:
                if domain == '__ALL__' and 'domain' not in i:
                    list_to_del.append(i)
                elif domain and 'domain=' + domain in i:
                    list_to_del.append(i)

            result_del = del_dyn_bl(key, list_to_del)
        elif ip:
            if domain == '__ALL__':
                del_ip = ip
            else:
                del_ip = 'domain=' + domain + ' ' + ip

            result_del = del_dyn_bl(key, [del_ip])
        else:
            messages.warning(request, pgettext('messages', "Invalid data"))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)
            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

        if result_del is False:
            return JsonResponse(status=401, data={'status': 'error', 'description': 'An occurred error in del bl list'})
        else:
            bl_list = get_and_search_bl(key, search_list)
            res_list, count_ips = bl_postprocessing_ips(
                bl_list, domain, search_type=search_type
            )
            ips_list = sort_bl(res_list)
            return JsonResponse(
                status=201,
                data={
                    'count': count_ips,
                    'domain': domain,
                    'ips': ips_list
                }
            )
    elif action == 'get':
        search_list = list()
        if search:
            if ';' in search:
                search = search.split(';')
            else:
                search = [search]
            for i in search:
                search_list.append(i)

        bl_list = get_and_search_bl(key, search_list)
        res_list, count_ips = bl_postprocessing_ips(
            bl_list, domain, search_type=search_type
        )
        ips_list = sort_bl(res_list)
        return JsonResponse(
            status=201,
            data={
                'count': count_ips,
                'domain': domain,
                'ips': ips_list
            }
        )
    else:
        return HttpResponse(401)


@require_POST
@my_waf_access
@login_required
@time_of_function
def delete_model(request):
    """
    Deleting model
    :param request: User request
    :return: Request status
    """
    key = request.POST.get('key')
    model = request.POST.get('model', '').replace('*.', 'sub_')
    key = LicKeys.objects.get(id=key)

    model = convert_to_punycode(model)

    delete_model_request = requests.post(
        url_normalize(api_url + '/del_models_uri'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': key.key,
              'vhost': model},

    )
    if delete_model_request.ok:

        return HttpResponse(status=200)
    else:
        model = convert_from_punycode(model)

        logger.info(
            'An error occurred while deleting model {model} with key: {key}'
            .format(model=model, key=key.get_formated_key())
        )
        messages.error(request, pgettext('messages', "Error with delete model {model}").format(model=model))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def model_training(request):
    """
    Training model
    :param request: User request
    :return: Request status
    """
    key = request.POST.get('key')
    hashed_key = LicKeys.objects.get(id=key).key
    model = request.POST.get('model')
    user = WAFUSER.objects.get(id=request.user.id)

    model = convert_to_punycode(model)

    if hashed_key and len(hashed_key) > 6:
        format_key = 'xxx' + hashed_key[-6:]
    else:
        format_key = hashed_key

    try:
        duration = ModelInterval.objects.get(user=user, model_name=model, key=LicKeys.objects.get(id=key)).interval
    except ModelInterval.DoesNotExist:
        duration = 4

    if duration == '':
        duration = 4

    response_list_vhost = requests.post(
        url_normalize(api_url + '/get_vhosts_list'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': hashed_key},
    )

    if response_list_vhost.ok:
        response_list_vhost_dict = ast.literal_eval(response_list_vhost.text)
        vhost_list = str(response_list_vhost_dict['vhosts_list']).strip('][').replace('\"', '')
        vhost_list = [i.strip() for i in vhost_list.split(', ')]
    else:
        logger.info('Response code {code} has been received from /get_vhosts_list for key: {key}'
                    .format(code=str(response_list_vhost.status_code), key=format_key))
        messages.error(request, pgettext('messages', "An error occurred while retrieving the list of model vhosts"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    vhost = None

    if model in vhost_list:
        vhost = model
    else:
        for i in vhost_list:
            if i.startswith('.') and (i in model or i[1:] == model):
                vhost = i
                break

    if vhost is None:
        model = convert_from_punycode(model)

        logger.info('The {model} model does not match any of the hosts from the list: {vhost_list} for key: {key}'.
                    format(model=str(model), vhost_list=vhost_list, key=format_key))
        messages.error(request,
                       pgettext('messages',
                                "The {model} model does not match any of the hosts from the list: {vhost_list}"
                                .format(model=str(model), vhost_list=vhost_list)))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    training_model_request = requests.post(
        url_normalize(api_url + '/set_training_uri'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': hashed_key,
              'vhost': vhost,
              'duration': int(duration),
              'complete': 'false',
              },
    )

    if training_model_request.ok and 'fail' not in training_model_request.text:
        return HttpResponse('{{"status": "success", "img": "{}" }}'.format(static('img/ai_off.png')),
                            content_type='application/json')
    else:
        if hashed_key and len(hashed_key) > 6:
            format_key = 'xxx' + hashed_key[-6:]
        else:
            format_key = hashed_key
        tmp_dict = json.loads(training_model_request.text.replace('\'', '\"'))
        if 'description' in tmp_dict:
            desc = ': ' + tmp_dict.get('description')
            end_log = 'with description: {}'.format(desc)
        else:
            desc = ''
            end_log = ''

        model = convert_from_punycode(model)
        logger.info('Error with training model {model} with key: {key} {end_log}'
                    .format(model=model, key=format_key, end_log=end_log))
        messages.error(request,
                       pgettext('messages', "Error with training model {model}{desc}").format(model=model, desc=desc))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def set_model_training_duration(request):
    """
    Set model training duration
    :param request: User request
    :return: Request status
    """
    user = WAFUSER.objects.get(id=request.user.id)
    model = request.POST.get('model')

    model = convert_to_punycode(model)

    duration = request.POST.get('duration')
    key_id = request.POST.get('key')
    key = LicKeys.objects.get(id=key_id)
    ModelInterval.objects.update_or_create(user=user, model_name=model, key=key, defaults={"interval": duration})
    return HttpResponse('{{"status": "success", "img": "{}" }}'.format(static('img/ai_off.png')),
                        content_type='application/json')


@require_POST
@my_waf_access
@login_required
@time_of_function
def get_model_duration(request):
    """
    Get model training duration
    :param request: User request
    :return: Request status
    """
    key = request.POST.get('key')
    hashed_key = LicKeys.objects.get(id=key).key
    model = request.POST.get('model')
    user = WAFUSER.objects.get(id=request.user.id)

    model = convert_to_punycode(model)

    try:
        duration = ModelInterval.objects.get(user=user, key=LicKeys.objects.get(id=key), model_name=model).interval
        return HttpResponse('{{"status": "success", "duration": "{}" }}'.format(str(duration)),
                            content_type='application/json')
    except ModelInterval.DoesNotExist:
        response_training_status = requests.post(
            url_normalize(api_url + '/get_training_uri'),
            headers={'User-agent': settings.USER_AGENT},
            proxies=get_proxy_api_dict(),
            timeout=30,
            data={'key': hashed_key,
                  'vhost': model},
        )
        if response_training_status.ok:
            response_training_status_dict = ast.literal_eval(response_training_status.text)
            try:
                duration = response_training_status_dict['duration']
            except KeyError:
                duration = 4
            return HttpResponse('{{"status": "success", "duration": "{}" }}'.format(str(duration)),
                                content_type='application/json')
        else:
            model = convert_from_punycode(model)

            if hashed_key and len(hashed_key) > 6:
                format_key = 'xxx' + hashed_key[-6:]
            else:
                format_key = hashed_key
            logger.info(
                'An error occurred while receiving training data for model {model} with key: {key}'
                .format(model=model, key=format_key)
            )
            messages.error(
                request,
                pgettext('messages', "An error occurred while receiving training data for model {model}").format(model=model)
            )
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def copy_model(request):
    """
    Copying model
    :param request: User request
    :return: Request status
    """
    key = request.POST.get('key')
    src_model = request.POST.get('src_model')
    dst_model = request.POST.get('dst_model')
    model_count = request.POST.get('models_count')
    key = LicKeys.objects.get(id=key)

    src_model = convert_to_punycode(src_model)
    dst_model = convert_to_punycode(dst_model)

    response_list_vhost = requests.post(
        url_normalize(api_url + '/get_vhosts_list'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': key.key},
    )
    if response_list_vhost.ok:
        response_list_vhost_dict = ast.literal_eval(response_list_vhost.text)
        vhost_list = str(response_list_vhost_dict['vhosts_list']).strip('][').replace('\"', '')
        vhost_list = vhost_list.split(', ')
    else:
        vhost_list = []
        logger.info('Response code {code} has been received from /get_vhosts_list for key: {key}'.
                    format(code=str(response_list_vhost.status_code), key=key.get_formated_key()))

    if src_model not in vhost_list or dst_model not in vhost_list:
        src_model = convert_from_punycode(src_model)
        dst_model = convert_from_punycode(dst_model)

        logger.info(
            'An error occurred when copying the model for the key {key}: the model "{model}" does not exist'
            .format(
                key=key.get_formated_key(),
                model=src_model if src_model not in vhost_list else dst_model
            )
        )
        return HttpResponse('{"status": "host does not exist"}', content_type='application/json')

    try:
        response_copy_model = requests.post(
            url_normalize(api_url + '/rep_models_uri'),
            headers={'User-agent': settings.USER_AGENT},
            proxies=get_proxy_api_dict(),
            timeout=30,
            data={'key': key.key,
                  'src': src_model,
                  'dst': dst_model
                  },
        )

        src_model = convert_from_punycode(src_model)
        dst_model = convert_from_punycode(dst_model)
        if response_copy_model.ok:

            context = {
                'key': key,
                'model_count': model_count,
                'model_name': dst_model,
            }
            return render(request, 'my_waf/new_model_row.html', context)
        else:
            src_model = convert_from_punycode(src_model)

            logger.info('An error occurred when copying model {model} for the key: {key}'.
                        format(model=src_model, key=key.get_formated_key()))

            messages.error(request,
                           pgettext('messages',
                                    "An error occurred when copying model {model}").format(model=src_model)
                           )
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    except Exception as e:
        src_model = convert_from_punycode(src_model)

        logger.error('An error occurred while getting information about the key: {key}. Error information: {err}'
                     .format(key=key.get_formated_key(), err=e))
        messages.error(request,
                       pgettext('messages',
                                "An error occurred when copying model {model}").format(model=src_model)
                       )
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@my_waf_access
@login_required
@time_of_function
def dyn_rules(request):
    """
    Adding, editing and deleting WL rules
    :param request: User request
    :return: Request status
    """
    url_name = resolve(request.path).url_name
    if url_name == 'Edit WL':
        action = 'upd'
    elif url_name == 'Add WL':
        action = 'add'
    else:
        action = 'del'
        selected_key = request.POST.get('lic_key')
        rid = request.POST.get('rid')
        hashed_key = LicKeys.objects.get(id=selected_key).key
        full_settings_dict = {"key": hashed_key, action: {'rid': int(rid)}}
        send_status = send_set_settings('WL', full_settings_dict)
        if send_status[0] == 'success':
            return HttpResponse(status=200)
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    get_form = request.POST.get('form')
    get_extension_list = request.POST.getlist('extension[]') or []
    get_form_list = str(get_form).split('&') or []
    get_mz = []
    for el in get_form_list:
        if 'mz=' in el:
            el = str(el).replace('mz=', '')
            for x, y in ("body", "BODY"), ("url", "URL"), ("args", "ARGS"), ("headers", "HEADERS"), \
                    ("mla", "MLA"), ("no", "No"), ('ban', 'Ban'):
                el = el.replace(x, y)
            get_mz.append(el)
    mz = '|'.join(get_mz)
    get_form_dict = {}
    for el in get_form_list:
        el = el.split('=')
        get_form_dict[el[0]] = el[1]
    extension_list = []
    for el in get_extension_list:
        el = str(el).split(';')
        if el[0] != 'none':
            if el[0].lower() == 'body' or el[0].lower() == 'url' \
                    or el[0].lower() == 'args' or el[0].lower() == 'headers':
                zone = '$' + el[0].upper()
            else:
                zone = '$' + el[0]
            if len(el) >= 3:
                if el[2] == 'true':
                    zone += '_X:'
                else:
                    zone += ':'
                extension_list.append(zone + el[1])
    if 'active' in get_form_dict and get_form_dict['active'] == 'true':
        get_form_dict['active'] = True
    else:
        get_form_dict['active'] = False

    if len(extension_list) > 0:
        extension = '|'.join(extension_list)
        get_form_dict['extension'] = extension
    get_form_dict['mz'] = mz
    if 'other_zone' in get_form_dict:
        if get_form_dict['other_zone'] != '':
            other_zone = '|'.join(unquote(get_form_dict['other_zone']).split(','))
            if get_form_dict['mz'] != '':
                get_form_dict['mz'] = get_form_dict['mz'] + '|' + other_zone
            else:
                get_form_dict['mz'] = other_zone
            get_form_dict.pop('other_zone')
        else:
            get_form_dict.pop('other_zone')
    selected_key = get_form_dict.get('lic_keys_select')

    try:
        hashed_key = LicKeys.objects.get(id=selected_key).key
    except ObjectDoesNotExist:
        messages.error(request, pgettext('messages', 'There is no record with such parameters.'))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    get_form_dict.pop('lic_keys_select')
    get_form_dict.pop('csrfmiddlewaretoken')

    if get_form_dict.get('domain'):
        get_form_dict['domain'] = convert_to_punycode(unquote(get_form_dict['domain']))

    full_settings_dict = {"key": hashed_key, action: get_form_dict}
    send_status = send_set_settings('WL', full_settings_dict)

    if 'active' in get_form_dict and get_form_dict['active'] is True:
        get_form_dict['active'] = 'True'
    else:
        get_form_dict['active'] = 'False'

    get_form_dict['mz'] = str(get_form_dict['mz']).upper()
    if 'extension' in get_form_dict:
        extension = get_form_dict['extension'].split('|$')
        extension_dict = {}
        for ext in extension:
            extension_dict[ext[:ext.rfind(':')].replace('$', '')] = ext[ext.find(':') + 1:]
        get_form_dict['extension_dict'] = extension_dict
        extension_str = re.sub(r'\|\$', '|', get_form_dict['extension'])
        get_form_dict['extension'] = extension_str[1:]
    key = {'id': selected_key}

    if get_form_dict.get('domain'):
        get_form_dict['domain'] = convert_from_punycode(get_form_dict['domain'])

    if action == 'add':
        if send_status[0] == 'success':
            get_form_dict['rid'] = send_status[1]
            get_form_dict['wl_id'] = get_form_dict['rl_id']
            get_form_dict['mz'] = get_form_dict['mz'].replace('COOKIE', 'Cookie').replace('URL', 'PATH')
            context = {
                'key': key,
                'rule': get_form_dict,
            }
            return render(request, 'my_waf/new_row_wl.html', context)
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    else:
        if send_status[0] == 'success':
            get_form_dict['wl_id'] = get_form_dict['rl_id']
            row = json.dumps(get_form_dict)
            key = json.dumps(key)
            return HttpResponse('{{"status": "success", "row": {}, "key": {} }}'.
                                format(row, key), content_type='application/json')
        else:
            messages.error(request, pgettext('messages', send_status[1]))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@require_POST
@login_required
@time_of_function
def settings_info(request):
    """
    Getting description of parameters from localization
    :param request: User request
    :return: JSON with descriptions
    """
    settings_info_text = {
        'vhosts_list': _('Vhost list info'),
        'models_list': _('Models list info'),
        'openapi_vhosts_list': _('Openapi vhosts list info'),
        'ai_extra': _('AI Extra info'),
        'ddos_enable': _('DDoS enable info'),
        'ddos_wl_ip': _('DDoS WL IP info'),
        'ddos_wl_url': _('DDoS WL URL info'),
        'ddos_interval': _('DDoS Interval info'),
        'ddos_latest_only': _('DDoS Latest only info'),
        'ddos_send_possible': _('DDoS Send possible info'),
        'brute_enable': _('Brute enable info'),
        'brute_detect': _('Brute detect info'),
        'brute_bf_detect': _('Brute bf detect info'),
        'brute_detect_type': _('BF_Type info'),
        'brute_detect_zone': _('BF_Zone info'),
        'brute_detect_condition': _('BF condition info'),
        'flood_detect': _('Flood detect info'),
        'brute_wl_host': _('Brute WL Host info'),
        'brute_wl_ip': _('Brute WL IP info'),
        'brute_interval': _('Brute interval info'),
        'brute_max_val': _('Brute max val info'),
        'brute_latest_only': _('Brute Latest only info'),
        'brute_send_possible': _('Brute Send possible info'),
        'nwaf_rfc_violation_wl': _('nwaf_rfc_violation_wl info'),
        'nwaf_rfc_violation_lm': _('nwaf_rfc_violation_lm info'),
        'nwaf_ai_extra_host_wl': _('nwaf_ai_extra_host_wl info'),
        'nwaf_ai_extra_host_lm': _('nwaf_ai_extra_host_lm info'),
        'nwaf_bf_detect_host_lm': _('nwaf_bf_detect_host_lm info'),
        'nwaf_body_exclude': _('nwaf_body_exclude info'),
        'nwaf_body_bin_exclude': _('nwaf_body_bin_exclude info'),
        'nwaf_ddos_detect_host_lm': _('nwaf_ddos_detect_host_lm info'),
        'nwaf_host_lm': _('nwaf_host_lm info'),
        'nwaf_host_wl': _('nwaf_host_wl info'),
        'nwaf_ip_lm': _('nwaf_ip_lm info'),
        'nwaf_ip_wl': _('nwaf_ip_wl info'),
        'nwaf_limit': _('nwaf_limit info'),
        'nwaf_wl': _('nwaf_wl info'),

        'nwaf_url': _('nwaf_url info'),
        'nwaf_domain': _('nwaf_host info'),
        'nwaf_args': _('nwaf_args info'),
        'nwaf_ua': _('nwaf_agent info'),
        'nwaf_cookie': _('nwaf_cookie info'),
        'nwaf_country': _('nwaf_country info'),
        'nwaf_body': _('nwaf_body info'),
        'nwaf_referer': _('nwaf_refer info'),
        'nwaf_other_headers': _('nwaf_other info'),
        'nwaf_ip': _('nwaf_ip info'),
        'nwaf_block_tor': _('nwaf_block_tor info'),
        'nwaf_block_proxy': _('nwaf_block_proxy info'),

        'nwaf_mla_host_lm': _('nwaf_mla_host_lm info'),
        'nwaf_put_body_exclude': _('nwaf_put_body_exclude info'),
        'nwaf_rmq_host_exclude': _('nwaf_rmq_host_exclude info'),
        'nwaf_openapi_url_wl': _('nwaf_openapi_url_wl info'),
        'nwaf_openapi_ip_wl': _('nwaf_openapi_ip_wl info'),
        'nwaf_openapi_url_lm': _('nwaf_openapi_url_lm info'),
        'nwaf_openapi_ip_lm': _('nwaf_openapi_ip_lm info'),
        'nwaf_post_body_exclude': _('nwaf_post_body_exclude info'),
        'nwaf_url_wl': _('nwaf_url_wl info'),
        'more_button_text': _('More button text'),
        'less_button_text': _('Less button text'),
        'erl_multiline_info': _('Erl multiline info'),
        'nwaf_b64_decode_disable': _('nwaf_b64_decode_disable info'),
        'nwaf_geoip_mla_disable': _('nwaf_geoip_mla_disable info'),
    }
    return JsonResponse(settings_info_text)


@require_POST
@my_waf_access
@login_required
@time_of_function
def get_vhost_list(request):
    """
    Getting list of vhost
    :param request: User request
    :return: JSON with vhosts
    """
    key = request.POST.get('key')
    hashed_key = LicKeys.objects.get(id=int(key)).key

    response_list_vhost = requests.post(
        url_normalize(api_url + '/get_vhosts_list'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': hashed_key},
    )
    if response_list_vhost.ok:
        response_list_vhost_dict = ast.literal_eval(response_list_vhost.text)
        vhost_list = str(response_list_vhost_dict['vhosts_list']).strip('][').replace('\"', '')
        vhost_list = vhost_list.split(', ')
    else:
        if hashed_key and len(hashed_key) > 6:
            format_key = 'xxx' + hashed_key[-6:]
        else:
            format_key = hashed_key
        logger.info('Response code {code} has been received from /get_vhosts_list for key: {key}'.
                    format(code=str(response_list_vhost.status_code), key=format_key))
        messages.error(request, pgettext('messages', "An error occurred while retrieving the list of model vhosts"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    res_list_vhost = []
    for i in vhost_list:
        res_list_vhost.append(convert_from_punycode(i))

    data = {'status': 'success', 'vhost_list': res_list_vhost}
    return JsonResponse(data)


@require_POST
@my_waf_access
@login_required
@time_of_function
def update_vhost_list(request):
    """
    Vhost list update
    :param request: User request
    :return: Request status
    """
    key = request.POST.get('key')
    action = request.POST.get('action')
    vhost = request.POST.get('vhost')
    key = LicKeys.objects.get(id=int(key))
    hashed_key = key.key
    vhost = convert_to_punycode(vhost)

    if not vhost:
        messages.error(request, pgettext('messages', "The list of vhosts is empty"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)

        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)

    response_list_vhost = requests.post(
        url_normalize(api_url + '/get_vhosts_list'),
        headers={'User-agent': settings.USER_AGENT},
        proxies=get_proxy_api_dict(),
        timeout=30,
        data={'key': hashed_key},
    )
    if response_list_vhost.ok:
        response_list_vhost_dict = ast.literal_eval(response_list_vhost.text)
        vhost_list = str(response_list_vhost_dict['vhosts_list']).strip('][').replace('\"', '')
        vhost_list = vhost_list.split(', ')
    else:
        vhost_list = []
        if hashed_key and len(hashed_key) > 6:
            format_key = 'xxx' + hashed_key[-6:]
        else:
            format_key = hashed_key
        logger.info('Response code {code} has been received from /get_vhosts_list for key: {key}'
                    .format(code=str(response_list_vhost.status_code), key=format_key))
    try:
        response_check_key = requests.post(
            auth_url,
            headers={'User-agent': settings.USER_AGENT},
            proxies=get_proxy_conf_dict(),
            timeout=30,
            data={'key': hashed_key}
        )
        if response_check_key.ok:
            check_key_result = json.loads(response_check_key.text.replace('\'', '\"'))
            if check_key_result['status'] == 'success':
                max_models = int(check_key_result['models_num'])
                vhost_list = list(filter(None, vhost_list))
                exist_errors = False
                if (len(vhost_list) + len(str(vhost).split(',')) > max_models) and action == 'add':
                    return HttpResponse('{"status": "exceeds the host limit"}', content_type='application/json')
                else:
                    if action == 'add':
                        vhost_list += [i.strip() for i in str(vhost).split(',')]
                    else:
                        vhost_split = [i.strip() for i in str(vhost).split(',')]
                        if len(vhost_split) > 0:
                            for host in vhost_split:
                                if host in vhost_list:
                                    vhost_list.remove(host)
                                else:
                                    host = convert_from_punycode(host)
                                    messages.error(
                                        request,
                                        pgettext(
                                            'messages',
                                            "The domain named {} does not exist in the list.".format(
                                                host
                                            )
                                        )
                                    )
                                    exist_errors = True
                        else:
                            vhost_list.remove(vhost)

                    set_vhosts = requests.post(
                        url_normalize(api_url + '/set_vhosts_list'),
                        headers={'User-agent': settings.USER_AGENT},
                        proxies=get_proxy_api_dict(),
                        timeout=30,
                        json={'key': hashed_key,
                              'vhosts_list': vhost_list},
                    )

                    if set_vhosts.text:
                        api_response = ast.literal_eval(set_vhosts.text)
                    else:
                        api_response = {'status': 'fail'}

                    if api_response['status'] == 'fail':
                        if exist_errors:
                            storage = messages.get_messages(request)
                            list_messages = get_list_messages(storage)

                            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
                        logger.error('Error with set vhost list: ' + str(vhost_list))
                        return HttpResponse('{"status": "error"}', content_type='application/json')
                    else:
                        if exist_errors:
                            storage = messages.get_messages(request)
                            list_messages = get_list_messages(storage)

                            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
                        return HttpResponse('{"status": "success"}', content_type='application/json')
        else:
            if hashed_key and len(hashed_key) > 6:
                format_key = 'xxx' + hashed_key[-6:]
            else:
                format_key = hashed_key
            logger.info('Response code {code} has been received while validating the key: {key}'.
                        format(code=str(response_check_key.status_code), key=format_key))
            messages.error(request, pgettext('messages', "An error occurred while getting information about the keys"))
            storage = messages.get_messages(request)
            list_messages = get_list_messages(storage)

            return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)
    except Exception as e:
        logger.error('An error occurred while getting information about the key: {key}. Error information: {err}'
                     .format(key=key.get_formated_key(), err=e))
        messages.error(request, pgettext('messages', "An error occurred while getting information about the keys"))
        storage = messages.get_messages(request)
        list_messages = get_list_messages(storage)
        return JsonResponse(data={'messages': list_messages, 'status': 'error'}, status=255)


@my_waf_access
@login_required
@time_of_function
def erl_rule(request):
    """
    Page with ERL rule
    :param request: User request
    """
    context = {}
    if request.GET.get('key_select'):
        key_id = request.GET.get('key_select')
        lic_key = LicKeys.objects.get(id=int(key_id))
        context = {
            'home': settings.HOME_PAGE + my_waf_url,
            'home_origin': settings.HOME_PAGE,
            'key': lic_key.id,
        }
    if request.GET.get('erl_id'):
        key_id = request.POST.get('lic_keys_select')
        lic_key = LicKeys.objects.get(id=int(key_id))
        context = {
            'home': settings.HOME_PAGE + my_waf_url,
            'home_origin': settings.HOME_PAGE,
            'key': lic_key.id,
        }
        get_erl = get_erl_rule(lic_key, request.GET.get('erl_id'))
        get_erl['lic_keys_select'] = key_id
        get_erl = {key: value for key, value in get_erl.items() if value is not None and value != ''
                   and value is not False}

        # ???
        if 'active' in get_erl and get_erl['active'] is True:
            get_erl['active'] = 'true'
        if 'api' in get_erl and get_erl['api'] is True:
            get_erl['api'] = 'true'
        if 'lm' in get_erl and get_erl['lm'] is True:
            get_erl['lm'] = 'true'
        if 'noban' in get_erl and get_erl['noban'] is True:
            get_erl['noban'] = 'true'
        if 'no_cookie' in get_erl and get_erl['no_cookie'] is True:
            get_erl['no_cookie'] = 'true'
        if 'no_ua' in get_erl and get_erl['no_ua'] is True:
            get_erl['no_ua'] = 'true'
        if 'no_referer' in get_erl and get_erl['no_referer'] is True:
            get_erl['no_referer'] = 'true'
        if 'block_proxy' in get_erl and get_erl['block_proxy'] is True:
            get_erl['block_proxy'] = 'true'
        if 'block_tor' in get_erl and get_erl['block_tor'] is True:
            get_erl['block_tor'] = 'true'

        get_erl['erl_id'] = str(get_erl['id'])
        get_erl.pop('id')
        if 'other_headers' in get_erl:
            if len(get_erl['other_headers']) == 0:
                get_erl.pop('other_headers')
            else:
                get_erl['other_headers'] = ','.join(get_erl['other_headers'])

        if 'other_headers_x' in get_erl:
            if len(get_erl['other_headers_x']) == 0:
                get_erl.pop('other_headers_x')
            else:
                get_erl['other_headers_x'] = ','.join(get_erl['other_headers_x'])

        # it for obfuscation????
        settings_dict = get_erl
        if 'active' in settings_dict and settings_dict['active'] == 'true':
            settings_dict['active'] = 'True'
        else:
            settings_dict['active'] = 'False'
        if 'api' in settings_dict and settings_dict['api'] == 'true':
            settings_dict['api'] = 'True'
        else:
            settings_dict['api'] = 'False'
        if 'lm' in settings_dict and settings_dict['lm'] == 'true':
            settings_dict['lm'] = 'True'
        else:
            settings_dict['lm'] = 'False'
        if 'noban' in settings_dict and settings_dict['noban'] == 'true':
            settings_dict['noban'] = 'True'
        else:
            settings_dict['noban'] = 'False'
        if 'no_cookie' in settings_dict and settings_dict['no_cookie'] == 'true':
            settings_dict['no_cookie'] = 'True'
        else:
            settings_dict['no_cookie'] = 'False'
        if 'no_ua' in settings_dict and settings_dict['no_ua'] == 'true':
            settings_dict['no_ua'] = 'True'
        else:
            settings_dict['no_ua'] = 'False'
        if 'no_referer' in settings_dict and settings_dict['no_referer'] == 'true':
            settings_dict['no_referer'] = 'True'
        else:
            settings_dict['no_referer'] = 'False'
        if 'block_proxy' in settings_dict and settings_dict['block_proxy'] == 'true':
            settings_dict['block_proxy'] = 'True'
        else:
            settings_dict['block_proxy'] = 'False'
        if 'block_tor' in settings_dict and settings_dict['block_tor'] == 'true':
            settings_dict['block_tor'] = 'True'
        else:
            settings_dict['block_tor'] = 'False'

        context['erl'] = settings_dict
        for key, value in settings_dict.items():
            if isinstance(value, str) and key not in ('active', 'ip', 'other_headers', 'other_headers_x', 'api', 'id', 'lm', 'noban', 'no_cookie', 'no_ua', 'no_referer', 'block_proxy', 'block_tor', 'erl_id', 'lic_keys_select'):
                settings_dict[key] = list(
                    filter(
                        None,
                        value.replace('|!', ',|!').replace('|&', ',|&')
                        .replace('|', ',|').replace('&', ',&')
                        .replace('!', '!').split(',')
                    )
                )

                result = []
                for el in settings_dict[key]:
                    if el[0] == '|' and el[1] == '!':
                        result.append(['|!', base64.b64decode(str(el[2:]).encode('utf-8')).decode('utf-8')])
                    elif el[0] == '&' and el[1] == '!':
                        result.append(['&!', base64.b64decode(str(el[2:]).encode('utf-8')).decode('utf-8')])
                    elif el[0] == '|' and el[1] != '!':
                        result.append(['|', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                    elif el[0] == '&' and el[1] != '!':
                        result.append(['&', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                    elif el[0] == '!':
                        result.append(['!', base64.b64decode(str(el[1:]).encode('utf-8')).decode('utf-8')])
                    else:
                        result.append([' ', base64.b64decode(str(el).encode('utf-8')).decode('utf-8')])
                settings_dict[key] = sorted(result, reverse=False, key=lambda x: [x[0], x[1]])
            if key == 'ip':
                settings_dict[key] = list(
                    filter(None, value.replace('|!', ',|!').replace('|&', ',|&').replace('|', ',|').replace('&', ',&').
                           replace('!', '!').split(',')))

                result = []
                for el in settings_dict[key]:
                    if el[0] == '|' and el[1] == '!':
                        result.append(['|!', el[2:]])
                    elif el[0] == '&' and el[1] == '!':
                        result.append(['&!', el[2:]])
                    elif el[0] == '|' and el[1] != '!':
                        result.append(['|', el[1:]])
                    elif el[0] == '&' and el[1] != '!':
                        result.append(['&', el[1:]])
                    elif el[0] == '!':
                        result.append(['!', el[1:]])
                    else:
                        result.append([' ', el])
                settings_dict[key] = result
            if key == 'other_headers' or key == 'other_headers_x':
                other_headers = []
                settings_dict[key] = settings_dict[key].split(',')

                for el in settings_dict[key]:
                    name = el[:el.find(':')]
                    values = el[el.find(':') + 1:]
                    if '!!' in name:
                        name = name.replace('!!', '')
                        name = base64.b64decode(str(name).encode('utf-8')).decode('utf-8')
                        name = '!!' + name
                    else:
                        name = base64.b64decode(str(name).encode('utf-8')).decode('utf-8')

                    result = []
                    last_end_op = 0
                    last_op = " "
                    for start_idx, end_idx in parse_values_erl_headers(values):
                        if start_idx != 0:
                            text = values[last_end_op: start_idx]
                            r_val = base64.b64decode(
                                str(text).encode('utf-8')
                            ).decode('utf-8')

                            result.append([last_op, r_val])

                        last_op = values[start_idx: end_idx]
                        last_end_op = end_idx
                    else:
                        text = values[last_end_op:]
                        r_val = base64.b64decode(
                            str(text).encode('utf-8')
                        ).decode('utf-8')
                        result.append([last_op, r_val])

                    header = {'key': name, 'value': result}
                    other_headers.append(header)
                settings_dict[key] = other_headers

            settings_dict[key] = recursion_convert_from_punycode_obj(settings_dict[key])

        try:
            settings_dict['end_of_life'] = ErlEndOfLife.objects.get(key=lic_key, rid=settings_dict['erl_id']).interval
        except ErlEndOfLife.DoesNotExist:
            settings_dict['end_of_life'] = None
    response = render(request, 'my_waf/erl_rule.html', context)
    response.set_cookie(key='settings_rule_page', value='ERL')
    return response
