import base64
import json
import logging
from json import JSONDecodeError
from urllib.parse import urlparse
import requests
from django.conf import settings
from django.utils.translation import pgettext
from main.utils import get_proxy_api_dict
from url_normalize import url_normalize


try:
    api_url = settings.API_URI
except AttributeError:
    api_url = 'http://localhost:8080/nw-api/'

base_nginx_url = url_normalize(api_url + '/nginx')
USER_AGENT = settings.USER_AGENT
logger = logging.getLogger(__name__)

list_type_bool_keys = ['proxy_intercept_error', 'ssl_stapling', 'backup', 'down', 'ssl']
list_type_list_keys = [
    'ssl_ciphers', 'ssl_protocols', 'dns', 'index', 'set_real_ip_from', 'proxy_set_header',
    'set_real_ip_from'
]
list_type_int_keys = [
    'fail_timeout', 'max_fails', 'weight', 'client_buffer_body_size', 'client_max_body_size', 'client_body_timeout',
    'listen_port', 'ssl_timeout', 'ssl_session_cache_size', 'proxy_connect_timeout', 'proxy_send_timeout',
    'proxy_read_timeout', 'port', 'size'
]
list_type_float_keys = ['proxy_http_version', ]
list_type_dynamic_keys = ['servers', ]


def get_nginx_json(key):
    request_json = send_get_request_nginx_json(key, 'get_nginx')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the nginx configuration for the key: {key}'
                .format(key=format_key)
            )
            return None
        else:
            return request_json.get('configuration')
    else:
        return None


def get_nginx_upstream_json(key):
    request_json = send_get_request_nginx_json(key, 'get_upstream')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the upstream configuration for the key: {key}'
                .format(key=format_key)
            )
            return None
    else:
        return None


def get_nginx_vhost_json(key):
    request_json = send_get_request_nginx_json(key, 'get_vhost')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the vhost configuration for the key: {key}'
                .format(key=format_key)
            )
            return None
        else:
            res_tmp = sorted(request_json.get('configuration'), key=lambda item: item['vhost'])

            return res_tmp
    else:
        return None


def get_nginx_ssl_json(key):
    request_json = send_get_request_nginx_json(key, 'get_vhost')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the ssl configuration for the key: {key}'
                .format(key=format_key)
            )
            return None
    else:
        return None


def get_nginx_location_json(key):
    request_json = send_get_request_nginx_json(key, 'get_vhost')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the ssl configuration for the key: {key}'
                .format(key=format_key)
            )
            return None
    else:
        return None


def send_get_request_nginx_json(key, path_url):
    try:
        response = requests.post(
            url_normalize(base_nginx_url + f'/{path_url}?format=json'),
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_api_dict(),
            json={"key": key},
            timeout=30
        )
    except Exception as e:
        logger.error(
            'An error occurred when receiving data in json format by URL \'{path_url}\': {error}'
            .format(path_url=base_nginx_url + f'/{path_url}', error=str(e))
        )
        return None

    if response.ok:
        try:
            return json.loads(response.text)
        except JSONDecodeError:
            logger.error(
                'An error occurred while get data in json format by URL \'{path_url}\': {code}'
                .format(path_url=base_nginx_url + f'/{path_url}', code='The response is not in json format')
            )
            return None
    else:
        logger.error(
            'An error occurred while get data in json format by URL \'{path_url}\' with response code: {code}'
            .format(path_url=base_nginx_url + f'/{path_url}', code=str(response.status_code))
        )

        return None


def send_get_request_nginx_conf(key, path_url):
    try:
        response = requests.post(
            url_normalize(base_nginx_url + f'/{path_url}'),
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_api_dict(),
            json={"key": key},
            timeout=30
        )
    except Exception as e:
        logger.error(
            'An error occurred when receiving data by URL \'{path_url}\': {error}'
            .format(path_url=base_nginx_url + f'/{path_url}', error=str(e))
        )
        return None

    if response.ok:
        return json.loads(response.text)
    else:
        logger.error(
            'An error occurred while get data by URL \'{path_url}\' with response code: {code}'
            .format(path_url=base_nginx_url + f'/{path_url}', code=str(response.status_code))
        )

        return None


def send_set_request_nginx(key, data, path_url):
    try:
        response = requests.post(
            url_normalize(base_nginx_url + f'/{path_url}'),
            headers={'User-agent': USER_AGENT},
            proxies=get_proxy_api_dict(),
            json={"key": key, "set": data},
            timeout=30
        )
    except Exception as e:
        logger.error(
            'An error occurred when receiving data by URL \'{path_url}\': {error}'
            .format(path_url=base_nginx_url + f'/{path_url}', error=str(e))
        )
        return None

    if response.ok:
        return json.loads(response.text)
    else:
        logger.error(
            'An error occurred while get data by URL \'{path_url}\' with response code: {code}'
            .format(path_url=base_nginx_url + f'/{path_url}', code=str(response.status_code))
        )

        return None


def set_nginx_conf(key, data):
    data = processing_data(data)
    request_json = send_set_request_nginx(key, data, 'set_nginx')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the nginx configuration for the key: {key}'
                .format(key=format_key)
            )
            if request_json.get('description'):
                return False, request_json.get('description')
            else:
                return False, 'Unknown error'
        else:
            return True, 'Ok'
    else:
        return None, 'Unknown error'


def set_nginx_vhost_json(key, data):
    data = processing_data(data)
    request_json = send_set_request_nginx(key, data, 'set_vhost')
    if request_json is not None:
        status = request_json.get('status')
        if status != 'success':
            if key and len(key) > 6:
                format_key = 'xxx' + key[-6:]
            else:
                format_key = key
            logger.error(
                'The fail status was received when receiving the vhost configuration for the key: {key}'
                .format(key=format_key)
            )
            if request_json.get('description'):
                return False, request_json.get('description')
            else:
                return False, 'Unknown error'
        else:
            return True, 'Ok'
    else:
        return None, 'Unknown error'


def processing_data(data):
    res = dict()
    for key in data:
        if not data.get(key):
            data[key] = None
        elif key in list_type_list_keys:
            if not isinstance(data[key], list):
                data[key] = data[key].replace(' ', '')
                try:
                    data[key] = json.loads(data[key].replace('\'', '\"'))
                except Exception:
                    data[key] = data.get(key, '').replace('\'\"', "").strip('[]').split(',')
        elif key in list_type_float_keys:
            data[key] = float(data[key])
        elif key in list_type_int_keys:
            data[key] = int(data[key])
        elif key in list_type_bool_keys:
            data[key] = bool(data[key])

        res[key.replace(' ', '')] = data[key]

    return res
