import hashlib
import logging
from datetime import datetime
from pathlib import Path
from django.conf import settings
from django.db import close_old_connections
from django.core.cache import cache


from report.utils.pdf_generator import generate_pdf
from report.utils.utils import get_filtered_attacks
from utils.mail import is_smtp_enabled

logger = logging.getLogger(__name__)

try:
    from uwsgidecorators import spool

    spoolIsrun = True
except Exception:
    spoolIsrun = False


    def spool(func):
        def func_wrapper(**arguments):
            return func(arguments)

        return func_wrapper


class PdfGenerateStatus:
    status = int
    short_task = 0
    will_send_to_mail = 1
    duplicate = 2
    mail_not_allowed = 3


@spool
def _start_generate_with_spool(spool_kwargs):
    from django.core.cache import cache  # dont delete it!
    cache_key = spool_kwargs.get('cache_key')
    if not cache:
        logger.error('An error occurred while generating pdf: cache timeout. Task failed')
        return
    kwargs = cache.get(cache_key)
    try:
        generate_pdf(**kwargs)
    except Exception as err:
        logger.exception(f'An error occurred while generating pdf: {err}')
    cache.delete(cache_key)


def start_generate_pdf_task(start_date: datetime,
                            end_date: datetime,
                            pdf_path: Path,
                            request) -> PdfGenerateStatus.status:
    """ launch generation pdf. Allows to launch only one proc with
    specific proc_name. Return generate status  """

    task_status = _specify_run_status(start_date, end_date, request)
    send_mail = task_status == PdfGenerateStatus.will_send_to_mail

    hash_object = hashlib.md5(pdf_path.name.encode())  # nosec

    cache_key = 'gen_dict_' + hash_object.hexdigest()
    if cache.get(cache_key):
        return PdfGenerateStatus.duplicate

    kwarg_dict = {f'start_date': start_date,
                  'end_date': end_date,
                  'pdf_path': pdf_path,
                  'send_mail': send_mail,
                  'user_id': request.user.id,
                  'user_mail': request.user.email,
                  'host': request.get_host()
                  }
    cache.set(cache_key, kwarg_dict, 2 * 60 * 60)
    close_old_connections()
    _start_generate_with_spool(cache_key=cache_key)
    return task_status


def _specify_run_status(start_date, end_date, request):
    attack_num = len(get_filtered_attacks(start_date, end_date, user_id=request.user.id))
    if attack_num > settings.MIN_ATTACK_TO_SEND_REPORT_TO_MAIL:
        if is_smtp_enabled():
            return PdfGenerateStatus.will_send_to_mail
        else:
            return PdfGenerateStatus.mail_not_allowed
    return PdfGenerateStatus.short_task
