import io
import logging
import os
import pypdf

from datetime import datetime
from pathlib import Path
from django.conf import settings
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Table, PageTemplate, BaseDocTemplate, Frame, TableStyle, NextPageTemplate

from main.utils import convert_from_punycode
from report.utils.send_report import send_pdf_link_to_mail
from report.utils.counters import make_attack_type_counters, make_extra_counters, \
    AttackCounter
from report.utils.utils import date_to_str, get_filtered_attacks, \
    date_to_str_with_time, get_top_ip_list
from django.db import close_old_connections

logger = logging.getLogger(__name__)
width_a4, height_a4 = A4
FONT_PDF = settings.FONT_PDF
img_logo = ImageReader(Path(settings.BASE_DIR, 'report/static/img/pentestit_logo.png'))


def generate_pdf(start_date: datetime,
                 end_date: datetime,
                 pdf_path: Path,
                 send_mail=False,
                 **request_dict) -> io.FileIO:
    """ This func return io.FileIO and save pdf file to pdf_path """
    try:
        close_old_connections()
        start_date_str = date_to_str(start_date)
        end_date_str = date_to_str(end_date)
        attacks_logs = get_filtered_attacks(start_date, end_date, user_id=request_dict.get('user_id')).order_by(
            'timestamp')
        date_string = start_date_str + ' - ' + end_date_str
        attack_numbers = attacks_logs.count()
        logger.info(f'Generating PDF report {pdf_path.name} with {attack_numbers} strings')

        class TitleTemplate(PageTemplate):
            def beforeDrawPage(self, canv, doc):
                # todo find better way to draw this pages:
                draw_main_page(canv, date_string)
                draw_second_page(canv, attacks_logs)
                draw_third_page(canv)
                canv.setFont(FONT_PDF, settings.TABLE_FONT_SIZE)

        template_border = Frame(0.5 * cm, 72, width_a4 - cm, 700, id='normal')
        titles_template = TitleTemplate(id='titles', frames=template_border, onPage=draw_template)
        attack_table_template = PageTemplate(id='attacks', frames=template_border, onPage=draw_template)
        pdf_filename = str(pdf_path)

        if attack_numbers < settings.PDF_FILE_DIVIDER:
            # generate all in 1 file

            file = make_pdf_with_titles(attacks_logs, attack_numbers, titles_template, attack_table_template,
                                        pdf_filename)
        else:

            file = make_and_merge_several_pdfs(attacks_logs, attack_numbers, attack_table_template, pdf_filename,
                                               titles_template, )

        logger.info(f'PDF report has been generated: {pdf_path.name}')
        if send_mail:
            user_mail = request_dict.get('user_mail')
            host = request_dict.get('host')
            send_pdf_link_to_mail(user_mail, host, pdf_path, date_string)
        close_old_connections()
        return file
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def make_pdf_with_titles(attacks_logs, attack_numbers, titles_template, attack_table_template, pdf_filename):
    try:
        elems = []
        doc = BaseDocTemplate(pdf_filename, showBoundary=0)
        doc.addPageTemplates([titles_template, attack_table_template])
        elems.append(NextPageTemplate('attacks'))
        make_all_elems_tables(attack_numbers, attacks_logs, elems)
        doc.build(elems)
        return io.FileIO(pdf_filename)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def make_only_attacks_pdf(attacks_logs, attack_numbers, attack_table_template, pdf_filename):
    try:
        elems = []
        doc = BaseDocTemplate(pdf_filename, showBoundary=0)
        doc.addPageTemplates([attack_table_template])
        make_all_elems_tables(attack_numbers, attacks_logs, elems)
        doc.build(elems)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def make_and_merge_several_pdfs(attacks_logs, attack_numbers, attack_table_template, pdf_filename, titles_template):
    try:
        merge_file_list = []
        file_numbers = attack_numbers // settings.PDF_FILE_DIVIDER + 1
        attack_delta = int(attack_numbers / file_numbers)

        # create first file
        pdf_filename_part = pdf_filename.rstrip('.pdf') + '_part0' + '.pdf'
        merge_file_list.append(pdf_filename_part)
        make_pdf_with_titles(attacks_logs[:attack_delta], attack_numbers, titles_template, attack_table_template,
                             pdf_filename=pdf_filename_part)

        # create other files
        for i in range(1, file_numbers):
            pdf_filename_part = pdf_filename.rstrip('.pdf') + f'_part{i}' + '.pdf'
            merge_file_list.append(pdf_filename_part)
            start = attack_delta * i
            end = attack_delta * (i + 1)
            end = None if end >= attack_numbers else end
            make_only_attacks_pdf(attacks_logs[start:end], attack_numbers, attack_table_template, pdf_filename_part)

        return merge_pdfs(merge_file_list, res_filename=pdf_filename)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def merge_pdfs(merge_file_list, res_filename):
    try:
        writer = pypdf.PdfWriter()
        tmp_filename = res_filename + '.tmp'
        for filename in merge_file_list:
            writer.append(fileobj=open(Path(filename), 'rb'))
        writer.write(open(Path(tmp_filename), 'wb'))
        os.rename(tmp_filename, res_filename)
        # delete tmp file parts
        for filename in merge_file_list:
            Path.unlink(Path(filename))
        return io.FileIO(res_filename)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def make_all_elems_tables(attack_numbers, attacks_logs, elems):
    try:
        devider = settings.PDF_TABLE_DIVIDER

        for i in range(attack_numbers // devider + 1):
            start = i * devider
            end = (i + 1) * devider
            end = None if end >= attack_numbers else end
            if not attacks_logs:
                break
            data = build_table_data(attacks_logs[start:end])
            if data:
                table = make_table_of_attacks(data)
                elems.append(table)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def build_table_data(attacks_logs):
    try:
        data = []

        for attack in attacks_logs:
            attack_time = date_to_str_with_time(attack.get('timestamp')).replace(' ', '\n')

            mz = str(attack.get('mz'))
            if mz == 'URL':
                mz = 'Path'

            # attack_info = f'Blocked by {settings.BT_DICT.get(bt).description}'
            attack_info = f"Blocked by {attack['description']}"

            if mz:
                attack_info += f' in {mz}'

            attack_method = attack.get('method').strip()
            vhost = attack.get('vhost')
            ip = attack.get('ip').strip()
            request_info = f"Target: {convert_from_punycode(vhost)} \n" \
                           f"Source IP: {ip}"
            if attack_method:
                request_info += f" ({attack_method})"

            data.append([attack_time, attack_info, request_info])
        return data
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def make_table_of_attacks(data: list) -> Table:
    try:
        table_style = TableStyle([
            ('FONTNAME', (0, 0), (-1, -1), settings.FONT_TABLE_PDF),  # font
            ('FONTSIZE', (0, 0), (-1, -1), settings.TABLE_FONT_SIZE),
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
            # ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
            ('LINEBELOW', (0, 0), (-1, -1), 0.5, colors.grey),
        ])

        table = Table(data, colWidths=[3 * cm,
                                       8 * cm,
                                       9 * cm],
                      rowHeights=25,
                      splitByRow=1, style=table_style)

        return table
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def draw_main_page(pdf, date_string):
    try:
        draw_template(pdf)
        pdf.setFont(FONT_PDF, 22)
        pdf.setFillColor(colors.HexColor('#000000'))
        pdf.drawCentredString(width_a4 / 2, height_a4 / 1.5 - 50, 'Nemesida WAF')
        pdf.drawCentredString(width_a4 / 2, height_a4 / 1.5 - 80, 'Website Protection Summary Report')
        pdf.setFont(FONT_PDF, 26)
        pdf.drawCentredString(width_a4 / 2, height_a4 / 1.5 - 110, date_string)
        pdf.showPage()
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def draw_second_page(pdf, attacks_logs):
    try:
        draw_template(pdf)
        pdf.setFont(FONT_PDF, 26)
        pdf.setFillColor(colors.HexColor('#000000'))

        logs_count_all = len(attacks_logs)  # .count()
        if logs_count_all:
            # total number of attacks
            text = f'Nemesida WAF blocked {logs_count_all} attack'
            text = text + 's' if logs_count_all > 1 else text
            pdf.drawCentredString(width_a4 / 2, height_a4 / 1.15, text)

        pdf.setFont(FONT_PDF, 14)
        y_draw_position = 30
        counters = make_attack_type_counters(attacks_logs)
        y_draw_position = draw_counters(pdf, counters, y_draw_position)

        y_draw_position += 10
        counters = make_extra_counters(attacks_logs)
        draw_counters(pdf, counters, y_draw_position)
        write_top_ip(pdf, attacks_logs, )
        pdf.showPage()
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def write_top_ip(pdf, attacks_logs, ):
    try:
        top_ip = get_top_ip_list(attacks_logs)
        pdf.setFont(FONT_PDF, 20)
        pdf.drawCentredString(width_a4 / 2, height_a4 / 1.15 - 250, f'Attackers Top {len(top_ip)}')
        pdf.setFont(FONT_PDF, 14)

        table_style = [('TEXTCOLOR', (0, 0), (-1, -1), colors.HexColor('#000000')),
                       ('FONTSIZE', (0, 0), (-1, 0), 14),
                       ('BOTTOMPADDING', (0, 0), (-1, 0), 8),
                       ('ALIGN', (1, 0), (-1, -1), 'CENTER')]

        data = [['IP-address', 'Count', 'Country']]
        for ip_dict in top_ip:
            if len(ip_dict['ip']) > 18:
                ip_dict['ip'] = ip_dict['ip'][:18] + '...'
            if 'country' not in ip_dict:
                ip_dict['country'] = ''
            data.append([ip_dict['ip'], ip_dict['ip__count'], ip_dict['country']])

        table = Table(data, style=table_style, colWidths=[100, 70, 70])
        table.wrapOn(pdf, width_a4, height_a4)
        table.drawOn(pdf, width_a4 / 2 - table._width / 2, height_a4 / 1.15 - 260 - table._height)
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def draw_counters(pdf, counters, y_draw_position):
    try:
        for counter in counters:
            counter: AttackCounter
            if counter.numbers_of_attacks:
                text = f'{counter.description}: {counter.numbers_of_attacks} attempt'
                text = text + 's' if counter.numbers_of_attacks > 1 else text
                pdf.drawCentredString(width_a4 / 2, height_a4 / 1.15 - y_draw_position, text)
                y_draw_position += 20
        return y_draw_position
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def draw_third_page(pdf):
    try:
        draw_template(pdf)
        pdf.setFont(FONT_PDF, 26)
        pdf.setFillColor(colors.HexColor('#000000'))
        pdf.drawCentredString(width_a4 / 2, height_a4 / 1.5 - 120, 'List of Attacks')
        pdf.showPage()
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')


def draw_template(pdf, doc=None):
    try:
        # makes headers and footer
        # header
        pdf.drawImage(img_logo, 30, height_a4 - 70, width=55, height=50, mask='auto')
        pdf.setFillColor(colors.HexColor('0x134473'))
        pdf.setFont(FONT_PDF, 30)
        pdf.drawString(width_a4 - 180, height_a4 - 40, '    PENTESTIT')
        pdf.setFont(FONT_PDF, 9)
        pdf.drawString(width_a4 - 160, height_a4 - 60, 'Cybersecurity services and software')
        pdf.setFont(FONT_PDF, 8)
        pdf.drawString(width_a4 - 160, height_a4 - 70, '+7 (495) 204-19-72 || info@pentestit.ru')

        # footer
        pdf.setStrokeColor(colors.HexColor('0x134473'))
        pdf.setFont(FONT_PDF, 12)
        pdf.setFillColor(colors.HexColor('0x134473'))
        pdf.drawString(width_a4 / 2 - 50, 35, 'www.pentestit.ru')
    except Exception as e:
        logger.error(f'An error occurred error when generate PDF {e}')
