
from datetime import datetime, timedelta

from django.conf import settings
from django.db.models import Count
from django.db.models.query import QuerySet
from pymemcache.client.base import Client

from main.common import get_filtered_logs, setter_geotokens_for_logs, getting_geoip_multithread, get_location_by_ip
from main.models import WAFUSER, Logs

cache_ip, cache_port = settings.CACHES['default']['LOCATION'].split(':')
cache_address = (cache_ip, int(cache_port))


def get_start_datetime(request) -> datetime:
    if 'start' in request.POST.keys():
        return datetime.strptime(request.POST.get('start'), "%d.%m.%Y")
    else:
        return (datetime.now() + timedelta(days=-30)).replace(hour=00, minute=00, second=00, microsecond=0)


def get_end_datetime(request) -> datetime:
    if 'end' in request.POST.keys():
        return datetime.strptime(request.POST.get('end'), "%d.%m.%Y").replace(hour=23, minute=59, second=59)
    else:
        datetime.now()


def date_to_str(date: datetime) -> str:
    # format 12.01.2023
    return date.strftime("%d.%m.%Y")


def date_to_str_with_time(date: datetime) -> str:
    # format 14.12.23 15:09:50
    return date.strftime("%d.%m.%y %H:%M:%S")


def get_filtered_attacks(start_date, end_date, user_id):
    domain_from_wafuser = WAFUSER.objects.get(id=user_id).domain
    user_alias = WAFUSER.objects.get(id=user_id).waf_alias
    attacks_logs = Logs.objects.using('waf').filter(timestamp__gte=start_date, timestamp__lte=end_date)
    attacks_logs = get_filtered_logs(attacks_logs, user_alias, domain_from_wafuser)
    attacks_logs = attacks_logs.values(*settings.FIELDS_FOR_REPOST).exclude(
        bt__in=settings.BT_REPORT_EXCLUDE_NUMBERS).exclude(
        possible=True)
    return attacks_logs


def get_top_ip_list(attacks_logs) -> list:
    """ return list of dicts with ip, count, country"""
    top_ip = attacks_logs.values('ip', 'alias').annotate(Count('ip')).order_by('-ip__count')[:20]

    return _add_ip_countries(top_ip)


def _add_ip_countries(top_ip: QuerySet) -> list:
    """ add ip country to every dict of list"""
    mcache = Client(cache_address)

    setter_geotokens_for_logs(top_ip)
    getting_geoip_multithread(top_ip, mcache)

    return _get_country_by_ip(top_ip, mcache)


def _get_country_by_ip(top_id: QuerySet, mcache) -> list:
    res = []
    for i in top_id:
        info_geoip = get_location_by_ip(mcache, i.get("ip"))
        cc = info_geoip.get('cc', 'None')
        if cc == "None" and info_geoip.get("description") == "Private IP":
            cc = "Private IP"
        elif info_geoip.get("cc", "None") != "None":
            cc = info_geoip.get("cc")
        else:
            cc = ""
        ip_info = {
            "ip": i.get('ip'),
            "ip__count": i.get('ip__count'),
            "country": cc
        }

        res.append(ip_info)
    return res
