import io
import logging
import time
from pathlib import Path

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect, FileResponse
from django.shortcuts import redirect
from django.core.cache import cache
from django.contrib import messages

from main.decorators import time_of_function
from report.utils.csv_generator import generate_csv_report
from report.utils.utils import get_start_datetime, get_end_datetime, date_to_str
from .tasks import start_generate_pdf_task, PdfGenerateStatus
from django.utils.translation import gettext as _

logger = logging.getLogger(__name__)
tasks = []


@login_required
@time_of_function
def report(request):
    
    if request.method != 'POST':
        return HttpResponseRedirect(settings.HOME_PAGE)

    report_type = request.POST.get('report_type')
    if report_type in ('csv', None):
        return generate_csv_report(request)
    if report_type == 'pdf':
        return get_or_generate_pdf_report(request)


def get_or_generate_pdf_report(request):
    start_date = get_start_datetime(request)
    end_date = get_end_datetime(request)

    start_date_str = date_to_str(start_date)
    end_date_str = date_to_str(end_date)

    pdf_name_for_user = f'report({start_date_str} - {end_date_str}).pdf'
    pdf_name_on_disk = pdf_name_for_user
    folder = Path(settings.TMP_REPORT_FOLDER, f'user{request.user.id}')
    folder.mkdir(parents=True, exist_ok=True)
    pdf_path = Path(settings.TMP_REPORT_FOLDER, folder, pdf_name_on_disk)

    if pdf_path.exists():
        pdf_obj = io.FileIO(pdf_path)
        return FileResponse(pdf_obj, as_attachment=True, filename=pdf_name_for_user)

    generate_status = start_generate_pdf_task(
        start_date=start_date,
        end_date=end_date,
        pdf_path=pdf_path,
        request=request
    )

    if generate_status == PdfGenerateStatus.short_task:
        return _check_file_and_return(request, pdf_path, pdf_name_for_user)
    elif generate_status == PdfGenerateStatus.duplicate:
        # this means that the report is still being generated, and user order it again
        messages.add_message(request, messages.INFO, _('Pdf_gen: Duplicated request to generate pdf'))
    elif generate_status == PdfGenerateStatus.will_send_to_mail:
        messages.add_message(request, messages.INFO, _('Pdf_gen: Pdf link will be sent on email'))
    elif generate_status == PdfGenerateStatus.mail_not_allowed:
        messages.add_message(request, messages.INFO, _('Pdf_gen: long task. You can config mail'))
    return redirect(request.META.get('HTTP_REFERER', 'redirect_if_referer_not_found'))


def _check_file_and_return(request, pdf_path, pdf_name_for_user):
    time_limit = 45
    timer = 0
    delta_time = 0.2
    while timer <= time_limit:
        time.sleep(delta_time)
        timer += delta_time
        if pdf_path.exists():
            pdf_obj = io.FileIO(pdf_path)
            return FileResponse(pdf_obj, as_attachment=True, filename=pdf_name_for_user)
    messages.add_message(request, messages.ERROR, _('Pdf_gen: Something went wrong'))
    return redirect(request.META.get('HTTP_REFERER', 'redirect_if_referer_not_found'))


@login_required()
def get_pdf_file_by_link(request, slug):
    file_path = cache.get(slug)
    if not file_path or not Path(file_path).exists():
        messages.add_message(request, messages.ERROR, _('Link is not valid or file not found.'))
        return redirect('Logs')

    pdf_obj = io.FileIO(file_path)
    return FileResponse(pdf_obj, as_attachment=True, filename=file_path.name)
