import logging

import pyotp
from django.conf import settings
from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils import translation
from django.utils.crypto import get_random_string
from django.utils.translation import gettext as _
from django.core.cache import cache
from django.shortcuts import redirect

from main.common import get_client_ip
from main.decorators import time_of_function
from main.models import Config, WAFUSER
from main.utils import convert_to_punycode, convert_from_punycode
from user.decorators import rate_limit
from user.user_settings import check_and_set_lang

logger = logging.getLogger(__name__)


@rate_limit(limit=6, timeout=30)
def login_view(request):
    if request.user.is_authenticated:
        return redirect('Logs')

    if request.method == 'POST' and request.rate_limit_error:
        # check rate limit
        email = request.POST['email'].strip()
        logger.warning(f'login attempt {email}: exceeding the number of requests')
        return render(request, 'login.html', {'login_error': _('Limit is exceeded. Try again in 30 seconds')})

    enabl_email = Config.objects.only("send_email").first()
    if not enabl_email:
        enabl_email = False
    else:
        enabl_email = enabl_email.send_email

    context = dict()
    context['email_enable'] = enabl_email
    context['home'] = settings.HOME_PAGE if settings.HOME_PAGE else '/'
    context.update(csrf(request))

    if request.POST and 'email' in request.POST.keys() \
            and 'password' in request.POST.keys():

        email = request.POST['email'].strip()
        password = request.POST['password']

        user = auth.authenticate(email=convert_to_punycode(email), password=password)
        if user is not None:
            if not user.is_2fa_active:
                logger.info(f'{email}: successful login')
                auth.login(request, user)
                return redirect('Logs')
            else:
                token = generate_and_cache_tmp_token(request, 15)
                context['token'] = token
                context['email'] = email
                logger.info(f'{email}: successful login (1st factor)')
                return render(request, 'two_factor_auth.html', context)
        else:
            logger.warning(f'login attempt {email}: wrong password or email')
            context['login_error'] = _('Incorrect password')
            return render(request, 'login.html', context)
    else:
        return check_and_set_lang(context, request)


def generate_and_cache_tmp_token(request, length=10) -> str:
    token = get_random_string(length)
    cache_key = get_cache_key_for_auth_token(request)
    cache.set(cache_key, token, timeout=300)
    return token


def get_cache_key_for_auth_token(request):
    ip = get_client_ip(request)
    email = request.POST.get('email')
    return f'{ip}_{email}'


@time_of_function
@rate_limit(limit=6, timeout=30)
def two_factor_auth(request):
    email = request.POST.get('email')
    user = WAFUSER.objects.get(email=convert_to_punycode(email))
    code_input = request.POST.get('code')
    token = request.POST.get('token')
    token_cache_key = get_cache_key_for_auth_token(request)
    token_in_cache = cache.get(token_cache_key)

    context = {'email': email,
               'token': token,
               'home': settings.HOME_PAGE}

    if request.method == 'POST' and request.rate_limit_error:
        # check rate limit
        context['login_error'] = _('Limit is exceeded. Try again in 30 seconds')
        logger.warning(f'login attempt {email}: exceeding the number of requests')
        return render(request, 'two_factor_auth.html', context)

    if not token_in_cache or not token:
        return redirect('login')

    if token == token_in_cache and pyotp.TOTP(user.secret).verify(code_input):
        auth.login(request, user)
        cache.delete(token_cache_key)
        logger.info(f'{email}: successful login (2nd factor)')
        return redirect('Logs')
    else:
        logger.warning(f'login attempt {email}: wrong 2FA code')
        context['incorrect_code'] = 'incorrect_code'
        return render(request, 'two_factor_auth.html', context)


@login_required
@time_of_function
def logout_view(request):
    request.LANGUAGE_CODE = translation.get_language()
    translation.activate(request.LANGUAGE_CODE)
    response = redirect('Logs')
    response.set_cookie(settings.LANGUAGE_COOKIE_NAME, request.LANGUAGE_CODE)
    auth.logout(request)

    return response
