import os
import sys
import logging
import smtplib
import psycopg2
import psycopg2.extras

from django.conf import settings
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
from pathlib import Path

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)

from main.utils import init_db_connection

if os.environ.get('DJANGO_SETTINGS_MODULE') is None:
    os.environ['DJANGO_SETTINGS_MODULE'] = "cabinet.settings_extra"

# Postgres connect
try:
    DB_PORT_CABINET = settings.DB_PORT_CABINET
    if DB_PORT_CABINET == '':
        DB_PORT_CABINET = '5432'
except NameError:
    DB_PORT_CABINET = '5432'


def smtp_config():
    connect = init_db_connection(
        settings.DB_NAME_CABINET,
        settings.DB_USER_CABINET,
        settings.DB_PASS_CABINET,
        settings.DB_HOST_CABINET,
        DB_PORT_CABINET)
    cursor = connect.cursor(cursor_factory=psycopg2.extras.DictCursor)
    cursor.execute('SELECT smtp_host, smtp_port, smtp_user, smtp_password, encryption, send_email FROM config;')
    result = cursor.fetchone()
    return result


def send_mail_from_cabinet(send_to: str,
                           subject: str,
                           text: str,
                           files: list = None,
                           logger: logging = None) -> bool:
    """ to attach file use list/tuple of files path or single path, return True after success
    pass a logger if you want to log errors
    """

    mail_conf = smtp_config()
    host = mail_conf.get('smtp_host')
    port = mail_conf.get('smtp_port')
    username = mail_conf.get('smtp_user')
    encryption = mail_conf.get('encryption')
    password = mail_conf.get('smtp_password')

    message = MIMEMultipart()
    message['From'] = username
    message['To'] = send_to
    message['Subject'] = subject
    message['Date'] = formatdate(localtime=True)

    # Add the message text
    message.attach(MIMEText(text, 'plain', 'utf-8'))

    if not files:
        files = []
    elif not isinstance(files, (list, tuple)):
        # if only one file
        files = [files]

    try:
        for file in files:
            file = Path(file)
            # Set up the message objects
            with open(file, 'rb') as f:
                attachment = MIMEApplication(f.read(), _subtype='pdf')
                attachment.add_header('Content-Disposition', 'attachment', filename=file.name)
                message.attach(attachment)

        server = _get_smtp_connector(host, port, encryption)
        server.login(username, password)
        server.send_message(message)
        server.close()
        if logger:
            logger.info(f'The mail with subject "{subject}" has been sent to {send_to}')
        return True
    except Exception as err:
        if logger:
            logger.exception(f'An error occurred while sending the mail: {err}')
        return False


def _get_smtp_connector(host, port, encryption):
    if encryption.lower() == 'ssl':
        server = smtplib.SMTP_SSL(host, port)
    elif encryption.lower() == 'tls':
        server = smtplib.SMTP(host, port)
        server.starttls()
    else:
        server = smtplib.SMTP(host, port)
    return server


def is_smtp_enabled() -> bool:
    """ check if smtp enabled """
    mail_conf = smtp_config()
    return mail_conf.get('send_email')
